@@ -1,15 +1,20 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Behaviour do
   @moduledoc """
   Mechanism for handling behaviours.
 
   This module is deprecated. Instead of `defcallback/1` and
   `defmacrocallback/1`, the `@callback` and `@macrocallback`
-  module attributes can be used (respectively). See the
+  module attributes can be used respectively. See the
   documentation for `Module` for more information on these
   attributes.
 
   Instead of `MyModule.__behaviour__(:callbacks)`,
-  `MyModule.behaviour_info(:callbacks)` can be used.
+  `MyModule.behaviour_info(:callbacks)` can be used. `behaviour_info/1`
+  is documented in `Module`.
   """
 
   @moduledoc deprecated: "Use @callback and @macrocallback attributes instead"
@@ -17,6 +22,7 @@ defmodule Behaviour do
   @doc """
   Defines a function callback according to the given type specification.
   """
+  @deprecated "Use the @callback module attribute instead"
   defmacro defcallback(spec) do
     do_defcallback(:def, split_spec(spec, quote(do: term)))
   end
@@ -24,6 +30,7 @@ defmodule Behaviour do
   @doc """
   Defines a macro callback according to the given type specification.
   """
+  @deprecated "Use the @macrocallback module attribute instead"
   defmacro defmacrocallback(spec) do
     do_defcallback(:defmacro, split_spec(spec, quote(do: Macro.t())))
   end
