@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Exception do
   @moduledoc """
   Functions for dealing with throw/catch/exit and exceptions.
@@ -78,6 +82,11 @@ defmodule Exception do
 
   @doc """
   Gets the message for an `exception`.
+
+  This function will invoke the `c:message/1` callback on the exception
+  module to retrieve the message. If the callback raises an exception or
+  returns a non-binary value, this function will rescue the error and
+  return a descriptive error message instead.
   """
   @spec message(t) :: String.t()
   def message(%module{__exception__: true} = exception) do
@@ -149,7 +158,7 @@ defmodule Exception do
   end
 
   @doc """
-  Normalizes and formats throw/errors/exits and stacktraces.
+  Normalizes and formats throws/errors/exits and stacktraces.
 
   It relies on `format_banner/3` and `format_stacktrace/1`
   to generate the final format.
@@ -173,15 +182,33 @@ defmodule Exception do
     end
   end
 
+  @doc false
+  def __format_message_with_term__(message, term) do
+    inspected =
+      term
+      |> inspect(pretty: true)
+      |> String.split("\n")
+      |> Enum.map_intersperse("\n", fn
+        "" -> ""
+        line -> "    " <> line
+      end)
+
+    IO.iodata_to_binary([message, "\n\n", inspected, "\n"])
+  end
+
   @doc """
-  Attaches information to exceptions for extra debugging.
+  Attaches information to throws/errors/exits for extra debugging.
 
   This operation is potentially expensive, as it reads data
   from the file system, parses beam files, evaluates code and
   so on.
 
-  If the exception module implements the optional `c:blame/2`
-  callback, it will be invoked to perform the computation.
+  If `kind` argument is `:error` and the `error` is an Erlang exception, this function will
+  normalize it. If the `error` argument is an Elixir exception, this function will invoke
+  the optional `c:blame/2` callback on the exception module if it is implemented.
+  Unlike `message/1`, this function will not rescue errors - if the callback raises an exception,
+  the error will propagate to the caller. It is your choice if you want to rescue and return
+  the original exception, return a different exception, or let it cascade.
   """
   @doc since: "1.5.0"
   @spec blame(:error, any, stacktrace) :: {t, stacktrace}
@@ -1040,7 +1067,12 @@ defmodule ArgumentError do
   An exception raised when an argument to a function is invalid.
 
   You can raise this exception when you want to signal that an argument to
-  a function is invalid.
+  a function is invalid. For example, this exception is raised when calling
+  `Integer.to_string/1` with an invalid argument:
+
+      iex> Integer.to_string(1.0)
+      ** (ArgumentError) errors were found at the given arguments:
+      ...
 
   `ArgumentError` exceptions have a single field, `:message` (a `t:String.t/0`),
   which is public and can be accessed freely when reading or creating `ArgumentError`
@@ -1057,8 +1089,7 @@ defmodule ArithmeticError do
   For example, this exception is raised if you divide by `0`:
 
       iex> 1 / 0
-      ** (ArithmeticError) bad argument in arithmetic expression: 1 / 0
-
+      ** (ArithmeticError) bad argument in arithmetic expression
   """
 
   defexception message: "bad argument in arithmetic expression"
@@ -1104,7 +1135,10 @@ defmodule SystemLimitError do
   @moduledoc """
   An exception raised when a system limit has been reached.
 
-  For example, this can happen if you try to create an atom that is too large.
+  For example, this can happen if you try to create an atom that is too large:
+
+      iex> String.to_atom(String.duplicate("a", 100_000))
+      ** (SystemLimitError) a system limit has been reached
   """
 
   defexception message: "a system limit has been reached"
@@ -1116,8 +1150,9 @@ defmodule MismatchedDelimiterError do
 
   For example:
 
-    * `[1, 2, 3}`
-    * `fn a -> )`
+      iex> Code.eval_string("[1, 2, 3}")
+      ** (MismatchedDelimiterError) mismatched delimiter found on nofile:1:9:
+      ...
 
   The following fields of this exceptions are public and can be accessed freely:
 
@@ -1131,7 +1166,6 @@ defmodule MismatchedDelimiterError do
     * `:closing_delimiter` - an atom representing the mismatched closing delimiter
     * `:expected_delimiter` - an atom representing the closing delimiter
     * `:description` - a description of the mismatched delimiter error
-
   """
 
   defexception [
@@ -1190,6 +1224,12 @@ defmodule SyntaxError do
   @moduledoc """
   An exception raised when there's a syntax error when parsing code.
 
+  For example:
+
+      iex> Code.eval_string("5 + 5h")
+      ** (SyntaxError) invalid syntax found on nofile:1:5:
+      ...
+
   The following fields of this exceptions are public and can be accessed freely:
 
     * `:file` (`t:Path.t/0` or `nil`) - the file where the error occurred, or `nil` if
@@ -1197,7 +1237,6 @@ defmodule SyntaxError do
     * `:line` - the line where the error occurred
     * `:column` - the column where the error occurred
     * `:description` - a description of the syntax error
-
   """
 
   defexception [:file, :line, :column, :snippet, description: "syntax error"]
@@ -1241,6 +1280,12 @@ defmodule TokenMissingError do
   @moduledoc """
   An exception raised when a token is missing when parsing code.
 
+  For example:
+
+      iex> Code.eval_string("[1, 2, 3")
+      ** (TokenMissingError) token missing on nofile:1:9:
+      ...
+
   The following fields of this exceptions are public and can be accessed freely:
 
     * `:file` (`t:Path.t/0` or `nil`) - the file where the error occurred, or `nil` if
@@ -1252,6 +1297,8 @@ defmodule TokenMissingError do
     * `:opening_delimiter` - an atom representing the opening delimiter
     * `:expected_delimiter` - an atom representing the expected delimiter
     * `:description` - a description of the missing token error
+
+  This is mostly raised by Elixir tooling when compiling and evaluating code.
   """
 
   defexception [
@@ -1331,12 +1378,18 @@ defmodule CompileError do
   @moduledoc """
   An exception raised when there's an error when compiling code.
 
+  For example:
+
+      1 = y
+      ** (CompileError) iex:1: undefined variable "y"
+
   The following fields of this exceptions are public and can be accessed freely:
 
     * `:file` (`t:Path.t/0` or `nil`) - the file where the error occurred, or `nil` if
       the error occurred in code that did not come from a file
     * `:line` (`t:non_neg_integer/0`) - the line where the error occurred
 
+  This is mostly raised by Elixir tooling when compiling and evaluating code.
   """
 
   defexception [:file, :line, description: "compile error"]
@@ -1354,12 +1407,19 @@ defmodule Kernel.TypespecError do
   @moduledoc """
   An exception raised when there's an error in a typespec.
 
+  For example, if your typespec definition points to an invalid type, you get an exception:
+
+      @type my_type :: intger()
+
+  will raise:
+
+      ** (Kernel.TypespecError) type intger/0 undefined
+
   The following fields of this exceptions are public and can be accessed freely:
 
     * `:file` (`t:Path.t/0` or `nil`) - the file where the error occurred, or `nil` if
       the error occurred in code that did not come from a file
     * `:line` (`t:non_neg_integer/0`) - the line where the error occurred
-
   """
 
   defexception [:file, :line, :description]
@@ -1374,6 +1434,16 @@ defmodule Kernel.TypespecError do
 end
 
 defmodule BadFunctionError do
+  @moduledoc """
+  An exception raised when a function is expected, but something else was given.
+
+  For example:
+
+      iex> value = "hello"
+      iex> value.()
+      ** (BadFunctionError) expected a function, got: "hello"
+  """
+
   defexception [:term]
 
   @impl true
@@ -1387,37 +1457,61 @@ defmodule BadFunctionError do
 end
 
 defmodule BadStructError do
+  @moduledoc deprecated:
+               "This exception is deprecated alongside the struct update syntax that raises it"
   defexception [:struct, :term]
 
   @impl true
   def message(exception) do
-    "expected a struct named #{inspect(exception.struct)}, got: #{inspect(exception.term)}"
+    Exception.__format_message_with_term__(
+      "expected a struct named #{inspect(exception.struct)}, got:",
+      exception.term
+    )
   end
 end
 
 defmodule BadMapError do
   @moduledoc """
-  An exception raised when something expected a map, but received something else.
+  An exception raised when a map is expected, but something else was given.
+
+  For example:
+
+      iex> value = "hello"
+      iex> %{value | key: "value"}
+      ** (BadMapError) expected a map, got:
+      ...
   """
 
   defexception [:term]
 
   @impl true
   def message(exception) do
-    "expected a map, got: #{inspect(exception.term)}"
+    Exception.__format_message_with_term__(
+      "expected a map, got:",
+      exception.term
+    )
   end
 end
 
 defmodule BadBooleanError do
   @moduledoc """
-  An exception raised when an operator expected a boolean, but received something else.
+  An exception raised when a boolean is expected, but something else was given.
+
+  This exception is raised by `and` and `or` when the first argument is not a boolean:
+
+      iex> 123 and true
+      ** (BadBooleanError) expected a boolean on left-side of "and", got:
+      ...
   """
 
   defexception [:term, :operator]
 
   @impl true
   def message(exception) do
-    "expected a boolean on left-side of \"#{exception.operator}\", got: #{inspect(exception.term)}"
+    Exception.__format_message_with_term__(
+      "expected a boolean on left-side of \"#{exception.operator}\", got:",
+      exception.term
+    )
   end
 end
 
@@ -1425,21 +1519,26 @@ defmodule MatchError do
   @moduledoc """
   An exception raised when a pattern match (`=/2`) fails.
 
+  For example:
+
+      iex> [_ | _] = []
+      ** (MatchError) no match of right hand side value:
+      ...
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:term` (`t:term/0`) - the term that did not match the pattern
 
-  For example, this exception gets raised for code like this:
-
-      [_ | _] = []
-
   """
 
   defexception [:term]
 
   @impl true
   def message(exception) do
-    "no match of right hand side value: #{inspect(exception.term)}"
+    Exception.__format_message_with_term__(
+      "no match of right hand side value:",
+      exception.term
+    )
   end
 end
 
@@ -1448,24 +1547,28 @@ defmodule CaseClauseError do
   An exception raised when a term in a `case/2` expression
   does not match any of the defined `->` clauses.
 
+  For example:
+
+      iex> case System.unique_integer() do
+      ...>   bin when is_binary(bin) -> :oops
+      ...>   :ok -> :neither_this_one
+      ...> end
+      ** (CaseClauseError) no case clause matching:
+      ...
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:term` (`t:term/0`) - the term that did not match any of the clauses
-
-  For example, this exception gets raised for a `case/2` like the following:
-
-      case System.unique_integer() do
-        bin when is_binary(bin) -> :oops
-        :ok -> :neither_this_one
-      end
-
   """
 
   defexception [:term]
 
   @impl true
   def message(exception) do
-    "no case clause matching: #{inspect(exception.term)}"
+    Exception.__format_message_with_term__(
+      "no case clause matching:",
+      exception.term
+    )
   end
 end
 
@@ -1474,28 +1577,32 @@ defmodule WithClauseError do
   An exception raised when a term in a `with/1` expression
   does not match any of the defined `->` clauses in its `else`.
 
-  The following fields of this exception are public and can be accessed freely:
-
-    * `:term` (`t:term/0`) - the term that did not match any of the clauses
-
   For example, this exception gets raised for a `with/1` like the following, because
   the `{:ok, 2}` term does not match the `:error` or `{:error, _}` clauses in the
   `else`:
 
-      with {:ok, 1} <- {:ok, 2} do
-        :woah
-      else
-        :error -> :error
-        {:error, _} -> :error
-      end
+      iex> with {:ok, 1} <- {:ok, 2} do
+      ...>   :woah
+      ...> else
+      ...>   :error -> :error
+      ...>   {:error, _} -> :error
+      ...> end
+      ** (WithClauseError) no with clause matching:
+      ...
 
+  The following fields of this exception are public and can be accessed freely:
+
+    * `:term` (`t:term/0`) - the term that did not match any of the clauses
   """
 
   defexception [:term]
 
   @impl true
   def message(exception) do
-    "no with clause matching: #{inspect(exception.term)}"
+    Exception.__format_message_with_term__(
+      "no with clause matching:",
+      exception.term
+    )
   end
 end
 
@@ -1505,11 +1612,11 @@ defmodule CondClauseError do
 
   For example, this exception gets raised for a `cond/1` like the following:
 
-      cond do
-        1 + 1 == 3 -> :woah
-        nil -> "yeah this won't happen
-      end
-
+      iex> cond do
+      ...>   1 + 1 == 3 -> :woah
+      ...>   nil -> "yeah this won't happen"
+      ...> end
+      ** (CondClauseError) no cond clause evaluated to a truthy value
   """
 
   defexception []
@@ -1522,8 +1629,20 @@ end
 
 defmodule TryClauseError do
   @moduledoc """
-  An exception raised when a term in a `try/1` expression
-  does not match any of the defined `->` clauses in its `else`.
+  An exception raised when none of the `else` clauses in a `try/1` match.
+
+  For example:
+
+      iex> try do
+      ...>   :ok
+      ...> rescue
+      ...>   e -> e
+      ...> else
+      ...>   # :ok -> :ok is missing
+      ...>   :not_ok -> :not_ok
+      ...> end
+      ** (TryClauseError) no try clause matching:
+      ...
 
   The following fields of this exception are public and can be accessed freely:
 
@@ -1534,13 +1653,22 @@ defmodule TryClauseError do
 
   @impl true
   def message(exception) do
-    "no try clause matching: #{inspect(exception.term)}"
+    Exception.__format_message_with_term__(
+      "no try clause matching:",
+      exception.term
+    )
   end
 end
 
 defmodule BadArityError do
   @moduledoc """
   An exception raised when a function is called with the wrong number of arguments.
+
+  For example:
+
+      my_function = fn x, y -> x + y end
+      my_function.(42)
+      ** (BadArityError) #Function<41.39164016/2 in :erl_eval.expr/6> with arity 2 called with 1 argument (42)
   """
 
   defexception [:function, :args]
@@ -1563,22 +1691,17 @@ defmodule UndefinedFunctionError do
   @moduledoc """
   An exception raised when a function is invoked that is not defined.
 
+  For example:
+
+      # Let's use apply/3 as otherwise Elixir emits a compile-time warning
+      iex> apply(String, :non_existing_fun, ["hello"])
+      ** (UndefinedFunctionError) function String.non_existing_fun/1 is undefined or private
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:module` (`t:module/0`) - the module name
     * `:function` (`t:atom/0`) - the function name
     * `:arity` (`t:non_neg_integer/0`) - the arity of the function
-
-  For example, if you try to call `MyMod.non_existing_fun("hello", 1)`,
-  the error would look like:
-
-      %UndefinedFunctionError{
-        module: MyMod,
-        function: :non_existing_fun,
-        arity: 2,
-        # Other private fields...
-      }
-
   """
 
   @function_threshold 0.77
@@ -1788,7 +1911,7 @@ defmodule UndefinedFunctionError do
   end
 
   defp format_fa({_dist, fun, arity}) do
-    ["      * ", Macro.inspect_atom(:remote_call, fun), ?/, Integer.to_string(arity), ?\n]
+    ["    * ", Macro.inspect_atom(:remote_call, fun), ?/, Integer.to_string(arity), ?\n]
   end
 
   defp exports_for(module) do
@@ -1818,22 +1941,16 @@ defmodule FunctionClauseError do
   @moduledoc """
   An exception raised when a function call doesn't match any defined clause.
 
+  For example:
+
+      iex> URI.parse(:wrong_argument)
+      ** (FunctionClauseError) no function clause matching in URI.parse/1
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:module` (`t:module/0`) - the module name
     * `:function` (`t:atom/0`) - the function name
     * `:arity` (`t:non_neg_integer/0`) - the arity of the function
-
-  For example, if you try to call a function such as `URI.parse/1` with something
-  other than a string, the error would look like:
-
-      %FunctionClauseError{
-        module: URI,
-        function: :parse,
-        arity: 1,
-        # Other private fields...
-      }
-
   """
 
   defexception [:module, :function, :arity, :kind, :args, :clauses]
@@ -1953,11 +2070,15 @@ defmodule Code.LoadError do
   @moduledoc """
   An exception raised when a file cannot be loaded.
 
+  This is typically raised by functions in the `Code` module, for example:
+
+      Code.require_file("missing_file.exs")
+      ** (Code.LoadError) could not load missing_file.exs. Reason: enoent
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:file` (`t:String.t/0`) - the file name
     * `:reason` (`t:term/0`) - the reason why the file could not be loaded
-
   """
 
   defexception [:file, :message, :reason]
@@ -1974,23 +2095,16 @@ defmodule Protocol.UndefinedError do
   @moduledoc """
   An exception raised when a protocol is not implemented for a given value.
 
+  For example:
+
+      iex> Enum.at("A string!", 0)
+      ** (Protocol.UndefinedError) protocol Enumerable not implemented for BitString
+      ...
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:protocol` (`t:module/0`) - the protocol that is not implemented
     * `:value` (`t:term/0`) - the value that does not implement the protocol
-
-  For example, this code:
-
-      Enum.at("A string!", 0)
-
-  would raise the following exception:
-
-      %Protocol.UndefinedError{
-        protocol: Enumerable,
-        value: "A string!",
-        # ...
-      }
-
   """
 
   defexception [:protocol, :value, description: ""]
@@ -2003,7 +2117,7 @@ defmodule Protocol.UndefinedError do
       # Indent only lines with contents on them
       |> String.replace(~r/^(?=.+)/m, "    ")
 
-    "protocol #{inspect(protocol)} not implemented for type " <>
+    "protocol #{inspect(protocol)} not implemented for " <>
       value_type(value) <>
       maybe_description(description) <>
       maybe_available(protocol) <>
@@ -2038,7 +2152,7 @@ defmodule Protocol.UndefinedError do
         ". There are no implementations for this protocol."
 
       {:consolidated, types} ->
-        ". This protocol is implemented for the following type(s): " <>
+        ". This protocol is implemented for: " <>
           Enum.map_join(types, ", ", &inspect/1)
 
       :not_consolidated ->
@@ -2052,13 +2166,17 @@ defmodule KeyError do
   An exception raised when a key is not found in a data structure.
 
   For example, this is raised by `Map.fetch!/2` when the given key
-  cannot be found in the given map.
+  cannot be found in the given map:
+
+      iex> map = %{name: "Alice", age: 25}
+      iex> Map.fetch!(map, :first_name)
+      ** (KeyError) key :first_name not found in:
+      ...
 
   The following fields of this exception are public and can be accessed freely:
 
     * `:term` (`t:term/0`) - the data structure that was searched
     * `:key` (`t:term/0`) - the key that was not found
-
   """
 
   defexception [:key, :term, :message]
@@ -2080,7 +2198,10 @@ defmodule KeyError do
           "make sure to add parentheses after the function name)"
 
       true ->
-        message <> " in: #{inspect(term, pretty: true, limit: :infinity)}"
+        Exception.__format_message_with_term__(
+          message <> " in:",
+          term
+        )
     end
   end
 
@@ -2122,7 +2243,7 @@ defmodule KeyError do
 
     case suggestions do
       [] -> []
-      suggestions -> [". Did you mean:\n\n" | format_suggestions(suggestions)]
+      suggestions -> ["\nDid you mean:\n\n" | format_suggestions(suggestions)]
     end
   end
 
@@ -2131,11 +2252,20 @@ defmodule KeyError do
     |> Enum.sort(&(elem(&1, 0) >= elem(&2, 0)))
     |> Enum.take(@max_suggestions)
     |> Enum.sort(&(elem(&1, 1) <= elem(&2, 1)))
-    |> Enum.map(fn {_, key} -> ["      * ", inspect(key), ?\n] end)
+    |> Enum.map(fn {_, key} -> ["    * ", inspect(key), ?\n] end)
   end
 end
 
 defmodule UnicodeConversionError do
+  @moduledoc """
+  An exception raised when converting data to or from Unicode.
+
+  For example:
+
+      iex> String.to_charlist(<<0xFF>>)
+      ** (UnicodeConversionError) invalid encoding starting at <<255>>
+
+  """
   defexception [:encoded, :message]
 
   def exception(opts) do
@@ -2213,10 +2343,31 @@ defmodule Enum.OutOfBoundsError do
   An exception that is raised when a function expects an enumerable to have
   a certain size but finds that it is too small.
 
-  For example, this is raised by `Access.at!/1`.
+  For example:
+
+      iex> Enum.fetch!([1, 2, 3], 5)
+      ** (Enum.OutOfBoundsError) out of bounds error at position 5 when traversing enumerable [1, 2, 3]
   """
 
-  defexception message: "out of bounds error"
+  defexception [:enumerable, :index, :message]
+
+  @impl true
+  def message(exception = %{message: nil}), do: message(exception.index, exception.enumerable)
+  def message(%{message: message}), do: message
+
+  def message(index, enumerable) do
+    "out of bounds error" <>
+      if index do
+        " at position #{index}"
+      else
+        ""
+      end <>
+      if enumerable do
+        " when traversing enumerable #{inspect(enumerable)}"
+      else
+        ""
+      end
+  end
 end
 
 defmodule Enum.EmptyError do
@@ -2224,7 +2375,11 @@ defmodule Enum.EmptyError do
   An exception that is raised when something expects a non-empty enumerable
   but finds an empty one.
 
-  For example, this is raised by `Enum.min/3`.
+  For example:
+
+      iex> Enum.min([])
+      ** (Enum.EmptyError) empty error
+
   """
 
   defexception message: "empty error"
@@ -2234,6 +2389,11 @@ defmodule File.Error do
   @moduledoc """
   An exception that is raised when a file operation fails.
 
+  For example, this exception is raised, when trying to read a non existent file:
+
+      iex> File.read!("nonexistent_file.txt")
+      ** (File.Error) could not read file "nonexistent_file.txt": no such file or directory
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:path` (`t:Path.t/0`) - the path of the file that caused the error
@@ -2262,6 +2422,11 @@ defmodule File.CopyError do
   @moduledoc """
   An exception that is raised when copying a file fails.
 
+  For example, this exception is raised when trying to copy to file or directory that isn't present:
+
+      iex> File.cp_r!("non_existent", "source_dir/subdir")
+      ** (File.CopyError) could not copy recursively from "non_existent" to "source_dir/subdir". non_existent: no such file or directory
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:source` (`t:Path.t/0`) - the source path
@@ -2291,6 +2456,11 @@ defmodule File.RenameError do
   @moduledoc """
   An exception that is raised when renaming a file fails.
 
+  For example, this exception is raised when trying to rename a file that isn't present:
+
+    iex> File.rename!("source.txt", "target.txt")
+    ** (File.RenameError) could not rename from "source.txt" to "target.txt": no such file or directory
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:source` (`t:Path.t/0`) - the source path
@@ -2320,6 +2490,11 @@ defmodule File.LinkError do
   @moduledoc """
   An exception that is raised when linking a file fails.
 
+  For example, this exception is raised when trying to link to file that isn't present:
+
+      iex> File.ln!("existing.txt", "link.txt")
+      ** (File.LinkError) could not create hard link from "link.txt" to "existing.txt": no such file or directory
+
   The following fields of this exception are public and can be accessed freely:
 
     * `:existing` (`t:Path.t/0`) - the existing file to link
@@ -2334,12 +2509,25 @@ defmodule File.LinkError do
   def message(exception) do
     formatted = IO.iodata_to_binary(:file.format_error(exception.reason))
 
-    "could not #{exception.action} from #{inspect(exception.existing)} to " <>
-      "#{inspect(exception.new)}: #{formatted}"
+    "could not #{exception.action} from #{inspect(exception.new)} to " <>
+      "#{inspect(exception.existing)}: #{formatted}"
   end
 end
 
 defmodule ErlangError do
+  @moduledoc """
+  An exception raised when invoking an Erlang code that errors
+  with a value not handled by Elixir.
+
+  Most common error reasons, such as `:badarg` are automatically
+  converted into exceptions by Elixir. However, you may invoke some
+  code that emits a custom error reason and those get wrapped into
+  `ErlangError`:
+
+      iex> :erlang.error(:some_invalid_error)
+      ** (ErlangError) Erlang error: :some_invalid_error
+  """
+
   defexception [:original, :reason]
 
   @impl true
@@ -2362,7 +2550,7 @@ defmodule ErlangError do
             is_map(module) and is_atom(function) and is_map_key(module, function) ->
               "you attempted to apply a function named #{inspect(function)} on a map/struct. " <>
                 "If you are using Kernel.apply/3, make sure the module is an atom. " <>
-                if is_function(module[function]) do
+                if is_function(Map.get(module, function)) do
                   "If you are trying to invoke an anonymous function in a map/struct, " <>
                     "add a dot between the function name and the parenthesis: map.#{function}.()"
                 else
@@ -2547,61 +2735,3 @@ defmodule ErlangError do
   defp nth(3), do: "3rd"
   defp nth(n), do: "#{n}th"
 end
-
-defmodule Inspect.Error do
-  @moduledoc """
-  Raised when a struct cannot be inspected.
-  """
-  @enforce_keys [:exception_module, :exception_message, :stacktrace, :inspected_struct]
-  defexception @enforce_keys
-
-  @impl true
-  def exception(arguments) when is_list(arguments) do
-    exception = Keyword.fetch!(arguments, :exception)
-    exception_module = exception.__struct__
-    exception_message = Exception.message(exception) |> String.trim_trailing("\n")
-    stacktrace = Keyword.fetch!(arguments, :stacktrace)
-    inspected_struct = Keyword.fetch!(arguments, :inspected_struct)
-
-    %Inspect.Error{
-      exception_module: exception_module,
-      exception_message: exception_message,
-      stacktrace: stacktrace,
-      inspected_struct: inspected_struct
-    }
-  end
-
-  @impl true
-  def message(%__MODULE__{
-        exception_module: exception_module,
-        exception_message: exception_message,
-        inspected_struct: inspected_struct
-      }) do
-    ~s'''
-    got #{inspect(exception_module)} with message:
-
-        """
-    #{pad(exception_message, 4)}
-        """
-
-    while inspecting:
-
-    #{pad(inspected_struct, 4)}
-    '''
-  end
-
-  @doc false
-  def pad(message, padding_length)
-      when is_binary(message) and is_integer(padding_length) and padding_length >= 0 do
-    padding = String.duplicate(" ", padding_length)
-
-    message
-    |> String.split("\n")
-    |> Enum.map(fn
-      "" -> "\n"
-      line -> [padding, line, ?\n]
-    end)
-    |> IO.iodata_to_binary()
-    |> String.trim_trailing("\n")
-  end
-end
