@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule File do
   @moduledoc ~S"""
   This module contains functions to manipulate files.
@@ -101,6 +105,7 @@ defmodule File do
 
   @type posix :: :file.posix()
   @type io_device :: :file.io_device()
+  @type file_descriptor :: :file.fd()
   @type stat_options :: [time: :local | :universal | :posix]
   @type mode ::
           :append
@@ -261,8 +266,15 @@ defmodule File do
     * `:enotdir` - a component of `path` is not a directory;
       on some platforms, `:enoent` is returned instead
 
+  ## Examples
+
+      File.mkdir("test/unit")
+      #=> :ok
+
+      File.mkdir("non/existing")
+      #=> {:error, :enoent}
   """
-  @spec mkdir(Path.t()) :: :ok | {:error, posix}
+  @spec mkdir(Path.t()) :: :ok | {:error, posix | :badarg}
   def mkdir(path) do
     :file.make_dir(IO.chardata_to_string(path))
   end
@@ -270,6 +282,14 @@ defmodule File do
   @doc """
   Same as `mkdir/1`, but raises a `File.Error` exception in case of failure.
   Otherwise `:ok`.
+
+  ## Examples
+
+      File.mkdir!("test/unit")
+      #=> :ok
+
+      File.mkdir!("non/existing")
+      ** (File.Error) could not make directory "non/existing": no such file or directory
   """
   @spec mkdir!(Path.t()) :: :ok
   def mkdir!(path) do
@@ -297,9 +317,17 @@ defmodule File do
       directories of `path`
     * `:enospc`  - there is no space left on the device
     * `:enotdir` - a component of `path` is not a directory
+    * `:eperm`   - missed required permissions
 
+  ## Examples
+
+      File.mkdir_p("non/existing/parents")
+      #=> :ok
+
+      File.mkdir_p("/usr/sbin/temp")
+      #=> {:error, :eperm}
   """
-  @spec mkdir_p(Path.t()) :: :ok | {:error, posix}
+  @spec mkdir_p(Path.t()) :: :ok | {:error, posix | :badarg}
   def mkdir_p(path) do
     do_mkdir_p(IO.chardata_to_string(path))
   end
@@ -309,24 +337,23 @@ defmodule File do
   end
 
   defp do_mkdir_p(path) do
-    if dir?(path) do
+    parent = Path.dirname(path)
+
+    if parent == path do
       :ok
     else
-      parent = Path.dirname(path)
+      case do_mkdir_p(parent) do
+        :ok ->
+          case :file.make_dir(path) do
+            {:error, :eexist} ->
+              if dir?(path), do: :ok, else: {:error, :enotdir}
 
-      if parent == path do
-        # Protect against infinite loop
-        {:error, :einval}
-      else
-        _ = do_mkdir_p(parent)
+            other ->
+              other
+          end
 
-        case :file.make_dir(path) do
-          {:error, :eexist} = error ->
-            if dir?(path), do: :ok, else: error
-
-          other ->
-            other
-        end
+        e ->
+          e
       end
     end
   end
@@ -334,6 +361,14 @@ defmodule File do
   @doc """
   Same as `mkdir_p/1`, but raises a `File.Error` exception in case of failure.
   Otherwise `:ok`.
+
+  ## Examples
+
+      File.mkdir_p!("non/existing/parents")
+      #=> :ok
+
+      File.mkdir_p!("/usr/sbin/temp")
+      ** (File.Error) could not make directory (with -p) "/usr/sbin/temp": not owner
   """
   @spec mkdir_p!(Path.t()) :: :ok
   def mkdir_p!(path) do
@@ -364,8 +399,16 @@ defmodule File do
     * `:enomem`  - there is not enough memory for the contents of the file
 
   You can use `:file.format_error/1` to get a descriptive string of the error.
+
+  ## Examples
+
+      File.read("hello.txt")
+      #=> {:ok, "world"}
+
+      File.read("non_existing.txt")
+      #=> {:error, :enoent}
   """
-  @spec read(Path.t()) :: {:ok, binary} | {:error, posix}
+  @spec read(Path.t()) :: {:ok, binary} | {:error, posix | :badarg | :terminated | :system_limit}
   def read(path) do
     :file.read_file(IO.chardata_to_string(path))
   end
@@ -373,6 +416,14 @@ defmodule File do
   @doc """
   Returns a binary with the contents of the given filename,
   or raises a `File.Error` exception if an error occurs.
+
+  ## Examples
+
+      File.read!("hello.txt")
+      #=> "world"
+
+      File.read!("non_existing.txt")
+      ** (File.Error) could not read file "non_existing.txt": no such file or directory
   """
   @spec read!(Path.t()) :: binary
   def read!(path) do
@@ -406,8 +457,16 @@ defmodule File do
 
   Note: Since file times are stored in POSIX time format on most operating systems,
   it is faster to retrieve file information with the `time: :posix` option.
+
+  ## Examples
+
+      File.stat("hello.txt")
+      #=> {:ok, %File.Stat{...}}
+
+      File.stat("non_existing.txt", time: :posix)
+      #=> {:error, :enoent}
   """
-  @spec stat(Path.t(), stat_options) :: {:ok, File.Stat.t()} | {:error, posix}
+  @spec stat(Path.t(), stat_options) :: {:ok, File.Stat.t()} | {:error, posix | :badarg}
   def stat(path, opts \\ []) do
     opts = Keyword.put_new(opts, :time, :universal)
 
@@ -423,6 +482,14 @@ defmodule File do
   @doc """
   Same as `stat/2` but returns the `File.Stat` directly,
   or raises a `File.Error` exception if an error is returned.
+
+  ## Examples
+
+      File.stat!("hello.txt")
+      #=> %File.Stat{...}
+
+      File.stat!("non_existing.txt", time: :posix)
+      ** (File.Error) could not read file stats "non_existing.txt": no such file or directory
   """
   @spec stat!(Path.t(), stat_options) :: File.Stat.t()
   def stat!(path, opts \\ []) do
@@ -459,8 +526,16 @@ defmodule File do
 
   Note: Since file times are stored in POSIX time format on most operating systems,
   it is faster to retrieve file information with the `time: :posix` option.
+
+  ## Examples
+
+      File.lstat("link_to_hello")
+      #=> {:ok, %File.Stat{type: :symlink, ...}}
+
+      File.lstat("non_existing.txt", time: :posix)
+      #=> {:error, :enoent}
   """
-  @spec lstat(Path.t(), stat_options) :: {:ok, File.Stat.t()} | {:error, posix}
+  @spec lstat(Path.t(), stat_options) :: {:ok, File.Stat.t()} | {:error, posix | :badarg}
   def lstat(path, opts \\ []) do
     opts = Keyword.put_new(opts, :time, :universal)
 
@@ -476,6 +551,14 @@ defmodule File do
   @doc """
   Same as `lstat/2` but returns the `File.Stat` struct directly,
   or raises a `File.Error` exception if an error is returned.
+
+  ## Examples
+
+      File.lstat!("link_to_hello")
+      #=> %File.Stat{type: :symlink, ...}
+
+      File.lstat!("non_existing.txt", time: :posix)
+      ** (File.Error) could not read file stats "non_existing.txt": no such file or directory
   """
   @spec lstat!(Path.t(), stat_options) :: File.Stat.t()
   def lstat!(path, opts \\ []) do
@@ -505,9 +588,16 @@ defmodule File do
     * `:enoent` - path does not exist
     * `:enotsup` - symbolic links are not supported on the current platform
 
+  ## Examples
+
+      File.read_link("link_to_hello")
+      #=> {:ok, "hello.txt"}
+
+      File.read_link("hello.txt")
+      #=> {:error, :einval}
   """
   @doc since: "1.5.0"
-  @spec read_link(Path.t()) :: {:ok, binary} | {:error, posix}
+  @spec read_link(Path.t()) :: {:ok, binary} | {:error, posix | :badarg}
   def read_link(path) do
     case path |> IO.chardata_to_string() |> :file.read_link() do
       {:ok, target} -> {:ok, IO.chardata_to_string(target)}
@@ -518,6 +608,14 @@ defmodule File do
   @doc """
   Same as `read_link/1` but returns the target directly,
   or raises a `File.Error` exception if an error is returned.
+
+  ## Examples
+
+      File.read_link!("link_to_hello")
+      #=> "hello.txt"
+
+      File.read_link!("hello.txt")
+      ** (File.Error) could not read link "hello.txt": invalid argument
   """
   @doc since: "1.5.0"
   @spec read_link!(Path.t()) :: binary
@@ -534,8 +632,16 @@ defmodule File do
   @doc """
   Writes the given `File.Stat` back to the file system at the given
   path. Returns `:ok` or `{:error, reason}`.
+
+  ## Examples
+
+      File.write_stat("hello.txt", new_stat)
+      #=> :ok
+
+      File.write_stat("non_existing.txt", new_stat)
+      #=> {:error, :enoent}
   """
-  @spec write_stat(Path.t(), File.Stat.t(), stat_options) :: :ok | {:error, posix}
+  @spec write_stat(Path.t(), File.Stat.t(), stat_options) :: :ok | {:error, posix | :badarg}
   def write_stat(path, stat, opts \\ []) do
     opts = Keyword.put_new(opts, :time, :universal)
     :file.write_file_info(IO.chardata_to_string(path), File.Stat.to_record(stat), opts)
@@ -544,6 +650,14 @@ defmodule File do
   @doc """
   Same as `write_stat/3` but raises a `File.Error` exception if it fails.
   Returns `:ok` otherwise.
+
+  ## Examples
+
+      File.write_stat!("hello.txt", new_stat)
+      #=> :ok
+
+      File.write_stat!("non_existing.txt", new_stat)
+      ** (File.Error) could not write file stats "non_existing.txt": no such file or directory
   """
   @spec write_stat!(Path.t(), File.Stat.t(), stat_options) :: :ok
   def write_stat!(path, stat, opts \\ []) do
@@ -584,7 +698,8 @@ defmodule File do
       #=> :ok
 
   """
-  @spec touch(Path.t(), erlang_time() | posix_time()) :: :ok | {:error, posix}
+  @spec touch(Path.t(), erlang_time() | posix_time()) ::
+          :ok | {:error, posix | :badarg | :terminated | :system_limit}
   def touch(path, time \\ System.os_time(:second))
 
   def touch(path, time) when is_tuple(time) do
@@ -638,9 +753,17 @@ defmodule File do
   Returns `:ok` if successful, `{:error, reason}` otherwise.
   If the operating system does not support hard links, returns
   `{:error, :enotsup}`.
+
+  ## Examples
+
+      File.ln("hello.txt", "hard_link_to_hello")
+      #=> :ok
+
+      File.ln("non_existing.txt", "link")
+      #=> {:error, :enoent}
   """
   @doc since: "1.5.0"
-  @spec ln(Path.t(), Path.t()) :: :ok | {:error, posix}
+  @spec ln(Path.t(), Path.t()) :: :ok | {:error, posix | :badarg}
   def ln(existing, new) do
     :file.make_link(IO.chardata_to_string(existing), IO.chardata_to_string(new))
   end
@@ -648,6 +771,14 @@ defmodule File do
   @doc """
   Same as `ln/2` but raises a `File.LinkError` exception if it fails.
   Returns `:ok` otherwise.
+
+  ## Examples
+
+      File.ln!("hello.txt", "hard_link_to_hello")
+      #=> :ok
+
+      File.ln!("non_existing.txt", "link")
+      ** (File.LinkError) could not create hard link from "non_existing.txt" to "link": no such file or directory
   """
   @doc since: "1.5.0"
   @spec ln!(Path.t(), Path.t()) :: :ok
@@ -671,9 +802,23 @@ defmodule File do
   Returns `:ok` if successful, `{:error, reason}` otherwise.
   If the operating system does not support symlinks, returns
   `{:error, :enotsup}`.
+
+  Creates a symlink even if the `existing` target actually doesn't exist
+
+  ## Examples
+
+      File.ln_s("hello.txt", "link_to_hello")
+      #=> :ok
+
+      File.ln_s("non_existing.txt", "link")
+      #=> :ok
+
+      # Returns error if `new` file exists
+      File.ln_s("non_existing.txt", "existed_link")
+      #=> {:error, :eexist}
   """
   @doc since: "1.5.0"
-  @spec ln_s(Path.t(), Path.t()) :: :ok | {:error, posix}
+  @spec ln_s(Path.t(), Path.t()) :: :ok | {:error, posix | :badarg}
   def ln_s(existing, new) do
     :file.make_symlink(IO.chardata_to_string(existing), IO.chardata_to_string(new))
   end
@@ -681,6 +826,15 @@ defmodule File do
   @doc """
   Same as `ln_s/2` but raises a `File.LinkError` exception if it fails.
   Returns `:ok` otherwise.
+
+  ## Examples
+
+      File.ln_s!("hello.txt", "link_to_hello")
+      #=> :ok
+
+      # Raises if `new` file exists
+      File.ln_s!("non_existing.txt", "existed_link")
+      ** (File.LinkError) could not create symlink from "non_existing.txt" to "existed_link": file already exists
   """
   @spec ln_s!(Path.t(), Path.t()) :: :ok
   def ln_s!(existing, new) do
@@ -718,9 +872,17 @@ defmodule File do
 
   Typical error reasons are the same as in `open/2`,
   `read/1` and `write/3`.
+
+  ## Examples
+
+      File.copy("hello.txt", "hello_copy.txt")
+      #=> {:ok, 6}
+
+      File.copy("non_existing.txt", "copy.txt")
+      #=> {:error, :enoent}
   """
   @spec copy(Path.t() | io_device, Path.t() | io_device, pos_integer | :infinity) ::
-          {:ok, non_neg_integer} | {:error, posix}
+          {:ok, non_neg_integer} | {:error, posix | :badarg | :terminated}
   def copy(source, destination, bytes_count \\ :infinity) do
     source = normalize_path_or_io_device(source)
     destination = normalize_path_or_io_device(destination)
@@ -731,6 +893,14 @@ defmodule File do
   @doc """
   The same as `copy/3` but raises a `File.CopyError` exception if it fails.
   Returns the `bytes_copied` otherwise.
+
+  ## Examples
+
+      File.copy!("hello.txt", "hello_copy.txt")
+      #=> 6
+
+      File.copy!("non_existing.txt", "copy.txt")
+      ** (File.CopyError) could not copy from "non_existing.txt" to "copy.txt": no such file or directory
   """
   @spec copy!(Path.t() | io_device, Path.t() | io_device, pos_integer | :infinity) ::
           non_neg_integer
@@ -758,19 +928,23 @@ defmodule File do
 
   Note: The command `mv` in Unix-like systems behaves differently depending on
   whether `source` is a file and the `destination` is an existing directory.
-  We have chosen to explicitly disallow this behaviour.
+  We have chosen to explicitly disallow this behavior.
 
   ## Examples
 
       # Rename file "a.txt" to "b.txt"
       File.rename("a.txt", "b.txt")
+      #=> :ok
 
       # Rename directory "samples" to "tmp"
       File.rename("samples", "tmp")
+      #=> :ok
 
+      File.rename("non_existing.txt", "existing.txt")
+      #=> {:error, :enoent}
   """
   @doc since: "1.1.0"
-  @spec rename(Path.t(), Path.t()) :: :ok | {:error, posix}
+  @spec rename(Path.t(), Path.t()) :: :ok | {:error, posix | :badarg}
   def rename(source, destination) do
     source = IO.chardata_to_string(source)
     destination = IO.chardata_to_string(destination)
@@ -780,6 +954,14 @@ defmodule File do
   @doc """
   The same as `rename/2` but raises a `File.RenameError` exception if it fails.
   Returns `:ok` otherwise.
+
+  ## Examples
+
+      File.rename!("samples", "tmp")
+      #=> :ok
+
+      File.rename!("non_existing.txt", "existing.txt")
+      ** (File.RenameError) could not rename from "non_existing.txt" to "existing.txt": no such file or directory
   """
   @doc since: "1.9.0"
   @spec rename!(Path.t(), Path.t()) :: :ok
@@ -797,7 +979,7 @@ defmodule File do
     end
   end
 
-  @doc """
+  @doc ~S"""
   Copies the contents of `source_file` to `destination_file` preserving its modes.
 
   `source_file` must be a file or a symbolic link to one. `destination_file` must
@@ -822,14 +1004,35 @@ defmodule File do
       The function receives arguments for `source_file` and `destination_file`. It should
       return `true` if the existing file should be overwritten, `false` if otherwise.
       The default callback returns `true`. On earlier versions, this callback could be
-      given as third argument, but such behaviour is now deprecated.
+      given as third argument, but such behavior is now deprecated.
 
+  ## Examples
+
+      File.cp("hello.txt", "hello_copy.txt")
+      #=> :ok
+
+      File.cp("hello.txt", "hello_copy.txt", on_conflict: fn source, destination ->
+        IO.gets("Overwriting #{destination} by #{source}. Type y to confirm. ") == "y\n"
+      end)
+      #=> :ok
+
+      File.cp("non_existing.txt", "copy.txt")
+      #=> {:error, :enoent}
   """
-  @spec cp(Path.t(), Path.t(), on_conflict: on_conflict_callback) :: :ok | {:error, posix}
+  @spec cp(Path.t(), Path.t(), on_conflict: on_conflict_callback) ::
+          :ok | {:error, posix | :badarg | :terminated}
   def cp(source_file, destination_file, options \\ [])
 
   # TODO: Deprecate me on Elixir v1.19
   def cp(source_file, destination_file, callback) when is_function(callback, 2) do
+    IO.warn_once(
+      {__MODULE__, :cp},
+      fn ->
+        "passing a callback to File.cp/3 is deprecated, pass it as a on_conflict: callback option instead"
+      end,
+      3
+    )
+
     cp(source_file, destination_file, on_conflict: callback)
   end
 
@@ -850,9 +1053,22 @@ defmodule File do
     Path.expand(p1) !== Path.expand(p2)
   end
 
-  @doc """
+  @doc ~S"""
   The same as `cp/3`, but raises a `File.CopyError` exception if it fails.
   Returns `:ok` otherwise.
+
+  ## Examples
+
+      File.cp!("hello.txt", "hello_copy.txt")
+      #=> :ok
+
+      File.cp!("hello.txt", "hello_copy.txt", on_conflict: fn source, destination ->
+        IO.gets("Overwriting #{destination} by #{source}. Type y to confirm. ") == "y\n"
+      end)
+      #=> :ok
+
+      File.cp!("non_existing.txt", "copy.txt")
+      ** (File.CopyError) could not copy from "non_existing.txt" to "copy.txt": no such file or directory
   """
   @spec cp!(Path.t(), Path.t(), on_conflict: on_conflict_callback) :: :ok
   def cp!(source_file, destination_file, options \\ []) do
@@ -895,7 +1111,7 @@ defmodule File do
 
   Note: The command `cp` in Unix-like systems behaves differently depending on
   whether `destination` is an existing directory or not. We have chosen to
-  explicitly disallow this behaviour. If `source` is a `file` and `destination`
+  explicitly disallow this behavior. If `source` is a `file` and `destination`
   is a directory, `{:error, :eisdir}` will be returned.
 
   ## Options
@@ -904,7 +1120,7 @@ defmodule File do
       The function receives arguments for `source` and `destination`. It should return
       `true` if the existing file should be overwritten, `false` if otherwise. The default
       callback returns `true`. On earlier versions, this callback could be given as third
-      argument, but such behaviour is now deprecated.
+      argument, but such behavior is now deprecated.
 
     * `:dereference_symlinks` - (since v1.14.0) By default, this function will copy symlinks
       by creating symlinks that point to the same location. This option forces symlinks to be
@@ -915,26 +1131,39 @@ defmodule File do
 
       # Copies file "a.txt" to "b.txt"
       File.cp_r("a.txt", "b.txt")
+      #=> {:ok, ["b.txt"]}
 
       # Copies all files in "samples" to "tmp"
       File.cp_r("samples", "tmp")
+      #=> {:ok, ["z.txt", "y.txt", "x.txt]}
 
       # Same as before, but asks the user how to proceed in case of conflicts
       File.cp_r("samples", "tmp", on_conflict: fn source, destination ->
         IO.gets("Overwriting #{destination} by #{source}. Type y to confirm. ") == "y\n"
       end)
+      #=> {:ok, ["z.txt", "y.txt", "x.txt]}
 
+      File.cp_r("non_existing.txt", "copy.txt")
+      #=> {:error, :enoent}
   """
   @spec cp_r(Path.t(), Path.t(),
           on_conflict: on_conflict_callback,
           dereference_symlinks: boolean()
         ) ::
-          {:ok, [binary]} | {:error, posix, binary}
+          {:ok, [binary]} | {:error, posix | :badarg | :terminated, binary}
 
   def cp_r(source, destination, options \\ [])
 
   # TODO: Deprecate me on Elixir v1.19
   def cp_r(source, destination, callback) when is_function(callback, 2) do
+    IO.warn_once(
+      {__MODULE__, :cp_r},
+      fn ->
+        "passing a callback to File.cp_r/3 is deprecated, pass it as a on_conflict: callback option instead"
+      end,
+      3
+    )
+
     cp_r(source, destination, on_conflict: callback)
   end
 
@@ -961,6 +1190,17 @@ defmodule File do
   @doc """
   The same as `cp_r/3`, but raises a `File.CopyError` exception if it fails.
   Returns the list of copied files otherwise.
+
+  ## Examples
+
+      File.cp_r!("a.txt", "b.txt")
+      #=> ["b.txt"]
+
+      File.cp_r!("samples", "tmp")
+      #=> ["z.txt", "y.txt", "x.txt]
+
+      File.cp_r!("non_existing.txt", "copy.txt")
+      ** (File.CopyError) could not copy recursively from "non_existing.txt" to "copy.txt". non_existing.txt: no such file or directory
   """
   @spec cp_r!(Path.t(), Path.t(),
           on_conflict: on_conflict_callback,
@@ -1121,9 +1361,19 @@ defmodule File do
       the parent directories
     * `:eisdir`  - the named file is a directory
 
-  Check `File.open/2` for other available options.
+  Check `File.open/2` for the list of available `modes`.
+
+  ## Examples
+
+      File.write("hello.txt", "world!")
+      #=> :ok
+
+      File.write("temp", "world!")
+      #=> {:error, :eisdir}
+
   """
-  @spec write(Path.t(), iodata, [mode]) :: :ok | {:error, posix}
+  @spec write(Path.t(), iodata, [mode]) ::
+          :ok | {:error, posix | :badarg | :terminated | :system_limit}
   def write(path, content, modes \\ []) do
     modes = normalize_modes(modes, false)
     :file.write_file(IO.chardata_to_string(path), content, modes)
@@ -1132,6 +1382,14 @@ defmodule File do
   @doc """
   Same as `write/3` but raises a `File.Error` exception if it fails.
   Returns `:ok` otherwise.
+
+  ## Examples
+
+      File.write!("hello.txt", "world!")
+      #=> :ok
+
+      File.write!("temp", "world!")
+      ** (File.Error) could not write to file "temp": illegal operation on a directory
   """
   @spec write!(Path.t(), iodata, [mode]) :: :ok
   def write!(path, content, modes \\ []) do
@@ -1170,9 +1428,8 @@ defmodule File do
 
       File.rm("tmp_dir/")
       #=> {:error, :eperm}
-
   """
-  @spec rm(Path.t()) :: :ok | {:error, posix}
+  @spec rm(Path.t()) :: :ok | {:error, posix | :badarg}
   def rm(path) do
     path = IO.chardata_to_string(path)
 
@@ -1210,6 +1467,14 @@ defmodule File do
   @doc """
   Same as `rm/1`, but raises a `File.Error` exception in case of failure.
   Otherwise `:ok`.
+
+  ## Examples
+
+      File.rm!("file.txt")
+      #=> :ok
+
+      File.rm!("non_existing/")
+      ** (File.Error) could not remove file "non_existing/": no such file or directory
   """
   @spec rm!(Path.t()) :: :ok
   def rm!(path) do
@@ -1238,9 +1503,8 @@ defmodule File do
 
       File.rmdir("file.txt")
       #=> {:error, :enotdir}
-
   """
-  @spec rmdir(Path.t()) :: :ok | {:error, posix}
+  @spec rmdir(Path.t()) :: :ok | {:error, posix | :badarg}
   def rmdir(path) do
     :file.del_dir(IO.chardata_to_string(path))
   end
@@ -1248,8 +1512,19 @@ defmodule File do
   @doc """
   Same as `rmdir/1`, but raises a `File.Error` exception in case of failure.
   Otherwise `:ok`.
+
+  ## Examples
+
+      File.rmdir!("tmp_dir")
+      #=> :ok
+
+      File.rmdir!("non_empty_dir")
+      ** (File.Error) could not remove directory "non_empty_dir": directory is not empty
+
+      File.rmdir!("file.txt")
+      ** (File.Error) could not remove directory "file.txt": not a directory
   """
-  @spec rmdir!(Path.t()) :: :ok | {:error, posix}
+  @spec rmdir!(Path.t()) :: :ok
   def rmdir!(path) do
     case rmdir(path) do
       :ok ->
@@ -1280,8 +1555,10 @@ defmodule File do
       File.rm_rf("unknown")
       #=> {:ok, []}
 
+      File.rm_rf("/tmp")
+      #=> {:error, :eperm, "/tmp"}
   """
-  @spec rm_rf(Path.t()) :: {:ok, [binary]} | {:error, posix, binary}
+  @spec rm_rf(Path.t()) :: {:ok, [binary]} | {:error, posix | :badarg, binary}
   def rm_rf(path) do
     {major, _} = :os.type()
 
@@ -1363,6 +1640,11 @@ defmodule File do
       {:ok, _} ->
         {:ok, :regular}
 
+      {:error, :eio} when major == :win32 ->
+        # unix domain socket returns `{:error, :eio}`
+        # on other platforms the result is `{:ok, :regular}`
+        {:ok, :regular}
+
       {:error, reason} ->
         {:error, reason}
     end
@@ -1370,7 +1652,18 @@ defmodule File do
 
   @doc """
   Same as `rm_rf/1` but raises a `File.Error` exception in case of failures,
-  otherwise the list of files or directories removed.
+  otherwise returns the list of files or directories removed.
+
+  ## Examples
+
+      File.rm_rf!("samples")
+      #=> ["samples", "samples/1.txt"]
+
+      File.rm_rf!("unknown")
+      #=> []
+
+      File.rm_rf!("/tmp")
+      ** (File.Error) could not remove files and directories recursively from "/tmp": not owner
   """
   @spec rm_rf!(Path.t()) :: [binary]
   def rm_rf!(path) do
@@ -1389,13 +1682,6 @@ defmodule File do
   @doc ~S"""
   Opens the given `path`.
 
-  In order to write and read files, one must use the functions
-  in the `IO` module. By default, a file is opened in `:binary` mode,
-  which requires the functions `IO.binread/2` and `IO.binwrite/2`
-  to interact with the file. A developer may pass `:utf8` as an
-  option when opening the file and then all other functions from
-  `IO` are available, since they work directly with Unicode data.
-
   `modes_or_function` can either be a list of modes or a function. If it's a
   list, it's considered to be a list of modes (that are documented below). If
   it's a function, then it's equivalent to calling `open(path, [],
@@ -1404,8 +1690,8 @@ defmodule File do
 
   The allowed modes:
 
-    * `:binary` - opens the file in binary mode, disabling special handling of Unicode sequences
-      (default mode).
+    * `:binary` - opens the file in binary mode, disabling special handling of
+      Unicode sequences (default mode).
 
     * `:read` - the file, which must exist, is opened for reading.
 
@@ -1446,19 +1732,33 @@ defmodule File do
 
   This function returns:
 
-    * `{:ok, io_device}` - the file has been opened in the requested mode.
+    * `{:ok, io_device | file_descriptor}` - the file has been opened in
+      the requested mode. We explore the differences between these two results
+      in the following section
 
-      `io_device` is actually the PID of the process which handles the file.
-      This process monitors the process that originally opened the file (the
-      owner process). If the owner process terminates, the file is closed and
-      the process itself terminates too. If any process to which the `io_device`
-      is linked terminates, the file will be closed and the process itself will
-      be terminated.
+    * `{:error, reason}` - the file could not be opened due to `reason`.
 
-      An `io_device` returned from this call can be used as an argument to the
-      `IO` module functions.
+  ## IO devices
 
-    * `{:error, reason}` - the file could not be opened.
+  By default, this function returns an IO device. An `io_device` is
+  a process which handles the file and you can interact with it using
+  the functions in the `IO` module. By default, a file is opened in
+  `:binary` mode, which requires the functions `IO.binread/2` and
+  `IO.binwrite/2` to interact with the file. A developer may pass `:utf8`
+  as a mode when opening the file and then all other functions from
+  `IO` are available, since they work directly with Unicode data.
+
+  Given the IO device is a file, if the owner process terminates,
+  the file is closed and the process itself terminates too. If any
+  process to which the `io_device` is linked terminates, the file
+  will be closed and the process itself will be terminated.
+
+  ## File descriptors
+
+  When the `:raw` or `:ram` modes are given, this function returns
+  a low-level file descriptors. This avoids creating a process but
+  requires using the functions in the [`:file`](`:file`) module to
+  interact with it.
 
   ## Examples
 
@@ -1467,8 +1767,11 @@ defmodule File do
       File.close(file)
 
   """
-  @spec open(Path.t(), [mode | :ram]) :: {:ok, io_device} | {:error, posix}
-  @spec open(Path.t(), (io_device -> res)) :: {:ok, res} | {:error, posix} when res: var
+  @spec open(Path.t(), [mode | :ram]) ::
+          {:ok, io_device | file_descriptor} | {:error, posix | :badarg | :system_limit}
+  @spec open(Path.t(), (io_device | file_descriptor -> res)) ::
+          {:ok, res} | {:error, posix | :badarg | :system_limit}
+        when res: var
   def open(path, modes_or_function \\ [])
 
   def open(path, modes) when is_list(modes) do
@@ -1494,15 +1797,17 @@ defmodule File do
   is given. For this reason, we do not recommend passing
   `:delayed_write` to this function.
 
+  See `open/2` for the list of available `modes`.
+
   ## Examples
 
       File.open("file.txt", [:read, :write], fn file ->
         IO.read(file, :line)
       end)
-
-  See `open/2` for the list of available `modes`.
+      #=> {:ok, "file content"}
   """
-  @spec open(Path.t(), [mode | :ram], (io_device -> res)) :: {:ok, res} | {:error, posix}
+  @spec open(Path.t(), [mode | :ram], (io_device | file_descriptor -> res)) ::
+          {:ok, res} | {:error, posix | :badarg | :system_limit}
         when res: var
   def open(path, modes, function) when is_list(modes) and is_function(function, 1) do
     case open(path, modes) do
@@ -1523,9 +1828,16 @@ defmodule File do
   could not be opened. Returns the IO device otherwise.
 
   See `open/2` for the list of available modes.
+
+  ## Examples
+
+      File.open!("file.txt", fn file ->
+        IO.read(file, :line)
+      end)
+      #=> "file content"
   """
-  @spec open!(Path.t(), [mode | :ram]) :: io_device
-  @spec open!(Path.t(), (io_device -> res)) :: res when res: var
+  @spec open!(Path.t(), [mode | :ram]) :: io_device | file_descriptor
+  @spec open!(Path.t(), (io_device | file_descriptor -> res)) :: res when res: var
   def open!(path, modes_or_function \\ []) do
     case open(path, modes_or_function) do
       {:ok, io_device_or_function_result} ->
@@ -1543,8 +1855,15 @@ defmodule File do
   If it succeeds opening the file, it returns the `function` result on the IO device.
 
   See `open/2` for the list of available `modes`.
+
+  ## Examples
+
+      File.open!("file.txt", [:read, :write], fn file ->
+        IO.read(file, :line)
+      end)
+      #=> "file content"
   """
-  @spec open!(Path.t(), [mode | :ram], (io_device -> res)) :: res when res: var
+  @spec open!(Path.t(), [mode | :ram], (io_device | file_descriptor -> res)) :: res when res: var
   def open!(path, modes, function) do
     case open(path, modes, function) do
       {:ok, function_result} ->
@@ -1562,8 +1881,17 @@ defmodule File do
   if read permissions do not exist for the parent directories of the
   current directory. For this reason, returns `{:ok, cwd}` in case
   of success, `{:error, reason}` otherwise.
+
+  ## Examples
+
+      File.cwd()
+      #=> {:ok, "/Users/user/elixir/elixir_lang"}
+
+      # Missing read permission for one of the parents of the current directory
+      File.cwd()
+      #=> {:error, :eacces}
   """
-  @spec cwd() :: {:ok, binary} | {:error, posix}
+  @spec cwd() :: {:ok, binary} | {:error, posix | :badarg}
   def cwd() do
     case :file.get_cwd() do
       {:ok, base} -> {:ok, IO.chardata_to_string(fix_drive_letter(base))}
@@ -1582,6 +1910,11 @@ defmodule File do
 
   @doc """
   The same as `cwd/0`, but raises a `File.Error` exception if it fails.
+
+  ## Examples
+
+      File.cwd!()
+      #=> "/Users/user/elixir/elixir_lang"
   """
   @spec cwd!() :: binary
   def cwd!() do
@@ -1604,14 +1937,30 @@ defmodule File do
   of `System.cmd/3` and `Port.open/2`.
 
   Returns `:ok` if successful, `{:error, reason}` otherwise.
+
+  ## Examples
+
+      File.cd("bin")
+      #=> :ok
+
+      File.cd("non_existing_dir")
+      #=> {:error, :enoent}
   """
-  @spec cd(Path.t()) :: :ok | {:error, posix}
+  @spec cd(Path.t()) :: :ok | {:error, posix | :badarg | :no_translation}
   def cd(path) do
     :file.set_cwd(IO.chardata_to_string(path))
   end
 
   @doc """
   The same as `cd/1`, but raises a `File.Error` exception if it fails.
+
+  ## Examples
+
+      File.cd!("bin")
+      #=> :ok
+
+      File.cd!("non_existing_dir")
+      ** (File.Error) could not set current working directory to "non_existing_dir": no such file or directory
   """
   @spec cd!(Path.t()) :: :ok
   def cd!(path) do
@@ -1640,6 +1989,14 @@ defmodule File do
 
   Raises an error if retrieving or changing the current
   directory fails.
+
+  ## Examples
+
+      File.cd!("bin", fn -> do_something() end)
+      #=> :result_of_do_something
+
+      File.cd!("non_existing_dir", fn -> do_something() end)
+      ** (File.Error) could not set current working directory to "non_existing_dir": no such file or directory
   """
   @spec cd!(Path.t(), (-> res)) :: res when res: var
   def cd!(path, function) do
@@ -1663,8 +2020,16 @@ defmodule File do
 
   Returns `{:ok, files}` in case of success,
   `{:error, reason}` otherwise.
+
+  ## Examples
+
+      File.ls("bin")
+      #=> {:ok, ["iex", "elixir"]}
+
+      File.ls("non_existing_dir")
+      #=> {:error, :enoent}
   """
-  @spec ls(Path.t()) :: {:ok, [binary]} | {:error, posix}
+  @spec ls(Path.t()) :: {:ok, [binary]} | {:error, posix | :badarg | {:no_translation, binary}}
   def ls(path \\ ".") do
     case :file.list_dir(IO.chardata_to_string(path)) do
       {:ok, file_list} -> {:ok, Enum.map(file_list, &IO.chardata_to_string/1)}
@@ -1674,6 +2039,14 @@ defmodule File do
 
   @doc """
   The same as `ls/1` but raises a `File.Error` exception in case of an error.
+
+  ## Examples
+
+      File.ls!("bin")
+      #=> ["iex", "elixir"]
+
+      File.ls!("non_existing_dir")
+      ** (File.Error) could not list directory "non_existing_dir": no such file or directory
   """
   @spec ls!(Path.t()) :: [binary]
   def ls!(path \\ ".") do
@@ -1696,6 +2069,15 @@ defmodule File do
   Note that if the option `:delayed_write` was used when opening the file,
   `close/1` might return an old write error and not even try to close the file.
   See `open/2` for more information.
+
+  ## Examples
+
+      {:ok, file} = File.open("hello.txt")
+      File.close(file)
+      #=> :ok
+
+      File.close(:not_an_io_device)
+      #=> {:error, :badarg}
   """
   @spec close(io_device) :: :ok | {:error, posix | :badarg | :terminated}
   def close(io_device) do
@@ -1761,15 +2143,17 @@ defmodule File do
   that is skipped whenever enumerating the stream (if both `:read_offset`
   and `:trim_bom` are given, the offset is skipped after the BOM).
 
+  See `Stream.run/1` for an example of streaming into a file.
+
   ## Examples
 
       # Read a utf8 text file which may include BOM
       File.stream!("./test/test.txt", [:trim_bom, encoding: :utf8])
+      #=> %File.Stream{path: "./test/test.txt", ...}
 
       # Read in 2048 byte chunks rather than lines
       File.stream!("./test/test.data", 2048)
-
-  See `Stream.run/1` for an example of streaming into a file.
+      #=> %File.Stream{path: "./test/test.data", ...}
   """
   @spec stream!(Path.t(), :line | pos_integer, [stream_mode]) :: File.Stream.t()
   def stream!(path, line_or_bytes, modes)
@@ -1809,8 +2193,16 @@ defmodule File do
   write, read and execute permission to the owner
   and both read and execute permission to group
   and others.
+
+  ## Examples
+
+      File.chmod("hello.txt", 0o755)
+      #=> :ok
+
+      File.chmod("non_existing.txt", 0o755)
+      #=> {:error, :enoent}
   """
-  @spec chmod(Path.t(), non_neg_integer) :: :ok | {:error, posix}
+  @spec chmod(Path.t(), non_neg_integer) :: :ok | {:error, posix | :badarg}
   def chmod(path, mode) do
     :file.change_mode(IO.chardata_to_string(path), mode)
   end
@@ -1818,6 +2210,14 @@ defmodule File do
   @doc """
   Same as `chmod/2`, but raises a `File.Error` exception in case of failure.
   Otherwise `:ok`.
+
+  ## Examples
+
+      File.chmod!("hello.txt", 0o755)
+      #=> :ok
+
+      File.chmod!("non_existing.txt", 0o755)
+      ** (File.Error) could not change mode for "non_existing.txt": no such file or directory
   """
   @spec chmod!(Path.t(), non_neg_integer) :: :ok
   def chmod!(path, mode) do
@@ -1837,8 +2237,16 @@ defmodule File do
   Changes the group given by the group ID `gid`
   for a given `file`. Returns `:ok` on success, or
   `{:error, reason}` on failure.
+
+  ## Examples
+
+      File.chgrp("hello.txt", 10)
+      #=> :ok
+
+      File.chgrp("non_existing.txt", 10)
+      #=> {:error, :enoent}
   """
-  @spec chgrp(Path.t(), non_neg_integer) :: :ok | {:error, posix}
+  @spec chgrp(Path.t(), non_neg_integer) :: :ok | {:error, posix | :badarg}
   def chgrp(path, gid) do
     :file.change_group(IO.chardata_to_string(path), gid)
   end
@@ -1846,6 +2254,14 @@ defmodule File do
   @doc """
   Same as `chgrp/2`, but raises a `File.Error` exception in case of failure.
   Otherwise `:ok`.
+
+  ## Examples
+
+    File.chgrp!("hello.txt", 10)
+    #=> :ok
+
+    File.chgrp!("non_existing.txt", 10)
+    ** (File.Error) could not change group for "non_existing.txt": no such file or directory
   """
   @spec chgrp!(Path.t(), non_neg_integer) :: :ok
   def chgrp!(path, gid) do
@@ -1865,8 +2281,16 @@ defmodule File do
   Changes the owner given by the user ID `uid`
   for a given `file`. Returns `:ok` on success,
   or `{:error, reason}` on failure.
+
+  ## Examples
+
+      File.chown("hello.txt", 15)
+      #=> :ok
+
+      File.chown("secret.txt", 15)
+      #=> {:error, :eperm}
   """
-  @spec chown(Path.t(), non_neg_integer) :: :ok | {:error, posix}
+  @spec chown(Path.t(), non_neg_integer) :: :ok | {:error, posix | :badarg}
   def chown(path, uid) do
     :file.change_owner(IO.chardata_to_string(path), uid)
   end
@@ -1874,6 +2298,14 @@ defmodule File do
   @doc """
   Same as `chown/2`, but raises a `File.Error` exception in case of failure.
   Otherwise `:ok`.
+
+  ## Examples
+
+      File.chown!("hello.txt", 15)
+      #=> :ok
+
+      File.chown!("secret.txt", 15)
+      ** (File.Error) could not change owner for "secret.txt": not owner
   """
   @spec chown!(Path.t(), non_neg_integer) :: :ok
   def chown!(path, uid) do
