@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 import Kernel, except: [round: 1]
 
 defmodule Float do
@@ -139,6 +143,8 @@ defmodule Float do
       iex> Float.parse("56.5xyz")
       {56.5, "xyz"}
 
+      iex> Float.parse(".12")
+      :error
       iex> Float.parse("pi")
       :error
       iex> Float.parse("1.7976931348623159e+308")
@@ -162,40 +168,50 @@ defmodule Float do
   end
 
   defp parse_unsigned(<<digit, rest::binary>>) when digit in ?0..?9,
-    do: parse_unsigned(rest, false, false, <<digit>>)
+    do: parse_unsigned(rest, false, false, [digit])
 
   defp parse_unsigned(binary) when is_binary(binary), do: :error
 
   defp parse_unsigned(<<digit, rest::binary>>, dot?, e?, acc) when digit in ?0..?9,
-    do: parse_unsigned(rest, dot?, e?, <<acc::binary, digit>>)
+    do: parse_unsigned(rest, dot?, e?, [digit | acc])
 
   defp parse_unsigned(<<?., digit, rest::binary>>, false, false, acc) when digit in ?0..?9,
-    do: parse_unsigned(rest, true, false, <<acc::binary, ?., digit>>)
+    do: parse_unsigned(rest, true, false, [digit, ?. | acc])
 
   defp parse_unsigned(<<exp_marker, digit, rest::binary>>, dot?, false, acc)
        when exp_marker in ~c"eE" and digit in ?0..?9,
-       do: parse_unsigned(rest, true, true, <<add_dot(acc, dot?)::binary, ?e, digit>>)
+       do: parse_unsigned(rest, true, true, [digit, ?e | add_dot(acc, dot?)])
 
   defp parse_unsigned(<<exp_marker, sign, digit, rest::binary>>, dot?, false, acc)
        when exp_marker in ~c"eE" and sign in ~c"-+" and digit in ?0..?9,
-       do: parse_unsigned(rest, true, true, <<add_dot(acc, dot?)::binary, ?e, sign, digit>>)
+       do: parse_unsigned(rest, true, true, [digit, sign, ?e | add_dot(acc, dot?)])
 
   # When floats are expressed in scientific notation, :erlang.binary_to_float/1 can raise an
   # ArgumentError if the e exponent is too big. For example, "1.0e400". Because of this, we
   # rescue the ArgumentError here and return an error.
   defp parse_unsigned(rest, dot?, true = _e?, acc) do
-    :erlang.binary_to_float(add_dot(acc, dot?))
+    acc
+    |> add_dot(dot?)
+    |> :lists.reverse()
+    |> :erlang.list_to_float()
   rescue
     ArgumentError -> :error
   else
     float -> {float, rest}
   end
 
-  defp parse_unsigned(rest, dot?, false = _e?, acc),
-    do: {:erlang.binary_to_float(add_dot(acc, dot?)), rest}
+  defp parse_unsigned(rest, dot?, false = _e?, acc) do
+    float =
+      acc
+      |> add_dot(dot?)
+      |> :lists.reverse()
+      |> :erlang.list_to_float()
+
+    {float, rest}
+  end
 
   defp add_dot(acc, true), do: acc
-  defp add_dot(acc, false), do: acc <> ".0"
+  defp add_dot(acc, false), do: [?0, ?. | acc]
 
   @doc """
   Rounds a float to the largest float less than or equal to `number`.
@@ -575,18 +591,14 @@ defmodule Float do
   Returns a charlist which corresponds to the shortest text representation
   of the given float.
 
-  The underlying algorithm changes depending on the Erlang/OTP version:
-
-    * For OTP >= 24, it uses the algorithm presented in "Ryū: fast
-      float-to-string conversion" in Proceedings of the SIGPLAN '2018
-      Conference on Programming Language Design and Implementation.
-
-    * For OTP < 24, it uses the algorithm presented in "Printing Floating-Point
-      Numbers Quickly and Accurately" in Proceedings of the SIGPLAN '1996
-      Conference on Programming Language Design and Implementation.
+  It uses the algorithm presented in "Ryū: fast float-to-string conversion"
+  in Proceedings of the SIGPLAN '2018 Conference on Programming Language
+  Design and Implementation.
 
   For a configurable representation, use `:erlang.float_to_list/2`.
 
+  Inlined by the compiler.
+
   ## Examples
 
       iex> Float.to_charlist(7.0)
@@ -594,7 +606,7 @@ defmodule Float do
 
   """
   @spec to_charlist(float) :: charlist
-  def to_charlist(float) when is_float(float) do
+  def to_charlist(float) do
     :erlang.float_to_list(float, [:short])
   end
 
@@ -614,6 +626,8 @@ defmodule Float do
 
   For a configurable representation, use `:erlang.float_to_binary/2`.
 
+  Inlined by the compiler.
+
   ## Examples
 
       iex> Float.to_string(7.0)
@@ -621,7 +635,7 @@ defmodule Float do
 
   """
   @spec to_string(float) :: String.t()
-  def to_string(float) when is_float(float) do
+  def to_string(float) do
     :erlang.float_to_binary(float, [:short])
   end
 
