@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule IO do
   @moduledoc ~S"""
   Functions handling input/output (IO).
@@ -12,6 +16,8 @@ defmodule IO do
   (as shown in typespecs). For more information on chardata, see the
   "IO data" section below.
 
+  The functions of this module use UNIX-style naming where possible.
+
   ## IO devices
 
   An IO device may be an atom or a PID. In case it is an atom,
@@ -122,14 +128,39 @@ defmodule IO do
   @type nodata :: {:error, term} | :eof
   @type chardata :: String.t() | maybe_improper_list(char | chardata, String.t() | [])
 
+  @type inspect_opts :: [Inspect.Opts.new_opt() | {:label, term}]
+
+  @typedoc """
+  Stacktrace information as keyword options for `warn/2`.
+
+  At least `:file` is required. Other options are optional and used
+  to provide more precise location information.
+  """
+  @type warn_stacktrace_opts :: [
+          file: String.t(),
+          line: pos_integer(),
+          column: pos_integer(),
+          module: module(),
+          function: {atom(), arity()}
+        ]
+
   defguardp is_device(term) when is_atom(term) or is_pid(term)
   defguardp is_iodata(data) when is_list(data) or is_binary(data)
 
-  @doc """
+  @doc ~S"""
   Reads from the IO `device`.
 
-  The `device` is iterated by the given number of characters, line by line if
-  `:line` is given, or until `:eof`.
+  The `device` is iterated as specified by the `line_or_chars` argument:
+
+    * if `line_or_chars` is an integer, it represents a number of bytes. The device is
+      iterated by that number of bytes. This should be the preferred mode for reading
+      non-textual inputs.
+
+    * if `line_or_chars` is `:line`, the device is iterated line by line.
+      CRLF newlines  ("\r\n") are automatically normalized to "\n".
+
+    * if `line_or_chars` is `:eof` (since v1.13), the device is iterated until `:eof`.
+      If the device is already at the end, it returns `:eof` itself.
 
   It returns:
 
@@ -145,8 +176,10 @@ defmodule IO do
   @spec read(device, :eof | :line | non_neg_integer) :: chardata | nodata
   def read(device \\ :stdio, line_or_chars)
 
-  # TODO: Deprecate me on v1.17
+  # TODO: Remove me on v2.0
   def read(device, :all) do
+    IO.warn("IO.read(device, :all) is deprecated, use IO.read(device, :eof) instead")
+
     with :eof <- read(device, :eof) do
       with [_ | _] = opts <- :io.getopts(device),
            false <- Keyword.get(opts, :binary, true) do
@@ -169,20 +202,20 @@ defmodule IO do
     :io.get_chars(map_dev(device), ~c"", count)
   end
 
-  @doc """
+  @doc ~S"""
   Reads from the IO `device`. The operation is Unicode unsafe.
 
   The `device` is iterated as specified by the `line_or_chars` argument:
 
     * if `line_or_chars` is an integer, it represents a number of bytes. The device is
-      iterated by that number of bytes.
+      iterated by that number of bytes. This should be the preferred mode for reading
+      non-textual inputs.
 
     * if `line_or_chars` is `:line`, the device is iterated line by line.
+      CRLF newlines  ("\r\n") are automatically normalized to "\n".
 
-    * if `line_or_chars` is `:eof`, the device is iterated until `:eof`. `line_or_chars`
-      can only be `:eof` since Elixir 1.13.0. `:eof` replaces the deprecated `:all`,
-      with the difference that `:all` returns `""` on end of file, while `:eof` returns
-      `:eof` itself.
+    * if `line_or_chars` is `:eof` (since v1.13), the device is iterated until `:eof`.
+      If the device is already at the end, it returns `:eof` itself.
 
   It returns:
 
@@ -200,8 +233,9 @@ defmodule IO do
   @spec binread(device, :eof | :line | non_neg_integer) :: iodata | nodata
   def binread(device \\ :stdio, line_or_chars)
 
-  # TODO: Deprecate me on v1.17
+  # TODO: Remove me on v2.0
   def binread(device, :all) do
+    IO.warn("IO.binread(device, :all) is deprecated, use IO.binread(device, :eof) instead")
     with :eof <- binread(device, :eof), do: ""
   end
 
@@ -282,6 +316,8 @@ defmodule IO do
   By default, the `device` is the standard output. It returns `:ok`
   if it succeeds.
 
+  Trivia: `puts` is shorthand for `put string`.
+
   ## Examples
 
       IO.puts("Hello World!")
@@ -326,7 +362,10 @@ defmodule IO do
       #=>   my_app.ex:4: MyApp.main/1
 
   """
-  @spec warn(chardata | String.Chars.t(), Exception.stacktrace() | keyword() | Macro.Env.t()) ::
+  @spec warn(
+          chardata | String.Chars.t(),
+          Exception.stacktrace() | warn_stacktrace_opts() | Macro.Env.t()
+        ) ::
           :ok
   def warn(message, stacktrace_info)
 
@@ -378,7 +417,7 @@ defmodule IO do
     stacktrace = Enum.drop(stacktrace, stacktrace_drop_levels)
 
     if :elixir_config.warn(key, stacktrace) do
-      warn(message, stacktrace)
+      warn(message.(), stacktrace)
     else
       :ok
     end
@@ -408,7 +447,7 @@ defmodule IO do
   end
 
   @doc """
-  Inspects and writes the given `item` to the device.
+  Inspects and writes the given `item` to the standard output.
 
   It's important to note that it returns the given `item` unchanged.
   This makes it possible to "spy" on values by inserting an
@@ -424,16 +463,19 @@ defmodule IO do
   The label will be printed before the inspected `item`.
 
   See `Inspect.Opts` for a full list of remaining formatting options.
+  To print to other IO devices, see `IO.inspect/3`
 
   ## Examples
 
+  The following code:
+
       IO.inspect(<<0, 1, 2>>, width: 40)
 
   Prints:
 
       <<0, 1, 2>>
 
-  We can use the `:label` option to decorate the output:
+  You can use the `:label` option to decorate the output:
 
       IO.inspect(1..100, label: "a wonderful range")
 
@@ -441,21 +483,23 @@ defmodule IO do
 
       a wonderful range: 1..100
 
-  The `:label` option is especially useful with pipelines:
+  Inspect truncates large inputs by default. The `:printable_limit` controls
+  the limit for strings and other string-like constructs (such as charlists):
 
-      [1, 2, 3]
-      |> IO.inspect(label: "before")
-      |> Enum.map(&(&1 * 2))
-      |> IO.inspect(label: "after")
-      |> Enum.sum()
+      "abc"
+      |> String.duplicate(9001)
+      |> IO.inspect(printable_limit: :infinity)
 
-  Prints:
+  For containers such as lists, maps, and tuples, the number of entries
+  is managed by the `:limit` option:
 
-      before: [1, 2, 3]
-      after: [2, 4, 6]
+      1..100
+      |> Enum.map(& {&1, &1})
+      |> Enum.into(%{})
+      |> IO.inspect(limit: :infinity)
 
   """
-  @spec inspect(item, keyword) :: item when item: var
+  @spec inspect(item, inspect_opts) :: item when item: var
   def inspect(item, opts \\ []) do
     inspect(:stdio, item, opts)
   end
@@ -465,7 +509,7 @@ defmodule IO do
 
   See `inspect/2` for a full list of options.
   """
-  @spec inspect(device, item, keyword) :: item when item: var
+  @spec inspect(device, item, inspect_opts) :: item when item: var
   def inspect(device, item, opts) when is_device(device) and is_list(opts) do
     label = if label = opts[:label], do: [to_chardata(label), ": "], else: []
     opts = Inspect.Opts.new(opts)
@@ -556,6 +600,8 @@ defmodule IO do
       for instance, `{:error, :estale}` if reading from an
       NFS volume
 
+  Trivia: `gets` is shorthand for `get string`.
+
   ## Examples
 
   To display "What is your name?" as a prompt and await user input:
@@ -580,15 +626,16 @@ defmodule IO do
   @spec stream() :: Enumerable.t(String.t())
   def stream, do: stream(:stdio, :line)
 
-  @doc """
+  @doc ~S"""
   Converts the IO `device` into an `IO.Stream`.
 
   An `IO.Stream` implements both `Enumerable` and
   `Collectable`, allowing it to be used for both read
   and write.
 
-  The `device` is iterated by the given number of characters or line by line if
-  `:line` is given.
+  The `device` is iterated by the given number of characters
+  or line by line if `:line` is given. In case `:line` is given,
+  "\r\n" is automatically normalized to "\n".
 
   This reads from the IO as UTF-8. Check out
   `IO.binstream/2` to handle the IO as a raw binary.
@@ -608,11 +655,11 @@ defmodule IO do
 
   Another example where you might want to collect a user input
   every new line and break on an empty line, followed by removing
-  redundant new line characters (`"\\n"`):
+  redundant new line characters (`"\n"`):
 
       IO.stream(:stdio, :line)
-      |> Enum.take_while(&(&1 != "\\n"))
-      |> Enum.map(&String.replace(&1, "\\n", ""))
+      |> Enum.take_while(&(&1 != "\n"))
+      |> Enum.map(&String.replace(&1, "\n", ""))
 
   """
   @spec stream(device, :line | pos_integer) :: Enumerable.t()
@@ -634,21 +681,24 @@ defmodule IO do
   @spec binstream() :: Enumerable.t(binary)
   def binstream, do: binstream(:stdio, :line)
 
-  @doc """
+  @doc ~S"""
   Converts the IO `device` into an `IO.Stream`. The operation is Unicode unsafe.
 
   An `IO.Stream` implements both `Enumerable` and
   `Collectable`, allowing it to be used for both read
   and write.
 
-  The `device` is iterated by the given number of bytes or line by line if
-  `:line` is given. This reads from the IO device as a raw binary.
+  The `device` is iterated by the given number of bytes or line
+  by line if `:line` is given. In case `:line` is given, "\r\n"
+  is automatically normalized to "\n". Passing the number of bytes
+  should be the preferred mode for reading non-textual inputs.
 
   Note that an IO stream has side effects and every time
   you go over the stream you may get different results.
 
-  Finally, do not use this function on IO devices in Unicode
-  mode as it will return the wrong result.
+  This reads from the IO device as a raw binary. Therefore,
+  do not use this function on IO devices in Unicode mode as
+  it will return the wrong result.
 
   `binstream/0` has been introduced in Elixir v1.12.0,
   while `binstream/2` has been available since v1.0.0.
