@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Kernel.ParallelCompiler do
   @moduledoc """
   A module responsible for compiling and requiring files in parallel.
@@ -12,6 +16,38 @@ defmodule Kernel.ParallelCompiler do
   @type warning() :: {file :: Path.t(), Code.position(), message :: String.t()}
   @type error() :: {file :: Path.t(), Code.position(), message :: String.t()}
 
+  @typedoc """
+  Options for parallel compilation functions.
+  """
+  @type compile_opts :: [
+          after_compile: (-> term()),
+          each_file: (Path.t() -> term()),
+          each_long_compilation: (Path.t() -> term()) | (Path.t(), pid() -> term()),
+          each_long_verification: (module() -> term()) | (module(), pid() -> term()),
+          each_module: (Path.t(), module(), binary() -> term()),
+          each_cycle: ([module()], [Code.diagnostic(:warning)] ->
+                         {:compile, [module()], [Code.diagnostic(:warning)]}
+                         | {:runtime, [module()], [Code.diagnostic(:warning)]}),
+          long_compilation_threshold: pos_integer(),
+          long_verification_threshold: pos_integer(),
+          verification: boolean(),
+          profile: :time,
+          dest: Path.t(),
+          beam_timestamp: term(),
+          return_diagnostics: boolean(),
+          max_concurrency: pos_integer()
+        ]
+
+  @typedoc """
+  Options for requiring files in parallel.
+  """
+  @type require_opts :: [
+          each_file: (Path.t() -> term()),
+          each_module: (Path.t(), module(), binary() -> term()),
+          max_concurrency: pos_integer(),
+          return_diagnostics: boolean()
+        ]
+
   @doc """
   Starts a task for parallel compilation.
   """
@@ -78,7 +114,7 @@ defmodule Kernel.ParallelCompiler do
     {compiler_pid, file_pid} = :erlang.get(:elixir_compiler_info)
     defining = :elixir_module.compiler_modules()
     on = Enum.map(refs_tasks, fn {_ref, %{pid: pid}} -> pid end)
-    send(compiler_pid, {:waiting, :pmap, self(), ref, file_pid, on, defining, :raise})
+    send(compiler_pid, {:waiting, :pmap, self(), ref, file_pid, nil, on, defining, :raise})
 
     # Now we allow the tasks to run. This step is not strictly
     # necessary but it makes compilation more deterministic by
@@ -110,10 +146,9 @@ defmodule Kernel.ParallelCompiler do
   the current file stops being compiled until the dependency is
   resolved.
 
-  It returns `{:ok, modules, warnings}` or `{:error, errors, warnings}`
-  by default but we recommend using `return_diagnostics: true` so it returns
-  diagnostics as maps as well as a map of compilation information.
-  The map has the shape of:
+  It must be invoked with `return_diagnostics: true` as option, so it returns
+  `{:ok, modules, warnings_info}` or `{:error, errors, warnings_info}`,
+  where `warnings_info` has the shape:
 
       %{
         runtime_warnings: [warning],
@@ -122,12 +157,22 @@ defmodule Kernel.ParallelCompiler do
 
   ## Options
 
+    * `:after_compile` - invoked after all modules are compiled, but before
+      they are verified. If the files are being written to disk, such as in
+      `compile_to_path/3`, this will be invoked after the files are written
+
     * `:each_file` - for each file compiled, invokes the callback passing the
       file
 
     * `:each_long_compilation` - for each file that takes more than a given
       timeout (see the `:long_compilation_threshold` option) to compile, invoke
-      this callback passing the file as its argument
+      this callback passing the file as its argument (and optionally the PID
+      of the process compiling the file)
+
+    * `:each_long_verification` (since v1.19.0) - for each file that takes more
+      than a given timeout (see the `:long_verification_threshold` option) to
+      compile, invoke this callback passing the module as its argument (and
+      optionally the PID of the process verifying the module)
 
     * `:each_module` - for each module compiled, invokes the callback passing
       the file, module and the module bytecode
@@ -139,11 +184,18 @@ defmodule Kernel.ParallelCompiler do
       * `{:runtime, modules, warnings}` - to stop compilation and verify the list
         of modules because dependent modules have changed
 
-    * `:long_compilation_threshold` - the timeout (in seconds) to check for modules
+    * `:long_compilation_threshold` - the timeout (in seconds) to check for files
       taking too long to compile. For each file that exceeds the threshold, the
-      `:each_long_compilation` callback is invoked. From Elixir v1.11, only the time
-      spent compiling the actual module is taken into account by the threshold, the
-      time spent waiting is not considered. Defaults to `10` seconds.
+      `:each_long_compilation` callback is invoked. Defaults to `10` seconds.
+
+    * `:long_verification_threshold` (since v1.19.0) - the timeout (in seconds) to
+      check for modules taking too long to compile. For each module that exceeds the
+      threshold, the `:each_long_verification` callback is invoked. Defaults to
+      `10` seconds.
+
+    * `:verification` (since v1.19.0) - if code verification, such as unused functions,
+      deprecation warnings, and type checking should run. Defaults to `true`.
+      We recommend disabling it only for debugging purposes.
 
     * `:profile` - if set to `:time` measure the compilation time of each compilation cycle
        and group pass checker
@@ -156,15 +208,16 @@ defmodule Kernel.ParallelCompiler do
     * `:beam_timestamp` - the modification timestamp to give all BEAM files
 
     * `:return_diagnostics` (since v1.15.0) - returns maps with information instead of
-      a list of warnings and returns diagnostics as maps instead of tuples
+      a list of warnings and returns diagnostics as maps instead of tuples.
+      This option must be set to true, except for backwards compatibibility reasons.
 
     * `:max_concurrency` - the maximum number of files to compile in parallel.
       Setting this option to 1 will compile files sequentially.
-      Defaults to the number of schedulers online, or at least 2.
+      Defaults to the number of schedulers online, or at least `2`.
 
   """
   @doc since: "1.6.0"
-  @spec compile([Path.t()], keyword()) ::
+  @spec compile([Path.t()], compile_opts()) ::
           {:ok, [atom], [warning] | info()}
           | {:error, [error] | [Code.diagnostic(:error)], [warning] | info()}
   def compile(files, options \\ []) when is_list(options) do
@@ -177,11 +230,11 @@ defmodule Kernel.ParallelCompiler do
   See `compile/2` for more information.
   """
   @doc since: "1.6.0"
-  @spec compile_to_path([Path.t()], Path.t(), keyword()) ::
+  @spec compile_to_path([Path.t()], Path.t(), compile_opts()) ::
           {:ok, [atom], [warning] | info()}
           | {:error, [error] | [Code.diagnostic(:error)], [warning] | info()}
   def compile_to_path(files, path, options \\ []) when is_binary(path) and is_list(options) do
-    spawn_workers(files, {:compile, path}, options)
+    spawn_workers(files, {:compile, path}, Keyword.put(options, :dest, path))
   end
 
   @doc """
@@ -190,10 +243,9 @@ defmodule Kernel.ParallelCompiler do
   Opposite to compile, dependencies are not attempted to be
   automatically solved between files.
 
-  It returns `{:ok, modules, warnings}` or `{:error, errors, warnings}`
-  by default but we recommend using `return_diagnostics: true` so it returns
-  diagnostics as maps as well as a map of compilation information.
-  The map has the shape of:
+  It must be invoked with `return_diagnostics: true` as option, so it returns
+  `{:ok, modules, warnings_info}` or `{:error, errors, warnings_info}`,
+  where `warnings_info` has the shape:
 
       %{
         runtime_warnings: [warning],
@@ -210,11 +262,15 @@ defmodule Kernel.ParallelCompiler do
 
     * `:max_concurrency` - the maximum number of files to compile in parallel.
       Setting this option to 1 will compile files sequentially.
-      Defaults to the number of schedulers online, or at least 2.
+      Defaults to the number of schedulers online, or at least `2`.
+
+    * `:return_diagnostics` (since v1.15.0) - returns maps with information instead of
+      a list of warnings and returns diagnostics as maps instead of tuples.
+      This option must be set to true, except for backwards compatibibility reasons.
 
   """
   @doc since: "1.6.0"
-  @spec require([Path.t()], keyword()) ::
+  @spec require([Path.t()], require_opts()) ::
           {:ok, [atom], [warning] | info()}
           | {:error, [error] | [Code.diagnostic(:error)], [warning] | info()}
   def require(files, options \\ []) when is_list(options) do
@@ -222,7 +278,7 @@ defmodule Kernel.ParallelCompiler do
   end
 
   @doc false
-  # TODO: Deprecate me on Elixir v1.19
+  @deprecated "Use Code.print_diagnostic/2 instead"
   def print_warning({file, location, warning}) do
     :elixir_errors.print_warning(location, file, warning)
   end
@@ -253,7 +309,7 @@ defmodule Kernel.ParallelCompiler do
         max(:erlang.system_info(:schedulers_online), 2)
       end)
 
-    {:ok, cache} = Module.ParallelChecker.start_link(schedulers)
+    {:ok, cache} = Module.ParallelChecker.start_link(options)
 
     {status, modules_or_errors, info} =
       try do
@@ -262,10 +318,13 @@ defmodule Kernel.ParallelCompiler do
         Module.ParallelChecker.stop(cache)
       end
 
-    # TODO: Require this to be set from Elixir v1.19
     if Keyword.get(options, :return_diagnostics, false) do
       {status, modules_or_errors, info}
     else
+      IO.warn(
+        "you must pass return_diagnostics: true when invoking Kernel.ParallelCompiler functions"
+      )
+
       to_tuples = &Enum.map(&1, fn diag -> {diag.file, diag.position, diag.message} end)
 
       modules_or_errors =
@@ -283,6 +342,7 @@ defmodule Kernel.ParallelCompiler do
       spawn_workers(files, %{}, %{}, [], %{}, [], [], %{
         beam_timestamp: Keyword.get(options, :beam_timestamp),
         dest: Keyword.get(options, :dest),
+        after_compile: Keyword.get(options, :after_compile, fn -> :ok end),
         each_cycle: Keyword.get(options, :each_cycle, fn -> {:runtime, [], []} end),
         each_file: Keyword.get(options, :each_file, fn _, _ -> :ok end) |> each_file(),
         each_long_compilation: Keyword.get(options, :each_long_compilation, fn _file -> :ok end),
@@ -292,7 +352,8 @@ defmodule Kernel.ParallelCompiler do
         timer_ref: timer_ref,
         long_compilation_threshold: threshold,
         schedulers: schedulers,
-        checker: checker
+        checker: checker,
+        verification?: Keyword.get(options, :verification, true)
       })
 
     Process.cancel_timer(state.timer_ref)
@@ -320,46 +381,45 @@ defmodule Kernel.ParallelCompiler do
   end
 
   defp write_module_binaries(result, {:compile, path}, timestamp) do
-    Enum.flat_map(result, fn
-      {{:module, module}, binary} when is_binary(binary) ->
-        full_path = Path.join(path, Atom.to_string(module) <> ".beam")
-        File.write!(full_path, binary)
-        if timestamp, do: File.touch!(full_path, timestamp)
-        [module]
+    File.mkdir_p!(path)
+    Code.prepend_path(path)
 
-      _ ->
-        []
-    end)
+    for {{:module, module}, {binary, _}} when is_binary(binary) <- result do
+      full_path = Path.join(path, Atom.to_string(module) <> ".beam")
+      File.write!(full_path, binary)
+      if timestamp, do: File.touch!(full_path, timestamp)
+      module
+    end
   end
 
   defp write_module_binaries(result, _output, _timestamp) do
-    for {{:module, module}, binary} when is_binary(binary) <- result, do: module
+    for {{:module, module}, {binary, _}} when is_binary(binary) <- result, do: module
   end
 
   ## Verification
 
   defp verify_modules(result, compile_warnings, dependent_modules, state) do
     modules = write_module_binaries(result, state.output, state.beam_timestamp)
-    runtime_warnings = maybe_check_modules(result, dependent_modules, state)
+    _ = state.after_compile.()
+
+    runtime_warnings =
+      if state.verification? do
+        profile(
+          state,
+          fn ->
+            num_modules = length(modules) + length(dependent_modules)
+            "group pass check of #{num_modules} modules"
+          end,
+          fn -> Module.ParallelChecker.verify(state.checker, dependent_modules) end
+        )
+      else
+        []
+      end
+
     info = %{compile_warnings: Enum.reverse(compile_warnings), runtime_warnings: runtime_warnings}
     {{:ok, modules, info}, state}
   end
 
-  defp maybe_check_modules(result, runtime_modules, state) do
-    compiled_modules =
-      for {{:module, module}, binary} when is_binary(binary) <- result,
-          do: module
-
-    profile(
-      state,
-      fn ->
-        num_modules = length(compiled_modules) + length(runtime_modules)
-        "group pass check of #{num_modules} modules"
-      end,
-      fn -> Module.ParallelChecker.verify(state.checker, runtime_modules) end
-    )
-  end
-
   defp profile_init(:time), do: {:time, System.monotonic_time(), 0}
   defp profile_init(nil), do: :none
 
@@ -394,7 +454,7 @@ defmodule Kernel.ParallelCompiler do
   defp spawn_workers([{pid, found} | t], spawned, waiting, files, result, warnings, errors, state) do
     {files, waiting} =
       case Map.pop(waiting, pid) do
-        {{kind, ref, file_pid, on, _defining, _deadlock}, waiting} ->
+        {%{kind: kind, ref: ref, file_pid: file_pid, on: on}, waiting} ->
           send(pid, {ref, found})
           {update_timing(files, file_pid, {:waiting, kind, on}), waiting}
 
@@ -420,8 +480,8 @@ defmodule Kernel.ParallelCompiler do
 
         try do
           case output do
-            {:compile, path} -> compile_file(file, path, parent)
-            :compile -> compile_file(file, dest, parent)
+            {:compile, _} -> compile_file(file, dest, false, parent)
+            :compile -> compile_file(file, dest, true, parent)
             :require -> require_file(file, parent)
           end
         catch
@@ -527,9 +587,9 @@ defmodule Kernel.ParallelCompiler do
     wait_for_messages([], spawned, waiting, files, result, warnings, errors, state)
   end
 
-  defp compile_file(file, path, parent) do
+  defp compile_file(file, path, force_load?, parent) do
     :erlang.process_flag(:error_handler, Kernel.ErrorHandler)
-    :erlang.put(:elixir_compiler_dest, path)
+    :erlang.put(:elixir_compiler_dest, {path, force_load?})
     :elixir_compiler.file(file, &each_file(&1, &2, parent))
   end
 
@@ -563,7 +623,7 @@ defmodule Kernel.ParallelCompiler do
   end
 
   defp count_modules(result) do
-    Enum.count(result, &match?({{:module, _}, binary} when is_binary(binary), &1))
+    Enum.count(result, &match?({{:module, _}, {binary, _}} when is_binary(binary), &1))
   end
 
   defp each_cycle_return({kind, modules, warnings}), do: {kind, modules, warnings}
@@ -585,7 +645,7 @@ defmodule Kernel.ParallelCompiler do
   defp without_definition(waiting_list, files) do
     nilify_empty_or_sort(
       for %{pid: file_pid} <- files,
-          {pid, {_, _, ^file_pid, on, _, _}} <- waiting_list,
+          {pid, %{file_pid: ^file_pid, on: on}} <- waiting_list,
           is_atom(on) and not defining?(on, waiting_list),
           do: {pid, :not_found}
     )
@@ -593,14 +653,14 @@ defmodule Kernel.ParallelCompiler do
 
   defp deadlocked(waiting_list, type, defining?) do
     nilify_empty_or_sort(
-      for {pid, {_, _, _, on, _, ^type}} <- waiting_list,
+      for {pid, %{on: on, deadlock: ^type}} <- waiting_list,
           is_atom(on) and defining?(on, waiting_list) == defining?,
           do: {pid, :deadlock}
     )
   end
 
   defp defining?(on, waiting_list) do
-    Enum.any?(waiting_list, fn {_, {_, _, _, _, defining, _}} -> on in defining end)
+    Enum.any?(waiting_list, fn {_, %{defining: defining}} -> on in defining end)
   end
 
   defp nilify_empty_or_sort([]), do: nil
@@ -630,30 +690,49 @@ defmodule Kernel.ParallelCompiler do
           state
         )
 
-      {:module_available, child, ref, file, module, binary} ->
-        state.each_module.(file, module, binary)
+      {{:module_loaded, module}, _ref, _type, _pid, _reason} ->
+        result =
+          Map.update!(result, {:module, module}, fn {binary, _loader} -> {binary, true} end)
 
-        # Release the module loader which is waiting for an ack
+        spawn_workers(queue, spawned, waiting, files, result, warnings, errors, state)
+
+      {:module_available, child, ref, file, module, binary, loaded?} ->
+        state.each_module.(file, module, binary)
         send(child, {ref, :ack})
-        {available, result} = update_result(result, :module, module, binary)
+
+        {available, load_status} =
+          case Map.get(result, {:module, module}) do
+            [_ | _] = pids when loaded? ->
+              {Enum.map(pids, &{&1, :found}), loaded?}
+
+            # When compiling files to disk, we only load the module
+            # if other modules are waiting for it.
+            [_ | _] = pids ->
+              pid = load_module(module, binary, state.dest)
+              {Enum.map(pids, &{&1, {:loading, pid}}), pid}
+
+            _ ->
+              {[], loaded?}
+          end
 
         spawn_workers(
           available ++ queue,
           spawned,
           waiting,
           files,
-          result,
+          Map.put(result, {:module, module}, {binary, load_status}),
           warnings,
           errors,
           state
         )
 
       # If we are simply requiring files, we do not add to waiting.
-      {:waiting, _kind, child, ref, _file_pid, _on, _defining, _deadlock} when output == :require ->
+      {:waiting, _kind, child, ref, _file_pid, _position, _on, _defining, _deadlock}
+      when output == :require ->
         send(child, {ref, :not_found})
         spawn_workers(queue, spawned, waiting, files, result, warnings, errors, state)
 
-      {:waiting, kind, child_pid, ref, file_pid, on, defining, deadlock} ->
+      {:waiting, kind, child_pid, ref, file_pid, position, on, defining, deadlock} ->
         # If we already got what we were waiting for, do not put it on waiting.
         # If we're waiting on ourselves, send :found so that we can crash with
         # a better error.
@@ -661,10 +740,22 @@ defmodule Kernel.ParallelCompiler do
 
         {waiting, files, result} =
           if not is_list(available_or_pending) or on in defining do
-            send(child_pid, {ref, :found})
+            # If what we are waiting on was defined but not loaded, we do it now.
+            {reply, result} = load_pending(kind, on, result, state)
+            send(child_pid, {ref, reply})
             {waiting, files, result}
           else
-            waiting = Map.put(waiting, child_pid, {kind, ref, file_pid, on, defining, deadlock})
+            waiting =
+              Map.put(waiting, child_pid, %{
+                kind: kind,
+                ref: ref,
+                file_pid: file_pid,
+                position: position,
+                on: on,
+                defining: defining,
+                deadlock: deadlock
+              })
+
             files = update_timing(files, file_pid, :compiling)
             result = Map.put(result, {kind, on}, [child_pid | available_or_pending])
             {waiting, files, result}
@@ -718,7 +809,7 @@ defmodule Kernel.ParallelCompiler do
         terminate(new_files)
 
         return_error(warnings, errors, state, fn ->
-          print_error(file, kind, reason, stack)
+          print_error(file, nil, kind, reason, stack)
           [to_error(file, kind, reason, stack)]
         end)
 
@@ -732,7 +823,7 @@ defmodule Kernel.ParallelCompiler do
           terminate(files)
 
           return_error(warnings, errors, state, fn ->
-            print_error(file.file, :exit, reason, [])
+            print_error(file.file, nil, :exit, reason, [])
             [to_error(file.file, :exit, reason, [])]
           end)
         else
@@ -755,6 +846,55 @@ defmodule Kernel.ParallelCompiler do
     {{:error, Enum.reverse(errors, fun.()), info}, state}
   end
 
+  defp load_pending(kind, module, result, state) do
+    case result do
+      %{{:module, ^module} => {binary, load_status}}
+      when kind in [:module, :struct] and is_binary(binary) ->
+        case load_status do
+          true ->
+            {:found, result}
+
+          false ->
+            pid = load_module(module, binary, state.dest)
+            result = Map.put(result, {:module, module}, {binary, pid})
+            {{:loading, pid}, result}
+
+          pid when is_pid(pid) ->
+            {{:loading, pid}, result}
+        end
+
+      _ ->
+        {:found, result}
+    end
+  end
+
+  # We load modules in a separate process to avoid blocking
+  # the parallel compiler. We store the PID of this process and
+  # all entries monitor it to know once the module is loaded.
+  defp load_module(module, binary, dest) do
+    {pid, _ref} =
+      :erlang.spawn_opt(
+        fn ->
+          beam_location =
+            case dest do
+              nil ->
+                []
+
+              dest ->
+                :filename.join(
+                  :elixir_utils.characters_to_list(dest),
+                  Atom.to_charlist(module) ++ ~c".beam"
+                )
+            end
+
+          :code.load_binary(module, beam_location, binary)
+        end,
+        monitor: [tag: {:module_loaded, module}]
+      )
+
+    pid
+  end
+
   defp update_result(result, kind, module, value) do
     available =
       case Map.get(result, {kind, module}) do
@@ -785,7 +925,12 @@ defmodule Kernel.ParallelCompiler do
     compiling = System.convert_time_unit(data.compiling, :native, :millisecond)
 
     if not data.warned and compiling >= state.long_compilation_threshold do
-      state.each_long_compilation.(data.file)
+      if is_function(state.each_long_compilation, 2) do
+        state.each_long_compilation.(data.file, data.pid)
+      else
+        state.each_long_compilation.(data.file)
+      end
+
       %{data | warned: true}
     else
       data
@@ -853,11 +998,11 @@ defmodule Kernel.ParallelCompiler do
         {:current_stacktrace, stacktrace} = Process.info(pid, :current_stacktrace)
         Process.exit(pid, :kill)
 
-        {kind, _, _, on, _, _} = Map.fetch!(waiting, pid)
+        %{kind: kind, on: on, position: position} = Map.fetch!(waiting, pid)
         description = "deadlocked waiting on #{kind} #{inspect(on)}"
         error = CompileError.exception(description: description, file: nil, line: nil)
-        print_error(file, :error, error, stacktrace)
-        {Path.relative_to_cwd(file), on, description, stacktrace}
+        print_error(file, position, :error, error, stacktrace)
+        {Path.relative_to_cwd(file), position, on, description, stacktrace}
       end
 
     IO.puts(:stderr, """
@@ -871,24 +1016,25 @@ defmodule Kernel.ParallelCompiler do
       |> Enum.map(&(&1 |> elem(0) |> String.length()))
       |> Enum.max()
 
-    for {file, mod, _, _} <- deadlock do
+    for {file, _, mod, _, _} <- deadlock do
       IO.puts(:stderr, ["  ", String.pad_leading(file, max), " => " | inspect(mod)])
     end
 
     IO.puts(
       :stderr,
       "\nEnsure there are no compile-time dependencies between those files " <>
-        "and that the modules they reference exist and are correctly named\n"
+        "(such as structs or macros) and that the modules they reference exist " <>
+        "and are correctly named\n"
     )
 
-    for {file, _, description, stacktrace} <- deadlock do
+    for {file, position, _, description, stacktrace} <- deadlock do
       file = Path.absname(file)
 
       %{
         severity: :error,
         file: file,
         source: file,
-        position: nil,
+        position: position,
         message: description,
         stacktrace: stacktrace,
         span: nil
@@ -908,9 +1054,11 @@ defmodule Kernel.ParallelCompiler do
     :ok
   end
 
-  defp print_error(file, kind, reason, stack) do
+  defp print_error(file, position, kind, reason, stack) do
+    position = if is_integer(position), do: ":#{position}", else: ""
+
     IO.write(:stderr, [
-      "\n== Compilation error in file #{Path.relative_to_cwd(file)} ==\n",
+      "\n== Compilation error in file #{Path.relative_to_cwd(file)}#{position} ==\n",
       Kernel.CLI.format_error(kind, reason, stack)
     ])
   end
