@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Kernel.SpecialForms do
   @moduledoc """
   Special forms are the basic building blocks of Elixir, and therefore
@@ -98,8 +102,8 @@ defmodule Kernel.SpecialForms do
 
       %User{}
 
-  Underneath a struct is just a map with a `:__struct__` key
-  pointing to the `User` module:
+  Underneath a struct is a map with a `:__struct__` key pointing
+  to the `User` module, where the keys are validated at compile-time:
 
       %User{} == %{__struct__: User, name: "john", age: 27}
 
@@ -112,26 +116,15 @@ defmodule Kernel.SpecialForms do
 
       %User{age: age} = user
 
-  An update operation specific for structs is also available:
-
-      %User{user | age: 28}
-
   The advantage of structs is that they validate that the given
   keys are part of the defined struct. The example below will fail
   because there is no key `:full_name` in the `User` struct:
 
       %User{full_name: "john doe"}
 
-  The syntax above will guarantee the given keys are valid at
-  compilation time and it will guarantee at runtime the given
-  argument is a struct, failing with `BadStructError` otherwise.
+  The map update syntax can also be used for updating structs:
 
-  Although structs are maps, by default structs do not implement
-  any of the protocols implemented for maps. Check
-  `Kernel.defprotocol/2` for more information on how structs
-  can be used with protocols for polymorphic dispatch. Also
-  see `Kernel.struct/2` and `Kernel.struct!/2` for examples on
-  how to create and update structs dynamically.
+      %{user | age: 28}
 
   ## Pattern matching on struct names
 
@@ -722,7 +715,7 @@ defmodule Kernel.SpecialForms do
   Returns the stacktrace for the currently handled exception.
 
   It is available only in the `catch` and `rescue` clauses of `try/1`
-  expressions.
+  expressions and function definitions.
 
   To retrieve the stacktrace of the current process, use
   `Process.info(self(), :current_stacktrace)` instead.
@@ -749,7 +742,8 @@ defmodule Kernel.SpecialForms do
       iex> x = 1
       iex> ^x = List.first([1])
       iex> ^x = List.first([2])
-      ** (MatchError) no match of right hand side value: 2
+      ** (MatchError) no match of right hand side value:
+      ...
 
   Note that `^x` always refers to the value of `x` prior to the match. The
   following example will match:
@@ -840,8 +834,8 @@ defmodule Kernel.SpecialForms do
     * `:context` - sets the resolution context.
 
     * `:generated` - marks the given chunk as generated so it does not emit warnings.
-      It is also useful to avoid dialyzer reporting errors when macros generate
-      unused clauses.
+      It is also useful to prevent the type system or dialyzer from reporting errors
+      when macros generate unused clauses.
 
     * `:file` - sets the quoted expressions to have the given file.
 
@@ -1048,7 +1042,7 @@ defmodule Kernel.SpecialForms do
         end
       end
 
-      require Hygiene
+      require ContextHygiene
       ContextHygiene.write()
       ContextHygiene.read()
       #=> 1
@@ -1248,8 +1242,11 @@ defmodule Kernel.SpecialForms do
       end)
 
   In the example above, we have generated the functions `foo/0` and
-  `bar/0` dynamically. Now, imagine that we want to convert this
-  functionality into a macro:
+  `bar/0` dynamically. Note the parentheses in `unquote(k)()` are important,
+  otherwise we would try to define a function as `def :foo` instead of
+  `def foo()`.
+
+  Now, imagine that we want to convert this functionality into a macro:
 
       defmacro defkv(kv) do
         Enum.map(kv, fn {k, v} ->
@@ -1274,8 +1271,9 @@ defmodule Kernel.SpecialForms do
   code fails.
 
   This is actually a common pitfall when developing macros. We are
-  assuming a particular shape in the macro. We can work around it
-  by unquoting the variable inside the quoted expression:
+  assuming a particular shape at compilation time, within the macro
+  implementation. One may try to work around it by unquoting the
+  variable inside the quoted expression:
 
       defmacro defkv(kv) do
         quote do
@@ -1287,9 +1285,10 @@ defmodule Kernel.SpecialForms do
 
   If you try to run our new macro, you will notice it won't
   even compile, complaining that the variables `k` and `v`
-  do not exist. This is because of the ambiguity: `unquote(k)`
-  can either be an unquote fragment, as previously, or a regular
-  unquote as in `unquote(kv)`.
+  do not exist. This is because the two `unquote`s in the call
+  above are meant to run at distinct moments: `unquote(kv)`
+  applies to the immediate quote, `unquote(k)` is an unquote
+  fragment.
 
   One solution to this problem is to disable unquoting in the
   macro, however, doing that would make it impossible to inject the
@@ -1378,6 +1377,18 @@ defmodule Kernel.SpecialForms do
       ...> end
       {:sum, [], [1, 2, 3, 4, 5]}
 
+  Also can be used in block context, outside of function arguments.
+  Though, it is still required to be wrapped into parentheses.
+
+      iex> requires = for module <- [Integer, Logger] do
+      ...>   quote do
+      ...>     require unquote(module)
+      ...>   end
+      ...> end
+      iex> block = quote do: (unquote_splicing(requires))
+      iex> Macro.to_string(block)
+      "require Integer\\nrequire Logger"
+
   """
   defmacro unquote(:unquote_splicing)(expr), do: error!([expr])
 
@@ -1541,7 +1552,7 @@ defmodule Kernel.SpecialForms do
   @doc """
   Combine matching clauses.
 
-  One of the ways to understand with is to show which code
+  One of the ways to understand `with` is to show which code
   patterns it improves. Imagine you have a map where the fields
   `width` and `height` are optional and you want to compute its
   area, as `{:ok, area}` or return `:error`. We could implement
@@ -1663,7 +1674,7 @@ defmodule Kernel.SpecialForms do
       ...> end
       {:error, :wrong_data}
 
-  The `else` block works like a `case` clause: it can have multiple clauses,
+  The `else` block works like a `case`: it can have multiple clauses,
   and the first match will be used. Variables bound inside `with` (such as
   `width` in this example) are not available in the `else` block.
 
@@ -1788,6 +1799,15 @@ defmodule Kernel.SpecialForms do
 
       &local_function/1
 
+  Note that `&local_function/1` creates a local capture, but
+  `&__MODULE__.local_function/1` or `&imported_function/1` create a remote
+  capture. For more information, refer to the ["Functions" section in the Erlang Reference Manual](https://www.erlang.org/doc/system/eff_guide_functions.html#function-calls).
+
+  Whether a capture is local or remote has implications when using hot code
+  reloading: local captures dispatch to the version of the module that existed
+  at the time they were created, while remote captures dispatch to the current
+  version of the module.
+
   See also `Function.capture/3`.
 
   ## Anonymous functions
@@ -1894,33 +1914,30 @@ defmodule Kernel.SpecialForms do
 
   ## Examples
 
-      case File.read(file) do
-        {:ok, contents} when is_binary(contents) ->
-          String.split(contents, "\n")
+      iex> string_date = "2015-01-23"
+      iex> case Date.from_iso8601(string_date) do
+      ...>   {:ok, date} -> date
+      ...>   {:error, _reason} -> Date.utc_today()
+      ...> end
+      ~D[2015-01-23]
 
-        {:error, _reason} ->
-          Logger.warning "could not find #{file}, assuming empty..."
-          []
-      end
-
-  In the example above, we match the result of `File.read/1`
+  In the example above, we match the result of `Date.from_iso8601/1`
   against each clause "head" and execute the clause "body"
-  corresponding to the first clause that matches.
+  corresponding to the first clause that matches. In our case
+  `string_date` contains a string with a valid ISO 8601 representation
+  of date. The function returns `{:ok, ~D[2015-01-23]}`, so the
+  `{:ok, date}` clause is matched.
 
   If no clause matches, an error is raised. For this reason,
   it may be necessary to add a final catch-all clause (like `_`)
   which will always match.
 
-      x = 10
-
-      case x do
-        0 ->
-          "This clause won't match"
-
-        _ ->
-          "This clause would match any value (x = #{x})"
-      end
-      #=> "This clause would match any value (x = 10)"
+      iex> x = 10
+      iex> case x do
+      ...>   0 -> "This clause won't match"
+      ...>   _ -> "This clause would match any value (x = #{x})"
+      ...> end
+      "This clause would match any value (x = 10)"
 
   If you find yourself nesting `case` expressions inside
   `case` expressions, consider using `with/1`.
@@ -1929,54 +1946,54 @@ defmodule Kernel.SpecialForms do
 
   Note that variables bound in a clause do not leak to the outer context:
 
-      case data do
-        {:ok, value} -> value
-        :error -> nil
-      end
+      iex> case {:ok, 7} do
+      ...>   {:ok, value} -> value
+      ...>   :error -> nil
+      ...> end
 
-      value
-      #=> unbound variable value
+      ...> value
+      ** (CompileError) undefined variable "value"
 
   Variables in the outer context cannot be overridden either:
 
-      value = 7
+      iex> value = 7
+      iex> case 3 > 5 do
+      ...>   false ->
+      ...>     value = 3
+      ...>     value + 2
+      ...>   true ->
+      ...>     3
+      ...> end
+      iex> value
+      7
 
-      case lucky? do
-        false -> value = 13
-        true -> true
-      end
-
-      value
-      #=> 7
-
-  In the example above, `value` is going to be `7` regardless of the value of
-  `lucky?`. The variable `value` bound in the clause and the variable `value`
-  bound in the outer context are two entirely separate variables.
+  In the example above, `value` is going to be `7` regardless of
+  which clause matched. The variable `value` bound in the clause
+  and the variable `value` bound in the outer context are two
+  entirely separate variables.
 
   If you want to pattern match against an existing variable,
   you need to use the `^/1` operator:
 
-      x = 1
-
-      case 10 do
-        ^x -> "Won't match"
-        _ -> "Will match"
-      end
-      #=> "Will match"
+      iex> x = 1
+      iex> case 10 do
+      ...>   ^x -> "Won't match"
+      ...>   _ -> "Will match"
+      ...> end
+      "Will match"
 
   ## Using guards to match against multiple values
 
   While it is not possible to match against multiple patterns in a single
   clause, it's possible to match against multiple values by using guards:
 
-      case data do
-        value when value in [:one, :two] ->
-          "#{value} has been matched"
-
-        :three ->
-          "three has been matched"
-      end
-
+      iex> case :two do
+      ...>   value when value in [:one, :two] ->
+      ...>     "#{value} has been matched"
+      ...>   :three ->
+      ...>     "three has been matched"
+      ...> end
+      "two has been matched"
   """
   defmacro case(condition, clauses), do: error!([condition, clauses])
 
@@ -1989,25 +2006,21 @@ defmodule Kernel.SpecialForms do
   The following example has a single clause that always evaluates
   to true:
 
-      cond do
-        hd([1, 2, 3]) ->
-          "1 is considered as true"
-      end
-      #=> "1 is considered as true"
+      iex> cond do
+      ...>   hd([1, 2, 3]) -> "1 is considered as true"
+      ...> end
+      "1 is considered as true"
 
   If all clauses evaluate to `nil` or `false`, `cond` raises an error.
   For this reason, it may be necessary to add a final always-truthy condition
   (anything non-`false` and non-`nil`), which will always match:
 
-      cond do
-        1 + 1 == 1 ->
-          "This will never match"
-        2 * 2 != 4 ->
-          "Nor this"
-        true ->
-          "This will"
-      end
-      #=> "This will"
+      iex> cond do
+      ...>   1 + 1 == 1 -> "This will never match"
+      ...>   2 * 2 != 4 -> "Nor this"
+      ...>   true -> "This will"
+      ...> end
+      "This will"
 
 
   If your `cond` has two clauses, and the last one falls back to
@@ -2053,68 +2066,86 @@ defmodule Kernel.SpecialForms do
   exception by its name. All the following formats are valid patterns
   in `rescue` clauses:
 
-      # Rescue a single exception without binding the exception
-      # to a variable
-      try do
-        UndefinedModule.undefined_function
-      rescue
-        UndefinedFunctionError -> nil
-      end
+  Rescue a single exception without binding the exception to a variable:
 
-      # Rescue any of the given exception without binding
-      try do
-        UndefinedModule.undefined_function
-      rescue
-        [UndefinedFunctionError, ArgumentError] -> nil
-      end
+      iex> try do
+      ...>   1 / 0
+      ...> rescue
+      ...>   ArithmeticError -> :rescued
+      ...> end
+      :rescued
 
-      # Rescue and bind the exception to the variable "x"
-      try do
-        UndefinedModule.undefined_function
-      rescue
-        x in [UndefinedFunctionError] -> nil
-      end
+  Rescue any of the given exception without binding:
+
+      iex> try do
+      ...>   1 / 0
+      ...> rescue
+      ...>   [ArithmeticError, ArgumentError] -> :rescued
+      ...> end
+      :rescued
+
+  Rescue and bind the exception to the variable `x`:
+
+      iex> try do
+      ...>   1 / 0
+      ...> rescue
+      ...>   x in [ArithmeticError] -> [:rescued, is_exception(x)]
+      ...> end
+      [:rescued, true]
+
+  Rescue different errors with separate clauses:
+
+      iex> try do
+      ...>   1 / 0
+      ...> rescue
+      ...>   ArgumentError -> :rescued_argument_error
+      ...>   ArithmeticError -> :rescued_arithmetic_error
+      ...> end
+      :rescued_arithmetic_error
+
+  Rescue all kinds of exceptions and bind the rescued exception
+  to the variable `x`:
+
+      iex> try do
+      ...>   1 / 0
+      ...> rescue
+      ...>   x -> [:rescued, is_exception(x)]
+      ...> end
+      [:rescued, true]
 
-      # Rescue all kinds of exceptions and bind the rescued exception
-      # to the variable "x"
-      try do
-        UndefinedModule.undefined_function
-      rescue
-        x -> nil
-      end
 
   ### Erlang errors
 
   Erlang errors are transformed into Elixir ones when rescuing:
 
-      try do
-        :erlang.error(:badarg)
-      rescue
-        ArgumentError -> :ok
-      end
-      #=> :ok
+      iex> try do
+      ...>   :erlang.error(:badarg)
+      ...> rescue
+      ...>   ArgumentError -> :rescued
+      ...> end
+      :rescued
 
   The most common Erlang errors will be transformed into their
   Elixir counterpart. Those which are not will be transformed
   into the more generic `ErlangError`:
 
-      try do
-        :erlang.error(:unknown)
-      rescue
-        ErlangError -> :ok
-      end
-      #=> :ok
+      iex> try do
+      ...>   :erlang.error(:unknown)
+      ...> rescue
+      ...>   ErlangError -> :rescued
+      ...> end
+      :rescued
 
   In fact, `ErlangError` can be used to rescue any error that is
   not a proper Elixir error. For example, it can be used to rescue
   the earlier `:badarg` error too, prior to transformation:
 
-      try do
-        :erlang.error(:badarg)
-      rescue
-        ErlangError -> :ok
-      end
-      #=> :ok
+      iex> try do
+      ...>   :erlang.error(:badarg)
+      ...> rescue
+      ...>   ErlangError -> :rescued
+      ...> end
+      :rescued
 
   ## `catch` clauses
 
@@ -2124,12 +2155,13 @@ defmodule Kernel.SpecialForms do
 
   `catch` can be used to catch values thrown by `Kernel.throw/1`:
 
-      try do
-        throw(:some_value)
-      catch
-        thrown_value ->
-          IO.puts("A value was thrown: #{inspect(thrown_value)}")
-      end
+      iex> try do
+      ...>   throw(:some_value)
+      ...> catch
+      ...>   thrown_value ->
+      ...>     "Thrown value: #{inspect(thrown_value)}"
+      ...> end
+      "Thrown value: :some_value"
 
   ### Catching values of any kind
 
@@ -2137,31 +2169,33 @@ defmodule Kernel.SpecialForms do
   allows matching on both the *kind* of the caught value as well as the value
   itself:
 
-      try do
-        exit(:shutdown)
-      catch
-        :exit, value ->
-          IO.puts("Exited with value #{inspect(value)}")
-      end
+      iex> try do
+      ...>   exit(:shutdown)
+      ...> catch
+      ...>   :exit, value ->
+      ...>     "Exited with value #{inspect(value)}"
+      ...> end
+      "Exited with value :shutdown"
 
-      try do
-        exit(:shutdown)
-      catch
-        kind, value when kind in [:exit, :throw] ->
-          IO.puts("Caught exit or throw with value #{inspect(value)}")
-      end
+      iex> try do
+      ...>   exit(:shutdown)
+      ...> catch
+      ...>   kind, value when kind in [:exit, :throw] ->
+      ...>     "Caught exit or throw with value #{inspect(value)}"
+      ...> end
+      "Caught exit or throw with value :shutdown"
 
   The `catch` clause also supports `:error` alongside `:exit` and `:throw` as
   in Erlang, although this is commonly avoided in favor of `raise`/`rescue` control
   mechanisms. One reason for this is that when catching `:error`, the error is
   not automatically transformed into an Elixir error:
 
-      try do
-        :erlang.error(:badarg)
-      catch
-        :error, :badarg -> :ok
-      end
-      #=> :ok
+      iex> try do
+      ...>   :erlang.error(:badarg)
+      ...> catch
+      ...>   :error, :badarg -> :rescued
+      ...> end
+      :rescued
 
   ## `after` clauses
 
@@ -2181,95 +2215,92 @@ defmodule Kernel.SpecialForms do
       end
 
   Although `after` clauses are invoked whether or not there was an error, they do not
-  modify the return value. All of the following examples return `:return_me`:
+  modify the return value. Both of the following examples print a message to STDOUT
+  and return `:returned`:
 
       try do
-        :return_me
+        :returned
       after
-        IO.puts("I will be printed")
+        IO.puts("This message will be printed")
         :not_returned
       end
+      #=> :returned
 
       try do
         raise "boom"
       rescue
-        _ -> :return_me
+        _ -> :returned
       after
-        IO.puts("I will be printed")
+        IO.puts("This message will be printed")
         :not_returned
       end
+      #=> :returned
 
   ## `else` clauses
 
   `else` clauses allow the result of the body passed to `try/1` to be pattern
   matched on:
 
-      x = 2
-      try do
-        1 / x
-      rescue
-        ArithmeticError ->
-          :infinity
-      else
-        y when y < 1 and y > -1 ->
-          :small
-        _ ->
-          :large
-      end
+      iex> x = 2
+      ...> try do
+      ...>   1 / x
+      ...> rescue
+      ...>   ArithmeticError -> :infinity
+      ...> else
+      ...>   y when y < 1 and y > -1 -> :small
+      ...>   _ -> :large
+      ...> end
+      :small
 
   If an `else` clause is not present and no exceptions are raised,
   the result of the expression will be returned:
 
-      x = 1
-      ^x =
-        try do
-          1 / x
-        rescue
-          ArithmeticError ->
-            :infinity
-        end
+      iex> x = 5
+      iex> try do
+      ...>   1 / x
+      ...> rescue
+      ...>   ArithmeticError -> :infinity
+      ...> end
+      0.2
 
   However, when an `else` clause is present but the result of the expression
   does not match any of the patterns then an exception will be raised. This
   exception will not be caught by a `catch` or `rescue` in the same `try`:
 
-      x = 1
-      try do
-        try do
-          1 / x
-        rescue
-          # The TryClauseError cannot be rescued here:
-          TryClauseError ->
-            :error_a
-        else
-          0 ->
-            :small
-        end
-      rescue
-        # The TryClauseError is rescued here:
-        TryClauseError ->
-          :error_b
-      end
+      iex> x = 1
+      iex> try do
+      ...>   try do
+      ...>     1 / x
+      ...>   rescue
+      ...>     # The TryClauseError cannot be rescued here:
+      ...>     TryClauseError -> :error_a
+      ...>   else
+      ...>      0.5 -> :small
+      ...>   end
+      ...> rescue
+      ...>   # The TryClauseError is rescued here:
+      ...>   TryClauseError -> :error_b
+      ...> end
+      :error_b
 
   Similarly, an exception inside an `else` clause is not caught or rescued
   inside the same `try`:
 
-      try do
-        try do
-          nil
-        catch
-          # The exit(1) call below can not be caught here:
-          :exit, _ ->
-            :exit_a
-        else
-          _ ->
-            exit(1)
-        end
-      catch
-        # The exit is caught here:
-        :exit, _ ->
-          :exit_b
-      end
+      iex> x = 1
+      iex> try do
+      ...>   try do
+      ...>     1 / x
+      ...>   catch
+      ...>     # The exit(1) call below can not be caught here:
+      ...>     :exit, _ -> :exit_a
+      ...>   else
+      ...>     _ -> exit(1)
+      ...>   end
+      ...> catch
+      ...>   # The exit is caught here:
+      ...>   :exit, _ -> :exit_b
+      ...> end
+      :exit_b
 
   This means the VM no longer needs to keep the stacktrace once inside
   an `else` clause and so tail recursion is possible when using a `try`
@@ -2280,16 +2311,15 @@ defmodule Kernel.SpecialForms do
   If the `try` ends up in the `rescue` or `catch` clauses, their result
   will not fall down to `else`:
 
-      try do
-        throw(:catch_this)
-      catch
-        :throw, :catch_this ->
-          :it_was_caught
-      else
-        # :it_was_caught will not fall down to this "else" clause.
-        other ->
-          {:else, other}
-      end
+      iex> try do
+      ...>   throw(:catch_this)
+      ...> catch
+      ...>   :throw, :catch_this -> :it_was_caught
+      ...> else
+      ...>   # :it_was_caught will not fall down to this "else" clause.
+      ...>   other -> {:else, other}
+      ...> end
+      :it_was_caught
 
   ## Variable handling
 
@@ -2334,30 +2364,32 @@ defmodule Kernel.SpecialForms do
   Any new and existing messages that do not match will remain in the mailbox.
 
   ## Examples
-
-      receive do
-        {:selector, number, name} when is_integer(number) ->
-          name
-        name when is_atom(name) ->
-          name
-        _ ->
-          IO.puts(:stderr, "Unexpected message received")
-      end
+      iex> send(self(), {:selector, 5, :quantity})
+      iex> receive do
+      ...>   {:selector, number, name} when is_integer(number) ->
+      ...>     name
+      ...>   name when is_atom(name) ->
+      ...>     name
+      ...>   _ ->
+      ...>     IO.puts(:stderr, "Unexpected message received")
+      ...> end
+      :quantity
 
   An optional `after` clause can be given in case no matching message is
   received during the given timeout period, specified in milliseconds:
 
-      receive do
-        {:selector, number, name} when is_integer(number) ->
-          name
-        name when is_atom(name) ->
-          name
-        _ ->
-          IO.puts(:stderr, "Unexpected message received")
-      after
-        5000 ->
-          IO.puts(:stderr, "No message in 5 seconds")
-      end
+      iex> receive do
+      ...>   {:selector, number, name} when is_integer(number) ->
+      ...>     name
+      ...>   name when is_atom(name) ->
+      ...>     name
+      ...>   _ ->
+      ...>     IO.puts(:stderr, "Unexpected message received")
+      ...> after
+      ...>   10 ->
+      ...>     "No message in 10 milliseconds"
+      ...> end
+      "No message in 10 milliseconds"
 
   The `after` clause can be specified even if there are no match clauses.
   The timeout value given to `after` can be any expression evaluating to
