@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Kernel.Typespec do
   @moduledoc false
 
@@ -493,6 +497,24 @@ defmodule Kernel.Typespec do
      state}
   end
 
+  defp typespec(
+         {:<<>>, meta, [{:"::", _, [{:_, _, ctx1}, {:*, prod_meta, [size, unit]}]}]},
+         _,
+         _,
+         state
+       )
+       when is_atom(ctx1) and is_integer(size) and size >= 0 and unit in 1..256 do
+    location = location(meta)
+    prod_location = location(prod_meta)
+
+    {{:type, location, :binary,
+      [
+        {:op, prod_location, :*, {:integer, prod_location, size},
+         {:integer, prod_location, unit}},
+        {:integer, location, 0}
+      ]}, state}
+  end
+
   defp typespec({:<<>>, meta, [{:"::", size_meta, [{:_, _, ctx}, size]}]}, _, _, state)
        when is_atom(ctx) and is_integer(size) and size >= 0 do
     location = location(meta)
@@ -608,12 +630,12 @@ defmodule Kernel.Typespec do
     typespec({:record, meta, [atom, []]}, vars, caller, state)
   end
 
-  defp typespec({:record, meta, [tag, field_specs]}, vars, caller, state)
-       when is_atom(tag) and is_list(field_specs) do
+  defp typespec({:record, meta, [name, field_specs]}, vars, caller, state)
+       when is_atom(name) and is_list(field_specs) do
     # We cannot set a function name to avoid tracking
     # as a compile time dependency because for records it actually is one.
-    case Macro.expand({tag, [], [{:{}, [], []}]}, caller) do
-      {_, _, [name, fields | _]} when is_list(fields) ->
+    case Macro.expand({name, [], [{:{}, [], []}]}, caller) do
+      {_, _, [tag, fields | _]} when is_list(fields) ->
         types =
           :lists.map(
             fn {field, _} ->
@@ -628,19 +650,19 @@ defmodule Kernel.Typespec do
 
         fun = fn {field, _} ->
           if not Keyword.has_key?(fields, field) do
-            compile_error(caller, "undefined field #{field} on record #{inspect(tag)}")
+            compile_error(caller, "undefined field #{field} on record #{inspect(name)}")
           end
         end
 
         :lists.foreach(fun, field_specs)
-        typespec({:{}, meta, [name | types]}, vars, caller, state)
+        typespec({:{}, meta, [tag | types]}, vars, caller, state)
 
       _ ->
-        compile_error(caller, "unknown record #{inspect(tag)}")
+        compile_error(caller, "unknown record #{inspect(name)}")
     end
   end
 
-  defp typespec({:record, _meta, [_tag, _field_specs]}, _vars, caller, _state) do
+  defp typespec({:record, _meta, [_name, _field_specs]}, _vars, caller, _state) do
     message = "invalid record specification, expected the record name to be an atom literal"
     compile_error(caller, message)
   end
@@ -855,7 +877,15 @@ defmodule Kernel.Typespec do
 
   defp typespec({:fun, meta, args}, vars, caller, state) do
     {args, state} = :lists.mapfoldl(&typespec(&1, vars, caller, &2), state, args)
-    {{:type, location(meta), :fun, args}, state}
+
+    if args != [] do
+      IO.warn(
+        "fun/#{length(args)} is not valid in typespecs. Either specify fun() or use (... -> return) instead",
+        caller
+      )
+    end
+
+    {{:type, location(meta), :fun, []}, state}
   end
 
   defp typespec({:..., _meta, _args}, _vars, caller, _state) do
