@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Module.ParallelChecker do
   @moduledoc false
 
@@ -5,13 +9,24 @@ defmodule Module.ParallelChecker do
 
   @type cache() :: {pid(), :ets.tid()}
   @type warning() :: term()
-  @type mode() :: :elixir | :erlang
+  @type mode() :: :erlang | :elixir | :protocol
+
+  @typedoc """
+  Options for `start_link/1`.
+  """
+  @type start_link_opts :: [
+          {:max_concurrency, pos_integer()}
+          | {:long_verification_threshold, pos_integer()}
+          | {:each_long_verification, (module() -> term()) | (module(), pid() -> term())}
+          | {atom(), term()}
+        ]
 
   @doc """
   Initializes the parallel checker process.
   """
-  def start_link(schedulers \\ nil) do
-    :proc_lib.start_link(__MODULE__, :init, [schedulers])
+  @spec start_link(start_link_opts()) :: {:ok, cache()}
+  def start_link(opts \\ []) do
+    :proc_lib.start_link(__MODULE__, :init, [opts])
   end
 
   @doc """
@@ -47,8 +62,17 @@ defmodule Module.ParallelChecker do
   @doc """
   Spawns a process that runs the parallel checker.
   """
-  def spawn({pid, {checker, table}}, module, module_map, log?) do
-    inner_spawn(pid, checker, table, module, cache_from_module_map(table, module_map), log?)
+  def spawn({pid, {checker, table}}, module, module_map, beam_location, log?) do
+    # Protocols may have been consolidated. So if we know their beam location,
+    # we discard their module map on purpose and start from file.
+    info =
+      if beam_location != [] and Keyword.has_key?(module_map.attributes, :__protocol__) do
+        List.to_string(beam_location)
+      else
+        cache_from_module_map(table, module_map)
+      end
+
+    inner_spawn(pid, checker, table, module, info, log?)
   end
 
   defp inner_spawn(pid, checker, table, module, info, log?) do
@@ -62,22 +86,30 @@ defmodule Module.ParallelChecker do
           {^ref, :cache} ->
             Process.link(pid)
 
-            module_tuple =
+            {mode, module_tuple} =
               cond do
-                is_tuple(info) ->
-                  info
-
                 is_binary(info) ->
-                  with {:ok, binary} <- File.read(info),
+                  location =
+                    case :code.which(module) do
+                      [_ | _] = path -> path
+                      _ -> info
+                    end
+
+                  with {:ok, binary} <- File.read(location),
                        {:ok, {_, [debug_info: chunk]}} <- :beam_lib.chunks(binary, [:debug_info]),
                        {:debug_info_v1, backend, data} = chunk,
                        {:ok, module_map} <- backend.debug_info(:elixir_v1, module, data, []) do
                     cache_from_module_map(table, module_map)
                   else
-                    _ -> nil
+                    _ -> {:not_found, nil}
                   end
+
+                is_tuple(info) ->
+                  info
               end
 
+            # We only make the module available now, so they are not visible during inference
+            :ets.insert(table, {module, mode})
             send(checker, {ref, :cached})
 
             receive do
@@ -93,7 +125,7 @@ defmodule Module.ParallelChecker do
                   end
 
                 send(pid, {__MODULE__, module, warnings})
-                send(checker, {__MODULE__, :done})
+                send(checker, {__MODULE__, :done, module})
             end
 
           {:DOWN, ^mon_ref, _, _, _} ->
@@ -101,7 +133,7 @@ defmodule Module.ParallelChecker do
         end
       end)
 
-    register(checker, spawned, ref)
+    register(checker, module, spawned, ref)
     :ok
   end
 
@@ -195,25 +227,34 @@ defmodule Module.ParallelChecker do
 
   @doc """
   Returns the export kind and deprecation reason for the given MFA from
-  the cache. If the module does not exist return `{:error, :module}`,
-  or if the function does not exist return `{:error, :function}`.
+  the cache. If the module does not exist return `:badmodule`,
+  or if the function does not exist return `{:badfunction, mode}`.
   """
-  @spec fetch_export(cache(), module(), atom(), arity()) ::
-          {:ok, mode(), binary() | nil, {:infer, [term()]} | :none}
-          | {:error, :function | :module}
-  def fetch_export({checker, table}, module, fun, arity) do
+  @spec fetch_export(cache(), module(), atom(), arity(), boolean()) ::
+          {:ok, mode(), binary() | nil, {:infer, [term()] | nil, [term()]} | :none}
+          | :badmodule
+          | {:badfunction, mode()}
+  def fetch_export({checker, table}, module, fun, arity, force?) do
     case :ets.lookup(table, module) do
       [] ->
-        cache_module({checker, table}, module)
-        fetch_export({checker, table}, module, fun, arity)
+        if force? do
+          cache_module({checker, table}, module)
+          fetch_export({checker, table}, module, fun, arity, false)
+        else
+          :badmodule
+        end
 
-      [{_key, false}] ->
-        {:error, :module}
+      [{_key, :uncached}] ->
+        cache_module({checker, table}, module)
+        fetch_export({checker, table}, module, fun, arity, false)
+
+      [{_key, :not_found}] ->
+        :badmodule
 
       [{_key, mode}] ->
         case :ets.lookup(table, {module, {fun, arity}}) do
           [{_key, reason, signature}] -> {:ok, mode, reason, signature}
-          [] -> {:error, :function}
+          [] -> {:badfunction, mode}
         end
     end
   end
@@ -221,7 +262,7 @@ defmodule Module.ParallelChecker do
   ## Module checking
 
   defp check_module(module_tuple, cache, log?) do
-    {module, file, line, definitions, no_warn_undefined, behaviours, impls, after_verify} =
+    {module, file, line, definitions, no_warn_undefined, behaviours, impls, attrs, after_verify} =
       module_tuple
 
     behaviour_warnings =
@@ -236,10 +277,10 @@ defmodule Module.ParallelChecker do
 
     diagnostics =
       module
-      |> Module.Types.warnings(file, definitions, no_warn_undefined, cache)
+      |> Module.Types.warnings(file, attrs, definitions, no_warn_undefined, cache)
       |> Kernel.++(behaviour_warnings)
       |> group_warnings()
-      |> emit_warnings(log?)
+      |> emit_warnings(file, log?)
 
     Enum.each(after_verify, fn {verify_mod, verify_fun} ->
       apply(verify_mod, verify_fun, [module])
@@ -273,7 +314,10 @@ defmodule Module.ParallelChecker do
       |> extract_no_warn_undefined()
       |> merge_compiler_no_warn_undefined()
 
-    {module, file, line, definitions, no_warn_undefined, behaviours, impls, after_verify}
+    attributes = Keyword.take(attributes, [:__protocol__, :__impl__])
+
+    {module, file, line, definitions, no_warn_undefined, behaviours, impls, attributes,
+     after_verify}
   end
 
   defp extract_no_warn_undefined(compile_opts) do
@@ -320,9 +364,9 @@ defmodule Module.ParallelChecker do
     Enum.sort(ungrouped ++ grouped)
   end
 
-  defp emit_warnings(warnings, log?) do
+  defp emit_warnings(warnings, file, log?) do
     Enum.flat_map(warnings, fn {locations, diagnostic} ->
-      diagnostics = Enum.map(locations, &to_diagnostic(diagnostic, &1))
+      diagnostics = Enum.map(locations, &to_diagnostic(diagnostic, file, &1))
       log? and print_diagnostics(diagnostics)
       diagnostics
     end)
@@ -336,10 +380,12 @@ defmodule Module.ParallelChecker do
     :elixir_errors.print_diagnostics(diagnostics)
   end
 
-  defp to_diagnostic(diagnostic, {file, position, mfa}) when is_list(position) do
+  defp to_diagnostic(diagnostic, source, {file, position, mfa}) when is_list(position) do
+    file = Path.absname(file)
+
     %{
       severity: :warning,
-      source: file,
+      source: source,
       file: file,
       position: position_to_tuple(position),
       stacktrace: [to_stacktrace(file, position, mfa)],
@@ -374,40 +420,48 @@ defmodule Module.ParallelChecker do
     if lock(checker, module) do
       object_code = :code.get_object_code(module)
 
-      # The chunk has more information, so that's our preference
-      with {^module, binary, _filename} <- object_code,
-           {:ok, {^module, [{~c"ExCk", chunk}]}} <- :beam_lib.chunks(binary, [~c"ExCk"]),
-           {:elixir_checker_v1, contents} <- :erlang.binary_to_term(chunk) do
-        cache_chunk(table, module, contents.exports)
-      else
-        _ ->
-          # Otherwise, if the module is loaded, use its info
-          case :erlang.module_loaded(module) do
-            true ->
-              {mode, exports} = info_exports(module)
-              deprecated = info_deprecated(module)
-              cache_info(table, module, exports, deprecated, %{}, mode)
+      mode =
+        with {^module, binary, _filename} <- object_code,
+             {:ok, {^module, [{~c"ExCk", chunk}]}} <- :beam_lib.chunks(binary, [~c"ExCk"]),
+             {:elixir_checker_v3, contents} <- :erlang.binary_to_term(chunk) do
+          # The chunk has more information, so that's our preference
+          cache_chunk(table, module, contents)
+        else
+          _ ->
+            # Otherwise, if the module is loaded, use its info
+            case :erlang.module_loaded(module) do
+              true ->
+                {mode, exports} = info_exports(module)
+                deprecated = info_deprecated(module)
+                cache_info(table, module, exports, deprecated, %{})
+                mode
 
-            false ->
-              # Or load exports from chunk
-              with {^module, binary, _filename} <- object_code,
-                   {:ok, {^module, [exports: exports]}} <- :beam_lib.chunks(binary, [:exports]) do
-                cache_info(table, module, exports, %{}, %{}, :erlang)
-              else
-                _ ->
-                  :ets.insert(table, {module, false})
-              end
-          end
-      end
+              false ->
+                # Or load exports from chunk
+                with {^module, binary, _filename} <- object_code,
+                     {:ok, {^module, [exports: exports]}} <- :beam_lib.chunks(binary, [:exports]) do
+                  cache_info(table, module, exports, %{}, %{})
+                  :erlang
+                else
+                  _ -> :not_found
+                end
+            end
+        end
 
-      unlock(checker, module)
+      :ets.insert(table, {module, mode})
+      unlock(checker, module, mode)
     end
   end
 
   defp info_exports(module) do
-    {:elixir, behaviour_exports(module) ++ [{:__info__, 1} | module.__info__(:functions)]}
-  rescue
-    _ -> {:erlang, module.module_info(:exports)}
+    try do
+      module.__info__(:functions)
+    rescue
+      _ -> {:erlang, module.module_info(:exports)}
+    else
+      functions ->
+        {elixir_mode(module.module_info(:attributes)), behaviour_exports(module) ++ functions}
+    end
   end
 
   defp info_deprecated(module) do
@@ -416,36 +470,41 @@ defmodule Module.ParallelChecker do
     _ -> %{}
   end
 
-  defp cache_from_module_map(table, map) do
-    exports =
-      [{:__info__, 1}] ++
-        behaviour_exports(map) ++
-        for({function, :def, _meta, _clauses} <- map.definitions, do: function)
-
-    cache_info(table, map.module, exports, Map.new(map.deprecated), map.signatures, :elixir)
-    module_map_to_module_tuple(map)
+  defp elixir_mode(attributes) do
+    if Keyword.has_key?(attributes, :__protocol__), do: :protocol, else: :elixir
   end
 
-  defp cache_info(table, module, exports, deprecated, sigs, mode) do
+  defp cache_from_module_map(table, map) do
+    exports =
+      behaviour_exports(map) ++
+        for({function, :def, _meta, _clauses} <- map.definitions, do: function)
+
+    cache_info(table, map.module, exports, Map.new(map.deprecated), map.signatures)
+    {elixir_mode(map.attributes), module_map_to_module_tuple(map)}
+  end
+
+  defp cache_info(table, module, exports, deprecated, sigs) do
     Enum.each(exports, fn fa ->
       reason = Map.get(deprecated, fa)
       :ets.insert(table, {{module, fa}, reason, Map.get(sigs, fa, :none)})
     end)
-
-    :ets.insert(table, {module, mode})
   end
 
-  defp cache_chunk(table, module, exports) do
-    Enum.each(exports, fn {{fun, arity}, info} ->
-      # TODO: Match on signature directly in Elixir v1.22+
+  defp cache_chunk(table, module, contents) do
+    Enum.each(contents.exports, fn {{fun, arity}, info} ->
+      sig =
+        case info do
+          %{sig: {key, _, _} = sig} when key in [:infer, :strong] -> sig
+          _ -> :none
+        end
+
       :ets.insert(
         table,
-        {{module, {fun, arity}}, Map.get(info, :deprecated), Map.get(info, :sig, :none)}
+        {{module, {fun, arity}}, Map.get(info, :deprecated), sig}
       )
     end)
 
-    :ets.insert(table, {{module, {:__info__, 1}}, nil, :none})
-    :ets.insert(table, {module, :elixir})
+    Map.get(contents, :mode, :elixir)
   end
 
   defp behaviour_exports(%{defines_behaviour: true}), do: [{:behaviour_info, 1}]
@@ -463,42 +522,80 @@ defmodule Module.ParallelChecker do
     :gen_server.call(server, {:lock, module}, :infinity)
   end
 
-  defp unlock(server, module) do
-    :gen_server.call(server, {:unlock, module}, :infinity)
+  defp unlock(server, module, mode) do
+    :gen_server.call(server, {:unlock, module, mode}, :infinity)
   end
 
-  defp register(server, pid, ref) do
-    :gen_server.cast(server, {:register, pid, ref})
+  defp register(server, module, pid, ref) do
+    :gen_server.cast(server, {:register, module, pid, ref})
   end
 
   ## Server callbacks
 
-  def init(schedulers) do
+  def init(options) do
     table = :ets.new(__MODULE__, [:set, :public, {:read_concurrency, true}])
     :proc_lib.init_ack({:ok, {self(), table}})
 
+    case :elixir_config.get(:infer_signatures) do
+      false ->
+        false
+
+      applications ->
+        for application <- applications do
+          case :application.get_key(application, :modules) do
+            {:ok, modules} ->
+              :ets.insert(table, Enum.map(modules, &{&1, :uncached}))
+
+            :undefined ->
+              IO.warn(
+                "cannot infer signatures from #{inspect(application)} because it is not loaded",
+                []
+              )
+          end
+        end
+    end
+
+    schedulers =
+      Keyword.get_lazy(options, :max_concurrency, fn ->
+        max(:erlang.system_info(:schedulers_online), 2)
+      end)
+
+    threshold = Keyword.get(options, :long_verification_threshold, 10) * 1000
+
+    callback =
+      case Keyword.get(options, :each_long_verification, fn _module, _pid -> :ok end) do
+        fun when is_function(fun, 1) -> fn module, _pid -> fun.(module) end
+        fun when is_function(fun, 2) -> fun
+      end
+
     state = %{
       waiting: %{},
       modules: [],
-      spawned: 0,
-      schedulers: schedulers || max(:erlang.system_info(:schedulers_online), 2)
+      spawned: %{},
+      schedulers: schedulers,
+      threshold: threshold,
+      callback: callback,
+      protocols: [],
+      table: table
     }
 
     :gen_server.enter_loop(__MODULE__, [], state)
   end
 
-  def handle_call(:start, _from, %{modules: modules} = state) do
-    for {pid, ref} <- modules do
+  def handle_call(:start, _from, %{modules: modules, protocols: protocols, table: table} = state) do
+    :ets.insert(table, Enum.map(protocols, &{&1, :uncached}))
+
+    for {_module, pid, ref} <- modules do
       send(pid, {ref, :cache})
     end
 
-    for {_pid, ref} <- modules do
+    for {_module, _pid, ref} <- modules do
       receive do
         {^ref, :cached} -> :ok
       end
     end
 
-    {:reply, length(modules), run_checkers(state)}
+    {:reply, length(modules), run_checkers(%{state | protocols: []})}
   end
 
   def handle_call({:lock, module}, from, %{waiting: waiting} = state) do
@@ -513,24 +610,35 @@ defmodule Module.ParallelChecker do
     end
   end
 
-  def handle_call({:unlock, module}, _from, %{waiting: waiting} = state) do
+  def handle_call({:unlock, module, mode}, _from, state) do
+    %{waiting: waiting, protocols: protocols} = state
     from_list = Map.fetch!(waiting, module)
     Enum.each(from_list, &:gen_server.reply(&1, false))
+
     waiting = Map.delete(waiting, module)
-    {:reply, :ok, %{state | waiting: waiting}}
+    protocols = if mode == :protocol, do: [module | protocols], else: protocols
+    {:reply, :ok, %{state | waiting: waiting, protocols: protocols}}
   end
 
-  def handle_info({__MODULE__, :done}, state) do
-    state = %{state | spawned: state.spawned - 1}
-    {:noreply, run_checkers(state)}
+  def handle_info({__MODULE__, :timeout, module, pid}, state) do
+    state.callback.(module, pid)
+    {:noreply, state}
+  end
+
+  def handle_info({__MODULE__, :done, module}, state) do
+    # Unfortunately we cannot assume uniqueness because the same module
+    # may be defined by mistake several times
+    {timer, spawned} = Map.pop(state.spawned, module)
+    timer && Process.cancel_timer(timer)
+    {:noreply, run_checkers(%{state | spawned: spawned})}
   end
 
   def handle_info({__MODULE__, :stop}, state) do
     {:stop, :normal, state}
   end
 
-  def handle_cast({:register, pid, ref}, %{modules: modules} = state) do
-    {:noreply, %{state | modules: [{pid, ref} | modules]}}
+  def handle_cast({:register, module, pid, ref}, %{modules: modules} = state) do
+    {:noreply, %{state | modules: [{module, pid, ref} | modules]}}
   end
 
   defp run_checkers(%{modules: []} = state) do
@@ -538,12 +646,14 @@ defmodule Module.ParallelChecker do
   end
 
   defp run_checkers(%{spawned: spawned, schedulers: schedulers} = state)
-       when spawned >= schedulers do
+       when map_size(spawned) >= schedulers do
     state
   end
 
-  defp run_checkers(%{modules: [{pid, ref} | modules]} = state) do
+  defp run_checkers(%{modules: [{module, pid, ref} | modules]} = state) do
     send(pid, {ref, :check})
-    run_checkers(%{state | modules: modules, spawned: state.spawned + 1})
+    timer = Process.send_after(self(), {__MODULE__, :timeout, module, pid}, state.threshold)
+    spawned = Map.put(state.spawned, module, timer)
+    run_checkers(%{state | modules: modules, spawned: spawned})
   end
 end
