@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Module.Types do
   @moduledoc false
   alias Module.Types.{Descr, Expr, Pattern, Helpers}
@@ -30,17 +34,26 @@ defmodule Module.Types do
   @modes [:static, :dynamic, :infer, :traversal]
 
   # These functions are not inferred because they are added/managed by the compiler
-  @no_infer [__protocol__: 1, behaviour_info: 1]
+  @no_infer [behaviour_info: 1]
 
   @doc false
-  def infer(module, file, defs, private, used_private, env, {_, cache}) do
-    infer_signatures? = :elixir_config.get(:infer_signatures) and cache != nil
+  def infer(module, file, attrs, defs, private, used_private, env, {_, cache}) do
+    # We don't care about inferring signatures for protocols,
+    # those will be replaced anyway. There is also nothing to
+    # infer if there is no cache system, we only do traversals.
+    infer_signatures? =
+      :elixir_config.get(:infer_signatures) != false and cache != nil and not protocol?(attrs)
+
+    impl = impl_for(attrs)
 
     finder =
       fn fun_arity ->
         case :lists.keyfind(fun_arity, 1, defs) do
-          {_, kind, _, _} = clause -> {infer_mode(kind, infer_signatures?), clause}
-          false -> false
+          {_, kind, _, _} = def ->
+            default_domain(infer_mode(kind, infer_signatures?), def, fun_arity, impl)
+
+          false ->
+            false
         end
       end
 
@@ -67,7 +80,11 @@ defmodule Module.Types do
           kind in [:def, :defmacro],
           reduce: {[], context()} do
         {types, context} ->
-          finder = fn _ -> {infer_mode(kind, infer_signatures?), def} end
+          # Optimized version of finder, since we already the definition
+          finder = fn _ ->
+            default_domain(infer_mode(kind, infer_signatures?), def, fun_arity, impl)
+          end
+
           {_kind, inferred, context} = local_handler(meta, fun_arity, stack, context, finder)
 
           if infer_signatures? and kind == :def and fun_arity not in @no_infer do
@@ -111,6 +128,45 @@ defmodule Module.Types do
     if infer_signatures? and kind in [:def, :defp], do: :infer, else: :traversal
   end
 
+  defp protocol?(attrs) do
+    List.keymember?(attrs, :__protocol__, 0)
+  end
+
+  defp impl_for(attrs) do
+    case List.keyfind(attrs, :__impl__, 0) do
+      {:__impl__, [protocol: protocol, for: for]} ->
+        if Code.ensure_loaded?(protocol) and function_exported?(protocol, :behaviour_info, 1) do
+          {for, protocol.behaviour_info(:callbacks)}
+        else
+          nil
+        end
+
+      _ ->
+        nil
+    end
+  end
+
+  defp default_domain(mode, def, {_, arity} = fun_arity, impl) do
+    with {for, callbacks} <- impl,
+         true <- fun_arity in callbacks do
+      args = [
+        Descr.dynamic(Module.Types.Of.impl(for))
+        | List.duplicate(Descr.dynamic(), arity - 1)
+      ]
+
+      {fun_arity, kind, meta, clauses} = def
+
+      clauses =
+        for {meta, args, guards, body} <- clauses do
+          {[type_check: {:impl, for}] ++ meta, args, guards, body}
+        end
+
+      {mode, {fun_arity, kind, meta, clauses}, args}
+    else
+      _ -> {mode, def, List.duplicate(Descr.dynamic(), arity)}
+    end
+  end
+
   defp undefined_function!(reason, meta, {fun, arity}, stack, env) do
     env = %{env | function: stack.function, file: stack.file}
     tuple = {reason, {fun, arity}, stack.module}
@@ -159,10 +215,12 @@ defmodule Module.Types do
   end
 
   @doc false
-  def warnings(module, file, defs, no_warn_undefined, cache) do
+  def warnings(module, file, attrs, defs, no_warn_undefined, cache) do
+    impl = impl_for(attrs)
+
     finder = fn fun_arity ->
       case :lists.keyfind(fun_arity, 1, defs) do
-        {_, _, _, _} = clause -> {:dynamic, clause}
+        {_, _, _, _} = def -> default_domain(:dynamic, def, fun_arity, impl)
         false -> false
       end
     end
@@ -172,7 +230,8 @@ defmodule Module.Types do
 
     context =
       Enum.reduce(defs, context(), fn {fun_arity, _kind, meta, _clauses} = def, context ->
-        finder = fn _ -> {:dynamic, def} end
+        # Optimized version of finder, since we already the definition
+        finder = fn _ -> default_domain(:dynamic, def, fun_arity, impl) end
         {_kind, _inferred, context} = local_handler(meta, fun_arity, stack, context, finder)
         context
       end)
@@ -188,10 +247,13 @@ defmodule Module.Types do
         not Keyword.get(meta, :from_super, false),
         reduce: context do
       context ->
-        {_kind, _inferred, mapping} = Map.fetch!(context.local_sigs, fun_arity)
+        {_kind, info, mapping} = Map.fetch!(context.local_sigs, fun_arity)
 
         clauses_indexes =
-          for type_index <- pending, {clause_index, ^type_index} <- mapping, do: clause_index
+          for type_index <- pending,
+              not skip_unused_clause?(info, type_index),
+              {clause_index, ^type_index} <- mapping,
+              do: clause_index
 
         Enum.reduce(clauses_indexes, context, fn clause_index, context ->
           {meta, _args, _guards, _body} = Enum.fetch!(clauses, clause_index)
@@ -201,6 +263,20 @@ defmodule Module.Types do
     end
   end
 
+  defp skip_unused_clause?(info, type_index) do
+    case info do
+      # If an inferred clause returns an empty type, then the reverse arrow
+      # will never propagate its domain up, which may lead to the clause never
+      # being invoked.
+      {:infer, _, inferred} ->
+        {_args_types, return} = Enum.fetch!(inferred, type_index)
+        Descr.empty?(return)
+
+      _ ->
+        false
+    end
+  end
+
   defp local_handler(_meta, fun_arity, stack, context, finder) do
     case context.local_sigs do
       %{^fun_arity => {kind, inferred, _mapping}} ->
@@ -211,11 +287,11 @@ defmodule Module.Types do
 
       local_sigs ->
         case finder.(fun_arity) do
-          {mode, {fun_arity, kind, meta, clauses}} ->
+          {mode, {fun_arity, kind, meta, clauses}, expected} ->
             context = put_in(context.local_sigs, Map.put(local_sigs, fun_arity, kind))
 
             {inferred, mapping, context} =
-              local_handler(fun_arity, kind, meta, clauses, mode, stack, context)
+              local_handler(fun_arity, kind, meta, clauses, expected, mode, stack, context)
 
             context =
               update_in(context.local_sigs, &Map.put(&1, fun_arity, {kind, inferred, mapping}))
@@ -228,8 +304,8 @@ defmodule Module.Types do
     end
   end
 
-  defp local_handler({fun, arity} = fun_arity, kind, meta, clauses, mode, stack, context) do
-    expected = List.duplicate(Descr.dynamic(), arity)
+  defp local_handler(fun_arity, kind, meta, clauses, expected, mode, stack, context) do
+    {fun, _arity} = fun_arity
     stack = stack |> fresh_stack(mode, fun_arity) |> with_file_meta(meta)
 
     {_, _, mapping, clauses_types, clauses_context} =
@@ -238,11 +314,18 @@ defmodule Module.Types do
           context = fresh_context(context)
 
           try do
-            {args_types, context} =
-              Pattern.of_head(args, guards, expected, :default, meta, stack, context)
+            {trees, context} =
+              Pattern.of_head(args, guards, expected, {:infer, expected}, meta, stack, context)
 
             {return_type, context} =
-              Expr.of_expr(body, stack, context)
+              Expr.of_expr(body, Descr.term(), body, stack, context)
+
+            args_types =
+              if stack.mode == :traversal do
+                expected
+              else
+                Pattern.of_domain(trees, expected, context)
+              end
 
             {type_index, inferred} =
               add_inferred(inferred, args_types, return_type, total - 1, [])
@@ -258,8 +341,19 @@ defmodule Module.Types do
           end
       end)
 
-    inferred = {:infer, Enum.reverse(clauses_types)}
-    {inferred, mapping, restore_context(context, clauses_context)}
+    domain =
+      case clauses_types do
+        [_] ->
+          nil
+
+        _ ->
+          clauses_types
+          |> Enum.map(fn {args, _} -> args end)
+          |> Enum.zip_with(fn types -> Enum.reduce(types, &Descr.union/2) end)
+      end
+
+    inferred = {:infer, domain, Enum.reverse(clauses_types)}
+    {inferred, mapping, restore_context(clauses_context, context)}
   end
 
   # We check for term equality of types as an optimization
@@ -319,14 +413,25 @@ defmodule Module.Types do
       module: module,
       # Current function
       function: function,
-      # List of calls to not warn on as undefined or :all or Macro.Env indicating limited remotes
+      # Handling of undefined remote calls. May be one of:
+      #
+      # * List of calls to not warn on as undefined
+      #
+      # * The atom `:all` to not mark anything as undefined
+      #
+      # * A Macro.Env struct to not mark anything as undefined
+      #   also used to extract structs from
+      #
       no_warn_undefined: no_warn_undefined,
       # A tuple with cache information (may be nil)
       cache: cache,
       # The mode to be used, see the @modes attribute
       mode: mode,
       # The function for handling local calls
-      local_handler: handler
+      local_handler: handler,
+      # Control if variable refinement is enabled.
+      # It is disabled only on dynamic dispatches.
+      refine_vars: false
     }
   end
 
@@ -356,7 +461,7 @@ defmodule Module.Types do
     %{context | vars: %{}, failed: false}
   end
 
-  defp restore_context(%{vars: vars, failed: failed}, later_context) do
+  defp restore_context(later_context, %{vars: vars, failed: failed}) do
     %{later_context | vars: vars, failed: failed}
   end
 
@@ -371,14 +476,16 @@ defmodule Module.Types do
   ## Module errors
 
   def format_error({:unused_args, {name, arity}}),
-    do: "default values for the optional arguments in #{name}/#{arity} are never used"
+    do:
+      "default values for the optional arguments in the private function #{name}/#{arity} are never used"
 
   def format_error({:unused_args, {name, arity}, count}) when arity - count == 1,
-    do: "the default value for the last optional argument in #{name}/#{arity} is never used"
+    do:
+      "the default value for the last optional argument in the private function #{name}/#{arity} is never used"
 
   def format_error({:unused_args, {name, arity}, count}),
     do:
-      "the default values for the last #{arity - count} optional arguments in #{name}/#{arity} are never used"
+      "the default values for the last #{arity - count} optional arguments in the private function #{name}/#{arity} are never used"
 
   def format_error({:unused_def, {name, arity}, :defp}),
     do: "function #{name}/#{arity} is unused"
