@@ -1,3 +1,6 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+
 defmodule PartitionSupervisor do
   @moduledoc """
   A supervisor that starts multiple partitions of the same child.
@@ -96,7 +99,7 @@ defmodule PartitionSupervisor do
 
       DynamicSupervisor.start_child(MyApp.DynamicSupervisor, {Agent, fn -> %{} end})
 
-  You can do start the dynamic supervisors under a `PartitionSupervisor`:
+  You can start the dynamic supervisors under a `PartitionSupervisor`:
 
       children = [
         {PartitionSupervisor,
@@ -148,10 +151,29 @@ defmodule PartitionSupervisor do
   @typedoc since: "1.14.0"
   @type name :: atom() | {:via, module(), term()}
 
+  @typedoc """
+  The "identifier" of a partition.
+  """
+  @typedoc since: "1.19.0"
+  @type partition() :: non_neg_integer()
+
+  @typedoc """
+  The possible options to give to `start_link/0`.
+  """
+  @typedoc since: "1.19.0"
+  @type start_link_option ::
+          {:name, name}
+          | {:child_spec, Supervisor.child_spec() | Supervisor.module_spec()}
+          | {:partitions, pos_integer()}
+          | {:strategy, Supervisor.strategy()}
+          | {:max_restarts, non_neg_integer()}
+          | {:max_seconds, non_neg_integer()}
+          | {:with_arguments, (args :: [term()], partition() -> updated_args :: [term()])}
+
   @doc false
   def child_spec(opts) when is_list(opts) do
     id =
-      case Keyword.get(opts, :name, DynamicSupervisor) do
+      case Keyword.get(opts, :name, PartitionSupervisor) do
         name when is_atom(name) -> name
         {:via, _module, name} -> name
       end
@@ -185,16 +207,19 @@ defmodule PartitionSupervisor do
 
   ## Options
 
+  See `t:start_link_option/0` for the type of each option.
+
     * `:name` - an atom or via tuple representing the name of the partition
-      supervisor (see `t:name/0`).
+      supervisor. *Required*.
 
-    * `:child_spec` - the child spec to be used when starting the partitions.
+    * `:child_spec` - the child spec to be used when starting the partitions. *Required*.
 
-    * `:partitions` - a positive integer with the number of partitions.
-      Defaults to `System.schedulers_online()` (typically the number of cores).
+    * `:partitions` - the number of partitions.
+      Defaults to `System.schedulers_online/0` (typically the number of cores).
 
-    * `:strategy` - the restart strategy option, defaults to `:one_for_one`.
+    * `:strategy` - the restart strategy option.
       You can learn more about strategies in the `Supervisor` module docs.
+      Defaults to `:one_for_one`.
 
     * `:max_restarts` - the maximum number of restarts allowed in
       a time frame. Defaults to `3`.
@@ -203,7 +228,9 @@ defmodule PartitionSupervisor do
       Defaults to `5`.
 
     * `:with_arguments` - a two-argument anonymous function that allows
-      the partition to be given to the child starting function. See the
+      the partition to be given to the child starting function. It takes the list of arguments
+      passed to the child start function and the partition itself, and must return
+      possibly-updated arguments to give to the child start function. See the
       `:with_arguments` section below.
 
   ## `:with_arguments`
@@ -224,30 +251,30 @@ defmodule PartitionSupervisor do
 
   """
   @doc since: "1.14.0"
-  @spec start_link(keyword) :: Supervisor.on_start()
+  @spec start_link([start_link_option()]) :: Supervisor.on_start()
   def start_link(opts) when is_list(opts) do
     name = opts[:name]
 
-    unless name do
+    if !name do
       raise ArgumentError, "the :name option must be given to PartitionSupervisor"
     end
 
     {child_spec, opts} = Keyword.pop(opts, :child_spec)
 
-    unless child_spec do
+    if !child_spec do
       raise ArgumentError, "the :child_spec option must be given to PartitionSupervisor"
     end
 
     {partitions, opts} = Keyword.pop(opts, :partitions, System.schedulers_online())
 
-    unless is_integer(partitions) and partitions >= 1 do
+    if not (is_integer(partitions) and partitions >= 1) do
       raise ArgumentError,
             "the :partitions option must be a positive integer, got: #{inspect(partitions)}"
     end
 
     {with_arguments, opts} = Keyword.pop(opts, :with_arguments, fn args, _partition -> args end)
 
-    unless is_function(with_arguments, 2) do
+    if not is_function(with_arguments, 2) do
       raise ArgumentError,
             "the :with_arguments option must be a function that receives two arguments, " <>
               "the current call arguments and the partition, got: #{inspect(with_arguments)}"
@@ -260,17 +287,17 @@ defmodule PartitionSupervisor do
       for partition <- 0..(partitions - 1) do
         args = with_arguments.(args, partition)
 
-        unless is_list(args) do
+        if not is_list(args) do
           raise "the call to the function in :with_arguments must return a list, got: #{inspect(args)}"
         end
 
-        start = {__MODULE__, :start_child, [mod, fun, args, name, partition]}
+        start = {__MODULE__, :start_child, [mod, fun, args, partition]}
         Map.merge(map, %{id: partition, start: start, modules: modules})
       end
 
     auto_shutdown = Keyword.get(opts, :auto_shutdown, :never)
 
-    unless auto_shutdown == :never do
+    if auto_shutdown != :never do
       raise ArgumentError,
             "the :auto_shutdown option must be :never, got: #{inspect(auto_shutdown)}"
     end
@@ -282,14 +309,14 @@ defmodule PartitionSupervisor do
   end
 
   @doc false
-  def start_child(mod, fun, args, name, partition) do
+  def start_child(mod, fun, args, partition) do
     case apply(mod, fun, args) do
       {:ok, pid} ->
-        register_child(name, partition, pid)
+        register_child(partition, pid)
         {:ok, pid}
 
       {:ok, pid, info} ->
-        register_child(name, partition, pid)
+        register_child(partition, pid)
         {:ok, pid, info}
 
       other ->
@@ -297,33 +324,97 @@ defmodule PartitionSupervisor do
     end
   end
 
-  defp register_child(name, partition, pid) when is_atom(name) do
-    :ets.insert(name, {partition, pid})
-  end
-
-  defp register_child({:via, _, _}, partition, pid) do
-    Registry.register(@registry, {self(), partition}, pid)
+  defp register_child(partition, pid) do
+    :ets.insert(Process.get(:ets_table), {partition, pid})
   end
 
   @impl true
   def init({name, partitions, children, init_opts}) do
-    init_partitions(name, partitions)
+    table = init_table(name)
+    :ets.insert(table, {:partitions, partitions, partitions})
+    Process.put(:ets_table, table)
     Supervisor.init(children, Keyword.put_new(init_opts, :strategy, :one_for_one))
   end
 
-  defp init_partitions(name, partitions) when is_atom(name) do
-    :ets.new(name, [:set, :named_table, :protected, read_concurrency: true])
-    :ets.insert(name, {:partitions, partitions})
+  defp init_table(name) when is_atom(name) do
+    :ets.new(name, [:set, :named_table, :public, read_concurrency: true])
   end
 
-  defp init_partitions({:via, _, _}, partitions) do
-    child_spec = {Registry, keys: :unique, name: @registry}
+  defp init_table({:via, _, _}) do
+    table = :ets.new(__MODULE__, [:set, :public, read_concurrency: true])
+    ensure_registry()
+    Registry.register(@registry, self(), table)
+    table
+  end
 
-    unless Process.whereis(@registry) do
-      Supervisor.start_child(:elixir_sup, child_spec)
+  defp ensure_registry do
+    if Process.whereis(@registry) == nil do
+      Supervisor.start_child(:elixir_sup, {Registry, keys: :unique, name: @registry})
     end
+  end
 
-    Registry.register(@registry, self(), partitions)
+  @doc """
+  Resizes the number of partitions in the PartitionSupervisor.
+
+  This is done by starting or stopping a given number of
+  partitions in the supervisor. All of the child specifications
+  are kept in the `PartitionSupervisor` itself.
+
+  The final number of partitions cannot be less than zero and
+  cannot be more than the number of partitions the supervisor
+  started with.
+  """
+  @doc since: "1.18.0"
+  @spec resize!(name(), non_neg_integer()) :: non_neg_integer()
+  def resize!(name, partitions) when is_integer(partitions) do
+    supervisor =
+      GenServer.whereis(name) || exit({:noproc, {__MODULE__, :resize!, [name, partitions]}})
+
+    table = table(name)
+    ensure_registry()
+
+    Registry.lock(@registry, supervisor, fn ->
+      case :ets.lookup(table, :partitions) do
+        [{:partitions, _current, max}] when partitions not in 0..max//1 ->
+          raise ArgumentError,
+                "the number of partitions to resize to must be a number between 0 and #{max}, got: #{partitions}"
+
+        [{:partitions, current, max}] when partitions > current ->
+          for id <- current..(partitions - 1) do
+            case Supervisor.restart_child(supervisor, id) do
+              {:ok, _} ->
+                :ok
+
+              {:ok, _, _} ->
+                :ok
+
+              {:error, reason} ->
+                raise "cannot restart partition #{id} of PartitionSupervisor #{inspect(name)} due to reason #{inspect(reason)}"
+            end
+          end
+
+          :ets.insert(table, {:partitions, partitions, max})
+          current
+
+        [{:partitions, current, max}] when partitions < current ->
+          :ets.insert(table, {:partitions, partitions, max})
+
+          for id <- partitions..(current - 1) do
+            case Supervisor.terminate_child(supervisor, id) do
+              :ok ->
+                :ok
+
+              {:error, reason} ->
+                raise "cannot terminate partition #{id} of PartitionSupervisor #{inspect(name)} due to reason #{inspect(reason)}"
+            end
+          end
+
+          current
+
+        [{:partitions, current, _max}] ->
+          current
+      end
+    end)
   end
 
   @doc """
@@ -332,24 +423,27 @@ defmodule PartitionSupervisor do
   @doc since: "1.14.0"
   @spec partitions(name()) :: pos_integer()
   def partitions(name) do
-    {_name, partitions} = name_partitions(name)
-    partitions
+    name |> table() |> partitions(name)
   end
 
-  # For whereis_name, we want to lookup on GenServer.whereis/1
-  # just once, so we lookup the name and partitions together.
-  defp name_partitions(name) when is_atom(name) do
+  defp partitions(table, name) do
     try do
-      {name, :ets.lookup_element(name, :partitions, 2)}
+      :ets.lookup_element(table, :partitions, 2)
     rescue
       _ -> exit({:noproc, {__MODULE__, :partitions, [name]}})
     end
   end
 
-  defp name_partitions(name) when is_tuple(name) do
+  defp table(name) when is_atom(name) do
+    name
+  end
+
+  # For whereis_name, we want to lookup on GenServer.whereis/1
+  # just once, so we lookup the name and partitions together.
+  defp table(name) when is_tuple(name) do
     with pid when is_pid(pid) <- GenServer.whereis(name),
-         [name_partitions] <- Registry.lookup(@registry, pid) do
-      name_partitions
+         [{_, table}] <- Registry.lookup(@registry, pid) do
+      table
     else
       _ -> exit({:noproc, {__MODULE__, :partitions, [name]}})
     end
@@ -374,7 +468,7 @@ defmodule PartitionSupervisor do
   @doc since: "1.14.0"
   @spec which_children(name()) :: [
           # Inlining [module()] | :dynamic here because :supervisor.modules() is not exported
-          {:undefined, pid | :restarting, :worker | :supervisor, [module()] | :dynamic}
+          {integer(), pid | :restarting, :worker | :supervisor, [module()] | :dynamic}
         ]
   def which_children(name) when is_atom(name) or elem(name, 0) == :via do
     Supervisor.which_children(name)
@@ -428,22 +522,17 @@ defmodule PartitionSupervisor do
 
   @doc false
   def whereis_name({name, key}) when is_atom(name) or is_tuple(name) do
-    {name, partitions} = name_partitions(name)
+    table = table(name)
+    partitions = partitions(table, name)
+
+    if partitions == 0 do
+      raise ArgumentError, "PartitionSupervisor #{inspect(name)} has zero partitions"
+    end
 
     partition =
       if is_integer(key), do: rem(abs(key), partitions), else: :erlang.phash2(key, partitions)
 
-    whereis_name(name, partition)
-  end
-
-  defp whereis_name(name, partition) when is_atom(name) do
-    :ets.lookup_element(name, partition, 2)
-  end
-
-  defp whereis_name(name, partition) when is_pid(name) do
-    @registry
-    |> Registry.values({name, partition}, name)
-    |> List.first(:undefined)
+    :ets.lookup_element(table, partition, 2)
   end
 
   @doc false
