@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Record do
   @moduledoc """
   Module to work with, define, and import records.
@@ -34,8 +38,20 @@ defmodule Record do
         # expands to: "@type user :: {:user, String.t(), integer}"
       end
 
+  ## Reflection
+
+  The record tag and its fields are stored as metadata in the "Docs" chunk
+  of the record definition macro. You can retrieve the documentation for
+  a module by calling `Code.fetch_docs/1`.
   """
 
+  @type extract_opts :: [
+          from: binary(),
+          from_lib: binary(),
+          includes: [binary()],
+          macros: keyword()
+        ]
+
   @doc """
   Extracts record information from an Erlang file.
 
@@ -47,8 +63,8 @@ defmodule Record do
 
   ## Options
 
-  This function accepts the following options, which are exclusive to each other
-  (i.e., only one of them can be used in the same call):
+  This function requires one of the following options, which are exclusive to each
+  other (i.e., only one of them can be used in the same call):
 
     * `:from` - (binary representing a path to a file) path to the Erlang file
       that contains the record definition to extract; with this option, this
@@ -60,6 +76,8 @@ defmodule Record do
       this function uses the same path lookup used by the `-include_lib`
       attribute used in Erlang modules.
 
+  It additionally accepts the following optional, non-exclusive options:
+
     * `:includes` - (a list of directories as binaries) if the record being
       extracted depends on relative includes, this option allows developers
       to specify the directory where those relative includes exist.
@@ -91,7 +109,7 @@ defmodule Record do
       ]
 
   """
-  @spec extract(name :: atom, keyword) :: keyword
+  @spec extract(name :: atom, extract_opts) :: keyword
   def extract(name, opts) when is_atom(name) and is_list(opts) do
     Record.Extractor.extract(name, opts)
   end
@@ -105,23 +123,10 @@ defmodule Record do
 
   ## Options
 
-  This function accepts the following options, which are exclusive to each other
-  (i.e., only one of them can be used in the same call):
+  Accepts the same options as listed for `Record.extract/2`.
 
-    * `:from` - (binary representing a path to a file) path to the Erlang file
-      that contains the record definitions to extract; with this option, this
-      function uses the same path lookup used by the `-include` attribute used in
-      Erlang modules.
-
-    * `:from_lib` - (binary representing a path to a file) path to the Erlang
-      file that contains the record definitions to extract; with this option,
-      this function uses the same path lookup used by the `-include_lib`
-      attribute used in Erlang modules.
-
-  These options are expected to be literals (including the binary values) at
-  compile time.
   """
-  @spec extract_all(keyword) :: [{name :: atom, keyword}]
+  @spec extract_all(extract_opts) :: [{name :: atom, keyword}]
   def extract_all(opts) when is_list(opts) do
     Record.Extractor.extract_all(opts)
   end
@@ -255,11 +260,10 @@ defmodule Record do
   """
   defmacro defrecord(name, tag \\ nil, kv) do
     quote bind_quoted: [name: name, tag: tag, kv: kv] do
-      fields = Record.__fields__(:defrecord, kv)
-      Record.__validate__(__MODULE__, name, fields)
-
       tag = tag || name
+      fields = Record.__record__(__MODULE__, :defrecord, name, tag, kv)
 
+      @doc record: {tag, fields}
       defmacro unquote(name)(args \\ []) do
         Record.__access__(unquote(tag), unquote(fields), args, __CALLER__)
       end
@@ -275,10 +279,8 @@ defmodule Record do
   """
   defmacro defrecordp(name, tag \\ nil, kv) do
     quote bind_quoted: [name: name, tag: tag, kv: kv] do
-      fields = Record.__fields__(:defrecordp, kv)
-      Record.__validate__(__MODULE__, name, fields)
-
       tag = tag || name
+      fields = Record.__record__(__MODULE__, :defrecordp, name, tag, kv)
 
       defmacrop unquote(name)(args \\ []) do
         Record.__access__(unquote(tag), unquote(fields), args, __CALLER__)
@@ -290,13 +292,6 @@ defmodule Record do
     end
   end
 
-  @doc false
-  def __validate__(module, name, fields) do
-    error_on_duplicate_record(module, name)
-    # TODO: Make it raise on v2.0
-    warn_on_duplicate_key(:lists.keysort(1, fields))
-  end
-
   defp error_on_duplicate_record(module, name) do
     defined_arity =
       Enum.find(0..2, fn arity ->
@@ -322,9 +317,30 @@ defmodule Record do
     warn_on_duplicate_key(rest)
   end
 
-  # Normalizes of record fields to have default values.
+  # Callback invoked from the record/2 macro.
   @doc false
-  def __fields__(type, fields) do
+  def __record__(module, kind, name, tag, kv) do
+    error_on_duplicate_record(module, name)
+
+    fields = fields(kind, kv)
+
+    # TODO: Remove me on Elixir v2.0
+    Module.register_attribute(module, :__records__, accumulate: true)
+
+    Module.put_attribute(module, :__records__, %{
+      kind: kind,
+      name: name,
+      tag: tag,
+      fields: :lists.map(&elem(&1, 0), fields)
+    })
+
+    # TODO: Make it raise on v2.0
+    warn_on_duplicate_key(:lists.keysort(1, fields))
+    fields
+  end
+
+  # Normalizes of record fields to have default values.
+  defp fields(kind, fields) do
     normalizer_fun = fn
       {key, value} when is_atom(key) ->
         try do
@@ -340,7 +356,7 @@ defmodule Record do
         {key, nil}
 
       other ->
-        raise ArgumentError, "#{type} fields must be atoms, got: #{inspect(other)}"
+        raise ArgumentError, "#{kind} fields must be atoms, got: #{inspect(other)}"
     end
 
     :lists.map(normalizer_fun, fields)
@@ -431,7 +447,7 @@ defmodule Record do
 
     if Keyword.has_key?(keyword, :_) do
       message = "updating a record with a default (:_) is equivalent to creating a new record"
-      IO.warn(message, Macro.Env.stacktrace(caller))
+      IO.warn(message, caller)
       create(tag, fields, keyword, caller)
     else
       updates =
@@ -447,12 +463,12 @@ defmodule Record do
     end
   end
 
-  defp hoist_expressions(keyword, %{context: nil}) do
+  defp hoist_expressions(keyword, %{context: nil, module: module}) do
     Enum.map_reduce(keyword, [], fn {key, expr}, acc ->
       if simple_argument?(expr) do
         {{key, expr}, acc}
       else
-        var = Macro.var(key, __MODULE__)
+        var = Macro.unique_var(key, module)
         {{key, var}, [{:=, [], [var, expr]} | acc]}
       end
     end)
