@@ -1,4 +1,10 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Regex do
+  # TODO: Remove the "Starting from Erlang/OTP 28" part in the Modifiers'
+  #       section once Erlang/OTP 28+ is exclusively supported.
   @moduledoc ~S"""
   Provides regular expressions for Elixir.
 
@@ -15,15 +21,11 @@ defmodule Regex do
       # A regular expression with case insensitive and Unicode options
       ~r/foo/iu
 
-  Regular expressions created via sigils are pre-compiled and stored
-  in the `.beam` file. Note that this may be a problem if you are precompiling
-  Elixir, see the "Precompilation" section for more information.
-
   A Regex is represented internally as the `Regex` struct. Therefore,
   `%Regex{}` can be used whenever there is a need to match on them.
-  Keep in mind that all of the structs fields are private. There is
-  also not guarantee two regular expressions from the same source are
-  equal, for example:
+  Keep in mind that all of the structs fields are private. And since
+  regexes are compiled, there is no guarantee two regular expressions
+  from the same source are equal, for example:
 
       ~r/(?<foo>.)(?<bar>.)/ == ~r/(?<foo>.)(?<bar>.)/
 
@@ -76,9 +78,16 @@ defmodule Regex do
 
     * `:caseless` (i) - adds case insensitivity
 
-    * `:dotall` (s) - causes dot to match newlines and also set newline to
-      anycrlf; the new line setting can be overridden by setting `(*CR)` or
-      `(*LF)` or `(*CRLF)` or `(*ANY)` according to `:re` documentation
+    * `:dotall` (s) - causes dot to match newlines and also sets newline to
+      `(*ANYCRLF)`.\
+      The new line setting, as described in the [`:re` documentation](`:re`),
+      can be overridden by starting the regular expression pattern with:
+      * `(*CR)` - carriage return
+      * `(*LF)` - line feed
+      * `(*CRLF)` - carriage return, followed by line feed
+      * `(*ANYCRLF)` - any of the three above
+      * `(*ANY)` - all Unicode newline sequences
+      * _Starting from Erlang/OTP 28, `(*NUL)` - the NUL character (binary zero)_
 
     * `:multiline` (m) - causes `^` and `$` to mark the beginning and end of
       each line; use `\A` and `\z` to match the end or beginning of the string
@@ -92,15 +101,6 @@ defmodule Regex do
     * `:ungreedy` (U) - inverts the "greediness" of the regexp
       (the previous `r` option is deprecated in favor of `U`)
 
-  The options not available are:
-
-    * `:anchored` - not available, use `^` or `\A` instead
-    * `:dollar_endonly` - not available, use `\z` instead
-    * `:no_auto_capture` - not available, use `?:` instead
-    * `:newline` - not available, use `(*CR)` or `(*LF)` or `(*CRLF)` or
-      `(*ANYCRLF)` or `(*ANY)` at the beginning of the regexp according to the
-      `:re` documentation
-
   ## Captures
 
   Many functions in this module handle what to capture in a regex
@@ -167,26 +167,16 @@ defmodule Regex do
       iex> Regex.replace(~r/\s/u, "Unicode\u00A0spaces", "-")
       "Unicode-spaces"
 
-  ## Precompilation
-
-  Regular expressions built with sigil are precompiled and stored in `.beam`
-  files. Precompiled regexes will be checked in runtime and may work slower
-  between operating systems and OTP releases. This is rarely a problem, as most Elixir code
-  shared during development is compiled on the target (such as dependencies,
-  archives, and escripts) and, when running in production, the code must either
-  be compiled on the target (via `mix compile` or similar) or released on the
-  host (via `mix releases` or similar) with a matching OTP, operating system
-  and architecture as the target.
-
-  If you know you are running on a different system than the current one and
-  you are doing multiple matches with the regex, you can manually invoke
-  `Regex.recompile/1` or `Regex.recompile!/1` to perform a runtime version
-  check and recompile the regex if necessary.
   """
 
-  defstruct re_pattern: nil, source: "", opts: [], re_version: ""
+  defstruct re_pattern: nil, source: "", opts: []
 
-  @type t :: %__MODULE__{re_pattern: term, source: binary, opts: binary | [term]}
+  @type t :: %__MODULE__{re_pattern: term, source: binary, opts: [term]}
+
+  @type named_captures_opts :: [
+          return: :binary | :index,
+          offset: non_neg_integer()
+        ]
 
   defmodule CompileError do
     @moduledoc """
@@ -209,38 +199,35 @@ defmodule Regex do
 
   ## Examples
 
-      iex> Regex.compile("foo")
-      {:ok, ~r/foo/}
+      Regex.compile("foo")
+      #=> {:ok, ~r/foo/}
 
-      iex> Regex.compile("*foo")
-      {:error, {~c"nothing to repeat", 0}}
+      Regex.compile("foo", "i")
+      #=> {:ok, ~r/foo/i}
 
-      iex> Regex.compile("foo", "i")
-      {:ok, ~r/foo/i}
-
-      iex> Regex.compile("foo", [:caseless])
-      {:ok, Regex.compile!("foo", [:caseless])}
+      Regex.compile("*foo")
+      #=> {:error, {~c"quantifier does not follow a repeatable item", 0}}
 
   """
-  @spec compile(binary, binary | [term]) :: {:ok, t} | {:error, any}
+  @spec compile(binary, binary | [term]) :: {:ok, t} | {:error, term}
   def compile(source, opts \\ "") when is_binary(source) do
-    compile(source, opts, version())
+    do_compile(source, opts)
   end
 
-  defp compile(source, opts, version) when is_binary(opts) do
+  defp do_compile(source, opts) when is_binary(opts) do
     case translate_options(opts, []) do
       {:error, rest} ->
         {:error, {:invalid_option, rest}}
 
       translated_opts ->
-        compile(source, translated_opts, version)
+        do_compile(source, translated_opts)
     end
   end
 
-  defp compile(source, opts, version) when is_list(opts) do
+  defp do_compile(source, opts) when is_list(opts) do
     case :re.compile(source, opts) do
       {:ok, re_pattern} ->
-        {:ok, %Regex{re_pattern: re_pattern, re_version: version, source: source, opts: opts}}
+        {:ok, %Regex{re_pattern: re_pattern, source: source, opts: opts}}
 
       error ->
         error
@@ -264,38 +251,29 @@ defmodule Regex do
   This checks the version stored in the regular expression
   and recompiles the regex in case of version mismatch.
   """
+  # Remove me on Elixir v1.22
+  @doc deprecated: "It can be removed and it has no effect"
   @doc since: "1.4.0"
-  @spec recompile(t) :: {:ok, t} | {:error, any}
   def recompile(%Regex{} = regex) do
-    version = version()
-
-    case regex do
-      %{re_version: ^version} ->
-        {:ok, regex}
-
-      _ ->
-        %{source: source, opts: opts} = regex
-        compile(source, opts, version)
-    end
+    {:ok, regex}
   end
 
   @doc """
   Recompiles the existing regular expression and raises `Regex.CompileError` in case of errors.
   """
+  # Remove me on Elixir v1.22
+  @doc deprecated: "It can be removed and it has no effect"
   @doc since: "1.4.0"
-  @spec recompile!(t) :: t
   def recompile!(regex) do
-    case recompile(regex) do
-      {:ok, regex} -> regex
-      {:error, {reason, at}} -> raise Regex.CompileError, "#{reason} at position #{at}"
-    end
+    regex
   end
 
   @doc """
   Returns the version of the underlying Regex engine.
   """
+  # Remove me on Elixir v1.22
+  @doc deprecated: "Use :re.version() instead"
   @doc since: "1.4.0"
-  @spec version :: term()
   def version do
     {:re.version(), :erlang.system_info(:endian)}
   end
@@ -337,7 +315,7 @@ defmodule Regex do
     * `:capture` - what to capture in the result. See the ["Captures" section](#module-captures)
       to see the possible capture values.
     * `:offset` - (since v1.12.0) specifies the starting offset to match in the given string.
-      Defaults to zero.
+      Defaults to `0`.
 
   ## Examples
 
@@ -357,7 +335,7 @@ defmodule Regex do
       ["d", ""]
 
   """
-  @spec run(t, binary, [term]) :: nil | [binary] | [{integer, integer}]
+  @spec run(t, binary, capture_opts) :: nil | [binary] | [{integer, integer}]
   def run(regex, string, options \\ [])
 
   def run(%Regex{} = regex, string, options) when is_binary(string) do
@@ -379,6 +357,8 @@ defmodule Regex do
 
     * `:return` - when set to `:index`, returns byte index and match length.
       Defaults to `:binary`.
+    * `:offset` - (since v1.12.0) specifies the starting offset to match in the given string.
+      Defaults to `0`.
 
   ## Examples
 
@@ -391,8 +371,15 @@ defmodule Regex do
       iex> Regex.named_captures(~r/a(?<foo>b)c(?<bar>d)/, "efgh")
       nil
 
+  You can also retrieve indexes from the named captures. This is particularly
+  useful if you want to know if a named capture matched or not:
+
+      iex> Regex.named_captures(~r/a(?<foo>b)c(?<bar>d)?/, "abc", return: :index)
+      %{"bar" => {-1, 0}, "foo" => {1, 1}}
+
+  You can then use `binary_part/3` to fetch the relevant part from the given string.
   """
-  @spec named_captures(t, String.t(), [term]) :: map | nil
+  @spec named_captures(t, String.t(), named_captures_opts) :: map | nil
   def named_captures(regex, string, options \\ []) when is_binary(string) do
     names = names(regex)
     options = Keyword.put(options, :capture, names)
@@ -441,6 +428,82 @@ defmodule Regex do
     opts
   end
 
+  @doc """
+  Returns the pattern as an embeddable string.
+
+  If the pattern was compiled with an option which cannot be represented
+  as an embeddable modifier in the current version of PCRE and strict is true
+  (the default) then an ArgumentError exception will be raised.
+
+  When the `:strict` option is false the pattern will be returned as though
+  any offending options had not be used and the function will not raise any
+  exceptions.
+
+  Embeddable modifiers/options are currently:
+
+    * 'i' - `:caseless`
+    * 'm' - `:multiline`
+    * 's' - `:dotall, {:newline, :anycrlf}`
+    * 'x' - `:extended`
+
+  Unembeddable modifiers are:
+
+    * 'f' - `:firstline`
+    * 'U' - `:ungreedy`
+    * 'u' - `:unicode, :ucp`
+
+  Any other regex compilation option not listed here is considered unembeddable
+  and will raise an exception unless the `:strict` option is false.
+
+  ## Examples
+      iex> Regex.to_embed(~r/foo/)
+      "(?-imsx:foo)"
+
+      iex> Regex.to_embed(~r/^foo/m)
+      "(?m-isx:^foo)"
+
+      iex> Regex.to_embed(~r/foo # comment/ix)
+      "(?ix-ms:foo # comment\\n)"
+
+      iex> Regex.to_embed(~r/foo/iu)
+      ** (ArgumentError) regex compiled with options [:ucp, :unicode] which cannot be represented as an embedded pattern in this version of PCRE
+
+      iex> Regex.to_embed(~r/foo/imsxu, strict: false)
+      "(?imsx:foo\\n)"
+
+  """
+  @doc since: "1.19.0"
+  @spec to_embed(t, strict: boolean()) :: String.t()
+  def to_embed(%Regex{source: source, opts: regex_opts}, embed_opts \\ []) do
+    strict = Keyword.get(embed_opts, :strict, true)
+
+    modifiers =
+      case embeddable_modifiers(regex_opts) do
+        {:ok, modifiers} ->
+          modifiers
+
+        {:error, modifiers, untranslatable} ->
+          if strict do
+            raise ArgumentError,
+                  "regex compiled with options #{inspect(untranslatable)} which cannot be " <>
+                    "represented as an embedded pattern in this version of PCRE"
+          else
+            modifiers
+          end
+      end
+
+    disabled = [?i, ?m, ?s, ?x] -- modifiers
+
+    disabled = if disabled != [], do: "-#{disabled}", else: ""
+
+    # Future proof option ordering consistency by sorting
+    modifiers = Enum.sort(modifiers)
+
+    nl = if Enum.member?(regex_opts, :extended), do: "\n", else: ""
+
+    "(?#{modifiers}#{disabled}:#{source}#{nl})"
+  end
+
   @doc """
   Returns a list of names in the regex.
 
@@ -451,17 +514,7 @@ defmodule Regex do
 
   """
   @spec names(t) :: [String.t()]
-  def names(%Regex{re_pattern: compiled, re_version: version, source: source}) do
-    re_pattern =
-      case version() do
-        ^version ->
-          compiled
-
-        _ ->
-          {:ok, recompiled} = :re.compile(source)
-          recompiled
-      end
-
+  def names(%Regex{re_pattern: re_pattern}) do
     {:namelist, names} = :re.inspect(re_pattern, :namelist)
     names
   end
@@ -479,7 +532,7 @@ defmodule Regex do
     * `:capture` - what to capture in the result. See the ["Captures" section](#module-captures)
       to see the possible capture values.
     * `:offset` - (since v1.12.0) specifies the starting offset to match in the given string.
-      Defaults to zero.
+      Defaults to `0`.
 
   ## Examples
 
@@ -508,7 +561,7 @@ defmodule Regex do
       [["cd"], ["ce"]]
 
   """
-  @spec scan(t(), String.t(), [term()]) :: [[String.t()]] | [[{integer(), integer()}]]
+  @spec scan(t(), String.t(), capture_opts) :: [[String.t()]] | [[{integer(), integer()}]]
   def scan(regex, string, options \\ [])
 
   def scan(%Regex{} = regex, string, options) when is_binary(string) do
@@ -524,24 +577,37 @@ defmodule Regex do
     end
   end
 
-  defp safe_run(
-         %Regex{re_pattern: compiled, source: source, re_version: version, opts: compile_opts},
-         string,
-         options
-       ) do
-    case version() do
-      ^version ->
-        :re.run(string, compiled, options)
-
-      _ when is_list(compile_opts) ->
-        :re.run(string, source, compile_opts ++ options)
-
-      # TODO: This clause is kept for compatibility with previous Elixir versions. Remove on v2.0+.
-      _ when is_binary(compile_opts) ->
-        :re.run(string, source, translate_options(compile_opts, options))
+  defp safe_run(%Regex{re_pattern: re_pattern} = regex, string, options) do
+    # TODO: Remove me when Erlang/OTP 28+ is required
+    # This allows regexes precompiled on Erlang/OTP 27- to work on Erlang/OTP 28+
+    with true <- :erlang.system_info(:otp_release) >= [?2, ?8],
+         {:re_pattern, _, _, _, <<_::bitstring>>} <- re_pattern do
+      %Regex{source: source, opts: compile_opts} = regex
+      :re.run(string, source, compile_opts ++ options)
+    else
+      _ -> :re.run(string, re_pattern, options)
     end
   end
 
+  @typedoc """
+  Options for regex functions that capture matches.
+  """
+  @type capture_opts :: [
+          return: :binary | :index,
+          capture: :all | :first | :all_but_first | :none | :all_names | [binary() | atom()],
+          offset: non_neg_integer()
+        ]
+
+  @typedoc """
+  Options for `split/3`.
+  """
+  @type split_opts :: [
+          parts: pos_integer() | :infinity,
+          trim: boolean(),
+          on: :first | :all | :all_but_first | :none | :all_names | [atom() | integer()],
+          include_captures: boolean()
+        ]
+
   @doc """
   Splits the given target based on the given pattern and in the given number of
   parts.
@@ -567,35 +633,35 @@ defmodule Regex do
 
   ## Examples
 
-      iex> Regex.split(~r{-}, "a-b-c")
+      iex> Regex.split(~r/-/, "a-b-c")
       ["a", "b", "c"]
 
-      iex> Regex.split(~r{-}, "a-b-c", parts: 2)
+      iex> Regex.split(~r/-/, "a-b-c", parts: 2)
       ["a", "b-c"]
 
-      iex> Regex.split(~r{-}, "abc")
+      iex> Regex.split(~r/-/, "abc")
       ["abc"]
 
-      iex> Regex.split(~r{}, "abc")
+      iex> Regex.split(~r//, "abc")
       ["", "a", "b", "c", ""]
 
-      iex> Regex.split(~r{a(?<second>b)c}, "abc")
+      iex> Regex.split(~r/a(?<second>b)c/, "abc")
       ["", ""]
 
-      iex> Regex.split(~r{a(?<second>b)c}, "abc", on: [:second])
+      iex> Regex.split(~r/a(?<second>b)c/, "abc", on: [:second])
       ["a", "c"]
 
-      iex> Regex.split(~r{(x)}, "Elixir", include_captures: true)
+      iex> Regex.split(~r/(x)/, "Elixir", include_captures: true)
       ["Eli", "x", "ir"]
 
-      iex> Regex.split(~r{a(?<second>b)c}, "abc", on: [:second], include_captures: true)
+      iex> Regex.split(~r/a(?<second>b)c/, "abc", on: [:second], include_captures: true)
       ["a", "b", "c"]
 
-      iex> Regex.split(~r{-}, "-a-b--c", trim: true)
+      iex> Regex.split(~r/-/, "-a-b--c", trim: true)
       ["a", "b", "c"]
 
   """
-  @spec split(t, String.t(), [term]) :: [String.t()]
+  @spec split(t, String.t(), split_opts) :: [String.t()]
   def split(regex, string, options \\ [])
 
   def split(%Regex{}, "", opts) do
@@ -654,10 +720,15 @@ defmodule Regex do
     <<_::binary-size(^offset), part::binary-size(^keep), match::binary-size(^length), _::binary>> =
       string
 
-    if keep == 0 and trim do
-      [match | do_split([h | t], string, new_offset, counter - 1, trim, true)]
-    else
-      [part, match | do_split([h | t], string, new_offset, counter - 1, trim, true)]
+    cond do
+      keep == 0 and (offset != 0 and length == 0) ->
+        do_split([h | t], string, new_offset, counter - 1, trim, true)
+
+      keep == 0 and trim ->
+        [match | do_split([h | t], string, new_offset, counter - 1, trim, true)]
+
+      true ->
+        [part, match | do_split([h | t], string, new_offset, counter - 1, trim, true)]
     end
   end
 
@@ -665,7 +736,7 @@ defmodule Regex do
     new_offset = pos + length
     keep = pos - offset
 
-    if keep == 0 and trim do
+    if keep == 0 and (trim or (offset != 0 and length == 0)) do
       do_split([h | t], string, new_offset, counter, trim, false)
     else
       <<_::binary-size(^offset), part::binary-size(^keep), _::binary>> = string
@@ -898,6 +969,29 @@ defmodule Regex do
 
   # Helpers
 
+  # translate options to modifiers as required for emedding
+  defp embeddable_modifiers(list), do: embeddable_modifiers(list, [], [])
+
+  defp embeddable_modifiers([:dotall, {:newline, :anycrlf} | t], acc, err),
+    do: embeddable_modifiers(t, [?s | acc], err)
+
+  defp embeddable_modifiers([:caseless | t], acc, err),
+    do: embeddable_modifiers(t, [?i | acc], err)
+
+  defp embeddable_modifiers([:extended | t], acc, err),
+    do: embeddable_modifiers(t, [?x | acc], err)
+
+  defp embeddable_modifiers([:multiline | t], acc, err),
+    do: embeddable_modifiers(t, [?m | acc], err)
+
+  defp embeddable_modifiers([option | t], acc, err),
+    do: embeddable_modifiers(t, acc, [option | err])
+
+  defp embeddable_modifiers([], acc, []), do: {:ok, acc}
+  defp embeddable_modifiers([], acc, err), do: {:error, acc, err}
+
+  # translate modifiers to options
+
   defp translate_options(<<?s, t::binary>>, acc),
     do: translate_options(t, [:dotall, {:newline, :anycrlf} | acc])
 
@@ -915,4 +1009,58 @@ defmodule Regex do
 
   defp translate_options(<<>>, acc), do: acc
   defp translate_options(t, _acc), do: {:error, t}
+
+  @doc false
+  def __escape__(%{__struct__: Regex} = regex) do
+    # OTP 28.0 introduced refs in patterns, which can't be used in AST anymore
+    # OTP 28.1 introduced :re.import/1 which allows us to work with pre-compiled binaries again
+
+    pattern_ast =
+      cond do
+        # TODO: Remove this when we require Erlang/OTP 28+
+        # Before OTP 28.0, patterns did not contain any refs and could be safely be escaped
+        :erlang.system_info(:otp_release) < [?2, ?8] ->
+          Macro.escape(regex.re_pattern)
+
+        # OTP 28.1+ introduced the ability to export and import regexes from compiled binaries
+        Code.ensure_loaded?(:re) and function_exported?(:re, :import, 1) ->
+          {:ok, exported} = :re.compile(regex.source, [:export] ++ regex.opts)
+
+          quote do
+            Regex.__import_pattern__(unquote(Macro.escape(exported)))
+          end
+          # we now that the Regex module is defined at this stage, so this macro can be safely called
+          |> Macro.update_meta(&([required: true] ++ &1))
+
+        # TODO: Remove this when we require Erlang/OTP 28.1+
+        # OTP 28.0 works in degraded mode performance-wise, we need to recompile from the source
+        true ->
+          quote do
+            {:ok, pattern} =
+              :re.compile(unquote(Macro.escape(regex.source)), unquote(Macro.escape(regex.opts)))
+
+            pattern
+          end
+      end
+
+    quote do
+      %{
+        __struct__: unquote(Regex),
+        re_pattern: unquote(pattern_ast),
+        source: unquote(Macro.escape(regex.source)),
+        opts: unquote(Macro.escape(regex.opts))
+      }
+    end
+  end
+
+  @doc false
+  defmacro __import_pattern__(pattern) do
+    if __CALLER__.context in [:match, :guard] do
+      raise ArgumentError, "escaped Regex structs are not allowed in match or guards"
+    end
+
+    quote do
+      :re.import(unquote(pattern))
+    end
+  end
 end
