@@ -1,3 +1,7 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+%% SPDX-FileCopyrightText: 2012 Plataformatec
+
 %% Main entry point for Elixir functions. All of those functions are
 %% private to the Elixir compiler and reserved to be used by Elixir only.
 -module(elixir).
@@ -26,27 +30,23 @@
 -type struct() :: #{'__struct__' := atom(), atom() => any()}.
 
 %% OTP Application API
-%% TODO: Remove Erlang/OTP 26+ checks
-
-load_paths(OTP, Paths) when OTP >= 26 -> code:add_pathsa(Paths, cache);
-load_paths(_OTP, Paths) -> code:add_pathsa(Paths).
 
 start(_Type, _Args) ->
-  OTP = parse_otp_release(),
+  _OTP = parse_otp_release(),
   preload_common_modules(),
-  set_stdio_and_stderr_to_binary_and_maybe_utf8(OTP),
+  ok = io:setopts(standard_io, [binary]),
   check_file_encoding(file:native_name_encoding()),
 
   case init:get_argument(elixir_root) of
     {ok, [[Root]]} ->
-      load_paths(OTP, [
+      code:add_pathsa([
         Root ++ "/eex/ebin",
         Root ++ "/ex_unit/ebin",
         Root ++ "/iex/ebin",
         Root ++ "/logger/ebin",
         Root ++ "/mix/ebin",
         Root ++ "/elixir/ebin"
-      ]);
+      ], cache);
     _ ->
       ok
   end,
@@ -57,11 +57,11 @@ start(_Type, _Args) ->
   end,
 
   case application:get_env(elixir, ansi_enabled) of
-    {ok, _} -> ok;
+    {ok, _} ->
+      ok;
+
     undefined ->
-      %% Remove prim_tty module check as well as checks from scripts on Erlang/OTP 26
-      ANSIEnabled = erlang:module_loaded(prim_tty) andalso (prim_tty:isatty(stdout) == true),
-      application:set_env(elixir, ansi_enabled, ANSIEnabled)
+      application:set_env(elixir, ansi_enabled, prim_tty:isatty(stdout) == true)
   end,
 
   Tokenizer = case code:ensure_loaded('Elixir.String.Tokenizer') of
@@ -90,10 +90,10 @@ start(_Type, _Args) ->
     {docs, true},
     {ignore_already_consolidated, false},
     {ignore_module_conflict, false},
+    {infer_signatures, [elixir]},
     {on_undefined_variable, raise},
     {parser_options, [{columns, true}]},
     {debug_info, true},
-    {warnings_as_errors, false},
     {relative_paths, true},
     {no_warn_undefined, []},
     {tracers, []}
@@ -117,19 +117,6 @@ stop(Tab) ->
 config_change(_Changed, _New, _Remove) ->
   ok.
 
-set_stdio_and_stderr_to_binary_and_maybe_utf8(OTP) when OTP >= 26 ->
-  ok = io:setopts(standard_io, [binary]),
-  ok;
-set_stdio_and_stderr_to_binary_and_maybe_utf8(_OTP) ->
-  Opts =
-    case init:get_argument(noshell) of
-      {ok, _} -> [binary, {encoding, utf8}];
-      error   -> [binary]
-    end,
-  ok = io:setopts(standard_io, Opts),
-  ok = io:setopts(standard_error, [{encoding, utf8}]),
-  ok.
-
 preload_common_modules() ->
   %% We attempt to load those modules here so throughout
   %% the codebase we can avoid code:ensure_loaded/1 checks.
@@ -140,10 +127,20 @@ preload_common_modules() ->
 parse_otp_release() ->
   %% Whenever we change this check, we should also change Makefile.
   case string:to_integer(erlang:system_info(otp_release)) of
-    {Num, _} when Num >= 25 ->
+    {Num, _} when Num >= 26 ->
+      case Num == 28 andalso (code:ensure_loaded(re) == {module, re}) andalso not erlang:function_exported(re, import, 1) of
+        true ->
+          io:format(standard_error,
+            "warning! Erlang/OTP 28.0 detected.~n"
+            "Regexes will be re-compiled from source at runtime, which will cause degraded performance.~n"
+            "This can be fixed by using Erlang OTP 28.1+ or 27-.~n"
+          , []);
+        false ->
+          ok
+      end,
       Num;
     _ ->
-      io:format(standard_error, "ERROR! Unsupported Erlang/OTP version, expected Erlang/OTP 25+~n", []),
+      io:format(standard_error, "ERROR! Unsupported Erlang/OTP version, expected Erlang/OTP 26+~n", []),
       erlang:halt(1)
   end.
 
@@ -177,19 +174,9 @@ check_file_encoding(Encoding) ->
   end.
 
 %% Boot and process given options. Invoked by Elixir's script.
-%% TODO: Delete prim_tty branches on Erlang/OTP 26.
 
 start() ->
-  case code:ensure_loaded(prim_tty) of
-    {module, _} ->
-      user_drv:start(#{initial_shell => iex:shell()});
-    {error, _} ->
-      case init:get_argument(elixir_root) of
-        {ok, [[Root]]} -> code:add_patha(Root ++ "/iex/ebin");
-        _ -> ok
-      end,
-      'Elixir.IEx.CLI':deprecated()
-  end.
+  user_drv:start(#{initial_shell => iex:shell()}).
 
 start_cli() ->
   {ok, _} = application:ensure_all_started(elixir),
@@ -211,7 +198,6 @@ start_cli() ->
 env_for_eval(#{lexical_tracker := Pid} = Env) ->
   NewEnv = Env#{
     context := nil,
-    context_modules := [],
     macro_aliases := [],
     versioned_vars := #{}
   },
@@ -233,16 +219,8 @@ env_for_eval(#{lexical_tracker := Pid} = Env) ->
   end;
 env_for_eval(Opts) when is_list(Opts) ->
   Env = elixir_env:new(),
-
-  Line = case lists:keyfind(line, 1, Opts) of
-    {line, LineOpt} when is_integer(LineOpt) -> LineOpt;
-    false -> ?key(Env, line)
-  end,
-
-  File = case lists:keyfind(file, 1, Opts) of
-    {file, FileOpt} when is_binary(FileOpt) -> FileOpt;
-    false -> ?key(Env, file)
-  end,
+  Line = elixir_utils:get_line(Opts, Env),
+  File = elixir_utils:get_file(Opts, Env),
 
   Module = case lists:keyfind(module, 1, Opts) of
     {module, ModuleOpt} when is_atom(ModuleOpt) -> ModuleOpt;
@@ -260,6 +238,7 @@ env_for_eval(Opts) when is_list(Opts) ->
     false -> []
   end,
 
+  %% TODO: Remove the following deprecations in future releases
   Aliases = case lists:keyfind(aliases, 1, Opts) of
     {aliases, AliasesOpt} when is_list(AliasesOpt) ->
       'Elixir.IO':warn(<<":aliases option in eval is deprecated">>),
@@ -460,8 +439,8 @@ quoted_to_erl(Quoted, ErlS, ExS, Env) ->
 
 string_to_tokens(String, StartLine, StartColumn, File, Opts) when is_integer(StartLine), is_binary(File) ->
   case elixir_tokenizer:tokenize(String, StartLine, StartColumn, Opts) of
-    {ok, _Line, _Column, [], Tokens, Terminators} ->
-      {ok, lists:reverse(Tokens, Terminators)};
+    {ok, _Line, _Column, [], Tokens, []} ->
+      {ok, lists:reverse(Tokens)};
     {ok, _Line, _Column, Warnings, Tokens, Terminators} ->
       (lists:keyfind(emit_warnings, 1, Opts) /= {emit_warnings, false}) andalso
         [elixir_errors:erl_warn(L, File, M) || {L, M} <- lists:reverse(Warnings)],
@@ -513,10 +492,14 @@ parser_location(Meta) ->
         {ok, Forms} ->
           Forms;
         {error, {Meta, Error, Token}} ->
-          elixir_errors:parse_error(Meta, File, Error, Token, {String, StartLine, StartColumn})
+          Indentation = proplists:get_value(indentation, Opts, 0),
+          Input = {String, StartLine, StartColumn, Indentation},
+          elixir_errors:parse_error(Meta, File, Error, Token, Input)
       end;
     {error, {Meta, Error, Token}} ->
-      elixir_errors:parse_error(Meta, File, Error, Token, {String, StartLine, StartColumn})
+      Indentation = proplists:get_value(indentation, Opts, 0),
+      Input = {String, StartLine, StartColumn, Indentation},
+      elixir_errors:parse_error(Meta, File, Error, Token, Input)
   end.
 
 to_binary(List) when is_list(List) -> elixir_utils:characters_to_binary(List);
