@@ -1,3 +1,7 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+%% SPDX-FileCopyrightText: 2012 Plataformatec
+
 %% Compiler backend to Erlang.
 
 -module(elixir_erl).
@@ -13,10 +17,10 @@ debug_info(elixir_v1, _Module, none, _Opts) ->
 debug_info(elixir_v1, _Module, {elixir_v1, Map, _Specs}, _Opts) ->
   {ok, Map};
 debug_info(erlang_v1, _Module, {elixir_v1, Map, Specs}, _Opts) ->
-  {Prefix, Forms, _, _, _} = dynamic_form(Map),
+  {Prefix, Forms, _, _, _} = dynamic_form(Map, nil),
   {ok, Prefix ++ Specs ++ Forms};
 debug_info(core_v1, _Module, {elixir_v1, Map, Specs}, Opts) ->
-  {Prefix, Forms, _, _, _} = dynamic_form(Map),
+  {Prefix, Forms, _, _, _} = dynamic_form(Map, nil),
   #{compile_opts := CompileOpts} = Map,
   AllOpts = CompileOpts ++ Opts,
 
@@ -128,9 +132,10 @@ scope(_Meta, ExpandCaptures) ->
 
 %% Static compilation hook, used in protocol consolidation
 
-consolidate(Map, TypeSpecs, Chunks) ->
-  {Prefix, Forms, _Def, _Defmacro, _Macros} = dynamic_form(Map),
-  load_form(Map, Prefix, Forms, TypeSpecs, Chunks).
+consolidate(Map, TypeSpecs, DocsChunk) ->
+  {Prefix, Forms, Def, _Defmacro, _Macros} = dynamic_form(Map, nil),
+  CheckerChunk = checker_chunk(Map, Def, chunk_opts(Map)),
+  load_form(Map, Prefix, Forms, TypeSpecs, DocsChunk ++ CheckerChunk).
 
 %% Dynamic compilation hook, used in regular compiler
 
@@ -150,22 +155,24 @@ compile(#{module := Module, anno := Anno} = BaseMap) ->
       false -> ?typespecs:translate_typespecs_for_module(Set, Bag)
     end,
 
-  {Prefix, Forms, Def, Defmacro, Macros} = dynamic_form(Map),
+  MD5 = ets:lookup_element(Set, exports_md5, 2),
+  {Prefix, Forms, Def, Defmacro, Macros} = dynamic_form(Map, MD5),
   {Types, Callbacks, TypeSpecs} = typespecs_form(Map, TranslatedTypespecs, Macros),
 
-  ChunkOpts =
-    case lists:member(deterministic, ?key(Map, compile_opts)) of
-      true -> [deterministic];
-      false -> []
-    end,
-
+  ChunkOpts = chunk_opts(Map),
   DocsChunk = docs_chunk(Map, Set, Module, Anno, Def, Defmacro, Types, Callbacks, ChunkOpts),
   CheckerChunk = checker_chunk(Map, Def, ChunkOpts),
   load_form(Map, Prefix, Forms, TypeSpecs, DocsChunk ++ CheckerChunk).
 
+chunk_opts(Map) ->
+  case lists:member(deterministic, ?key(Map, compile_opts)) of
+    true -> [deterministic];
+    false -> []
+  end.
+
 dynamic_form(#{module := Module, relative_file := RelativeFile,
                attributes := Attributes, definitions := Definitions, unreachable := Unreachable,
-               deprecated := Deprecated, compile_opts := Opts} = Map) ->
+               deprecated := Deprecated, compile_opts := Opts} = Map, MD5) ->
   %% TODO: Match on anno directly in Elixir v1.22+
   Line = case Map of
     #{anno := AnnoValue} -> erl_anno:line(AnnoValue);
@@ -175,12 +182,7 @@ dynamic_form(#{module := Module, relative_file := RelativeFile,
   {Def, Defmacro, Macros, Exports, Functions} =
     split_definition(Definitions, Unreachable, Line, [], [], [], [], {[], []}),
 
-  FilteredOpts = lists:filter(fun(
-    {no_warn_undefined, _}) -> false;
-    (debug_info) -> false;
-    (_) -> true
-  end, Opts),
-
+  FilteredOpts = proplists:delete(debug_info, proplists:delete(no_warn_undefined, Opts)),
   Location = {elixir_utils:characters_to_list(RelativeFile), Line},
 
   Prefix = [{attribute, Line, file, Location},
@@ -188,7 +190,7 @@ dynamic_form(#{module := Module, relative_file := RelativeFile,
             {attribute, Line, compile, [no_auto_import | FilteredOpts]}],
 
   Struct = maps:get(struct, Map, nil),
-  Forms0 = functions_form(Line, Module, Def, Defmacro, Exports, Functions, Deprecated, Struct),
+  Forms0 = functions_form(Line, Module, Def, Defmacro, Exports, Functions, Deprecated, Struct, MD5),
   Forms1 = attributes_form(Line, Attributes, Forms0),
   {Prefix, Forms1, Def, Defmacro, Macros}.
 
@@ -292,11 +294,11 @@ is_macro(_)         -> false.
 
 % Functions
 
-functions_form(Line, Module, Def, Defmacro, Exports, Body, Deprecated, Struct) ->
-  {Spec, Info} = add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct),
+functions_form(Line, Module, Def, Defmacro, Exports, Body, Deprecated, Struct, MD5) ->
+  {Spec, Info} = add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct, MD5),
   [{attribute, Line, export, lists:usort([{'__info__', 1} | Exports])}, Spec, Info | Body].
 
-add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct) ->
+add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct, MD5) ->
   DefNA = [NA || {NA, _Meta} <- Def],
   DefmacroNA = [NA || {NA, _Meta} <- Defmacro],
 
@@ -319,7 +321,7 @@ add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct) ->
       functions_info(DefNA),
       macros_info(DefmacroNA),
       struct_info(Struct),
-      exports_md5_info(Struct, DefNA, DefmacroNA),
+      exports_md5_info(MD5, DefNA, DefmacroNA, Struct),
       get_module_info(Module, attributes),
       get_module_info(Module, compile),
       get_module_info(Module, md5),
@@ -331,11 +333,12 @@ add_info_function(Line, Module, Def, Defmacro, Deprecated, Struct) ->
 get_module_info(Module) ->
   {clause, 0, [{atom, 0, module}], [], [{atom, 0, Module}]}.
 
-exports_md5_info(Struct, Def, Defmacro) ->
-  %% Deprecations do not need to be part of exports_md5 because it is always
-  %% checked by the runtime pass, so it is not really part of compilation.
-  Md5 = erlang:md5(term_to_binary({Def, Defmacro, Struct}, [deterministic])),
-  {clause, 0, [{atom, 0, exports_md5}], [], [elixir_to_erl(Md5)]}.
+exports_md5_info(MD5Attr, Def, Defmacro, Struct) ->
+  MD5 = if
+    is_binary(MD5Attr) -> MD5Attr;
+    MD5Attr =:= nil -> elixir_module:exports_md5(Def, Defmacro, Struct)
+  end,
+  {clause, 0, [{atom, 0, exports_md5}], [], [elixir_to_erl(MD5)]}.
 
 functions_info(Def) ->
   {clause, 0, [{atom, 0, functions}], [], [elixir_to_erl(Def)]}.
@@ -590,15 +593,12 @@ get_moduledoc_meta(Set) ->
 
 get_docs(Set, Module, Definitions, Kind) ->
   [{Key,
-    maybe_generated(erl_anno:new(Line), Ctx),
+    erl_anno:new(Line),
     [signature_to_binary(Module, Name, Signature)],
     doc_value(Doc, Name),
     Meta#{source_annos => [?ann(DefinitionMeta)]}
    } || {{Name, Arity}, DefinitionMeta} <- Definitions,
-        {Key, Ctx, Line, Signature, Doc, Meta} <- ets:lookup(Set, {Kind, Name, Arity})].
-
-maybe_generated(Ann, nil) -> Ann;
-maybe_generated(Ann, _Ctx) -> erl_anno:set_generated(true, Ann).
+        {Key, _Ctx, Line, Signature, Doc, Meta} <- ets:lookup(Set, {Kind, Name, Arity})].
 
 get_callback_docs(Set, Callbacks) ->
   [{Key,
@@ -640,7 +640,8 @@ signature_to_binary(_, Name, Signature) ->
   'Elixir.IO':iodata_to_binary(Doc).
 
 checker_chunk(Map, Def, ChunkOpts) ->
-  #{deprecated := Deprecated, defines_behaviour := DefinesBehaviour, signatures := Signatures} = Map,
+  #{deprecated := Deprecated, defines_behaviour := DefinesBehaviour,
+    signatures := Signatures, attributes := Attributes} = Map,
   DeprecatedMap = maps:from_list(Deprecated),
 
   Exports =
@@ -654,10 +655,14 @@ checker_chunk(Map, Def, ChunkOpts) ->
     end || {FA, _Meta} <- prepend_behaviour_info(DefinesBehaviour, Def)],
 
   Contents = #{
-    exports => Exports
+    exports => Exports,
+    mode => case lists:keymember('__protocol__', 1, Attributes) of
+      true -> protocol;
+      false -> elixir
+    end
   },
 
-  [{<<"ExCk">>, term_to_binary({elixir_checker_v1, Contents}, ChunkOpts)}].
+  [{<<"ExCk">>, term_to_binary({elixir_checker_v3, Contents}, ChunkOpts)}].
 
 prepend_behaviour_info(true, Def) -> [{{behaviour_info, 1}, []} | Def];
 prepend_behaviour_info(false, Def) -> Def.
