@@ -1,3 +1,7 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+%% SPDX-FileCopyrightText: 2012 Plataformatec
+
 %% A bunch of helpers to help to deal with errors in Elixir source code.
 %% This is not exposed in the Elixir language.
 %%
@@ -92,7 +96,10 @@ emit_diagnostic(Severity, Position, File, Message, Stacktrace, Options) ->
 
   Diagnostic = #{
     severity => Severity,
-    source => File,
+    source => case get(elixir_compiler_file) of
+      undefined -> File;
+      CompilerFile -> CompilerFile
+    end,
     file => File,
     position => Position,
     message => unicode:characters_to_binary(Message),
@@ -439,11 +446,12 @@ cut_snippet(Location, Input) ->
       nil
   end.
 
-cut_snippet({InputString, StartLine, StartColumn}, Line, Span) ->
+cut_snippet({InputString, StartLine, StartColumn, Indentation}, Line, Span) ->
   %% In case the code is indented, we need to add the indentation back
   %% for the snippets to match the reported columns.
-  Indent = binary:copy(<<" ">>, StartColumn - 1),
-  Lines = string:split(InputString, "\n", all),
+  Prelude = lists:duplicate(max(StartColumn - Indentation - 1, 0), " "),
+  Lines = string:split(Prelude ++ InputString, "\n", all),
+  Indent = binary:copy(<<" ">>, Indentation),
   [Head | Tail] = lists:nthtail(Line - StartLine, Lines),
   IndentedTail = indent_n(Tail, Span - 1, <<"\n", Indent/binary>>),
   elixir_utils:characters_to_binary([Indent, Head, IndentedTail]).
@@ -456,18 +464,16 @@ indent_n([H | T], Count, Indent) -> [Indent, H | indent_n(T, Count - 1, Indent)]
 
 prefix(warning) -> highlight(<<"warning:">>, warning);
 prefix(error) -> highlight(<<"error:">>, error);
-prefix(hint) -> highlight(<<"hint:">>, hint).
+prefix(hint) -> <<"hint:">>.
 
 highlight(Message, Severity) ->
   case {Severity, application:get_env(elixir, ansi_enabled, false)} of
     {warning, true} -> yellow(Message);
     {error, true} -> red(Message);
-    {hint, true} -> blue(Message);
     _ -> Message
   end.
 
 yellow(Msg) -> ["\e[33m", Msg, "\e[0m"].
-blue(Msg) -> ["\e[34m", Msg, "\e[0m"].
 red(Msg) -> ["\e[31m", Msg, "\e[0m"].
 
 env_format(Meta, #{file := EnvFile} = E) ->
