@@ -1,3 +1,7 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+%% SPDX-FileCopyrightText: 2012 Plataformatec
+
 -module(elixir_expand).
 -export([expand/3, expand_args/3, expand_arg/3, format_error/1]).
 -import(elixir_errors, [file_error/4, module_error/4, function_error/4]).
@@ -253,11 +257,12 @@ expand({quote, Meta, [Opts, Do]}, S, E) when is_list(Do) ->
   {EContext, SC, EC} = expand(QContext, SP, EP),
   Quoted = elixir_quote:quote(Exprs, Q),
   {EQuoted, ES, EQ} = expand(Quoted, SC, EC),
+  BindingMeta = lists:keydelete(column, 1, Meta),
 
   EBinding =
       [{'{}', [],
         ['=', [], [
-          {'{}', [], [K, Meta, EContext]},
+          {'{}', [], [K, BindingMeta, EContext]},
           V
         ]
        ]} || {K, V} <- Binding],
@@ -419,8 +424,8 @@ expand({Name, Meta, Kind}, S, E) when is_atom(Name), is_atom(Kind) ->
                 {ok, CurrentVersion};
 
               is_map_key(Pair, Pre) ->
-                %% TODO: Enable this warning on Elixir v1.19
                 %% TODO: Remove me on Elixir 2.0
+                %% TODO: Deprecate me on Elixir 1.20
                 %% elixir_errors:file_warn(Meta, E, ?MODULE, {unpinned_bitsize_var, Name, Kind}),
                 {ok, CurrentVersion};
 
@@ -570,18 +575,24 @@ escape_map(Map) -> {'%{}', [], lists:sort(maps:to_list(Map))}.
 expand_multi_alias_call(Kind, Meta, Base, Refs, Opts, S, E) ->
   {BaseRef, SB, EB} = expand_without_aliases_report(Base, S, E),
 
-  Fun = fun
-    ({'__aliases__', _, Ref}, SR, ER) ->
-      expand({Kind, Meta, [elixir_aliases:concat([BaseRef | Ref]), Opts]}, SR, ER);
+  case is_atom(BaseRef) of
+    true ->
+      Fun = fun
+        ({'__aliases__', _, Ref}, SR, ER) ->
+          expand({Kind, Meta, [elixir_aliases:concat([BaseRef | Ref]), Opts]}, SR, ER);
 
-    (Ref, SR, ER) when is_atom(Ref) ->
-      expand({Kind, Meta, [elixir_aliases:concat([BaseRef, Ref]), Opts]}, SR, ER);
+        (Ref, SR, ER) when is_atom(Ref) ->
+          expand({Kind, Meta, [elixir_aliases:concat([BaseRef, Ref]), Opts]}, SR, ER);
 
-    (Other, _SR, _ER) ->
-      file_error(Meta, E, ?MODULE, {expected_compile_time_module, Kind, Other})
-  end,
+        (Other, _SR, _ER) ->
+          file_error(Meta, E, ?MODULE, {expected_compile_time_module, Kind, Other})
+      end,
 
-  mapfold(Fun, SB, EB, Refs).
+      mapfold(Fun, SB, EB, Refs);
+
+    false ->
+      file_error(Meta, E, ?MODULE, {invalid_alias, Base})
+  end.
 
 resolve_super(Meta, Arity, E) ->
   Module = assert_module_scope(Meta, super, E),
