@@ -1,3 +1,7 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+%% SPDX-FileCopyrightText: 2012 Plataformatec
+
 % Handle string and string-like interpolations.
 -module(elixir_interpolation).
 -export([extract/6, unescape_string/1, unescape_string/2,
@@ -64,7 +68,8 @@ extract([$#, ${ | Rest], Buffer, Output, Line, Column, Scope, true, Last) ->
       {error, Reason};
     {ok, EndLine, EndColumn, Warnings, Tokens, Terminators} when Scope#elixir_tokenizer.cursor_completion /= false ->
       NewScope = Scope#elixir_tokenizer{warnings=Warnings, cursor_completion=noprune},
-      Output2 = build_interpol(Line, Column, EndLine, EndColumn, lists:reverse(Tokens, Terminators), Output1),
+      {CursorTerminators, _} = cursor_complete(EndLine, EndColumn, Terminators),
+      Output2 = build_interpol(Line, Column, EndLine, EndColumn, lists:reverse(Tokens, CursorTerminators), Output1),
       extract([], [], Output2, EndLine, EndColumn, NewScope, true, Last);
     {ok, _, _, _, _, _} ->
       {error, {string, Line, Column, "missing interpolation terminator: \"}\"", []}}
@@ -84,13 +89,22 @@ extract(Rest, Buffer, Output, Line, Column, Scope, Interpol, Last) ->
 
 extract_char(Rest, Buffer, Output, Line, Column, Scope, Interpol, Last) ->
   case unicode_util:gc(Rest) of
+    [Char | NewRest] when ?break(Char) ->
+      Token = io_lib:format("\\u~4.16.0B", [Char]),
+      Msg = io_lib:format("invalid bidirectional formatting character: ~ts. If you want to use such character, use it in its escaped ~ts form instead", [[Char], Token]),
+      WarningScope = prepend_warning(Line, Column, Msg, Scope),
+      extract(NewRest, [Char | Buffer], Output, Line, Column + 1, WarningScope, Interpol, Last);
+
     [Char | _] when ?bidi(Char) ->
       Token = io_lib:format("\\u~4.16.0B", [Char]),
       Pre = "invalid bidirectional formatting character in string: ",
       Pos = io_lib:format(". If you want to use such character, use it in its escaped ~ts form instead", [Token]),
       {error, {?LOC(Line, Column), {Pre, Pos}, Token}};
 
-    [Char | NewRest] ->
+    [Char | NewRest] when is_list(Char) ->
+      extract(NewRest, lists:reverse(Char, Buffer), Output, Line, Column + 1, Scope, Interpol, Last);
+
+    [Char | NewRest] when is_integer(Char) ->
       extract(NewRest, [Char | Buffer], Output, Line, Column + 1, Scope, Interpol, Last);
 
     [] ->
@@ -114,6 +128,16 @@ strip_horizontal_space([H | T], Buffer, Counter) when H =:= $\s; H =:= $\t ->
 strip_horizontal_space(T, Buffer, Counter) ->
   {T, Buffer, Counter}.
 
+cursor_complete(Line, Column, Terminators) ->
+  lists:mapfoldl(
+    fun({Start, _, _}, AccColumn) ->
+      End = elixir_tokenizer:terminator(Start),
+      {{End, {Line, AccColumn, nil}}, AccColumn + length(erlang:atom_to_list(End))}
+    end,
+    Column,
+    Terminators
+  ).
+
 %% Unescape a series of tokens as returned by extract.
 
 unescape_tokens(Tokens) ->
