@@ -1,6 +1,10 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+%% SPDX-FileCopyrightText: 2012 Plataformatec
+
 -module(elixir_module).
 -export([file/1, data_tables/1, is_open/1, mode/1, delete_definition_attributes/6,
-         compile/6, expand_callback/6, format_error/1, compiler_modules/0,
+         compile/6, expand_callback/6, format_error/1, compiler_modules/0, exports_md5/3,
          write_cache/3, read_cache/2, next_counter/1, taint/1, cache_env/1, get_cached_env/1]).
 -include("elixir.hrl").
 -define(counter_attr, {elixir, counter}).
@@ -19,6 +23,9 @@ put_compiler_modules([]) ->
 put_compiler_modules(M) when is_list(M) ->
   erlang:put(elixir_compiler_modules, M).
 
+exports_md5(Def, Defmacro, Struct) ->
+  erlang:md5(term_to_binary({lists:sort(Def), lists:sort(Defmacro), Struct}, [deterministic])).
+
 %% Table functions
 
 file(Module) ->
@@ -155,7 +162,7 @@ compile(Meta, Module, ModuleAsCharlist, Block, Vars, Prune, E) ->
     put_compiler_modules([Module | CompilerModules]),
     {Result, ModuleE, CallbackE} = eval_form(Line, Module, DataBag, Block, Vars, Prune, E),
     CheckerInfo = checker_info(),
-    BeamLocation = beam_location(ModuleAsCharlist),
+    {BeamLocation, Forceload} = beam_location(ModuleAsCharlist),
 
     {Binary, PersistedAttributes, Autoload} =
       elixir_erl_compiler:spawn(fun() ->
@@ -189,15 +196,18 @@ compile(Meta, Module, ModuleAsCharlist, Block, Vars, Prune, E) ->
             true -> {#{}, []};
             false ->
               UsedPrivate = bag_lookup_element(DataBag, used_private, 2),
-              'Elixir.Module.Types':infer(Module, File, AllDefinitions, Private, UsedPrivate, E, CheckerInfo)
+              'Elixir.Module.Types':infer(Module, File, Attributes, AllDefinitions, Private, UsedPrivate, E, CheckerInfo)
           end,
 
         RawCompileOpts = bag_lookup_element(DataBag, {accumulate, compile}, 2),
         CompileOpts = validate_compile_opts(RawCompileOpts, AllDefinitions, Unreachable, Line, E),
         Impls = bag_lookup_element(DataBag, impls, 2),
 
+        Struct = get_struct(DataSet),
+        set_exports_md5(DataSet, AllDefinitions, Struct),
+
         ModuleMap = #{
-          struct => get_struct(DataSet),
+          struct => Struct,
           module => Module,
           anno => Anno,
           file => File,
@@ -215,17 +225,17 @@ compile(Meta, Module, ModuleAsCharlist, Block, Vars, Prune, E) ->
 
         compile_error_if_tainted(DataSet, E),
         Binary = elixir_erl:compile(ModuleMap),
-        Autoload = proplists:get_value(autoload, CompileOpts, true),
-        spawn_parallel_checker(CheckerInfo, Module, ModuleMap),
+        Autoload = Forceload or proplists:get_value(autoload, CompileOpts, false),
+        spawn_parallel_checker(CheckerInfo, Module, ModuleMap, BeamLocation),
         {Binary, PersistedAttributes, Autoload}
       end),
 
     Autoload andalso code:load_binary(Module, BeamLocation, Binary),
+    make_module_available(Module, Binary, Autoload),
     put_compiler_modules(CompilerModules),
     eval_callbacks(Line, DataBag, after_compile, [CallbackE, Binary], CallbackE),
     elixir_env:trace({on_module, Binary, none}, ModuleE),
     warn_unused_attributes(DataSet, DataBag, PersistedAttributes, E),
-    make_module_available(Module, Binary),
     (element(2, CheckerInfo) == nil) andalso
       [VerifyMod:VerifyFun(Module) ||
        {VerifyMod, VerifyFun} <- bag_lookup_element(DataBag, {accumulate, after_verify}, 2)],
@@ -253,6 +263,16 @@ compile_error_if_tainted(DataSet, E) ->
     false -> ok
   end.
 
+set_exports_md5(DataSet, AllDefinitions, Struct) ->
+  {Funs, Macros} =
+    lists:foldl(fun
+      ({Tuple, def, _Meta, _Clauses}, {Funs, Macros}) -> {[Tuple | Funs], Macros};
+      ({Tuple, defmacro, _Meta, _Clauses}, {Funs, Macros}) -> {Funs, [Tuple | Macros]};
+      ({_Tuple, _Kind, _Meta, _Clauses}, {Funs, Macros}) -> {Funs, Macros}
+    end, {[], []}, AllDefinitions),
+  MD5 = exports_md5(Funs, Macros, Struct),
+  ets:insert(DataSet, {exports_md5, MD5, nil, []}).
+
 validate_compile_opts(Opts, Defs, Unreachable, Line, E) ->
   lists:flatmap(fun (Opt) -> validate_compile_opt(Opt, Defs, Unreachable, Line, E) end, Opts).
 
@@ -406,9 +426,10 @@ build(Module, Line, File, E) ->
     {derive, [], accumulate, []},
     {dialyzer, [], accumulate, []},
     {external_resource, [], accumulate, []},
+    {nifs, [], accumulate, []},
     {on_definition, [], accumulate, []},
-    {type, [], accumulate, []},
     {opaque, [], accumulate, []},
+    {type, [], accumulate, []},
     {typep, [], accumulate, []},
     {spec, [], accumulate, []},
     {callback, [], accumulate, []},
@@ -420,7 +441,7 @@ build(Module, Line, File, E) ->
     {?counter_attr, 0}
   ]),
 
-  Persisted = [behaviour, dialyzer, external_resource, on_load, vsn, nifs],
+  Persisted = [behaviour, dialyzer, external_resource, nifs, on_load, vsn],
   ets:insert(DataBag, [{persisted_attributes, Attr} || Attr <- Persisted]),
 
   OnDefinition =
@@ -545,10 +566,11 @@ bag_lookup_element(Table, Name, Pos) ->
 
 beam_location(ModuleAsCharlist) ->
   case get(elixir_compiler_dest) of
-    Dest when is_binary(Dest) ->
-      filename:join(elixir_utils:characters_to_list(Dest), ModuleAsCharlist ++ ".beam");
+    {Dest, Forceload} when is_binary(Dest) ->
+      {filename:join(elixir_utils:characters_to_list(Dest), ModuleAsCharlist ++ ".beam"),
+       Forceload};
     _ ->
-      ""
+      {"", true}
   end.
 
 %% Integration with elixir_compiler that makes the module available
@@ -559,17 +581,17 @@ checker_info() ->
     _ -> 'Elixir.Module.ParallelChecker':get()
   end.
 
-spawn_parallel_checker({_, nil}, _Module, _ModuleMap) ->
+spawn_parallel_checker({_, nil}, _Module, _ModuleMap, _BeamLocation) ->
   ok;
-spawn_parallel_checker(CheckerInfo, Module, ModuleMap) ->
+spawn_parallel_checker(CheckerInfo, Module, ModuleMap, BeamLocation) ->
   Log =
     case erlang:get(elixir_code_diagnostics) of
       {_, false} -> false;
       _ -> true
     end,
-  'Elixir.Module.ParallelChecker':spawn(CheckerInfo, Module, ModuleMap, Log).
+  'Elixir.Module.ParallelChecker':spawn(CheckerInfo, Module, ModuleMap, BeamLocation, Log).
 
-make_module_available(Module, Binary) ->
+make_module_available(Module, Binary, Loaded) ->
   case get(elixir_module_binaries) of
     Current when is_list(Current) ->
       put(elixir_module_binaries, [{Module, Binary} | Current]);
@@ -582,7 +604,7 @@ make_module_available(Module, Binary) ->
       ok;
     {PID, _} ->
       Ref = make_ref(),
-      PID ! {module_available, self(), Ref, get(elixir_compiler_file), Module, Binary},
+      PID ! {module_available, self(), Ref, get(elixir_compiler_file), Module, Binary, Loaded},
       receive {Ref, ack} -> ok end
   end.
 
