@@ -1,17 +1,29 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+%% SPDX-FileCopyrightText: 2012 Plataformatec
+
 %% Convenience functions used throughout elixir source code
 %% for ast manipulation and querying.
 -module(elixir_utils).
--export([get_line/1, generated/1,
+-export([get_line/1, get_line/2, get_file/2, generated/1,
   split_last/1, split_opts/1, noop/0, var_context/2,
   characters_to_list/1, characters_to_binary/1, relative_to_cwd/1,
   macro_name/1, returns_boolean/1, caller/4, meta_keep/1,
   read_file_type/1, read_file_type/2, read_link_type/1, read_posix_mtime_and_size/1,
-  change_posix_time/2, change_universal_time/2,
-  guard_op/2, extract_splat_guards/1, extract_guards/1,
+  change_posix_time/2, change_universal_time/2, var_info/2,
+  guard_op/2, guard_info/1, extract_splat_guards/1, extract_guards/1,
   erlang_comparison_op_to_elixir/1, erl_fa_to_elixir_fa/2, jaro_similarity/2]).
 -include("elixir.hrl").
 -include_lib("kernel/include/file.hrl").
 
+var_info(Name, Kind) when Kind == nil; is_integer(Kind) ->
+  io_lib:format("\"~ts\"", [Name]);
+var_info(Name, Kind) ->
+  io_lib:format("\"~ts\" (context ~ts)", [Name, elixir_aliases:inspect(Kind)]).
+
+guard_info(#elixir_ex{prematch={_, _, {bitsize, _}}}) -> "bitstring size specifier";
+guard_info(_) -> "guard".
+
 macro_name(Macro) ->
   list_to_atom("MACRO-" ++ atom_to_list(Macro)).
 
@@ -164,6 +176,18 @@ get_line(Opts) when is_list(Opts) ->
     _ -> 0
   end.
 
+get_line(Meta, Env) when is_list(Meta) ->
+  case lists:keyfind(line, 1, Meta) of
+    {line, LineOpt} when is_integer(LineOpt) -> LineOpt;
+    false -> ?key(Env, line)
+  end.
+
+get_file(Meta, Env) when is_list(Meta) ->
+  case lists:keyfind(file, 1, Meta) of
+    {file, FileOpt} when is_binary(FileOpt) -> FileOpt;
+    false -> ?key(Env, file)
+  end.
+
 generated([{generated, true} | _] = Meta) -> Meta;
 generated(Meta) -> [{generated, true} | Meta].
 
