@@ -1,3 +1,6 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+
 -module(iex).
 -export([start/0, start/2, shell/0, sync_remote/2]).
 
@@ -72,7 +75,7 @@ sync_remote(Parent, Ref) ->
   receive {done, Ref} -> ok end.
 
 start_mfa(Args, MFA) ->
-  Opts = [{dot_iex_path, get_dot_iex(Args)}, {on_eof, halt}],
+  Opts = [{dot_iex, get_dot_iex(Args)}, {on_eof, halt}],
   {?MODULE, start, [Opts, MFA]}.
 
 get_dot_iex(["--dot-iex", H | _]) -> elixir_utils:characters_to_binary(H);
