@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule Code.Formatter.LiteralsTest do
@@ -6,7 +10,6 @@ defmodule Code.Formatter.LiteralsTest do
   import CodeFormatterHelpers
 
   @short_length [line_length: 10]
-  @keep_charlists [normalize_charlists_as_sigils: false]
 
   describe "integers" do
     test "in decimal base" do
@@ -112,10 +115,6 @@ defmodule Code.Formatter.LiteralsTest do
 
     test "quoted operators" do
       assert_same ~S[:"::"]
-      assert_same ~S[:"..//"]
-      assert_format ~S[:..//], ~S[:"..//"]
-      assert_format ~S{[..//: 1]}, ~S{["..//": 1]}
-      assert_same ~S{["..//": 1]}
     end
 
     test "uses double quotes even when single quotes are used" do
@@ -198,83 +197,49 @@ defmodule Code.Formatter.LiteralsTest do
 
   describe "charlists" do
     test "without escapes" do
-      assert_format ~S[''], ~S[~c""]
-      assert_format ~S[' '], ~S[~c" "]
-      assert_format ~S['foo'], ~S[~c"foo"]
-
-      assert_same ~S[''], @keep_charlists
-      assert_same ~S[' '], @keep_charlists
-      assert_same ~S['foo'], @keep_charlists
+      assert_same ~S['']
+      assert_same ~S[' ']
+      assert_same ~S['foo']
     end
 
     test "with escapes" do
-      assert_format ~S['f\a\b\ro'], ~S[~c"f\a\b\ro"]
-      assert_format ~S['single \' quote'], ~S[~c"single ' quote"]
-      assert_format ~S['double " quote'], ~S[~c'double " quote']
-      assert_format ~S['escaped \" quote'], ~S[~c'escaped \" quote']
-      assert_format ~S['\\"'], ~S[~c'\\"']
-
-      assert_same ~S['f\a\b\ro'], @keep_charlists
-      assert_same ~S['single \' quote'], @keep_charlists
-      assert_same ~S['double " quote'], @keep_charlists
-      assert_same ~S['escaped \" quote'], @keep_charlists
-      assert_same ~S['\\"'], @keep_charlists
+      assert_same ~S['f\a\b\ro']
+      assert_same ~S['single \' quote']
+      assert_same ~S['double " quote']
+      assert_same ~S['escaped \" quote']
+      assert_same ~S['\\"']
     end
 
     test "keeps literal new lines" do
-      assert_format """
-                    'fo
-                    o'
-                    """,
-                    """
-                    ~c"fo
-                    o"
-                    """
-
       assert_same """
-                  'fo
-                  o'
-                  """,
-                  @keep_charlists
+      'fo
+      o'
+      """
     end
 
     test "with interpolation" do
-      assert_format ~S['one #{2} three'], ~S[~c"one #{2} three"]
-      assert_format ~S['#{1}\n \\ " \"'], ~S[~c'#{1}\n \\ " \"']
-
-      assert_same ~S['one #{2} three'], @keep_charlists
-      assert_same ~S['#{1}\n \\ " \"'], @keep_charlists
+      assert_same ~S['one #{2} three']
+      assert_same ~S['#{1}\n \\ " \"']
     end
 
     test "with escape and interpolation" do
-      assert_format ~S['one\n\'#{2}\'\nthree'], ~S[~c"one\n'#{2}'\nthree"]
-      assert_format ~S['one\n"#{2}"\nthree'], ~S[~c'one\n"#{2}"\nthree']
-
-      assert_same ~S['one\n\'#{2}\'\nthree'], @keep_charlists
+      assert_same ~S['one\n\'#{2}\'\nthree']
     end
 
     test "with interpolation on line limit" do
-      assert_format ~S"""
-                    'one #{"two"} three'
-                    """,
-                    ~S"""
-                    ~c"one #{"two"} three"
-                    """,
-                    @short_length
+      assert_same ~S"""
+                  'one #{"two"} three'
+                  """,
+                  @short_length
     end
 
     test "literal new lines don't count towards line limit" do
-      assert_format ~S"""
-                    'one
-                    #{"two"}
-                    three'
-                    """,
-                    ~S"""
-                    ~c"one
-                    #{"two"}
-                    three"
-                    """,
-                    @short_length
+      assert_same ~S"""
+                  'one
+                  #{"two"}
+                  three'
+                  """,
+                  @short_length
     end
   end
 
@@ -403,149 +368,45 @@ defmodule Code.Formatter.LiteralsTest do
 
   describe "charlist heredocs" do
     test "without escapes" do
-      assert_format ~S"""
-                    '''
-                    hello
-                    '''
-                    """,
-                    ~S'''
-                    ~c"""
-                    hello
-                    """
-                    '''
-
       assert_same ~S"""
-                  '''
-                  hello
-                  '''
-                  """,
-                  @keep_charlists
+      '''
+      hello
+      '''
+      """
     end
 
     test "with escapes" do
-      assert_format ~S"""
-                    '''
-                    f\a\b\ro
-                    '''
-                    """,
-                    ~S'''
-                    ~c"""
-                    f\a\b\ro
-                    """
-                    '''
-
-      assert_format ~S"""
-                    '''
-                    multiple "\"" quotes
-                    '''
-                    """,
-                    ~S'''
-                    ~c"""
-                    multiple "\"" quotes
-                    """
-                    '''
+      assert_same ~S"""
+      '''
+      f\a\b\ro
+      '''
+      """
 
       assert_same ~S"""
-                  '''
-                  f\a\b\ro
-                  '''
-                  """,
-                  @keep_charlists
-
-      assert_same ~S"""
-                  '''
-                  multiple "\"" quotes
-                  '''
-                  """,
-                  @keep_charlists
+      '''
+      multiple "\"" quotes
+      '''
+      """
     end
 
     test "with interpolation" do
-      assert_format ~S"""
-                    '''
-                    one
-                    #{2}
-                    three
-                    '''
-                    """,
-                    ~S'''
-                    ~c"""
-                    one
-                    #{2}
-                    three
-                    """
-                    '''
-
-      assert_format ~S"""
-                    '''
-                    one
-                    "
-                    #{2}
-                    "
-                    three
-                    '''
-                    """,
-                    ~S'''
-                    ~c"""
-                    one
-                    "
-                    #{2}
-                    "
-                    three
-                    """
-                    '''
+      assert_same ~S"""
+      '''
+      one
+      #{2}
+      three
+      '''
+      """
 
       assert_same ~S"""
-                  '''
-                  one
-                  #{2}
-                  three
-                  '''
-                  """,
-                  @keep_charlists
-
-      assert_same ~S"""
-                  '''
-                  one
-                  "
-                  #{2}
-                  "
-                  three
-                  '''
-                  """,
-                  @keep_charlists
-    end
-
-    test "with interpolation on line limit" do
-      assert_format ~S"""
-                    '''
-                    one #{"two two"} three
-                    '''
-                    """,
-                    ~S'''
-                    ~c"""
-                    one #{"two two"} three
-                    """
-                    ''',
-                    @short_length
-    end
-
-    test "literal new lines don't count towards line limit" do
-      assert_format ~S"""
-                    '''
-                    one
-                    #{"two"}
-                    three
-                    '''
-                    """,
-                    ~S'''
-                    ~c"""
-                    one
-                    #{"two"}
-                    three
-                    """
-                    ''',
-                    @short_length
+      '''
+      one
+      "
+      #{2}
+      "
+      three
+      '''
+      """
     end
   end
 end
