@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("test_helper.exs", __DIR__)
 
 defmodule ExceptionTest do
@@ -9,6 +13,24 @@ defmodule ExceptionTest do
 
   doctest Exception
 
+  doctest RuntimeError
+  doctest SystemLimitError
+  doctest MismatchedDelimiterError
+  doctest SyntaxError
+  doctest TokenMissingError
+  doctest BadBooleanError
+  doctest UndefinedFunctionError
+  doctest FunctionClauseError
+  doctest Protocol.UndefinedError
+  doctest UnicodeConversionError
+  doctest Enum.OutOfBoundsError
+  doctest Enum.EmptyError
+  doctest File.Error
+  doctest File.CopyError
+  doctest File.RenameError
+  doctest File.LinkError
+  doctest ErlangError
+
   test "message/1" do
     defmodule BadException do
       def message(exception) do
@@ -552,20 +574,33 @@ defmodule ExceptionTest do
     end
 
     test "annotates function clause errors" do
-      assert blame_message(Access, & &1.fetch(:foo, :bar)) =~ """
-             no function clause matching in Access.fetch/2
+      import PathHelpers
 
-             The following arguments were given to Access.fetch/2:
+      write_beam(
+        defmodule ExampleModule do
+          def fun(arg1, arg2)
+          def fun(:one, :one), do: :ok
+          def fun(:two, :two), do: :ok
+        end
+      )
+
+      message = blame_message(ExceptionTest.ExampleModule, & &1.fun(:three, :four))
+
+      assert message =~ """
+             no function clause matching in ExceptionTest.ExampleModule.fun/2
+
+             The following arguments were given to ExceptionTest.ExampleModule.fun/2:
 
                  # 1
-                 :foo
+                 :three
 
                  # 2
-                 :bar
+                 :four
 
-             Attempted function clauses (showing 5 out of 5):
+             Attempted function clauses (showing 2 out of 2):
 
-                 def fetch(-%module{} = container-, key)
+                 def fun(-:one-, -:one-)
+                 def fun(-:two-, -:two-)
              """
     end
 
@@ -573,17 +608,17 @@ defmodule ExceptionTest do
       assert blame_message(Enum, & &1.map(:ok)) == """
              function Enum.map/1 is undefined or private. Did you mean:
 
-                   * map/2
+                 * map/2
              """
 
       assert blame_message(Enum, & &1.man(:ok)) == """
              function Enum.man/1 is undefined or private. Did you mean:
 
-                   * map/2
-                   * max/1
-                   * max/2
-                   * max/3
-                   * min/1
+                 * map/2
+                 * max/1
+                 * max/2
+                 * max/3
+                 * min/1
              """
 
       message = blame_message(:erlang, & &1.gt_cookie())
@@ -681,12 +716,6 @@ defmodule ExceptionTest do
                "function :erlang.hash/2 is undefined or private, use erlang:phash2/2 instead"
     end
 
-    test "annotates undefined key error with nil hints" do
-      assert blame_message(nil, & &1.foo) ==
-               "key :foo not found in: nil\n\nIf you are using the dot syntax, " <>
-                 "such as map.field, make sure the left-hand side of the dot is a map"
-    end
-
     test "annotates undefined function clause error with nil hints" do
       assert blame_message(nil, & &1.foo()) ==
                "function nil.foo/0 is undefined. If you are using the dot syntax, " <>
@@ -702,34 +731,71 @@ defmodule ExceptionTest do
     test "annotates key error with suggestions if keys are atoms" do
       message = blame_message(%{first: nil, second: nil}, fn map -> map.firts end)
 
-      assert message == """
-             key :firts not found in: %{first: nil, second: nil}. Did you mean:
+      assert message ==
+               """
+               key :firts not found in:
+
+                   %{first: nil, second: nil}
+
+               Did you mean:
 
                    * :first
-             """
+               """
 
       message = blame_message(%{"first" => nil, "second" => nil}, fn map -> map.firts end)
 
-      assert message == "key :firts not found in: %{\"first\" => nil, \"second\" => nil}"
+      assert message == """
+             key :firts not found in:
+
+                 %{"first" => nil, "second" => nil}
+             """
 
       message =
         blame_message(%{"first" => nil, "second" => nil}, fn map -> Map.fetch!(map, "firts") end)
 
-      assert message == "key \"firts\" not found in: %{\"first\" => nil, \"second\" => nil}"
+      assert message ==
+               """
+               key "firts" not found in:
+
+                   %{"first" => nil, "second" => nil}
+               """
 
       message =
-        blame_message([first: nil, second: nil], fn kwlist -> Keyword.fetch!(kwlist, :firts) end)
+        blame_message(
+          [
+            created_at: nil,
+            updated_at: nil,
+            deleted_at: nil,
+            started_at: nil,
+            finished_at: nil
+          ],
+          fn kwlist ->
+            Keyword.fetch!(kwlist, :inserted_at)
+          end
+        )
 
       assert message == """
-             key :firts not found in: [first: nil, second: nil]. Did you mean:
+             key :inserted_at not found in:
 
-                   * :first
+                 [
+                   created_at: nil,
+                   updated_at: nil,
+                   deleted_at: nil,
+                   started_at: nil,
+                   finished_at: nil
+                 ]
+
+             Did you mean:
+
+                 * :created_at
+                 * :finished_at
+                 * :started_at
              """
     end
 
     test "annotates key error with suggestions for structs" do
       message = blame_message(%URI{}, fn map -> map.schema end)
-      assert message =~ "key :schema not found in: %URI{"
+      assert message =~ "key :schema not found in:\n\n    %URI{"
       assert message =~ "Did you mean:"
       assert message =~ "* :scheme"
     end
@@ -818,15 +884,24 @@ defmodule ExceptionTest do
 
   describe "blaming unit tests" do
     test "annotates clauses errors" do
-      args = [%{}, :key, nil]
+      import PathHelpers
+
+      write_beam(
+        defmodule BlameModule do
+          def fun(arg), do: arg
+        end
+      )
+
+      args = [nil]
 
       {exception, stack} =
-        Exception.blame(:error, :function_clause, [{Keyword, :pop, args, [line: 13]}])
+        Exception.blame(:error, :function_clause, [{BlameModule, :fun, args, [line: 13]}])
 
       assert %FunctionClauseError{kind: :def, args: ^args, clauses: [_]} = exception
-      assert stack == [{Keyword, :pop, 3, [line: 13]}]
+      assert stack == [{BlameModule, :fun, 1, [line: 13]}]
     end
 
+    @tag :require_ast
     test "annotates args and clauses from mfa" do
       import PathHelpers
 
@@ -900,68 +975,68 @@ defmodule ExceptionTest do
     import Exception, only: [message: 1]
 
     test "RuntimeError" do
-      assert %RuntimeError{} |> message == "runtime error"
-      assert %RuntimeError{message: "unexpected roquefort"} |> message == "unexpected roquefort"
+      assert %RuntimeError{} |> message() == "runtime error"
+      assert %RuntimeError{message: "unexpected roquefort"} |> message() == "unexpected roquefort"
     end
 
     test "ArithmeticError" do
-      assert %ArithmeticError{} |> message == "bad argument in arithmetic expression"
+      assert %ArithmeticError{} |> message() == "bad argument in arithmetic expression"
 
       assert %ArithmeticError{message: "unexpected camembert"}
-             |> message == "unexpected camembert"
+             |> message() == "unexpected camembert"
     end
 
     test "ArgumentError" do
-      assert %ArgumentError{} |> message == "argument error"
-      assert %ArgumentError{message: "unexpected comté"} |> message == "unexpected comté"
+      assert %ArgumentError{} |> message() == "argument error"
+      assert %ArgumentError{message: "unexpected comté"} |> message() == "unexpected comté"
     end
 
     test "KeyError" do
-      assert %KeyError{} |> message == "key nil not found"
-      assert %KeyError{message: "key missed"} |> message == "key missed"
+      assert %KeyError{} |> message() == "key nil not found"
+      assert %KeyError{message: "key missed"} |> message() == "key missed"
     end
 
     test "Enum.OutOfBoundsError" do
-      assert %Enum.OutOfBoundsError{} |> message == "out of bounds error"
+      assert %Enum.OutOfBoundsError{} |> message() == "out of bounds error"
 
       assert %Enum.OutOfBoundsError{message: "the brie is not on the table"}
-             |> message == "the brie is not on the table"
+             |> message() == "the brie is not on the table"
     end
 
     test "Enum.EmptyError" do
-      assert %Enum.EmptyError{} |> message == "empty error"
+      assert %Enum.EmptyError{} |> message() == "empty error"
 
       assert %Enum.EmptyError{message: "there is no saint-nectaire left!"}
-             |> message == "there is no saint-nectaire left!"
+             |> message() == "there is no saint-nectaire left!"
     end
 
     test "UndefinedFunctionError" do
-      assert %UndefinedFunctionError{} |> message == "undefined function"
+      assert %UndefinedFunctionError{} |> message() == "undefined function"
 
       assert %UndefinedFunctionError{module: Kernel, function: :bar, arity: 1}
-             |> message == "function Kernel.bar/1 is undefined or private"
+             |> message() == "function Kernel.bar/1 is undefined or private"
 
       assert %UndefinedFunctionError{module: Foo, function: :bar, arity: 1}
-             |> message ==
+             |> message() ==
                "function Foo.bar/1 is undefined (module Foo is not available). " <>
                  "Make sure the module name is correct and has been specified in full (or that an alias has been defined)"
 
       assert %UndefinedFunctionError{module: nil, function: :bar, arity: 3}
-             |> message == "function nil.bar/3 is undefined"
+             |> message() == "function nil.bar/3 is undefined"
 
       assert %UndefinedFunctionError{module: nil, function: :bar, arity: 0}
-             |> message == "function nil.bar/0 is undefined"
+             |> message() == "function nil.bar/0 is undefined"
     end
 
     test "FunctionClauseError" do
-      assert %FunctionClauseError{} |> message == "no function clause matches"
+      assert %FunctionClauseError{} |> message() == "no function clause matches"
 
       assert %FunctionClauseError{module: Foo, function: :bar, arity: 1}
-             |> message == "no function clause matching in Foo.bar/1"
+             |> message() == "no function clause matching in Foo.bar/1"
     end
 
     test "ErlangError" do
-      assert %ErlangError{original: :sample} |> message == "Erlang error: :sample"
+      assert %ErlangError{original: :sample} |> message() == "Erlang error: :sample"
     end
 
     test "MissingApplicationsError" do
@@ -969,7 +1044,7 @@ defmodule ExceptionTest do
                apps: [{:logger, "~> 1.18"}, {:ex_unit, Version.parse_requirement!(">= 0.0.0")}],
                description: "applications are required"
              }
-             |> message == """
+             |> message() == """
              applications are required
 
              To address this, include these applications as your dependencies:
