@@ -1,9 +1,14 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule Kernel.DialyzerTest do
   use ExUnit.Case, async: true
 
   @moduletag :dialyzer
+  @moduletag :require_ast
   import PathHelpers
 
   setup_all do
@@ -17,7 +22,7 @@ defmodule Kernel.DialyzerTest do
       |> String.to_charlist()
 
     # Some OSs (like Windows) do not provide the HOME environment variable.
-    unless System.get_env("HOME") do
+    if !System.get_env("HOME") do
       System.put_env("HOME", System.user_home())
     end
 
@@ -26,6 +31,8 @@ defmodule Kernel.DialyzerTest do
       :elixir,
       :elixir_env,
       :elixir_erl_pass,
+      :maps,
+      :sets,
       ArgumentError,
       Atom,
       Code,
@@ -39,9 +46,13 @@ defmodule Kernel.DialyzerTest do
       List,
       Macro,
       Macro.Env,
+      MapSet,
       Module,
+      Protocol,
       String,
-      String.Chars
+      String.Chars,
+      Task,
+      Task.Supervisor
     ]
 
     files = Enum.map(mods, &:code.which/1)
@@ -49,7 +60,9 @@ defmodule Kernel.DialyzerTest do
 
     # Compile Dialyzer fixtures
     source_files = Path.wildcard(Path.join(fixture_path("dialyzer"), "*"))
-    {:ok, _, _} = Kernel.ParallelCompiler.compile_to_path(source_files, dir)
+
+    {:ok, _, _} =
+      Kernel.ParallelCompiler.compile_to_path(source_files, dir, return_diagnostics: true)
 
     {:ok, [base_dir: dir, base_plt: plt]}
   end
@@ -112,28 +125,6 @@ defmodule Kernel.DialyzerTest do
     assert_dialyze_no_warnings!(context)
   end
 
-  test "no warnings on struct update", context do
-    copy_beam!(context, Dialyzer.StructUpdate)
-    assert_dialyze_no_warnings!(context)
-  end
-
-  @tag warnings: [:specdiffs]
-  test "no warnings on protocol calls with opaque types", context do
-    alias Dialyzer.ProtocolOpaque
-
-    copy_beam!(context, ProtocolOpaque)
-    copy_beam!(context, ProtocolOpaque.Entity)
-    copy_beam!(context, ProtocolOpaque.Entity.Any)
-    copy_beam!(context, ProtocolOpaque.Duck)
-    assert_dialyze_no_warnings!(context)
-
-    # Also ensure no warnings after consolidation.
-    Code.prepend_path(context.base_dir)
-    {:ok, binary} = Protocol.consolidate(ProtocolOpaque.Entity, [ProtocolOpaque.Duck, Any])
-    File.write!(Path.join(context.outdir, "#{ProtocolOpaque.Entity}.beam"), binary)
-    assert_dialyze_no_warnings!(context)
-  end
-
   test "no warnings on and/2 and or/2", context do
     copy_beam!(context, Dialyzer.BooleanCheck)
     assert_dialyze_no_warnings!(context)
@@ -164,6 +155,11 @@ defmodule Kernel.DialyzerTest do
     assert_dialyze_no_warnings!(context)
   end
 
+  test "no warnings on with when multiple else clauses and one is a no_return", context do
+    copy_beam!(context, Dialyzer.WithThrowingElse)
+    assert_dialyze_no_warnings!(context)
+  end
+
   test "no warnings on defmacrop", context do
     copy_beam!(context, Dialyzer.Defmacrop)
     assert_dialyze_no_warnings!(context)
@@ -184,6 +180,11 @@ defmodule Kernel.DialyzerTest do
     assert_dialyze_no_warnings!(context)
   end
 
+  test "no warning due to opaqueness edge cases", context do
+    copy_beam!(context, Dialyzer.Opaqueness)
+    assert_dialyze_no_warnings!(context)
+  end
+
   defp copy_beam!(context, module) do
     name = "#{module}.beam"
     File.cp!(Path.join(context.base_dir, name), Path.join(context.outdir, name))
