@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule Kernel.DocsTest do
@@ -119,6 +123,14 @@ defmodule Kernel.DocsTest do
 
           def two_good_names(first, :ok), do: first
           def two_good_names(second, :error), do: second
+
+          def really_long_signature(
+                really_long_var_named_one,
+                really_long_var_named_two,
+                really_long_var_named_three
+              ) do
+            {really_long_var_named_one, really_long_var_named_two, really_long_var_named_three}
+          end
         end
       )
 
@@ -128,6 +140,7 @@ defmodule Kernel.DocsTest do
       assert [
                arg_names,
                only_underscore,
+               really_long_signature,
                two_good_names,
                with_defaults,
                with_map_and_default,
@@ -141,6 +154,12 @@ defmodule Kernel.DocsTest do
       # only_underscore/1
       assert {{:only_underscore, 1}, ["only_underscore(_)"]} = only_underscore
 
+      # really_long_signature/3
+      assert {{:really_long_signature, 3},
+              [
+                "really_long_signature(really_long_var_named_one, really_long_var_named_two, really_long_var_named_three)"
+              ]} = really_long_signature
+
       # two_good_names/2
       assert {{:two_good_names, 2}, ["two_good_names(first, atom)"]} = two_good_names
 
@@ -160,6 +179,8 @@ defmodule Kernel.DocsTest do
     end
 
     test "includes docs for functions, modules, types and callbacks" do
+      line = __ENV__.line
+
       write_beam(
         defmodule SampleDocs do
           @moduledoc "Module doc"
@@ -224,6 +245,12 @@ defmodule Kernel.DocsTest do
           module = Module
           module.add_doc(__MODULE__, __ENV__.line, :def, {:nullary, 0}, [], "add_doc")
           def nullary, do: 0
+
+          defmodule SampleBehaviour do
+            @callback foo(any()) :: any()
+          end
+
+          @behaviour SampleBehaviour
         end
       )
 
@@ -232,7 +259,19 @@ defmodule Kernel.DocsTest do
 
       assert module_doc == "Module doc"
 
-      assert %{authors: "Elixir Contributors", purpose: :test} = module_doc_meta
+      file = String.to_charlist(__ENV__.file)
+
+      source_annos = [:erl_anno.new({line + 3, 19})]
+
+      assert %{
+               # Generated meta
+               source_path: ^file,
+               source_annos: ^source_annos,
+               behaviours: [SampleDocs.SampleBehaviour],
+               # User meta
+               authors: "Elixir Contributors",
+               purpose: :test
+             } = module_doc_meta
 
       [
         callback_bar,
@@ -256,9 +295,16 @@ defmodule Kernel.DocsTest do
       assert {{:callback, :bar, 0}, _, [], :hidden, %{}} = callback_bar
       assert {{:callback, :baz, 2}, _, [], :none, %{}} = callback_baz
 
+      source_annos = [:erl_anno.new({line + 20, 21})]
+
       assert {{:callback, :foo, 1}, _, [], %{"en" => "Callback doc"},
-              %{since: "1.2.3", deprecated: "use baz/2 instead", color: :blue, stable: true}} =
-               callback_foo
+              %{
+                source_annos: ^source_annos,
+                since: "1.2.3",
+                deprecated: "use baz/2 instead",
+                color: :blue,
+                stable: true
+              }} = callback_foo
 
       assert {{:callback, :callback_multi, 1}, _, [], %{"en" => "Callback with multiple clauses"},
               %{}} = callback_multi
@@ -266,7 +312,8 @@ defmodule Kernel.DocsTest do
       assert {{:function, :__struct__, 0}, _, ["%Kernel.DocsTest.SampleDocs{}"],
               %{"en" => "My struct"}, %{}} = function_struct_0
 
-      assert {{:function, :__struct__, 1}, _, ["__struct__(kv)"], :none, %{}} = function_struct_1
+      assert {{:function, :__struct__, 1}, _, ["__struct__(kv)"], :hidden, %{}} =
+               function_struct_1
 
       assert {{:function, :bar, 1}, _, ["bar(arg)"], %{"en" => "Multiple function head doc"},
               %{deprecated: "something else"}} = function_bar
@@ -274,8 +321,11 @@ defmodule Kernel.DocsTest do
       assert {{:function, :baz, 1}, _, ["baz(arg)"], %{"en" => "Multiple function head and docs"},
               %{since: "1.2.3"}} = function_baz
 
+      source_annos = [:erl_anno.new({line + 42, 15})]
+
       assert {{:function, :foo, 1}, _, ["foo(arg \\\\ 0)"], %{"en" => "Function doc"},
               %{
+                source_annos: ^source_annos,
                 since: "1.2.3",
                 deprecated: "use baz/2 instead",
                 color: :blue,
@@ -288,8 +338,10 @@ defmodule Kernel.DocsTest do
 
       assert {{:function, :qux, 1}, _, ["qux(bool)"], :hidden, %{}} = function_qux
 
-      assert {{:macro, :is_zero, 1}, _, ["is_zero(v)"], %{"en" => "A guard"}, %{guard: true}} =
-               guard_is_zero
+      source_annos = [:erl_anno.new({line + 60, 20})]
+
+      assert {{:macro, :is_zero, 1}, _, ["is_zero(v)"], %{"en" => "A guard"},
+              %{source_annos: ^source_annos, guard: true}} = guard_is_zero
 
       assert {{:macrocallback, :macrocallback_multi, 1}, _, [],
               %{"en" => "Macrocallback with multiple clauses"}, %{}} = macrocallback_multi
@@ -299,8 +351,14 @@ defmodule Kernel.DocsTest do
 
       assert {{:type, :bar, 1}, _, [], %{"en" => "Opaque type doc"}, %{opaque: true}} = type_bar
 
-      assert {{:type, :foo, 1}, _, [], %{"en" => "Type doc"}, %{since: "1.2.3", color: :red}} =
-               type_foo
+      source_annos = [:erl_anno.new({line + 12, 17})]
+
+      assert {{:type, :foo, 1}, _, [], %{"en" => "Type doc"},
+              %{
+                source_annos: ^source_annos,
+                since: "1.2.3",
+                color: :red
+              }} = type_foo
     end
   end
 
@@ -366,11 +424,34 @@ defmodule Kernel.DocsTest do
            ] = Enum.sort(function_docs)
   end
 
-  describe "special signatures" do
-    test "fn" do
-      {:docs_v1, _, _, _, _, _, docs} = Code.fetch_docs(Kernel.SpecialForms)
-      {_, _, fn_docs, _, _} = Enum.find(docs, &match?({_, :fn, 1}, elem(&1, 0)))
-      assert fn_docs == ["fn"]
-    end
+  test "generated functions are annotated as such" do
+    line = __ENV__.line
+
+    write_beam(
+      defmodule ToBeUsed do
+        defmacro __using__(_) do
+          quote generated: true do
+            @doc "Hello"
+            def foo, do: :bar
+          end
+        end
+      end
+    )
+
+    write_beam(
+      defmodule WillBeUsing do
+        use ToBeUsed
+      end
+    )
+
+    {:docs_v1, _, _, _, _, _, docs} = Code.fetch_docs(WillBeUsing)
+
+    doc_anno = :erl_anno.new(line + 15)
+    source_anno = :erl_anno.set_generated(true, :erl_anno.new(line + 15))
+
+    assert [
+             {{:function, :foo, 0}, ^doc_anno, ["foo()"], %{"en" => "Hello"},
+              %{source_annos: [^source_anno]}}
+           ] = docs
   end
 end
