@@ -1,9 +1,14 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule Kernel.LexicalTrackerTest do
   use ExUnit.Case, async: true
 
   alias Kernel.LexicalTracker, as: D
+  defstruct used_by_tests: :ok
 
   setup do
     {:ok, pid} = D.start_link()
@@ -453,7 +458,7 @@ defmodule Kernel.LexicalTrackerTest do
         Code.eval_string("""
         defmodule Kernel.LexicalTrackerTest.PatternGuardsCompile do
           %URI{} = URI.parse("/")
-          case 1..3 do
+          case Range.new(1, 3) do
             range when is_struct(range, Range) -> :ok
           end
           Kernel.LexicalTracker.references(__ENV__.lexical_tracker)
@@ -567,5 +572,22 @@ defmodule Kernel.LexicalTrackerTest do
       refute Config in exports
       refute Config in runtime
     end
+
+    test "defimpl does not add dependencies on for only on impl" do
+      {{compile, exports, runtime, _}, _binding} =
+        Code.eval_string("""
+        defimpl String.Chars, for: Kernel.LexicalTrackerTest do
+          def to_string(val), do: val.used_by_tests
+          Kernel.LexicalTracker.references(__ENV__.lexical_tracker)
+        end |> elem(3)
+        """)
+
+      refute String.Chars in compile
+      assert String.Chars in exports
+
+      refute Kernel.LexicalTrackerTest in compile
+      refute Kernel.LexicalTrackerTest in exports
+      refute Kernel.LexicalTrackerTest in runtime
+    end
   end
 end
