@@ -1,3 +1,6 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule Kernel.ParserTest do
@@ -66,13 +69,13 @@ defmodule Kernel.ParserTest do
 
   describe "ternary ops" do
     test "root" do
-      assert parse!("1..2//3") == {:"..//", [line: 1], [1, 2, 3]}
-      assert parse!("(1..2)//3") == {:"..//", [line: 1], [1, 2, 3]}
+      assert parse!("1..2//3") == {:..//, [line: 1], [1, 2, 3]}
+      assert parse!("(1..2)//3") == {:..//, [line: 1], [1, 2, 3]}
     end
 
     test "with do-blocks" do
       assert parse!("foo do end..bar do end//baz do end") == {
-               :"..//",
+               :..//,
                [line: 1],
                [
                  {:foo, [line: 1], [[do: {:__block__, [], []}]]},
@@ -84,7 +87,7 @@ defmodule Kernel.ParserTest do
 
     test "with no parens" do
       assert parse!("1..foo do end//bar bat") == {
-               :"..//",
+               :..//,
                [line: 1],
                [
                  1,
@@ -126,6 +129,49 @@ defmodule Kernel.ParserTest do
       # that are Common-script and non-ASCII
       # assert Code.eval_string("_ℕ𝕩 = 1") == {1, [{:"_ℕ𝕩", 1}]}
     end
+
+    test "handles graphemes inside quoted identifiers" do
+      string_to_quoted =
+        fn code ->
+          Code.string_to_quoted!(code,
+            token_metadata: true,
+            literal_encoder: &{:ok, {:__block__, &2, [&1]}},
+            emit_warnings: false
+          )
+        end
+
+      assert {
+               {:., _, [{:foo, _, nil}, :"➡️"]},
+               [no_parens: true, delimiter: ~S["], line: 1],
+               []
+             } = string_to_quoted.(~S|foo."➡️"|)
+
+      assert {
+               {:., _, [{:foo, _, nil}, :"➡️"]},
+               [no_parens: true, delimiter: ~S['], line: 1],
+               []
+             } = string_to_quoted.(~S|foo.'➡️'|)
+
+      assert {:__block__, [delimiter: ~S["], line: 1], [:"➡️"]} = string_to_quoted.(~S|:"➡️"|)
+
+      assert {:__block__, [delimiter: ~S['], line: 1], [:"➡️"]} = string_to_quoted.(~S|:'➡️'|)
+
+      assert {:__block__, [closing: [line: 1], line: 1],
+              [
+                [
+                  {{:__block__, [delimiter: ~S["], format: :keyword, line: 1], [:"➡️"]},
+                   {:x, [line: 1], nil}}
+                ]
+              ]} = string_to_quoted.(~S|["➡️": x]|)
+
+      assert {:__block__, [closing: [line: 1], line: 1],
+              [
+                [
+                  {{:__block__, [delimiter: ~S['], format: :keyword, line: 1], [:"➡️"]},
+                   {:x, [line: 1], nil}}
+                ]
+              ]} = string_to_quoted.(~S|['➡️': x]|)
+    end
   end
 
   describe "strings/sigils" do
@@ -341,6 +387,16 @@ defmodule Kernel.ParserTest do
                 {[line: 1, column: 1], "atom length must be less than system limit: ", atom}}
     end
 
+    test "avoids crashes on invalid AST" do
+      encoder = fn atom, _meta -> {:ok, {:atom, [], [atom]}} end
+
+      assert {:error, {_, "missing terminator: )", ""}} =
+               Code.string_to_quoted("Module(", static_atoms_encoder: encoder)
+
+      assert {:error, {_, "syntax error before: ", "'('"}} =
+               Code.string_to_quoted("Module()", static_atoms_encoder: encoder)
+    end
+
     test "may return errors" do
       encoder = fn _atom, _meta ->
         {:error, "Invalid atom name"}
@@ -391,6 +447,43 @@ defmodule Kernel.ParserTest do
                {:ok, {:=, context, [List.to_string(nfd_abba), 1]}}
     end
 
+    test "not in" do
+      assert Code.string_to_quoted!("a not in b", columns: true) ==
+               {:not, [line: 1, column: 3],
+                [
+                  {:in, [line: 1, column: 7],
+                   [{:a, [line: 1, column: 1], nil}, {:b, [line: 1, column: 10], nil}]}
+                ]}
+
+      assert Code.string_to_quoted!("a not  in b", columns: true, token_metadata: true) ==
+               {:not, [line: 1, column: 3],
+                [
+                  {:in, [line: 1, column: 8],
+                   [{:a, [line: 1, column: 1], nil}, {:b, [line: 1, column: 11], nil}]}
+                ]}
+
+      assert Code.string_to_quoted!("a\nnot in b", columns: true, token_metadata: true) ==
+               {:not, [newlines: 1, line: 2, column: 1],
+                [
+                  {:in, [line: 2, column: 5],
+                   [{:a, [line: 1, column: 1], nil}, {:b, [line: 2, column: 8], nil}]}
+                ]}
+
+      assert Code.string_to_quoted!("a not in\nb", columns: true, token_metadata: true) ==
+               {:not, [newlines: 1, line: 1, column: 3],
+                [
+                  {:in, [line: 1, column: 7],
+                   [{:a, [line: 1, column: 1], nil}, {:b, [line: 2, column: 1], nil}]}
+                ]}
+
+      assert Code.string_to_quoted!("a\nnot in\nb", columns: true, token_metadata: true) ==
+               {:not, [newlines: 1, line: 2, column: 1],
+                [
+                  {:in, [line: 2, column: 5],
+                   [{:a, [line: 1, column: 1], nil}, {:b, [line: 3, column: 1], nil}]}
+                ]}
+    end
+
     test "handles maps and structs" do
       assert Code.string_to_quoted("%{}", columns: true) ==
                {:ok, {:%{}, [line: 1, column: 1], []}}
@@ -593,6 +686,161 @@ defmodule Kernel.ParserTest do
       assert string_to_quoted.("foo(\n) do\nend") ==
                {:foo, [do: [line: 2], end: [line: 3], newlines: 1, closing: [line: 2], line: 1],
                 [[do: {:__block__, [], []}]]}
+
+      assert string_to_quoted.("foo(\n)(\n)") ==
+               {{:foo, [newlines: 1, closing: [line: 2], line: 1], []},
+                [newlines: 1, closing: [line: 3], line: 1], []}
+    end
+
+    test "adds opening and closing information for single-expression block" do
+      file = "1 + (2 + 3)"
+
+      assert Code.string_to_quoted!(file, token_metadata: true, columns: true) ==
+               {:+, [line: 1, column: 3],
+                [
+                  1,
+                  {:+,
+                   [
+                     parens: [line: 1, column: 5, closing: [line: 1, column: 11]],
+                     line: 1,
+                     column: 8
+                   ], [2, 3]}
+                ]}
+
+      file = "1 + ((2 + 3))"
+
+      assert Code.string_to_quoted!(file, token_metadata: true, columns: true) ==
+               {:+, [line: 1, column: 3],
+                [
+                  1,
+                  {:+,
+                   [
+                     parens: [line: 1, column: 5, closing: [line: 1, column: 13]],
+                     parens: [line: 1, column: 6, closing: [line: 1, column: 12]],
+                     line: 1,
+                     column: 9
+                   ], [2, 3]}
+                ]}
+    end
+
+    test "adds opening and closing information for tuples" do
+      string_to_quoted = &Code.string_to_quoted!(&1, token_metadata: true, columns: true)
+
+      assert string_to_quoted.("{}") ==
+               {:{}, [closing: [line: 1, column: 2], line: 1, column: 1], []}
+
+      assert string_to_quoted.("{123}") ==
+               {:{}, [closing: [line: 1, column: 5], line: 1, column: 1], [123]}
+
+      assert string_to_quoted.("x.{}") ==
+               {{:., [line: 1, column: 2], [{:x, [line: 1, column: 1], nil}, :{}]},
+                [closing: [line: 1, column: 4], line: 1, column: 2], []}
+
+      assert string_to_quoted.("x.{123}") ==
+               {{:., [line: 1, column: 2], [{:x, [line: 1, column: 1], nil}, :{}]},
+                [closing: [line: 1, column: 7], line: 1, column: 2], [123]}
+    end
+
+    test "adds opening and closing information for empty block" do
+      string_to_quoted =
+        &Code.string_to_quoted!(&1, token_metadata: true, columns: true, emit_warnings: false)
+
+      file = "()"
+
+      assert string_to_quoted.(file) ==
+               {:__block__, [parens: [line: 1, column: 1, closing: [line: 1, column: 2]]], []}
+
+      file = "(())"
+
+      assert string_to_quoted.(file) ==
+               {:__block__,
+                [
+                  parens: [line: 1, column: 1, closing: [line: 1, column: 4]],
+                  parens: [line: 1, column: 2, closing: [line: 1, column: 3]]
+                ], []}
+
+      file = """
+      (
+        # Foo
+        (
+          # Bar
+        )
+      )
+      """
+
+      assert string_to_quoted.(file) ==
+               {:__block__,
+                [
+                  end_of_expression: [newlines: 1, line: 6, column: 2],
+                  parens: [line: 1, column: 1, closing: [line: 6, column: 1]],
+                  end_of_expression: [newlines: 1, line: 5, column: 4],
+                  parens: [line: 3, column: 3, closing: [line: 5, column: 3]]
+                ], []}
+    end
+
+    test "adds opening and closing information for stab arguments" do
+      file = "fn () -> x end "
+
+      assert Code.string_to_quoted!(file, token_metadata: true, columns: true) ==
+               {:fn, [closing: [line: 1, column: 12], line: 1, column: 1],
+                [
+                  {:->,
+                   [
+                     parens: [line: 1, column: 4, closing: [line: 1, column: 5]],
+                     line: 1,
+                     column: 7
+                   ], [[], {:x, [line: 1, column: 10], nil}]}
+                ]}
+
+      file = "fn (x, y) -> x end "
+
+      assert Code.string_to_quoted!(file, token_metadata: true, columns: true) ==
+               {
+                 :fn,
+                 [{:closing, [line: 1, column: 16]}, {:line, 1}, {:column, 1}],
+                 [
+                   {:->,
+                    [
+                      parens: [line: 1, column: 4, closing: [line: 1, column: 9]],
+                      line: 1,
+                      column: 11
+                    ],
+                    [
+                      [{:x, [line: 1, column: 5], nil}, {:y, [line: 1, column: 8], nil}],
+                      {:x, [line: 1, column: 14], nil}
+                    ]}
+                 ]
+               }
+
+      file = "if true do (x, y) -> x end"
+
+      assert Code.string_to_quoted!(file, token_metadata: true, columns: true) ==
+               {
+                 :if,
+                 [
+                   {:do, [line: 1, column: 9]},
+                   {:end, [line: 1, column: 24]},
+                   {:line, 1},
+                   {:column, 1}
+                 ],
+                 [
+                   true,
+                   [
+                     do: [
+                       {:->,
+                        [
+                          parens: [line: 1, column: 12, closing: [line: 1, column: 17]],
+                          line: 1,
+                          column: 19
+                        ],
+                        [
+                          [{:x, [line: 1, column: 13], nil}, {:y, [line: 1, column: 16], nil}],
+                          {:x, [line: 1, column: 22], nil}
+                        ]}
+                     ]
+                   ]
+                 ]
+               }
     end
 
     test "with :literal_encoder" do
@@ -609,6 +857,9 @@ defmodule Kernel.ParserTest do
       assert string_to_quoted.("nil") == {:__block__, [line: 1], [nil]}
       assert string_to_quoted.(":one") == {:__block__, [line: 1], [:one]}
 
+      assert string_to_quoted.("true") == {:__block__, [line: 1], [true]}
+      assert string_to_quoted.(":true") == {:__block__, [format: :atom, line: 1], [true]}
+
       assert string_to_quoted.("[one: :two]") == {
                :__block__,
                [{:closing, [line: 1]}, {:line, 1}],
@@ -627,18 +878,25 @@ defmodule Kernel.ParserTest do
       assert string_to_quoted.(~s("""\nhello\n""")) ==
                {:__block__, [delimiter: ~s["""], indentation: 0, line: 1], ["hello\n"]}
 
-      assert string_to_quoted.("'''\nhello\n'''") ==
-               {:__block__, [delimiter: ~s['''], indentation: 0, line: 1], [~c"hello\n"]}
-
       assert string_to_quoted.(~s[fn (1) -> "hello" end]) ==
                {:fn, [closing: [line: 1], line: 1],
                 [
                   {:->, [line: 1],
                    [
-                     [{:__block__, [token: "1", line: 1], [1]}],
+                     [
+                       {:__block__,
+                        [
+                          parens: [line: 1, closing: [line: 1]],
+                          token: "1",
+                          line: 1
+                        ], [1]}
+                     ],
                      {:__block__, [delimiter: "\"", line: 1], ["hello"]}
                    ]}
                 ]}
+
+      assert string_to_quoted.("(1)") ==
+               {:__block__, [parens: [line: 1, closing: [line: 1]], token: "1", line: 1], [1]}
     end
 
     test "adds identifier_location for qualified identifiers" do
@@ -677,6 +935,38 @@ defmodule Kernel.ParserTest do
       assert string_to_quoted.("foo.\nBar\n.\nBaz") ==
                {:__aliases__, [last: [line: 4], line: 1], [{:foo, [line: 1], nil}, :Bar, :Baz]}
     end
+
+    test "adds metadata about assoc operator position in maps" do
+      opts = [
+        literal_encoder: &{:ok, {:__block__, &2, [&1]}},
+        token_metadata: true,
+        columns: true
+      ]
+
+      string_to_quoted = &Code.string_to_quoted!(&1, opts)
+
+      file = "%{:key => 1, {} => {}}"
+
+      assert string_to_quoted.(file) ==
+               {
+                 :%{},
+                 [closing: [line: 1, column: 22], line: 1, column: 1],
+                 [
+                   {{:__block__, [assoc: [line: 1, column: 8], line: 1, column: 3], [:key]},
+                    {:__block__, [token: "1", line: 1, column: 11], [1]}},
+                   {
+                     {:{},
+                      [
+                        assoc: [line: 1, column: 17],
+                        closing: [line: 1, column: 15],
+                        line: 1,
+                        column: 14
+                      ], []},
+                     {:{}, [closing: [line: 1, column: 21], line: 1, column: 20], []}
+                   }
+                 ]
+               }
+    end
   end
 
   describe "syntax errors" do
@@ -747,6 +1037,18 @@ defmodule Kernel.ParserTest do
       )
     end
 
+    test "invalid newline in source" do
+      assert_syntax_error(
+        ["nofile:1:1:", ~s/invalid line break character in comment: \\u2028/],
+        ~c"# This is a \u2028"
+      )
+
+      assert_syntax_error(
+        ["nofile:1:5:", "invalid line break character in comment: \\u2028"],
+        ~c"foo. # This is a \u2028"
+      )
+    end
+
     test "reserved tokens" do
       assert_syntax_error(["nofile:1:1:", "reserved token: __aliases__"], ~c"__aliases__")
       assert_syntax_error(["nofile:1:1:", "reserved token: __block__"], ~c"__block__")
@@ -1194,6 +1496,19 @@ defmodule Kernel.ParserTest do
       assert_syntax_error(["nofile:1:5:", "syntax error before: ?す"], ~c":ok ?す")
     end
 
+    test "character literals take newlines into account" do
+      ExUnit.CaptureIO.capture_io(:stderr, fn ->
+        assert parse!("{?\n}\n{123}") ==
+                 {:__block__, [], [{:{}, [line: 1], ~c"\n"}, {:{}, [line: 3], ~c"{"}]}
+
+        assert parse!("{?\\n}\n{123}") ==
+                 {:__block__, [], [{:{}, [line: 1], ~c"\n"}, {:{}, [line: 2], ~c"{"}]}
+
+        assert parse!("{?\\\n}\n{123}") ==
+                 {:__block__, [], [{:{}, [line: 1], ~c"\n"}, {:{}, [line: 3], ~c"{"}]}
+      end)
+    end
+
     test "numbers are printed correctly in syntax errors" do
       assert_syntax_error(["nofile:1:5:", ~s/syntax error before: "12"/], ~c":ok 12")
       assert_syntax_error(["nofile:1:5:", ~s/syntax error before: "0b1"/], ~c":ok 0b1")
