@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule Kernel.QuoteTest do
@@ -383,8 +387,6 @@ defmodule Kernel.QuoteTest.ErrorsTest do
         mod = Kernel.QuoteTest.ErrorsTest
         file = __ENV__.file |> Path.relative_to_cwd() |> String.to_charlist()
         assert [{^mod, :will_raise, 2, [file: ^file, line: @line] ++ _} | _] = __STACKTRACE__
-    else
-      _ -> flunk("expected failure")
     end
   end
 
@@ -392,13 +394,12 @@ defmodule Kernel.QuoteTest.ErrorsTest do
   test "outside function error" do
     try do
       will_raise()
+      flunk("expected failure")
     rescue
       RuntimeError ->
         mod = Kernel.QuoteTest.ErrorsTest
         file = __ENV__.file |> Path.relative_to_cwd() |> String.to_charlist()
         assert [{^mod, _, _, [file: ^file, line: @line] ++ _} | _] = __STACKTRACE__
-    else
-      _ -> flunk("expected failure")
     end
   end
 end
@@ -789,4 +790,48 @@ defmodule Kernel.QuoteTest.HasUnquoteTest do
 
     refute :elixir_quote.has_unquotes(ast)
   end
+
+  test "unquote with invalid AST (shallow check)" do
+    for term <- [
+          %{unescaped: :map},
+          1..10,
+          {:bad_meta, nil, []},
+          {:bad_arg, nil, 1},
+          {:bad_tuple},
+          make_ref(),
+          [:improper | :list],
+          [nested: {}]
+        ] do
+      message = """
+      tried to unquote invalid AST: #{inspect(term)}
+      Did you forget to escape term using Macro.escape/1?\
+      """
+
+      assert_raise ArgumentError, message, fn -> quote do: unquote(term) end
+    end
+  end
+
+  test "unquote with invalid AST is not checked deeply" do
+    assert quote do: unquote(foo: [1 | 2]) == [foo: [1 | 2]]
+    assert quote do: unquote(foo: [bar: %{}]) == [foo: [bar: %{}]]
+  end
+
+  test "unquote_splicing with invalid AST" do
+    for args <- [
+          "not_a_list",
+          [:improper | :list],
+          [%{unescaped: :map}],
+          [1..10],
+          [{:bad_meta, nil, []}],
+          [{:bad_arg, nil, 1}],
+          [{:bad_tuple}],
+          [make_ref()],
+          [nested: {}]
+        ] do
+      message =
+        "expected a list with quoted expressions in unquote_splicing/1, got: #{inspect(args)}"
+
+      assert_raise ArgumentError, message, fn -> quote do: [unquote_splicing(args)] end
+    end
+  end
 end
