@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule Kernel.WithTest do
@@ -126,7 +130,7 @@ defmodule Kernel.WithTest do
     end
 
     assert_raise RuntimeError, fn ->
-      with({:ok, res} <- ok(42), res = res + oops(), do: res)
+      with({:ok, res} <- ok(42), oops(), do: res)
     end
   end
 
@@ -149,7 +153,7 @@ defmodule Kernel.WithTest do
   end
 
   test "else conditions with match error" do
-    assert_raise WithClauseError, "no with clause matching: :error", fn ->
+    assert_raise WithClauseError, "no with clause matching:\n\n    :error\n", fn ->
       with({:ok, res} <- error(), do: res, else: ({:error, error} -> error))
     end
   end
