@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("test_helper.exs", __DIR__)
 
 defmodule Macro.ExternalTest do
@@ -78,6 +82,14 @@ defmodule MacroTest do
 
       contents = quote(unquote: false, do: unquote(x))
       assert Macro.escape(contents, unquote: true) == {:x, [], MacroTest}
+
+      contents = %{foo: quote(unquote: false, do: unquote(1))}
+      assert Macro.escape(contents, unquote: true) == {:%{}, [], [foo: 1]}
+    end
+
+    test "with generated" do
+      assert Macro.escape(%{a: {}}, generated: true) ==
+               {:%{}, [generated: true], [a: {:{}, [], []}]}
     end
 
     defp eval_escaped(contents) do
@@ -88,6 +100,7 @@ defmodule MacroTest do
     test "with remote unquote" do
       contents = quote(unquote: false, do: Kernel.unquote(:is_atom)(:ok))
       assert eval_escaped(contents) == quote(do: Kernel.is_atom(:ok))
+      assert eval_escaped(%{foo: contents}) == %{foo: quote(do: Kernel.is_atom(:ok))}
     end
 
     test "with nested unquote" do
@@ -131,16 +144,68 @@ defmodule MacroTest do
         quote(unquote: false, do: [1, unquote_splicing([2]), 3, unquote_splicing([4]) | [5]])
 
       assert eval_escaped(contents) == [1, 2, 3, 4, 5]
+
+      contents = %{foo: quote(unquote: false, do: [1, 2, unquote_splicing([3, 4, 5])])}
+      assert eval_escaped(contents) == %{foo: [1, 2, 3, 4, 5]}
     end
 
     test "does not add context to quote" do
       assert Macro.escape({:quote, [], [[do: :foo]]}) == {:{}, [], [:quote, [], [[do: :foo]]]}
     end
 
-    test "inspects container when a reference cannot be escaped" do
-      assert_raise ArgumentError, ~r"~r/foo/ contains a reference", fn ->
-        Macro.escape(%{~r/foo/ | re_pattern: make_ref()})
-      end
+    test "escapes the content of :quote tuples" do
+      assert Macro.escape({:quote, [%{}], [{}]}) ==
+               {:{}, [], [:quote, [{:%{}, [], []}], [{:{}, [], []}]]}
+
+      assert Macro.escape([:foo, {:quote, [%{}], [{}]}]) ==
+               [:foo, {:{}, [], [:quote, [{:%{}, [], []}], [{:{}, [], []}]]}]
+    end
+
+    @tag :re_import
+    test "escape regex will remove references and replace it by a call to :re.import/1" do
+      assert {
+               :%{},
+               [],
+               [
+                 __struct__: Regex,
+                 re_pattern:
+                   {{:., [], [{:__aliases__, _, [:Regex]}, :__import_pattern__]},
+                    [required: true], [{:{}, [], [:re_exported_pattern | _]}]},
+                 source: "foo",
+                 opts: []
+               ]
+             } = Macro.escape(~r/foo/)
+    end
+
+    @tag :re_import
+    test "escape raises within structs fields" do
+      assert_raise ArgumentError,
+                   ~r"Regex defines custom escaping rules which are not supported in struct defaults",
+                   fn ->
+                     defmodule Test do
+                       defstruct my_regex: ~r/^hi$/
+                     end
+                   end
+    end
+
+    defmodule EscapedStruct do
+      defstruct [:ast, :ref]
+
+      def __escape__(%{ast: ast}), do: ast
+    end
+
+    test "escape struct with custom __escape__ (valid AST)" do
+      struct = %EscapedStruct{ast: {:valid_ast, [], []}, ref: make_ref()}
+
+      assert {:valid_ast, [], []} = Macro.escape(struct)
+    end
+
+    test "escape struct with custom __escape__ (shallow invalid AST)" do
+      struct = %EscapedStruct{ast: %{invalid: :ast}, ref: make_ref()}
+
+      assert_raise ArgumentError,
+                   "MacroTest.EscapedStruct.__escape__/1 returned invalid AST: %{invalid: :ast}",
+                   fn -> Macro.escape(struct) end
     end
   end
 
@@ -331,12 +396,22 @@ defmodule MacroTest do
   end
 
   describe "dbg/3" do
+    defmacrop dbg_format_no_newline(ast, options \\ quote(do: [syntax_colors: []])) do
+      quote do
+        ExUnit.CaptureIO.with_io(fn ->
+          try do
+            unquote(Macro.dbg(ast, options, __CALLER__))
+          rescue
+            e -> e
+          end
+        end)
+      end
+    end
+
     defmacrop dbg_format(ast, options \\ quote(do: [syntax_colors: []])) do
       quote do
         {result, formatted} =
-          ExUnit.CaptureIO.with_io(fn ->
-            unquote(Macro.dbg(ast, options, __CALLER__))
-          end)
+          dbg_format_no_newline(unquote(ast), unquote(options))
 
         # Make sure there's an empty line after the output.
         assert String.ends_with?(formatted, "\n\n") or
@@ -346,27 +421,27 @@ defmodule MacroTest do
       end
     end
 
-    test "with a simple expression" do
+    test "simple expression" do
       {result, formatted} = dbg_format(1 + 1)
       assert result == 2
       assert formatted =~ "1 + 1 #=> 2"
     end
 
-    test "with variables" do
+    test "variables" do
       my_var = 1 + 1
       {result, formatted} = dbg_format(my_var)
       assert result == 2
       assert formatted =~ "my_var #=> 2"
     end
 
-    test "with a function call" do
+    test "function call" do
       {result, formatted} = dbg_format(Atom.to_string(:foo))
 
       assert result == "foo"
       assert formatted =~ ~s[Atom.to_string(:foo) #=> "foo"]
     end
 
-    test "with a multiline input" do
+    test "multiline input" do
       {result, formatted} =
         dbg_format(
           case 1 + 1 do
@@ -387,7 +462,7 @@ defmodule MacroTest do
 
     defp abc, do: [:a, :b, :c]
 
-    test "with a pipeline on a single line" do
+    test "pipeline on a single line" do
       {result, formatted} = dbg_format(abc() |> tl() |> tl |> Kernel.hd())
       assert result == :c
 
@@ -405,7 +480,7 @@ defmodule MacroTest do
       assert formatted =~ ~r/[^\n]\n\n$/
     end
 
-    test "with a pipeline on multiple lines" do
+    test "pipeline on multiple lines" do
       {result, formatted} =
         dbg_format(
           abc()
@@ -430,7 +505,7 @@ defmodule MacroTest do
       assert formatted =~ ~r/[^\n]\n\n$/
     end
 
-    test "with simple boolean expressions" do
+    test "simple boolean expressions" do
       {result, formatted} = dbg_format(:rand.uniform() < 0.0 and length([]) == 0)
       assert result == false
 
@@ -442,7 +517,7 @@ defmodule MacroTest do
              """
     end
 
-    test "with left-associative operators" do
+    test "left-associative operators" do
       {result, formatted} = dbg_format(List.first([]) || "yes" || raise("foo"))
       assert result == "yes"
 
@@ -455,7 +530,7 @@ defmodule MacroTest do
              """
     end
 
-    test "with composite boolean expressions" do
+    test "composite boolean expressions" do
       true1 = length([]) == 0
       true2 = length([]) == 0
       {result, formatted} = dbg_format((true1 and true2) or (List.first([]) || true1))
@@ -471,7 +546,23 @@ defmodule MacroTest do
              """
     end
 
-    test "with block of code" do
+    test "boolean expressions that raise" do
+      falsy = length([]) != 0
+      x = 0
+
+      {result, formatted} = dbg_format_no_newline((falsy || x) && 1 / x)
+
+      assert %ArithmeticError{} = result
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             falsy #=> false
+             falsy || x #=> 0
+             """
+    end
+
+    test "block of code" do
       {result, formatted} =
         dbg_format(
           (
@@ -486,16 +577,33 @@ defmodule MacroTest do
       assert formatted =~ "macro_test.exs"
 
       assert formatted =~ """
-             Code block:
-             (
-               a = 1 #=> 1
-               b = a + 2 #=> 3
-               a + b #=> 4
-             )
+             a = 1 #=> 1
+             b = a + 2 #=> 3
+             a + b #=> 4
              """
     end
 
-    test "with case" do
+    test "block that raises" do
+      {result, formatted} =
+        dbg_format_no_newline(
+          (
+            a = 1
+            b = a - 1
+            a / b
+          )
+        )
+
+      assert result == %ArithmeticError{}
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             a = 1 #=> 1
+             b = a - 1 #=> 0
+             """
+    end
+
+    test "case" do
       list = List.flatten([1, 2, 3])
 
       {result, formatted} =
@@ -522,7 +630,28 @@ defmodule MacroTest do
              """
     end
 
-    test "with case - guard" do
+    test "case that raises" do
+      x = true
+
+      {result, formatted} =
+        dbg_format(
+          case true and x do
+            true -> raise "oops"
+            false -> :ok
+          end
+        )
+
+      assert %RuntimeError{} = result
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             Case argument:
+             true and x #=> true
+             """
+    end
+
+    test "case with guards" do
       {result, formatted} =
         dbg_format(
           case 0..100//5 do
@@ -555,7 +684,7 @@ defmodule MacroTest do
              """
     end
 
-    test "with cond" do
+    test "cond" do
       map = %{b: 5}
 
       {result, formatted} =
@@ -614,6 +743,26 @@ defmodule MacroTest do
              """
     end
 
+    test "if expression that raises" do
+      x = true
+
+      {result, formatted} =
+        dbg_format(
+          if true and x do
+            raise "oops"
+          end
+        )
+
+      assert %RuntimeError{} = result
+
+      assert formatted =~ "macro_test.exs"
+
+      assert formatted =~ """
+             If condition:
+             true and x #=> true
+             """
+    end
+
     test "if expression without else" do
       x = true
       map = %{a: 5, b: 1}
@@ -660,7 +809,7 @@ defmodule MacroTest do
              """
     end
 
-    test "with with/1 (all clauses match)" do
+    test "with/1 (all clauses match)" do
       opts = %{width: 10, height: 15}
 
       {result, formatted} =
@@ -693,7 +842,7 @@ defmodule MacroTest do
              """
     end
 
-    test "with with/1 (no else)" do
+    test "with/1 (no else)" do
       opts = %{width: 10}
 
       {result, formatted} =
@@ -721,7 +870,7 @@ defmodule MacroTest do
              """
     end
 
-    test "with with/1 (else clause)" do
+    test "with/1 (else clause)" do
       opts = %{width: 10}
 
       {result, formatted} =
@@ -752,7 +901,7 @@ defmodule MacroTest do
              """
     end
 
-    test "with with/1 (guard)" do
+    test "with/1 (guard)" do
       opts = %{width: 10, height: 0.0}
 
       {result, formatted} =
@@ -783,7 +932,7 @@ defmodule MacroTest do
              """
     end
 
-    test "with with/1 (guard in else)" do
+    test "with/1 (guard in else)" do
       opts = %{}
 
       {result, _formatted} =
@@ -799,18 +948,20 @@ defmodule MacroTest do
       assert result == :atom
     end
 
-    test "with with/1 respects the WithClauseError" do
+    test "with/1 respects the WithClauseError" do
       value = Enum.random([:unexpected])
 
       error =
         assert_raise WithClauseError, fn ->
-          dbg(
-            with :ok <- value do
-              true
-            else
-              :error -> false
-            end
-          )
+          ExUnit.CaptureIO.capture_io(fn ->
+            dbg(
+              with :ok <- value do
+                true
+              else
+                :error -> false
+              end
+            )
+          end)
         end
 
       assert error.term == :unexpected
@@ -1546,7 +1697,7 @@ defmodule MacroTest do
   test "generate_arguments/2" do
     assert Macro.generate_arguments(0, __MODULE__) == []
     assert Macro.generate_arguments(1, __MODULE__) == [{:arg1, [], __MODULE__}]
-    assert Macro.generate_arguments(4, __MODULE__) |> length == 4
+    assert Macro.generate_arguments(4, __MODULE__) |> length() == 4
   end
 
   defp postwalk(ast) do
@@ -1555,31 +1706,32 @@ defmodule MacroTest do
 
   test "struct_info!/2 expands structs multiple levels deep" do
     defmodule StructBang do
+      @enforce_keys [:b]
       defstruct [:a, :b]
 
       assert Macro.struct_info!(StructBang, __ENV__) == [
-               %{field: :a, default: nil},
-               %{field: :b, default: nil}
+               %{field: :a, default: nil, required: false},
+               %{field: :b, default: nil, required: true}
              ]
 
       def within_function do
         assert Macro.struct_info!(StructBang, __ENV__) == [
-                 %{field: :a, default: nil},
-                 %{field: :b, default: nil}
+                 %{field: :a, default: nil, required: false},
+                 %{field: :b, default: nil, required: true}
                ]
       end
 
       defmodule Nested do
         assert Macro.struct_info!(StructBang, __ENV__) == [
-                 %{field: :a, default: nil},
-                 %{field: :b, default: nil}
+                 %{field: :a, default: nil, required: false},
+                 %{field: :b, default: nil, required: true}
                ]
       end
     end
 
     assert Macro.struct_info!(StructBang, __ENV__) == [
-             %{field: :a, default: nil},
-             %{field: :b, default: nil}
+             %{field: :a, default: nil, required: false},
+             %{field: :b, default: nil, required: true}
            ]
   end
 
