@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("type_helper.exs", __DIR__)
 
 defmodule Module.Types.ExprTest do
@@ -21,8 +25,6 @@ defmodule Module.Types.ExprTest do
     assert typecheck!("foo") == binary()
     assert typecheck!([]) == empty_list()
     assert typecheck!(%{}) == closed_map([])
-    assert typecheck!(& &1) == fun()
-    assert typecheck!(fn -> :ok end) == fun()
   end
 
   test "generated" do
@@ -80,7 +82,7 @@ defmodule Module.Types.ExprTest do
       assert typecheck!([x = [123, :foo]], tl(x)) == dynamic(list(union(atom([:foo]), integer())))
 
       assert typecheck!([x = [123 | :foo]], tl(x)) ==
-               dynamic(union(atom([:foo]), list(integer(), atom([:foo]))))
+               dynamic(union(atom([:foo]), non_empty_list(integer(), atom([:foo]))))
 
       assert typeerror!(tl([])) |> strip_ansi() ==
                ~l"""
@@ -115,7 +117,51 @@ defmodule Module.Types.ExprTest do
   end
 
   describe "funs" do
-    test "incompatible" do
+    test "infers functions" do
+      assert typecheck!(& &1) |> equal?(fun([term()], dynamic()))
+
+      assert typecheck!(fn -> :ok end) |> equal?(fun([], dynamic(atom([:ok]))))
+
+      assert typecheck!(fn
+               <<"ok">>, {} -> :ok
+               <<"error">>, {} -> :error
+               [_ | _], %{} -> :list
+             end)
+             |> equal?(
+               intersection(
+                 fun(
+                   [non_empty_list(term(), term()), open_map()],
+                   dynamic(atom([:list]))
+                 ),
+                 fun(
+                   [binary(), tuple([])],
+                   dynamic(atom([:ok, :error]))
+                 )
+               )
+             )
+    end
+
+    test "application" do
+      assert typecheck!(
+               [map],
+               (fn
+                  %{a: a} = data -> %{data | b: a}
+                  data -> data
+                end).(map)
+             ) == dynamic()
+
+      assert typecheck!(
+               [],
+               [true, false]
+               |> Enum.random()
+               |> then(fn
+                 true -> :ok
+                 _ -> :error
+               end)
+             ) == dynamic(atom([:ok, :error]))
+    end
+
+    test "bad function" do
       assert typeerror!([%x{}, a1, a2], x.(a1, a2)) == ~l"""
              expected a 2-arity function on call:
 
@@ -132,6 +178,130 @@ defmodule Module.Types.ExprTest do
                  %x{}
              """
     end
+
+    test "bad arity" do
+      assert typeerror!([a1, a2], (&String.to_integer/1).(a1, a2)) == ~l"""
+             expected a 2-arity function on call:
+
+                 (&String.to_integer/1).(a1, a2)
+
+             but got function with arity 1:
+
+                 (binary() -> integer())
+             """
+    end
+
+    test "bad argument" do
+      assert typeerror!([], (&String.to_integer/1).(:foo))
+             |> strip_ansi() == ~l"""
+             incompatible types given on function application:
+
+                 (&String.to_integer/1).(:foo)
+
+             given types:
+
+                 :foo
+
+             but function has type:
+
+                 (binary() -> integer())
+             """
+
+      assert typeerror!(
+               [x],
+               (if x do
+                  &String.to_integer/1
+                else
+                  &List.to_integer/1
+                end).(:foo)
+             )
+             |> strip_ansi() == ~l"""
+             incompatible types given on function application:
+
+                 (if x do
+                    &String.to_integer/1
+                  else
+                    &List.to_integer/1
+                  end).(:foo)
+
+             given types:
+
+                 :foo
+
+             but function has type:
+
+                 (binary() -> integer()) or (non_empty_list(integer()) -> integer())
+
+             hint: the function has an empty domain and therefore cannot be applied to any argument. \
+             This may happen when you have a union of functions, which means the only valid argument \
+             to said function are types that satisfy all sides of the union (which may be none)
+             """
+    end
+
+    test "bad arguments from inferred type" do
+      assert typeerror!(
+               (
+                 fun = fn %{} -> :map end
+                 fun.(:error)
+               )
+             )
+             |> strip_ansi() == """
+             incompatible types given on function application:
+
+                 fun.(:error)
+
+             given types:
+
+                 :error
+
+             but function has type:
+
+                 (map() -> dynamic(:map))
+             """
+    end
+
+    test "capture printing" do
+      assert typeerror!(123 = &{:ok, &1}) == """
+             the following pattern will never match:
+
+                 123 = &{:ok, &1}
+
+             because the right-hand side has type:
+
+                 (term() -> dynamic({:ok, term()}))
+             """
+    end
+
+    test "works when there are multiple clauses with lists and maps" do
+      type =
+        typecheck!(fn
+          [:oban, :job, _event], _measure, _meta, _opts ->
+            :ok
+
+          [:oban, :notifier, :switch], _measure, %{status: _status}, _opts ->
+            :ok
+
+          [:oban, :peer, :election, :stop], _measure, _meta, _opts ->
+            :ok
+
+          [:oban, :plugin, :exception], _measure, _meta, _opts ->
+            :ok
+
+          [:oban, :plugin, :stop], _measure, _meta, _opts ->
+            :ok
+
+          [:oban, :queue, :shutdown], _measure, %{orphaned: [_ | _]}, _opts ->
+            :ok
+
+          [:oban, :stager, :switch], _measure, %{mode: _mode}, _opts ->
+            :ok
+
+          _event, _measure, _meta, _opts ->
+            :ok
+        end)
+
+      assert subtype?(type, fun([term(), term(), term(), term()], atom([:ok])))
+    end
   end
 
   describe "remotes" do
@@ -215,7 +385,7 @@ defmodule Module.Types.ExprTest do
                """
 
       assert typeerror!(
-               [<<x::integer>>, y = Atom, z],
+               [<<x::integer>>, y = SomeMod, z],
                (
                  mod =
                    cond do
@@ -233,7 +403,7 @@ defmodule Module.Types.ExprTest do
 
                where "mod" was given the type:
 
-                   # type: dynamic(Atom or integer()) or integer()
+                   # type: dynamic(SomeMod) or integer()
                    # from: types_test.ex:LINE-9
                    mod =
                      cond do
@@ -243,6 +413,35 @@ defmodule Module.Types.ExprTest do
                """
     end
 
+    test "calling a function with invalid arguments on variables" do
+      assert typeerror!(
+               (
+                 x = List
+                 x.to_tuple(123)
+               )
+             )
+             |> strip_ansi() ==
+               ~l"""
+               incompatible types given to List.to_tuple/1:
+
+                   x.to_tuple(123)
+
+               given types:
+
+                   integer()
+
+               but expected one of:
+
+                   list(term())
+
+               where "x" was given the type:
+
+                   # type: List
+                   # from: types_test.ex:LINE-5
+                   x = List
+               """
+    end
+
     test "capture a function with non atoms" do
       assert typeerror!([<<x::integer>>], &x.foo_bar/2) ==
                ~l"""
@@ -261,13 +460,84 @@ defmodule Module.Types.ExprTest do
                    <<x::integer>>
                """
     end
+
+    test "requires all combinations to be compatible (except refinements)" do
+      assert typecheck!(
+               [condition, arg],
+               (
+                 # While the code below may raise, it may also always succeed
+                 # if condition and arg are passed in tandem. Therefore, we
+                 # turn off refinement on dynamic calls.
+                 mod = if condition, do: String, else: List
+                 res = mod.to_integer(arg)
+                 {arg, res}
+               )
+             ) == tuple([dynamic(), integer()])
+
+      assert typeerror!(
+               [condition],
+               (
+                 arg = if condition, do: "foo", else: [?f, ?o, ?o]
+                 mod = if condition, do: String, else: List
+                 mod.to_integer(arg)
+               )
+             )
+             |> strip_ansi() == ~l"""
+             incompatible types given to List.to_integer/1:
+
+                 mod.to_integer(arg)
+                 #=> invoked as List.to_integer/1
+
+             given types:
+
+                 binary() or non_empty_list(integer())
+
+             but expected one of:
+
+                 non_empty_list(integer())
+
+             where "arg" was given the type:
+
+                 # type: binary() or non_empty_list(integer())
+                 # from: types_test.ex:LINE-5
+                 arg =
+                   if condition do
+                     "foo"
+                   else
+                     ~c"foo"
+                   end
+
+             where "mod" was given the type:
+
+                 # type: List or String
+                 # from: types_test.ex:LINE-4
+                 mod =
+                   if condition do
+                     String
+                   else
+                     List
+                   end
+             """
+    end
+  end
+
+  describe "remote capture" do
+    test "strong" do
+      assert typecheck!(&String.to_atom/1) == fun([binary()], atom())
+      assert typecheck!(&:erlang.element/2) == fun([integer(), open_tuple([])], dynamic())
+    end
+
+    test "unknown" do
+      assert typecheck!(&Module.Types.ExprTest.__ex_unit__/1) == dynamic(fun(1))
+      assert typecheck!([x], &x.something/1) == dynamic(fun(1))
+    end
   end
 
   describe "binaries" do
     test "warnings" do
       assert typeerror!([<<x::binary-size(2)>>], <<x::float>>) ==
                ~l"""
-               incompatible types in expression:
+               incompatible types in binary construction:
 
                    <<x::float>>
 
@@ -288,7 +558,7 @@ defmodule Module.Types.ExprTest do
 
       assert typeerror!([<<x::binary>>], <<x>>) ==
                ~l"""
-               incompatible types in expression:
+               incompatible types in binary construction:
 
                    <<x>>
 
@@ -311,7 +581,7 @@ defmodule Module.Types.ExprTest do
 
       assert typeerror!([<<x>>], <<x::binary>>) ==
                ~l"""
-               incompatible types in expression:
+               incompatible types in binary construction:
 
                    <<x::binary>>
 
@@ -340,7 +610,7 @@ defmodule Module.Types.ExprTest do
     test "size error" do
       assert typeerror!([<<x::binary>>, y], <<y::size(x)>>) ==
                ~l"""
-               incompatible types in expression:
+               expected an integer in binary size:
 
                    size(x)
 
@@ -348,10 +618,6 @@ defmodule Module.Types.ExprTest do
 
                    binary()
 
-               but expected type:
-
-                   integer()
-
                where "x" was given the type:
 
                    # type: binary()
@@ -490,7 +756,7 @@ defmodule Module.Types.ExprTest do
                """
     end
 
-    test "duplicate/2" do
+    test "Tuple.duplicate/2" do
       assert typecheck!(Tuple.duplicate(123, 0)) == tuple([])
       assert typecheck!(Tuple.duplicate(123, 1)) == tuple([integer()])
       assert typecheck!(Tuple.duplicate(123, 2)) == tuple([integer(), integer()])
@@ -532,7 +798,7 @@ defmodule Module.Types.ExprTest do
     end
 
     test "creating open maps" do
-      assert typecheck!(%{123 => 456}) == open_map()
+      assert typecheck!(%{123 => 456}) == dynamic(open_map())
       # Since key cannot override :foo, we preserve it
       assert typecheck!([key], %{key => 456, foo: :bar}) == dynamic(open_map(foo: atom([:bar])))
       # Since key can override :foo, we do not preserve it
@@ -659,6 +925,87 @@ defmodule Module.Types.ExprTest do
              """
     end
 
+    test "updating structs" do
+      integer_date_type =
+        dynamic(
+          closed_map(
+            __struct__: atom([Date]),
+            day: integer(),
+            calendar: term(),
+            month: term(),
+            year: term()
+          )
+        )
+
+      # When we know the type
+      assert typecheck!([], %Date{Date.new!(1, 1, 1) | day: 31}) ==
+               integer_date_type
+
+      assert typecheck!([], %Date{%Date{Date.new!(1, 1, 1) | day: 13} | day: 31}) ==
+               integer_date_type
+
+      # When we don't know the type of var
+      assert typeerror!([x], %Date{x | day: 31}) == ~l"""
+             a struct for Date is expected on struct update:
+
+                 %Date{x | day: 31}
+
+             but got type:
+
+                 dynamic()
+
+             where "x" was given the type:
+
+                 # type: dynamic()
+                 # from: types_test.ex:LINE
+                 x
+
+             when defining the variable "x", you must also pattern match on "%Date{}".
+
+             hint: given pattern matching is enough to catch typing errors, you may optionally convert the struct update into a map update. For example, instead of:
+
+                 user = some_function()
+                 %User{user | name: "John Doe"}
+
+             it is enough to write:
+
+                 %User{} = user = some_function()
+                 %{user | name: "John Doe"}
+             """
+
+      # When we don't know the type of capture
+      assert typeerror!([], &%Date{&1 | day: 31}) =~ ~l"""
+             a struct for Date is expected on struct update:
+
+                 %Date{&1 | day: 31}
+
+             but got type:
+
+                 dynamic()
+
+             where "capture" was given the type:
+
+                 # type: dynamic()
+                 # from: types_test.ex:LINE
+                 &1
+
+             instead of using &1, you must define an anonymous function, define a variable and pattern match on "%Date{}".
+             """
+
+      # When we don't know the type of expression
+      assert typeerror!([], %Date{SomeMod.fun() | day: 31}) =~ """
+             a struct for Date is expected on struct update:
+
+                 %Date{SomeMod.fun() | day: 31}
+
+             but got type:
+
+                 dynamic()
+
+             you must assign "SomeMod.fun()" to variable and pattern match on "%Date{}".
+             """
+    end
+
     test "updating to open maps" do
       assert typecheck!(
                [key],
@@ -717,42 +1064,8 @@ defmodule Module.Types.ExprTest do
              """
     end
 
-    test "updating structs" do
-      assert typecheck!([x], %Point{x | x: :zero}) ==
-               dynamic(open_map(__struct__: atom([Point]), x: atom([:zero])))
-
-      assert typecheck!([x], %Point{%Point{x | x: :zero} | y: :one}) ==
-               dynamic(open_map(__struct__: atom([Point]), x: atom([:zero]), y: atom([:one])))
-
-      assert typeerror!(
-               (
-                 x = %{x: 0}
-                 %Point{x | x: :zero}
-               )
-             ) ==
-               ~l"""
-               incompatible types in struct update:
-
-                   %Point{x | x: :zero}
-
-               expected type:
-
-                   %Point{x: term(), y: term(), z: term()}
-
-               but got type:
-
-                   %{x: integer()}
-
-               where "x" was given the type:
-
-                   # type: %{x: integer()}
-                   # from: types_test.ex:LINE-4
-                   x = %{x: 0}
-               """
-    end
-
     test "nested map" do
-      assert typecheck!([x = %{}], x.foo.bar) == dynamic()
+      assert typecheck!([x = %{}], x.foo.bar) |> equal?(dynamic())
     end
 
     test "accessing a field on not a map" do
@@ -801,14 +1114,14 @@ defmodule Module.Types.ExprTest do
              the given type does not have the given key:
 
                  dynamic(%URI{
-                   authority: term(),
-                   fragment: term(),
-                   host: term(),
-                   path: term(),
-                   port: term(),
-                   query: term(),
                    scheme: term(),
-                   userinfo: term()
+                   authority: term(),
+                   userinfo: term(),
+                   host: term(),
+                   port: term(),
+                   path: term(),
+                   query: term(),
+                   fragment: term()
                  })
 
              where "x" was given the type:
@@ -820,6 +1133,34 @@ defmodule Module.Types.ExprTest do
 
       assert [%{type: :variable, name: :x}] = diagnostic.details.typing_traces
     end
+
+    test "inspect struct definition" do
+      assert typeerror!(
+               (
+                 p = %Point{x: 123}
+                 Integer.to_string(p)
+               )
+             )
+             |> strip_ansi() == ~l"""
+             incompatible types given to Integer.to_string/1:
+
+                 Integer.to_string(p)
+
+             given types:
+
+                 %Point{x: integer(), y: nil, z: integer()}
+
+             but expected one of:
+
+                 integer()
+
+             where "p" was given the type:
+
+                 # type: %Point{x: integer(), y: nil, z: integer()}
+                 # from: types_test.ex:LINE-4
+                 p = %Point{..., x: 123}
+             """
+    end
   end
 
   describe "comparison" do
@@ -831,7 +1172,11 @@ defmodule Module.Types.ExprTest do
     test "in dynamic mode" do
       assert typedyn!([x = 123, y = 456.0], x < y) == dynamic(boolean())
       assert typedyn!([x = 123, y = 456.0], x == y) == dynamic(boolean())
-      assert typedyn!(123 == 456) == boolean()
+      assert typedyn!([x = 123, y = 456], x == y) == dynamic(boolean())
+    end
+
+    test "using literals" do
+      assert typecheck!(:foo == :bar) == boolean()
     end
 
     test "min/max" do
@@ -842,93 +1187,115 @@ defmodule Module.Types.ExprTest do
     end
 
     test "warns when comparison is constant" do
-      assert typewarn!([x = :foo, y = 321], min(x, y)) ==
-               {dynamic(union(integer(), atom([:foo]))),
-                ~l"""
-                comparison between distinct types found:
+      assert typeerror!([x = :foo, y = 321], min(x, y)) ==
+               ~l"""
+               comparison between distinct types found:
 
-                    min(x, y)
+                   min(x, y)
 
-                given types:
+               given types:
 
-                    min(dynamic(:foo), integer())
+                   min(dynamic(:foo), integer())
 
-                where "x" was given the type:
+               where "x" was given the type:
 
-                    # type: dynamic(:foo)
-                    # from: types_test.ex:LINE-2
-                    x = :foo
+                   # type: dynamic(:foo)
+                   # from: types_test.ex:LINE-1
+                   x = :foo
 
-                where "y" was given the type:
+               where "y" was given the type:
 
-                    # type: integer()
-                    # from: types_test.ex:LINE-2
-                    y = 321
+                   # type: integer()
+                   # from: types_test.ex:LINE-1
+                   y = 321
 
-                While Elixir can compare across all types, you are comparing across types \
-                which are always disjoint, and the result is either always true or always false
-                """}
+               While Elixir can compare across all types, you are comparing across types \
+               which are always disjoint, and the result is either always true or always false
+               """
 
-      assert typewarn!([x = 123, y = 456.0], x === y) ==
-               {boolean(),
-                ~l"""
-                comparison between distinct types found:
+      assert typeerror!([x = 123, y = 456.0], x === y) ==
+               ~l"""
+               comparison between distinct types found:
 
-                    x === y
+                   x === y
 
-                given types:
+               given types:
 
-                    integer() === float()
+                   integer() === float()
 
-                where "x" was given the type:
+               where "x" was given the type:
 
-                    # type: integer()
-                    # from: types_test.ex:LINE-2
-                    x = 123
+                   # type: integer()
+                   # from: types_test.ex:LINE-1
+                   x = 123
 
-                where "y" was given the type:
+               where "y" was given the type:
 
-                    # type: float()
-                    # from: types_test.ex:LINE-2
-                    y = 456.0
+                   # type: float()
+                   # from: types_test.ex:LINE-1
+                   y = 456.0
 
-                While Elixir can compare across all types, you are comparing across types \
-                which are always disjoint, and the result is either always true or always false
-                """}
+               While Elixir can compare across all types, you are comparing across types \
+               which are always disjoint, and the result is either always true or always false
+               """
     end
 
     test "warns on comparison with struct across dynamic call" do
-      assert typewarn!([x = :foo, y = %Point{}, mod = Kernel], mod.<=(x, y)) ==
-               {boolean(),
-                ~l"""
-                comparison with structs found:
+      assert typeerror!([x = %Point{}, y = %Point{}, mod = Kernel], mod.<=(x, y)) ==
+               ~l"""
+               comparison with structs found:
 
-                    mod.<=(x, y)
+                   mod.<=(x, y)
 
-                given types:
+               given types:
 
-                    dynamic(:foo) <= dynamic(%Point{})
+                   dynamic(%Point{}) <= dynamic(%Point{})
 
-                where "mod" was given the type:
+               where "mod" was given the type:
 
-                    # type: dynamic(Kernel)
-                    # from: types_test.ex:LINE-2
-                    mod = Kernel
+                   # type: dynamic(Kernel)
+                   # from: types_test.ex:LINE-1
+                   mod = Kernel
 
-                where "x" was given the type:
+               where "x" was given the type:
 
-                    # type: dynamic(:foo)
-                    # from: types_test.ex:LINE-2
-                    x = :foo
+                   # type: dynamic(%Point{})
+                   # from: types_test.ex:LINE-1
+                   x = %Point{}
 
-                where "y" was given the type:
+               where "y" was given the type:
 
-                    # type: dynamic(%Point{})
-                    # from: types_test.ex:LINE-2
-                    y = %Point{}
+                   # type: dynamic(%Point{})
+                   # from: types_test.ex:LINE-1
+                   y = %Point{}
 
-                Comparison operators (>, <, >=, <=, min, and max) perform structural and not semantic comparison. Comparing with a struct won't give meaningful results. Structs that can be compared typically define a compare/2 function within their modules that can be used for semantic comparison.
-                """}
+               Comparison operators (>, <, >=, <=, min, and max) perform structural and not semantic comparison. Comparing with a struct won't give meaningful results. Structs that can be compared typically define a compare/2 function within their modules that can be used for semantic comparison.
+               """
+
+      assert typeerror!(
+               [x = %Point{}, mod = Kernel, condition],
+               (
+                 y = if condition, do: 456, else: %Point{}
+                 mod.<=(x, y)
+               )
+             ) =~ "comparison with structs found:"
+
+      assert typecheck!(
+               [x = 123, mod = Kernel, condition],
+               (
+                 y = if condition, do: 456, else: %Point{}
+                 mod.<=(x, y)
+               )
+             ) == boolean()
+
+      assert typeerror!(
+               [mod = Kernel, condition],
+               (
+                 x = if condition, do: 123, else: %Point{}
+                 y = if condition, do: 456, else: %Point{}
+                 mod.<=(x, y)
+               )
+             ) =~ "comparison with structs found:"
     end
   end
 
@@ -1031,6 +1398,29 @@ defmodule Module.Types.ExprTest do
   end
 
   describe "case" do
+    test "does not type check literals" do
+      assert typecheck!(
+               case :dev do
+                 :dev -> :ok
+                 :prod -> :error
+               end
+             ) == atom([:ok, :error])
+    end
+
+    test "resets branches" do
+      assert typecheck!(
+               [x],
+               (
+                 case :rand.uniform() do
+                   y when y < 0.5 -> x.foo
+                   y when y > 0.5 -> x.bar()
+                 end
+
+                 x
+               )
+             ) == dynamic()
+    end
+
     test "returns unions of all clauses" do
       assert typecheck!(
                [x],
@@ -1086,7 +1476,7 @@ defmodule Module.Types.ExprTest do
   end
 
   describe "conditionals" do
-    test "if does not report on literal atoms" do
+    test "if does not report on literals" do
       assert typecheck!(
                if true do
                  :ok
@@ -1094,15 +1484,15 @@ defmodule Module.Types.ExprTest do
              ) == atom([:ok, nil])
     end
 
-    test "and does not report on literal atoms" do
+    test "and does not report on literals" do
       assert typecheck!(false and true) == boolean()
     end
 
-    test "and reports on non-atom literals" do
-      assert typeerror!(1 and true) == ~l"""
+    test "and reports violations" do
+      assert typeerror!([x = 123], x and true) =~ """
              the following conditional expression will always evaluate to integer():
 
-                 1
+                 x
              """
     end
   end
@@ -1128,6 +1518,22 @@ defmodule Module.Types.ExprTest do
              ) == dynamic(atom([:ok, :error, :timeout]))
     end
 
+    test "resets branches" do
+      assert typecheck!(
+               [x, timeout = :infinity],
+               (
+                 receive do
+                   y when y > 0.5 -> x.foo
+                   _ -> x.bar()
+                 after
+                   timeout -> <<^x::integer>> = :crypto.strong_rand_bytes(1)
+                 end
+
+                 x
+               )
+             ) == dynamic()
+    end
+
     test "errors on bad timeout" do
       assert typeerror!(
                [x = :timeout],
@@ -1150,6 +1556,17 @@ defmodule Module.Types.ExprTest do
                  # from: types_test.ex:LINE-5
                  x = :timeout
              """
+
+      # Check for compatibility, not subtyping
+      assert typeerror!(
+               [<<x::integer, y::float>>],
+               receive do
+               after
+                 if(:rand.uniform(), do: x, else: y) -> :ok
+               end
+             ) =~ "expected "
+    after
+      " timeout given to receive to be an integer"
     end
   end
 
@@ -1186,6 +1603,25 @@ defmodule Module.Types.ExprTest do
              ) == atom([:caught1, :caught2, :rescue, :else1, :else2])
     end
 
+    test "resets branches (except after)" do
+      assert typecheck!(
+               [x],
+               (
+                 try do
+                   <<^x::float>> = :crypto.strong_rand_bytes(8)
+                 rescue
+                   ArgumentError -> x.foo
+                 catch
+                   _, _ -> x.bar()
+                 after
+                   <<^x::integer>> = :crypto.strong_rand_bytes(8)
+                 end
+
+                 x
+               )
+             ) == dynamic(integer())
+    end
+
     test "reports error from clause that will never match" do
       assert typeerror!(
                [x],
@@ -1256,6 +1692,38 @@ defmodule Module.Types.ExprTest do
                )
     end
 
+    test "generates custom traces" do
+      assert typeerror!(
+               try do
+                 raise "oops"
+               rescue
+                 e ->
+                   Integer.to_string(e)
+               end
+             )
+             |> strip_ansi() == ~l"""
+             incompatible types given to Integer.to_string/1:
+
+                 Integer.to_string(e)
+
+             given types:
+
+                 %{..., __exception__: true, __struct__: atom()}
+
+             but expected one of:
+
+                 integer()
+
+             where "e" was given the type:
+
+                 # type: %{..., __exception__: true, __struct__: atom()}
+                 # from: types_test.ex
+                 rescue e
+
+             hint: when you rescue without specifying exception names, the variable is assigned a type of a struct but all of its fields are unknown. If you are trying to access an exception's :message key, either specify the exception names or use `Exception.message/1`.
+             """
+    end
+
     test "defines an open map of two fields in anonymous rescue" do
       assert typecheck!(
                try do
@@ -1367,6 +1835,20 @@ defmodule Module.Types.ExprTest do
                     y = false
                 """}
     end
+
+    test "resets branches" do
+      assert typecheck!(
+               [x],
+               (
+                 cond do
+                   :rand.uniform() > 0.5 -> x.foo
+                   true -> x.bar()
+                 end
+
+                 x
+               )
+             ) == dynamic()
+    end
   end
 
   describe "comprehensions" do
@@ -1389,17 +1871,106 @@ defmodule Module.Types.ExprTest do
 
                #{hints(:inferred_bitstring_spec)}
                """
+
+      # Check for compatibility, not subtyping
+      assert typeerror!(
+               [<<x::integer, y::binary>>],
+               for(<<i <- if(:rand.uniform() > 0.5, do: x, else: y)>>, do: i)
+             ) =~
+               ~l"""
+               expected the right side of <- in a binary generator to be a binary:
+
+                   if :rand.uniform() > 0.5 do
+                     x
+                   else
+                     y
+                   end
+
+               but got type:
+
+                   binary() or integer()
+
+               where "x" was given the type:
+
+                   # type: integer()
+                   # from: types_test.ex:LINE-3
+                   <<x::integer, ...>>
+
+               where "y" was given the type:
+
+                   # type: binary()
+                   # from: types_test.ex:LINE-3
+                   <<..., y::binary>>
+               """
+    end
+
+    test ":into" do
+      assert typecheck!([binary], for(<<x <- binary>>, do: x)) == list(integer())
+      assert typecheck!([binary], for(<<x <- binary>>, do: x, into: [])) == list(integer())
+      assert typecheck!([binary], for(<<x <- binary>>, do: x, into: "")) == binary()
+      assert typecheck!([binary, other], for(<<x <- binary>>, do: x, into: other)) == dynamic()
+
+      assert typecheck!([enum], for(x <- enum, do: x)) == list(dynamic())
+      assert typecheck!([enum], for(x <- enum, do: x, into: [])) == list(dynamic())
+      assert typecheck!([enum], for(x <- enum, do: x, into: "")) == binary()
+      assert typecheck!([enum, other], for(x <- enum, do: x, into: other)) == dynamic()
+
+      assert typecheck!(
+               [binary],
+               (
+                 into = if :rand.uniform() > 0.5, do: [], else: "0"
+                 for(<<x::float <- binary>>, do: x, into: into)
+               )
+             ) == union(binary(), list(float()))
+
+      assert typecheck!(
+               [binary, empty_list = []],
+               (
+                 into = if :rand.uniform() > 0.5, do: empty_list, else: "0"
+                 for(<<x::float <- binary>>, do: x, into: into)
+               )
+             ) == dynamic(union(binary(), list(float())))
+    end
+
+    test ":reduce checks" do
+      assert typecheck!(
+               [list],
+               for _ <- list, reduce: :ok do
+                 :ok -> 1
+                 _ -> 2.0
+               end
+             ) == union(atom([:ok]), union(integer(), float()))
     end
   end
 
   describe "info" do
     test "__info__/1" do
-      assert typecheck!([x], x.__info__(:functions)) == list(tuple([atom(), integer()]))
-
       assert typecheck!(GenServer.__info__(:functions)) == list(tuple([atom(), integer()]))
 
       assert typewarn!(:string.__info__(:functions)) ==
                {dynamic(), ":string.__info__/1 is undefined or private"}
+
+      assert typecheck!([x], x.__info__(:functions)) == list(tuple([atom(), integer()]))
+
+      assert typeerror!([x], x.__info__(:whatever)) |> strip_ansi() =~ """
+             incompatible types given to __info__/1:
+
+                 x.__info__(:whatever)
+
+             given types:
+
+                 :whatever
+             """
+    end
+
+    test "__info__/1 for struct information" do
+      assert typecheck!(GenServer.__info__(:struct)) == atom([nil])
+
+      assert typecheck!(URI.__info__(:struct)) ==
+               list(closed_map(default: if_set(term()), field: atom()))
+
+      assert typecheck!([x], x.__info__(:struct)) ==
+               list(closed_map(default: if_set(term()), field: atom())) |> union(atom([nil]))
     end
 
     test "behaviour_info/1" do
