@@ -1,14 +1,47 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("type_helper.exs", __DIR__)
 
 defmodule Module.Types.IntegrationTest do
   use ExUnit.Case
 
   import ExUnit.CaptureIO
+  import Module.Types.Descr
+
+  defp builtin_protocols do
+    [
+      Collectable,
+      Enumerable,
+      IEx.Info,
+      Inspect,
+      JSON.Encoder,
+      List.Chars,
+      String.Chars
+    ]
+  end
+
+  test "built-in protocols" do
+    builtin_protocols =
+      for app <- ~w[eex elixir ex_unit iex logger mix]a,
+          Application.ensure_loaded(app),
+          module <- Application.spec(app, :modules),
+          Code.ensure_loaded(module),
+          function_exported?(module, :__protocol__, 1),
+          do: module
+
+    # If this test fails, update:
+    # * lib/elixir/scripts/elixir_docs.ex
+    assert Enum.sort(builtin_protocols) == builtin_protocols()
+  end
 
   setup_all do
-    previous = Application.get_env(:elixir, :ansi_enabled, false)
     Application.put_env(:elixir, :ansi_enabled, false)
-    on_exit(fn -> Application.put_env(:elixir, :ansi_enabled, previous) end)
+
+    on_exit(fn ->
+      Application.put_env(:elixir, :ansi_enabled, true)
+    end)
   end
 
   describe "ExCk chunk" do
@@ -40,7 +73,7 @@ defmodule Module.Types.IntegrationTest do
 
       assert [
                {{:c, 0}, %{}},
-               {{:e, 0}, %{deprecated: "oops", sig: {:infer, _}}}
+               {{:e, 0}, %{deprecated: "oops", sig: {:infer, _, _}}}
              ] = read_chunk(modules[A]).exports
 
       assert read_chunk(modules[B]).exports == [
@@ -51,6 +84,105 @@ defmodule Module.Types.IntegrationTest do
                {{:behaviour_info, 1}, %{sig: :none}}
              ]
     end
+
+    test "writes exports with inferred function types" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          def captured, do: &to_capture/1
+          defp to_capture(<<"ok">>), do: :ok
+          defp to_capture(<<"error">>), do: :error
+          defp to_capture([_ | _]), do: :list
+        end
+        """
+      }
+
+      modules = compile_modules(files)
+      exports = read_chunk(modules[A]).exports |> Map.new()
+
+      return = fn name, arity ->
+        pair = {name, arity}
+        %{^pair => %{sig: {:infer, nil, [{_, return}]}}} = exports
+        return
+      end
+
+      assert return.(:captured, 0)
+             |> equal?(
+               fun_from_non_overlapping_clauses([
+                 {[binary()], dynamic(atom([:ok, :error]))},
+                 {[non_empty_list(term(), term())], dynamic(atom([:list]))}
+               ])
+             )
+    end
+
+    test "writes exports for implementations" do
+      files = %{
+        "pi.ex" => """
+        defprotocol Itself do
+          @fallback_to_any true
+          def itself(data)
+        end
+
+        defimpl Itself,
+          for: [
+            Atom,
+            BitString,
+            Float,
+            Function,
+            Integer,
+            List,
+            Map,
+            Port,
+            PID,
+            Reference,
+            Tuple,
+            Any,
+            Range,
+            Unknown
+          ] do
+          def itself(data), do: data
+          def this_wont_warn(:ok), do: :ok
+        end
+        """
+      }
+
+      {modules, stderr} = with_io(:stderr, fn -> compile_modules(files) end)
+
+      assert stderr =~
+               "you are implementing a protocol for Unknown but said module is not available"
+
+      refute stderr =~ "this_wont_warn"
+
+      itself_arg = fn mod ->
+        {_, %{sig: {:infer, nil, [{[value], value}]}}} =
+          List.keyfind(read_chunk(modules[mod]).exports, {:itself, 1}, 0)
+
+        value
+      end
+
+      assert itself_arg.(Itself.Atom) == dynamic(atom())
+      assert itself_arg.(Itself.BitString) == dynamic(binary())
+      assert itself_arg.(Itself.Float) == dynamic(float())
+      assert itself_arg.(Itself.Function) == dynamic(fun())
+      assert itself_arg.(Itself.Integer) == dynamic(integer())
+
+      assert itself_arg.(Itself.List) ==
+               dynamic(union(empty_list(), non_empty_list(term(), term())))
+
+      assert itself_arg.(Itself.Map) == dynamic(open_map(__struct__: if_set(negation(atom()))))
+      assert itself_arg.(Itself.Port) == dynamic(port())
+      assert itself_arg.(Itself.PID) == dynamic(pid())
+      assert itself_arg.(Itself.Reference) == dynamic(reference())
+      assert itself_arg.(Itself.Tuple) == dynamic(tuple())
+      assert itself_arg.(Itself.Any) == dynamic(term())
+
+      assert itself_arg.(Itself.Range) ==
+               dynamic(
+                 closed_map(__struct__: atom([Range]), first: term(), last: term(), step: term())
+               )
+
+      assert itself_arg.(Itself.Unknown) == dynamic(open_map(__struct__: atom([Unknown])))
+    end
   end
 
   describe "type checking" do
@@ -196,29 +328,6 @@ defmodule Module.Types.IntegrationTest do
       assert_warnings(files, warnings)
     end
 
-    test "unused generated private clauses" do
-      files = %{
-        "a.ex" => """
-        defmodule A do
-          use B
-          def public(x), do: private(List.to_tuple(x))
-        end
-        """,
-        "b.ex" => """
-        defmodule B do
-          defmacro __using__(_) do
-            quote generated: true do
-              defp private({:ok, ok}), do: ok
-              defp private(:error), do: :error
-            end
-          end
-        end
-        """
-      }
-
-      assert_no_warnings(files)
-    end
-
     test "unused overridable private clauses" do
       files = %{
         "a.ex" => """
@@ -244,6 +353,247 @@ defmodule Module.Types.IntegrationTest do
       assert_no_warnings(files)
     end
 
+    test "unused private clauses without warnings" do
+      files = %{
+        "a.ex" => """
+        defmodule A do
+          use B
+
+          # Not all clauses are invoked, but do not warn since they are generated
+          def public1(x), do: generated(List.to_tuple(x))
+
+          # Avoid false positives caused by inference
+          def public2(x), do: (:ok = raising_private(x))
+
+          defp raising_private(true), do: :ok
+          defp raising_private(false), do: raise "oops"
+        end
+        """,
+        "b.ex" => """
+        defmodule B do
+          defmacro __using__(_) do
+            quote generated: true do
+              defp generated({:ok, ok}), do: ok
+              defp generated(:error), do: :error
+            end
+          end
+        end
+        """
+      }
+
+      assert_no_warnings(files)
+    end
+
+    test "mismatched implementation" do
+      files = %{
+        "a.ex" => """
+        defprotocol Itself do
+          def itself(data)
+        end
+
+        defimpl Itself, for: Range do
+          def itself(nil), do: nil
+          def itself(range), do: range
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: the 1st pattern in clause will never match:
+
+                nil
+
+            because it is expected to receive type:
+
+                dynamic(%Range{})
+
+            hint: defimpl for Range requires its callbacks to match exclusively on %Range{}
+
+            typing violation found at:
+            │
+          6 │   def itself(nil), do: nil
+            │   ~~~~~~~~~~~~~~~~~~~~~~~~
+            │
+            └─ a.ex:6: Itself.Range.itself/1
+        """
+      ]
+
+      assert_warnings(files, warnings)
+    end
+
+    @tag :require_ast
+    test "no implementation" do
+      files = %{
+        "a.ex" => """
+        defprotocol NoImplProtocol do
+          def callback(data)
+        end
+        """,
+        "b.ex" => """
+        defmodule NoImplProtocol.Caller do
+          def run do
+            NoImplProtocol.callback(:hello)
+          end
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: incompatible types given to NoImplProtocol.callback/1:
+
+                NoImplProtocol.callback(:hello)
+
+            given types:
+
+                -:hello-
+
+            but the protocol was not yet implemented for any type and therefore will always fail. \
+        This error typically happens within libraries that define protocols and will disappear as \
+        soon as there is one implementation. If you expect the protocol to be implemented later on, \
+        you can define an implementation specific for development/test.
+
+            typing violation found at:
+            │
+          3 │     NoImplProtocol.callback(:hello)
+            │                    ~
+            │
+            └─ b.ex:3:20: NoImplProtocol.Caller.run/0
+        """
+      ]
+
+      assert_warnings(files, warnings, consolidate_protocols: true)
+    end
+
+    @tag :require_ast
+    test "String.Chars protocol dispatch" do
+      files = %{
+        "a.ex" => """
+        defmodule FooBar do
+          def example1(_.._//_ = data), do: to_string(data)
+          def example2(_.._//_ = data), do: "hello \#{data} world"
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: incompatible value given to string interpolation:
+
+                data
+
+            it has type:
+
+                -dynamic(%Range{})-
+
+            but expected a type that implements the String.Chars protocol, it must be one of:
+
+                dynamic(
+                  %Date{} or %DateTime{} or %NaiveDateTime{} or %Time{} or %URI{} or %Version{} or
+                    %Version.Requirement{}
+                ) or atom() or binary() or empty_list() or float() or integer() or non_empty_list(term(), term())
+
+            where "data" was given the type:
+
+                # type: dynamic(%Range{})
+                # from: a.ex:3:24
+                _.._//_ = data
+
+            hint: string interpolation uses the String.Chars protocol to convert a data structure into a string. Either convert the data type into a string upfront or implement the protocol accordingly
+        """,
+        """
+            warning: incompatible types given to String.Chars.to_string/1:
+
+                to_string(data)
+
+            given types:
+
+                -dynamic(%Range{})-
+
+            but expected a type that implements the String.Chars protocol, it must be one of:
+
+                dynamic(
+                  %Date{} or %DateTime{} or %NaiveDateTime{} or %Time{} or %URI{} or %Version{} or
+                    %Version.Requirement{}
+                ) or atom() or binary() or empty_list() or float() or integer() or non_empty_list(term(), term())
+
+            where "data" was given the type:
+
+                # type: dynamic(%Range{})
+                # from: a.ex:2:24
+                _.._//_ = data
+        """
+      ]
+
+      assert_warnings(files, warnings, consolidate_protocols: true)
+    end
+
+    @tag :require_ast
+    test "Enumerable protocol dispatch" do
+      files = %{
+        "a.ex" => """
+        defmodule FooBar do
+          def example1(%Date{} = date), do: for(x <- date, do: x)
+          def example2(), do: for(i <- [1, 2, 3], into: Date.utc_today(), do: i * 2)
+          def example3(), do: for(i <- [1, 2, 3], into: 456, do: i * 2)
+        end
+        """
+      }
+
+      warnings = [
+        """
+            warning: incompatible value given to for-comprehension:
+
+                x <- date
+
+            it has type:
+
+                -dynamic(%Date{})-
+
+            but expected a type that implements the Enumerable protocol, it must be one of:
+
+                dynamic(
+                  %Date.Range{} or %File.Stream{} or %GenEvent.Stream{} or %HashDict{} or %HashSet{} or
+                    %IO.Stream{} or %MapSet{} or %Range{} or %Stream{}
+                ) or empty_list() or fun() or non_empty_list(term(), term()) or non_struct_map()
+
+            where "date" was given the type:
+
+                # type: dynamic(%Date{})
+                # from: a.ex:2:24
+                %Date{} = date
+
+            hint: for-comprehensions use the Enumerable protocol to traverse data structures. Either convert the data type into a list (or another Enumerable) or implement the protocol accordingly
+        """,
+        """
+            warning: incompatible value given to :into option in for-comprehension:
+
+                into: Date.utc_today()
+
+            it has type:
+        """,
+        """
+            warning: incompatible value given to :into option in for-comprehension:
+
+                into: 456
+
+            it has type:
+
+                -integer()-
+
+            but expected a type that implements the Collectable protocol, it must be one of:
+
+                dynamic(%File.Stream{} or %HashDict{} or %HashSet{} or %IO.Stream{} or %MapSet{}) or binary() or
+                  empty_list() or non_empty_list(term(), term()) or non_struct_map()
+
+            hint: the :into option in for-comprehensions use the Collectable protocol to build its result. Either pass a valid data type or implement the protocol accordingly
+        """
+      ]
+
+      assert_warnings(files, warnings, consolidate_protocols: true)
+    end
+
     test "incompatible default argument" do
       files = %{
         "a.ex" => """
@@ -304,8 +654,42 @@ defmodule Module.Types.IntegrationTest do
       assert String.ends_with?(file, "generated.ex")
       assert Path.type(file) == :absolute
     after
-      :code.delete(A)
-      :code.purge(A)
+      purge(A)
+    end
+
+    @tag :require_ast
+    test "regressions" do
+      files = %{
+        # do not emit false positives from defguard
+        "a.ex" => """
+        defmodule A do
+          defguard is_non_nil_arity_function(fun, arity)
+                   when arity != nil and is_function(fun, arity)
+
+          def check(fun, args) do
+            is_non_nil_arity_function(fun, length(args))
+          end
+        end
+        """,
+        # do not parse binary segments as variables
+        "b.ex" => """
+        defmodule B do
+          def decode(byte) do
+            case byte do
+              enc when enc in [<<0x00>>, <<0x01>>] -> :ok
+            end
+          end
+        end
+        """,
+        # String.Chars protocol dispatch on improper lists
+        "c.ex" => """
+        defmodule C do
+          def example, do: to_string([?a, ?b | "!"])
+        end
+        """
+      }
+
+      assert_no_warnings(files, consolidate_protocols: true)
     end
   end
 
@@ -822,41 +1206,6 @@ defmodule Module.Types.IntegrationTest do
     end
   end
 
-  describe "regressions" do
-    test "does not emit false positives from defguard" do
-      files = %{
-        "a.ex" => """
-        defmodule A do
-          defguard is_non_nil_arity_function(fun, arity)
-                   when arity != nil and is_function(fun, arity)
-
-          def check(fun, args) do
-            is_non_nil_arity_function(fun, length(args))
-          end
-        end
-        """
-      }
-
-      assert_no_warnings(files)
-    end
-
-    test "do not parse binary segments as variables" do
-      files = %{
-        "a.ex" => """
-        defmodule A do
-          def decode(byte) do
-            case byte do
-              enc when enc in [<<0x00>>, <<0x01>>] -> :ok
-            end
-          end
-        end
-        """
-      }
-
-      assert_no_warnings(files)
-    end
-  end
-
   describe "after_verify" do
     test "reports functions" do
       files = %{
@@ -1025,40 +1374,42 @@ defmodule Module.Types.IntegrationTest do
     end
   end
 
-  defp assert_warnings(files, expected) when is_binary(expected) do
-    assert capture_compile_warnings(files) == expected
+  defp assert_warnings(files, expected, opts \\ [])
+
+  defp assert_warnings(files, expected, opts) when is_binary(expected) do
+    assert capture_compile_warnings(files, opts) == expected
   end
 
-  defp assert_warnings(files, expecteds) when is_list(expecteds) do
-    output = capture_compile_warnings(files)
+  defp assert_warnings(files, expecteds, opts) when is_list(expecteds) do
+    output = capture_compile_warnings(files, opts)
 
     Enum.each(expecteds, fn expected ->
       assert output =~ expected
     end)
   end
 
-  defp assert_no_warnings(files) do
-    assert capture_compile_warnings(files) == ""
+  defp assert_no_warnings(files, opts \\ []) do
+    assert capture_compile_warnings(files, opts) == ""
   end
 
-  defp capture_compile_warnings(files) do
+  defp capture_compile_warnings(files, opts) do
     in_tmp(fn ->
       paths = generate_files(files)
-      capture_io(:stderr, fn -> compile_to_path(paths) end)
+      capture_io(:stderr, fn -> compile_to_path(paths, opts) end)
     end)
   end
 
   defp with_compile_warnings(files) do
     in_tmp(fn ->
       paths = generate_files(files)
-      with_io(:stderr, fn -> compile_to_path(paths) end) |> elem(0)
+      with_io(:stderr, fn -> compile_to_path(paths, []) end) |> elem(0)
     end)
   end
 
   defp compile_modules(files) do
     in_tmp(fn ->
       paths = generate_files(files)
-      {modules, _warnings} = compile_to_path(paths)
+      {modules, _warnings} = compile_to_path(paths, [])
 
       Map.new(modules, fn module ->
         {^module, binary, _filename} = :code.get_object_code(module)
@@ -1067,13 +1418,43 @@ defmodule Module.Types.IntegrationTest do
     end)
   end
 
-  defp compile_to_path(paths) do
+  defp compile_to_path(paths, opts) do
+    if opts[:consolidate_protocols] do
+      Code.prepend_path(".")
+
+      result =
+        compile_to_path_with_after_compile(paths, fn ->
+          if Keyword.get(opts, :consolidate_protocols, false) do
+            paths = [".", Application.app_dir(:elixir, "ebin")]
+            protocols = Protocol.extract_protocols(paths)
+
+            for protocol <- protocols do
+              impls = Protocol.extract_impls(protocol, paths)
+              {:ok, binary} = Protocol.consolidate(protocol, impls)
+              File.write!(Atom.to_string(protocol) <> ".beam", binary)
+              purge(protocol)
+            end
+          end
+        end)
+
+      Code.delete_path(".")
+      Enum.each(builtin_protocols(), &purge/1)
+
+      result
+    else
+      compile_to_path_with_after_compile(paths, fn -> :ok end)
+    end
+  end
+
+  defp compile_to_path_with_after_compile(paths, callback) do
     {:ok, modules, warnings} =
-      Kernel.ParallelCompiler.compile_to_path(paths, ".", return_diagnostics: true)
+      Kernel.ParallelCompiler.compile_to_path(paths, ".",
+        return_diagnostics: true,
+        after_compile: callback
+      )
 
     for module <- modules do
-      :code.delete(module)
-      :code.purge(module)
+      purge(module)
     end
 
     {modules, warnings}
@@ -1088,10 +1469,15 @@ defmodule Module.Types.IntegrationTest do
 
   defp read_chunk(binary) do
     assert {:ok, {_module, [{~c"ExCk", chunk}]}} = :beam_lib.chunks(binary, [~c"ExCk"])
-    assert {:elixir_checker_v1, map} = :erlang.binary_to_term(chunk)
+    assert {:elixir_checker_v3, map} = :erlang.binary_to_term(chunk)
     map
   end
 
+  defp purge(mod) do
+    :code.delete(mod)
+    :code.purge(mod)
+  end
+
   defp in_tmp(fun) do
     path = PathHelpers.tmp_path("checker")
 
