@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("test_helper.exs", __DIR__)
 
 defmodule ProcessTest do
@@ -57,6 +61,12 @@ defmodule ProcessTest do
     assert Process.sleep(0) == :ok
   end
 
+  test "sleep/1 with 2^32" do
+    {pid, monitor_ref} = spawn_monitor(fn -> Process.sleep(2 ** 32) end)
+    refute_receive {:DOWN, ^monitor_ref, :process, ^pid, {:timeout_value, _trace}}, 100
+    Process.exit(pid, :kill)
+  end
+
   test "info/2" do
     pid = spawn(fn -> Process.sleep(1000) end)
     assert Process.info(pid, :priority) == {:priority, :normal}
@@ -120,7 +130,7 @@ defmodule ProcessTest do
       end)
 
     true = Process.exit(pid, :normal)
-    refute_receive {:EXIT, ^pid, :normal}
+    refute_receive {:EXIT, ^pid, :normal}, 100
     assert Process.alive?(pid)
 
     # now exit the process for real so it doesn't hang around
