@@ -1,3 +1,7 @@
+%% SPDX-License-Identifier: Apache-2.0
+%% SPDX-FileCopyrightText: 2021 The Elixir Team
+%% SPDX-FileCopyrightText: 2012 Plataformatec
+
 -module(tokenizer_test).
 -include_lib("eunit/include/eunit.hrl").
 
@@ -61,7 +65,7 @@ unquoted_atom_test() ->
   [{atom, {1, 1, _}, '&&'}] = tokenize(":&&").
 
 quoted_atom_test() ->
-  [{atom_quoted, {1, 1, nil}, 'foo bar'}] = tokenize(":\"foo bar\"").
+  [{atom_quoted, {1, 1, $"}, 'foo bar'}] = tokenize(":\"foo bar\"").
 
 oversized_atom_test() ->
   OversizedAtom = string:copies("a", 256),
