@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule ExUnit.AssertionError do
   @moduledoc """
   Raised to signal an assertion error.
@@ -34,6 +38,20 @@ defmodule ExUnit.AssertionError do
     @no_value
   end
 
+  @impl true
+  def exception(message) when is_binary(message) do
+    %ExUnit.AssertionError{message: message}
+  end
+
+  def exception(opts) do
+    with {:ok, message} when not is_binary(message) <- Keyword.fetch(opts, :message) do
+      raise ArgumentError,
+            ":message in ExUnit.Assertion must be a binary, got: #{inspect(message)}"
+    end
+
+    struct(__MODULE__, opts)
+  end
+
   @impl true
   def message(exception) do
     "\n\n" <> ExUnit.Formatter.format_assertion_error(exception)
@@ -610,7 +628,7 @@ defmodule ExUnit.Assertions do
   def __pins__(pins) do
     pins
     |> Enum.filter(fn {{_, ctx}, _} -> ctx == nil end)
-    |> Enum.reverse()
+    |> Enum.sort()
     |> Enum.map_join(@indent, fn {{name, _}, var} -> "#{name} = #{inspect(var)}" end)
     |> case do
       "" ->
@@ -786,7 +804,7 @@ defmodule ExUnit.Assertions do
   ## Examples
 
       assert_raise ArithmeticError, "bad argument in arithmetic expression", fn ->
-        1 + "test"
+        1 / 0
       end
 
       assert_raise RuntimeError, ~r/^today's lucky number is 0\.\d+!$/, fn ->
