@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule ExUnit.Formatter do
   @moduledoc """
   Helper functions for formatting and the formatting protocols.
@@ -94,6 +98,7 @@ defmodule ExUnit.Formatter do
           | :error_info
           | :test_module_info
           | :test_info
+          | :parameters_info
           | :location_info
           | :stacktrace_info
           | :blame_diff
@@ -119,7 +124,7 @@ defmodule ExUnit.Formatter do
     * `:diff_insert` and `:diff_insert_whitespace` - Should format a diff insertion,
       with or without whitespace respectively.
 
-    * `:extra_info` - Should format extra information, such as the `"code: "` label
+    * `:extra_info` - Should format optional extra labels, such as the `"code: "` label
       that precedes code to show.
 
     * `:error_info` - Should format error information.
@@ -129,6 +134,8 @@ defmodule ExUnit.Formatter do
 
     * `:test_info` - Should format test information.
 
+    * `:parameters_info` - Should format test parameters.
+
     * `:location_info` - Should format test location information.
 
     * `:stacktrace_info` - Should format stacktrace information.
@@ -266,9 +273,10 @@ defmodule ExUnit.Formatter do
         ) :: String.t()
         when failure: {atom, term, Exception.stacktrace()}
   def format_test_failure(test, failures, counter, width, formatter) do
-    %ExUnit.Test{name: name, module: module, tags: tags} = test
+    %ExUnit.Test{name: name, module: module, tags: tags, parameters: parameters} = test
 
     test_info(with_counter(counter, "#{name} (#{inspect(module)})"), formatter) <>
+      test_parameters(parameters, formatter) <>
       test_location(with_location(tags), formatter) <>
       Enum.map_join(Enum.with_index(failures), "", fn {{kind, reason, stack}, index} ->
         {text, stack} = format_kind_reason(test, kind, reason, stack, width, formatter)
@@ -305,9 +313,10 @@ defmodule ExUnit.Formatter do
         ) :: String.t()
         when failure: {atom, term, Exception.stacktrace()}
   def format_test_all_failure(test_module, failures, counter, width, formatter) do
-    name = test_module.name
+    %{name: name, parameters: parameters} = test_module
 
     test_module_info(with_counter(counter, "#{inspect(name)}: "), formatter) <>
+      test_parameters(parameters, formatter) <>
       Enum.map_join(Enum.with_index(failures), "", fn {{kind, reason, stack}, index} ->
         {text, stack} = format_kind_reason(test_module, kind, reason, stack, width, formatter)
         failure_header(failures, index) <> text <> format_stacktrace(stack, name, nil, formatter)
@@ -472,7 +481,19 @@ defmodule ExUnit.Formatter do
     content_width = if width == :infinity, do: width, else: width - padding_size
 
     case format_diff(left, right, context, formatter) do
-      {result, env} ->
+      {nil, hints} when is_atom(context) ->
+        {if_value(left, inspect), if_value(right, inspect), hints}
+
+      {nil, hints} ->
+        left =
+          Macro.prewalk(left, fn
+            {_, [original: original], _} -> original
+            other -> other
+          end)
+
+        {if_value(left, &code_multiline(&1, padding_size)), if_value(right, inspect), hints}
+
+      {result, hints} ->
         left =
           result.left
           |> Diff.to_algebra(&colorize_diff_delete(&1, formatter))
@@ -485,19 +506,18 @@ defmodule ExUnit.Formatter do
           |> Algebra.nest(padding_size)
           |> Algebra.format(content_width)
 
-        {left, right, Enum.map(env.hints, &{:hint, format_hint(&1)})}
+        {left, right, hints}
+    end
+  end
 
-      nil when is_atom(context) ->
-        {if_value(left, inspect), if_value(right, inspect), []}
-
-      nil ->
-        left =
-          Macro.prewalk(left, fn
-            {_, [original: original], _} -> original
-            other -> other
-          end)
-
-        {if_value(left, &code_multiline(&1, padding_size)), if_value(right, inspect), []}
+  defp format_diff(left, right, context, formatter) do
+    if has_value?(left) and has_value?(right) do
+      {result, env} = find_diff(left, right, context)
+      result = if formatter.(:diff_enabled?, false), do: result
+      hints = Enum.map(env.hints, &{:hint, format_hint(&1)})
+      {result, hints}
+    else
+      {nil, []}
     end
   end
 
@@ -505,12 +525,6 @@ defmodule ExUnit.Formatter do
     "you are comparing strings that have the same visual representation but are made of different Unicode codepoints"
   end
 
-  defp format_diff(left, right, context, formatter) do
-    if has_value?(left) and has_value?(right) and formatter.(:diff_enabled?, false) do
-      find_diff(left, right, context)
-    end
-  end
-
   defp colorize_diff_delete(doc, formatter) do
     format = colorize_format(doc, :diff_delete, :diff_delete_whitespace)
     formatter.(format, doc)
@@ -711,6 +725,15 @@ defmodule ExUnit.Formatter do
   defp test_info(msg, nil), do: msg <> "\n"
   defp test_info(msg, formatter), do: test_info(formatter.(:test_info, msg), nil)
 
+  defp test_parameters(params, _formatter) when params == %{}, do: ""
+  defp test_parameters(params, nil) when is_binary(params), do: "     " <> params <> "\n"
+
+  defp test_parameters(params, nil) when is_map(params),
+    do: test_parameters("Parameters: #{inspect(params)}", nil)
+
+  defp test_parameters(params, formatter),
+    do: test_parameters(formatter.(:parameters_info, params), nil)
+
   defp test_location(msg, nil), do: "     " <> msg <> "\n"
   defp test_location(msg, formatter), do: test_location(formatter.(:location_info, msg), nil)
 
