@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule ExUnit.BeamHelpers do
@@ -448,14 +452,16 @@ defmodule ExUnit.DocTestTest.Haiku do
   def crlf_test, do: :ok
 
   defimpl Inspect do
-    def inspect(haiku, _opts) do
-      author = if haiku.author == "", do: "", else: "\n  -- #{haiku.author}"
+    def inspect(haiku, opts) do
+      indent = Keyword.get(opts.custom_options, :indent, 2)
+      spaces = String.duplicate(" ", indent)
+      author = if haiku.author == "", do: "", else: "\n#{spaces}-- #{haiku.author}"
 
       """
       #Haiku<
-        #{haiku.first_phrase}
-        #{haiku.second_phrase}
-        #{haiku.third_phrase}#{author}
+      #{spaces}#{haiku.first_phrase}
+      #{spaces}#{haiku.second_phrase}
+      #{spaces}#{haiku.third_phrase}#{author}
       >
       """
       |> String.trim_trailing("\n")
@@ -464,6 +470,47 @@ defmodule ExUnit.DocTestTest.Haiku do
 end
 |> ExUnit.BeamHelpers.write_beam()
 
+defmodule ExUnit.DocTestTest.HaikuIndent4UsingInspectOpts do
+  @moduledoc """
+  This module is part of the DocTest test suite,
+  to ensure that DocTest respect the `:inspect_opts` options.
+  """
+
+  @doc """
+      # Simple Haiku, inspect output consists of multiple lines.
+      iex> ExUnit.DocTestTest.Haiku.new("Haikus are easy", "But sometimes they don't make sense", "Refrigerator")
+      #Haiku<
+          Haikus are easy
+          But sometimes they don't make sense
+          Refrigerator
+      >
+
+      # Haiku with Unicode characters (Japanese Kanji, em-dash).
+      iex> ExUnit.DocTestTest.Haiku.new("古池や", "蛙飛びこむ", "水の音", "Matsuo Basho")
+      #Haiku<
+          古池や
+          蛙飛びこむ
+          水の音
+          -- Matsuo Basho
+      >
+
+  """
+  def indent_4, do: :ok
+
+  @doc """
+       iex> ExUnit.DocTestTest.Haiku.new("古池や", "蛙飛びこむ", "水の音", "Matsuo Basho")
+       #Haiku<
+           古池や
+           蛙飛びこむ
+           水の音
+           -- Matsuo Basho
+       >
+       """
+       |> String.replace("\n", "\r\n")
+  def crlf_test, do: :ok
+end
+|> ExUnit.BeamHelpers.write_beam()
+
 defmodule ExUnit.DocTestTest.PatternMatching do
   def starting_line, do: __ENV__.line + 2
 
@@ -505,6 +552,34 @@ defmodule ExUnit.DocTestTest.PatternMatching do
 end
 |> ExUnit.BeamHelpers.write_beam()
 
+defmodule ExUnit.DocTestTest.Ellipsis do
+  @doc """
+      iex> ExUnit.DocTestTest.Ellipsis.same_line_err(self())
+      ** (ArgumentError) Unexpected: ...
+  """
+  def same_line_err(arg) do
+    raise ArgumentError, "Unexpected: #{inspect(arg)}"
+  end
+
+  @doc """
+      iex> ExUnit.DocTestTest.Ellipsis.multi_line_err(self())
+      ** (ArgumentError) Unexpected:
+      ...
+  """
+  def multi_line_err(arg) do
+    raise ArgumentError, "Unexpected:\n#{inspect(arg)}"
+  end
+
+  @doc """
+      iex> ExUnit.DocTestTest.Ellipsis.triple_dot_err(:foo)
+      ** (ArgumentError) :foo ...
+  """
+  def triple_dot_err(arg) do
+    raise ArgumentError, "#{inspect(arg)} ..."
+  end
+end
+|> ExUnit.BeamHelpers.write_beam()
+
 defmodule ExUnit.DocTestTest do
   use ExUnit.Case
 
@@ -528,6 +603,11 @@ defmodule ExUnit.DocTestTest do
   doctest ExUnit.DocTestTest.FencedHeredocs
   doctest ExUnit.DocTestTest.Haiku
 
+  doctest ExUnit.DocTestTest.HaikuIndent4UsingInspectOpts,
+    inspect_opts: [custom_options: [indent: 4]]
+
+  doctest ExUnit.DocTestTest.Ellipsis
+
   import ExUnit.CaptureIO
 
   test "multiple functions filtered with :only" do
@@ -578,22 +658,23 @@ defmodule ExUnit.DocTestTest do
     ExUnit.configure(seed: 0, colors: [enabled: false])
     output = capture_io(fn -> ExUnit.run() end)
     output = String.replace(output, [IO.ANSI.red(), IO.ANSI.reset()], "")
+    location = location(ExUnit.DocTestTest.Failure)
 
     assert output =~ """
              1) doctest module ExUnit.DocTestTest.Failure (1) (ExUnit.DocTestTest.FailureCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (SyntaxError) invalid syntax found on test/ex_unit/doc_test_test.exs:#{starting_line}:6:
+                Doctest did not compile, got: (SyntaxError) invalid syntax found on #{location}:#{starting_line}:6:
                      error: syntax error before: '*'
                      │
                  #{starting_line} │  1 + * 1
                      │      ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{starting_line}:6
+                     └─ #{location}:#{starting_line}:6
                 doctest:
                   iex> 1 + * 1
                   1
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{starting_line}: ExUnit.DocTestTest.Failure (module)
+                  #{stack(starting_line)}ExUnit.DocTestTest.Failure (module)
            """
 
     assert output =~ """
@@ -607,7 +688,7 @@ defmodule ExUnit.DocTestTest do
                 left:  2
                 right: 3
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{starting_line + 3}: ExUnit.DocTestTest.Failure (module)
+                  #{stack(starting_line + 3)}ExUnit.DocTestTest.Failure (module)
            """
 
     assert output =~ """
@@ -624,7 +705,7 @@ defmodule ExUnit.DocTestTest do
                 left:  "This is a slightly shorter text string."
                 right: "This is a much shorter text string."
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{starting_line + 6}: ExUnit.DocTestTest.Failure (module)
+                  #{stack(starting_line + 6)}ExUnit.DocTestTest.Failure (module)
            """
 
     assert output =~ """
@@ -638,16 +719,16 @@ defmodule ExUnit.DocTestTest do
                 left:  ":oops"
                 right: "#Inspect<[]>"
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{starting_line + 12}: ExUnit.DocTestTest.Failure (module)
+                  #{stack(starting_line + 12)}ExUnit.DocTestTest.Failure (module)
            """
 
     assert output =~ """
              5) doctest module ExUnit.DocTestTest.Failure (5) (ExUnit.DocTestTest.FailureCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                ** (UndefinedFunctionError) function Hello.world/0 is undefined (module Hello is not available)
+                ** (UndefinedFunctionError) function Hello.world/0 is undefined (module Hello is not available). Make sure the module name is correct and has been specified in full (or that an alias has been defined)
                 stacktrace:
                   Hello.world()
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 15}: (test)
+                  (for doctest at) #{location}:#{starting_line + 15}: (test)
            """
 
     assert output =~ """
@@ -658,7 +739,7 @@ defmodule ExUnit.DocTestTest do
                   iex> raise "oops"
                   ** (WhatIsThis) oops
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{starting_line + 18}: ExUnit.DocTestTest.Failure (module)
+                  #{stack(starting_line + 18)}ExUnit.DocTestTest.Failure (module)
            """
 
     assert output =~ """
@@ -673,14 +754,14 @@ defmodule ExUnit.DocTestTest do
                   iex> raise "oops"
                   ** (RuntimeError) hello
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{starting_line + 21}: ExUnit.DocTestTest.Failure (module)
+                  #{stack(starting_line + 21)}ExUnit.DocTestTest.Failure (module)
            """
 
     assert output =~ "8) doctest ExUnit.DocTestTest.Failure.raising_inspect/0"
     assert output =~ "iex> ExUnit.DocTestTest.Haiku.new(:this, :is, {:not, :a, :haiku})"
 
     assert output =~
-             "test/ex_unit/doc_test_test.exs:#{starting_line + 28}: ExUnit.DocTestTest.Failure (module)"
+             "#{stack(starting_line + 28)}ExUnit.DocTestTest.Failure (module)"
 
     assert output =~ "8 doctests, 8 failures"
   end
@@ -696,26 +777,29 @@ defmodule ExUnit.DocTestTest do
 
     doctest_line = __ENV__.line - 3
     starting_line = ExUnit.DocTestTest.Invalid.starting_line() + 1
+    location = location(ExUnit.DocTestTest.Invalid)
 
     ExUnit.configure(seed: 0, colors: [enabled: false])
     output = capture_io(fn -> ExUnit.run() end)
     output = String.replace(output, [IO.ANSI.red(), IO.ANSI.reset()], "")
 
+    line = starting_line
+
     assert output =~ """
              1) doctest ExUnit.DocTestTest.Invalid.a/0 (1) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (SyntaxError) invalid syntax found on test/ex_unit/doc_test_test.exs:#{starting_line}:6:
+                Doctest did not compile, got: (SyntaxError) invalid syntax found on #{location}:#{line}:6:
                      error: syntax error before: '*'
                      │
-                 #{starting_line} │  1 + * 1
+                 #{line} │  1 + * 1
                      │      ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{starting_line}:6
+                     └─ #{location}:#{line}:6
                 doctest:
                   iex> 1 + * 1
                   1
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{starting_line}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 6
@@ -723,18 +807,18 @@ defmodule ExUnit.DocTestTest do
     assert output =~ """
              2) doctest ExUnit.DocTestTest.Invalid.b/0 (2) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (SyntaxError) invalid syntax found on test/ex_unit/doc_test_test.exs:#{line}:6:
+                Doctest did not compile, got: (SyntaxError) invalid syntax found on #{location}:#{line}:6:
                      error: syntax error before: '*'
                      │
                  #{line} │  1 + * 1
                      │      ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{line}:6
+                     └─ #{location}:#{line}:6
                 doctest:
                   iex> 1 + * 1
                   1
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 15
@@ -742,19 +826,19 @@ defmodule ExUnit.DocTestTest do
     assert output =~ """
              3) doctest ExUnit.DocTestTest.Invalid.indented_too_much/0 (3) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (SyntaxError) invalid syntax found on test/ex_unit/doc_test_test.exs:#{line}:3:
+                Doctest did not compile, got: (SyntaxError) invalid syntax found on #{location}:#{line}:3:
                      error: unexpected token: "`" (column 3, code point U+0060)
                      │
                  #{line} │   ```
                      │   ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{line}:3
+                     └─ #{location}:#{line}:3
                 doctest:
                   iex> 1 + 2
                   3
                     ```
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line - 1}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line - 1)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 23
@@ -762,19 +846,19 @@ defmodule ExUnit.DocTestTest do
     assert output =~ """
              4) doctest ExUnit.DocTestTest.Invalid.dedented_past_fence/0 (4) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (SyntaxError) invalid syntax found on test/ex_unit/doc_test_test.exs:#{line}:5:
+                Doctest did not compile, got: (SyntaxError) invalid syntax found on #{location}:#{line}:5:
                      error: unexpected token: "`" (column 5, code point U+0060)
                      │
                  #{line} │     ```
                      │     ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{line}:5
+                     └─ #{location}:#{line}:5
                 doctest:
                   iex> 1 + 2
                   3
                       ```
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line - 1}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line - 1)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 28
@@ -784,7 +868,7 @@ defmodule ExUnit.DocTestTest do
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
                 Doctest did not compile, got: (UnicodeConversionError) invalid encoding starting at <<255, 34, 41>>
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 35
@@ -792,20 +876,20 @@ defmodule ExUnit.DocTestTest do
     assert output =~ """
            6) doctest ExUnit.DocTestTest.Invalid.misplaced_opaque_type/0 (6) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (TokenMissingError) token missing on test/ex_unit/doc_test_test.exs:#{line}:20:
+                Doctest did not compile, got: (TokenMissingError) token missing on #{location}:#{line}:20:
                      error: missing terminator: }
                      │
-                 228 │ {:ok, #Inspect<[]>}
+                 #{line} │ {:ok, #Inspect<[]>}
                      │ │                  └ missing closing delimiter (expected "}")
                      │ └ unclosed delimiter
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{line}:20
+                     └─ #{location}:#{line}:20
                 If you are planning to assert on the result of an iex> expression which contains a value inspected as #Name<...>, please make sure the inspected value is placed at the beginning of the expression, otherwise Elixir will treat it as a comment due to the leading sign #.
                 doctest:
                   iex> {:ok, :oops}
                   {:ok, #Inspect<[]>}
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 40
@@ -813,18 +897,18 @@ defmodule ExUnit.DocTestTest do
     assert output =~ """
              7) doctest ExUnit.DocTestTest.Invalid.t/0 (7) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (SyntaxError) invalid syntax found on test/ex_unit/doc_test_test.exs:#{line}:6:
+                Doctest did not compile, got: (SyntaxError) invalid syntax found on #{location}:#{line}:6:
                      error: syntax error before: '*'
                      │
                  #{line} │  1 + * 1
                      │      ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{line}:6
+                     └─ #{location}:#{line}:6
                 doctest:
                   iex> 1 + * 1
                   1
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 48
@@ -832,19 +916,19 @@ defmodule ExUnit.DocTestTest do
     assert output =~ """
              8) doctest ExUnit.DocTestTest.Invalid.result/0 (8) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (SyntaxError) invalid syntax found on test/ex_unit/doc_test_test.exs:#{line}:5:
+                Doctest did not compile, got: (SyntaxError) invalid syntax found on #{location}:#{line}:5:
                      error: syntax error before: '*'
                      │
                  #{line} │ 1 + * 1
                      │     ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{line}:5
+                     └─ #{location}:#{line}:5
                 doctest:
                   iex> :foo
                   iex> :bar
                   1 + * 1
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 53
@@ -852,18 +936,18 @@ defmodule ExUnit.DocTestTest do
     assert output =~ """
              9) doctest ExUnit.DocTestTest.Invalid.mixed/0 (9) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (TokenMissingError) token missing on test/ex_unit/doc_test_test.exs:#{line}:6:
+                Doctest did not compile, got: (TokenMissingError) token missing on #{location}:#{line}:6:
                      error: syntax error: expression is incomplete
                      │
                  #{line} │  123 +
                      │      ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{line}:6
+                     └─ #{location}:#{line}:6
                 doctest:
                   iex> 123 +
                   :mixed
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line)}ExUnit.DocTestTest.Invalid (module)
            """
 
     line = starting_line + 61
@@ -871,18 +955,18 @@ defmodule ExUnit.DocTestTest do
     assert output =~ """
             10) doctest ExUnit.DocTestTest.Invalid.invalid_second/0 (10) (ExUnit.DocTestTest.InvalidCompiled)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
-                Doctest did not compile, got: (TokenMissingError) token missing on test/ex_unit/doc_test_test.exs:#{line}:6:
+                Doctest did not compile, got: (TokenMissingError) token missing on #{location}:#{line}:6:
                      error: syntax error: expression is incomplete
                      │
                  #{line} │  123 +
                      │      ^
                      │
-                     └─ test/ex_unit/doc_test_test.exs:#{line}:6
+                     └─ #{location}:#{line}:6
                 doctest:
                   iex> 123 +
                   :mixed
                 stacktrace:
-                  test/ex_unit/doc_test_test.exs:#{line}: ExUnit.DocTestTest.Invalid (module)
+                  #{stack(line)}ExUnit.DocTestTest.Invalid (module)
            """
 
     assert output =~ "10 doctests, 10 failures"
@@ -896,6 +980,7 @@ defmodule ExUnit.DocTestTest do
 
     doctest_line = __ENV__.line - 3
     starting_line = ExUnit.DocTestTest.PatternMatching.starting_line()
+    location = location(ExUnit.DocTestTest.PatternMatching)
 
     ExUnit.configure(seed: 0, colors: [enabled: false])
     output = capture_io(fn -> ExUnit.run() end)
@@ -908,7 +993,7 @@ defmodule ExUnit.DocTestTest do
                 left:  {1, 2}
                 right: {1, 3}
                 stacktrace:
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 3}: (test)
+                  (for doctest at) #{location}:#{starting_line + 3}: (test)
            """
 
     assert output =~ """
@@ -919,7 +1004,7 @@ defmodule ExUnit.DocTestTest do
                 left:  {1, 2}
                 right: {1, 3}
                 stacktrace:
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 6}: (test)
+                  (for doctest at) #{location}:#{starting_line + 6}: (test)
            """
 
     assert output =~ """
@@ -930,7 +1015,7 @@ defmodule ExUnit.DocTestTest do
                 left:  "Hello, world"
                 right: "Hello world"
                 stacktrace:
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 8}: (test)
+                  (for doctest at) #{location}:#{starting_line + 8}: (test)
            """
 
     assert output =~ """
@@ -941,7 +1026,7 @@ defmodule ExUnit.DocTestTest do
                 left:  "Hello, " <> _
                 right: "Hello world"
                 stacktrace:
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 10}: (test)
+                  (for doctest at) #{location}:#{starting_line + 10}: (test)
            """
 
     assert output =~ """
@@ -952,7 +1037,7 @@ defmodule ExUnit.DocTestTest do
                 left:  [:a | _]
                 right: [:b, :a]
                 stacktrace:
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 12}: (test)
+                  (for doctest at) #{location}:#{starting_line + 12}: (test)
            """
 
     assert output =~ """
@@ -965,7 +1050,7 @@ defmodule ExUnit.DocTestTest do
                 left:  [^atom | _]
                 right: [:b, :a]
                 stacktrace:
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 15}: (test)
+                  (for doctest at) #{location}:#{starting_line + 15}: (test)
            """
 
     assert output =~ """
@@ -976,18 +1061,18 @@ defmodule ExUnit.DocTestTest do
                 left:  %{b: _, d: :e}
                 right: %{a: :c, d: :e}
                 stacktrace:
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 17}: (test)
+                  (for doctest at) #{location}:#{starting_line + 17}: (test)
            """
 
     assert output =~ """
              8) doctest module ExUnit.DocTestTest.PatternMatching (8) (ExUnit.DocTestTest.PatternMatchingRunner)
                 test/ex_unit/doc_test_test.exs:#{doctest_line}
                 match (=) failed
-                code:  %{year: 2001, day: 1} = ~D"2000-01-01"
+                code:  %{year: 2001, day: 1} = ~D[2000-01-01]
                 left:  %{year: 2001, day: 1}
                 right: ~D[2000-01-01]
                 stacktrace:
-                  (for doctest at) test/ex_unit/doc_test_test.exs:#{starting_line + 19}: (test)
+                  (for doctest at) #{location}:#{starting_line + 19}: (test)
            """
 
     assert output =~ "10 doctests, 8 failures"
@@ -1006,9 +1091,7 @@ defmodule ExUnit.DocTestTest do
     message =
       ~s[unknown IEx prompt: "iex(foo@bar)1> 1 +".\nAccepted formats are: iex>, iex(1)>, ...>, ...(1)>]
 
-    regex = ~r[test/ex_unit/doc_test_test\.exs:\d+: #{Regex.escape(message)}]
-
-    assert_raise ExUnit.DocTest.Error, regex, fn ->
+    assert_raise ExUnit.DocTest.Error, ~r/#{Regex.escape(message)}/, fn ->
       defmodule HostUsage do
         use ExUnit.Case
         doctest ExUnit.DocTestTest.Host
@@ -1030,22 +1113,22 @@ defmodule ExUnit.DocTestTest do
     assert %{
              test_type: :doctest,
              doctest: NoImport,
-             doctest_line: 133,
-             doctest_data: %{end_line: 134}
+             doctest_line: 137,
+             doctest_data: %{end_line: 138}
            } = test1.tags
 
     assert %{
              test_type: :doctest,
              doctest: NoImport,
-             doctest_line: 136,
-             doctest_data: %{end_line: 137}
+             doctest_line: 140,
+             doctest_data: %{end_line: 141}
            } = test2.tags
 
     assert %{
              test_type: :doctest,
              doctest: NoTrailing,
-             doctest_line: 145,
-             doctest_data: %{end_line: 145}
+             doctest_line: 149,
+             doctest_data: %{end_line: 149}
            } = test3.tags
   end
 
@@ -1074,7 +1157,7 @@ defmodule ExUnit.DocTestTest do
 
     test "fails when testing functions are not found" do
       message = """
-      test/ex_unit/doc_test_test\.exs: undefined or private functions given to doctest:
+      #{stack(0)}undefined or private functions given to doctest:
 
           ExUnit.DocTestTest.SomewhatGoodModuleWithOnly.three/0
           ExUnit.DocTestTest.SomewhatGoodModuleWithOnly.four/1
@@ -1105,7 +1188,7 @@ defmodule ExUnit.DocTestTest do
 
     test "fails in indentation mismatch" do
       message =
-        ~r[test/ex_unit/doc_test_test\.exs:\d+: indentation level mismatch on doctest line: \"   iex> bar = 2\".*is exactly 2 spaces]s
+        ~r[indentation level mismatch on doctest line: \"   iex> bar = 2\".*is exactly 2 spaces]s
 
       assert_raise ExUnit.DocTest.Error, message, fn ->
         defmodule NeverCompiled do
@@ -1114,8 +1197,7 @@ defmodule ExUnit.DocTestTest do
         end
       end
 
-      message =
-        ~r[test/ex_unit/doc_test_test\.exs:\d+: indentation level mismatch on doctest line: \"    3\".*is exactly 2 spaces]s
+      message = ~r[indentation level mismatch on doctest line: \"    3\".*is exactly 2 spaces]s
 
       assert_raise ExUnit.DocTest.Error, message, fn ->
         defmodule NeverCompiled do
@@ -1124,8 +1206,7 @@ defmodule ExUnit.DocTestTest do
         end
       end
 
-      message =
-        ~r[test/ex_unit/doc_test_test\.exs:\d+: indentation level mismatch on doctest line: \"  3\".*is exactly 4 spaces]s
+      message = ~r[indentation level mismatch on doctest line: \"  3\".*is exactly 4 spaces]s
 
       assert_raise ExUnit.DocTest.Error, message, fn ->
         defmodule NeverCompiled do
@@ -1134,8 +1215,7 @@ defmodule ExUnit.DocTestTest do
         end
       end
 
-      message =
-        ~r[test/ex_unit/doc_test_test\.exs:\d+: indentation level mismatch on doctest line: \"```\".*is exactly 2 spaces]s
+      message = ~r[indentation level mismatch on doctest line: \"```\".*is exactly 2 spaces]s
 
       assert_raise ExUnit.DocTest.Error, message, fn ->
         defmodule NeverCompiled do
@@ -1145,7 +1225,7 @@ defmodule ExUnit.DocTestTest do
       end
 
       message =
-        ~r[test/ex_unit/doc_test_test\.exs:\d+: indentation level mismatch on doctest line: \"world\\\"\".*is exactly 4 spaces]s
+        ~r[indentation level mismatch on doctest line: \"world\\\"\".*is exactly 4 spaces]s
 
       assert_raise ExUnit.DocTest.Error, message, fn ->
         defmodule NeverCompiled do
@@ -1211,4 +1291,20 @@ defmodule ExUnit.DocTestTest do
     [head | _] = Direct.__ex_unit__().tests
     assert apply(Direct, head.name, [%{}]) == {:ok, 2}
   end
+
+  defp location(module) do
+    if __MODULE__.module_info(:compile)[:source] do
+      "test/ex_unit/doc_test_test.exs"
+    else
+      inspect(module)
+    end
+  end
+
+  defp stack(line) do
+    if __MODULE__.module_info(:compile)[:source] do
+      [Exception.format_file_line("test/ex_unit/doc_test_test.exs", line), ?\s]
+    else
+      ""
+    end
+  end
 end
