@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule ExUnit.SupervisedTest do
@@ -149,7 +153,7 @@ defmodule ExUnit.SupervisedTest do
   end
 
   defp wait_until_registered(name) do
-    unless Process.whereis(name) do
+    if !Process.whereis(name) do
       wait_until_registered(name)
     end
   end
