@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("test_helper.exs", __DIR__)
 
 defmodule ExUnitTest do
@@ -29,7 +33,7 @@ defmodule ExUnitTest do
     assert capture_io(fn ->
              assert ExUnit.async_run() |> ExUnit.await_run() ==
                       %{failures: 0, skipped: 0, total: 0, excluded: 0}
-           end) =~ "\n0 failures\n"
+           end) =~ "\n0 tests, 0 failures\n"
   end
 
   test "supports rerunning given modules" do
@@ -137,7 +141,7 @@ defmodule ExUnitTest do
     assert result =~ """
            Showing results so far...
 
-           0 failures
+           0 tests, 0 failures
            """
   end
 
@@ -331,19 +335,19 @@ defmodule ExUnitTest do
 
     {result, output} = run_with_filter([exclude: [even: true]], [ParityTest])
     assert result == %{failures: 0, skipped: 0, excluded: 1, total: 4}
-    assert output =~ "\n4 tests, 0 failures, 1 excluded\n"
+    assert output =~ "\n3 tests, 0 failures (1 excluded)\n"
 
     {result, output} = run_with_filter([exclude: :even], [ParityTest])
     assert result == %{failures: 0, skipped: 0, excluded: 3, total: 4}
-    assert output =~ "\n4 tests, 0 failures, 3 excluded\n"
+    assert output =~ "\n1 test, 0 failures (3 excluded)\n"
 
     {result, output} = run_with_filter([exclude: :even, include: [even: true]], [ParityTest])
     assert result == %{failures: 1, skipped: 0, excluded: 2, total: 4}
-    assert output =~ "\n4 tests, 1 failure, 2 excluded\n"
+    assert output =~ "\n2 tests, 1 failure (2 excluded)\n"
 
     {result, output} = run_with_filter([exclude: :test, include: [even: true]], [ParityTest])
     assert result == %{failures: 1, skipped: 0, excluded: 3, total: 4}
-    assert output =~ "\n4 tests, 1 failure, 3 excluded\n"
+    assert output =~ "\n1 test, 1 failure (3 excluded)\n"
   end
 
   test "log capturing" do
@@ -508,8 +512,9 @@ defmodule ExUnitTest do
 
     # Empty because it is already loaded
     {result, output} = run_with_filter([exclude: :module], [])
+
     assert result == %{failures: 0, skipped: 0, excluded: 2, total: 2}
-    assert output =~ "\n2 tests, 0 failures, 2 excluded\n"
+    assert output =~ "\n0 tests, 0 failures (2 excluded)\n"
 
     {result, output} =
       [exclude: :test, include: [module: "ExUnitTest.SecondTestModule"]]
@@ -517,7 +522,7 @@ defmodule ExUnitTest do
 
     assert result == %{failures: 1, skipped: 0, excluded: 1, total: 2}
     assert output =~ "\n  1) test false (ExUnitTest.SecondTestModule)\n"
-    assert output =~ "\n2 tests, 1 failure, 1 excluded\n"
+    assert output =~ "\n1 test, 1 failure (1 excluded)\n"
   end
 
   test "raises on reserved tag :file in module" do
@@ -680,7 +685,7 @@ defmodule ExUnitTest do
       end)
 
     refute output =~ max_failures_reached_msg()
-    assert output =~ "\n6 tests, 0 failures, 1 excluded, 4 invalid, 1 skipped\n"
+    assert output =~ "\n5 tests, 0 failures, 4 invalid, 1 skipped (1 excluded)\n"
   end
 
   test "parameterized tests" do
@@ -715,6 +720,30 @@ defmodule ExUnitTest do
     assert_receive {:tmp_dir, tmp_dir2} when tmp_dir1 != tmp_dir2
   end
 
+  test "empty parameterized tests" do
+    defmodule EmptyParameterizedTests do
+      use ExUnit.Case, async: true, parameterize: []
+
+      test "hello world" do
+        assert false
+      end
+    end
+
+    configure_and_reload_on_exit(trace: true)
+    assert capture_io(fn -> ExUnit.run() end) =~ "0 tests"
+
+    defmodule EmptyGroupedParameterizedTests do
+      use ExUnit.Case, async: true, parameterize: [], group: :example
+
+      test "hello world" do
+        assert false
+      end
+    end
+
+    configure_and_reload_on_exit(trace: true)
+    assert capture_io(fn -> ExUnit.run() end) =~ "0 tests"
+  end
+
   describe "after_suite/1" do
     test "executes all callbacks set in reverse order" do
       Process.register(self(), :after_suite_test_process)
@@ -787,7 +816,7 @@ defmodule ExUnitTest do
         end)
 
       assert output =~ max_failures_reached_msg()
-      assert output =~ "\n6 tests, 2 failures, 1 excluded, 1 skipped\n"
+      assert output =~ "\n5 tests, 2 failures, 1 skipped (1 excluded)\n"
     end
 
     test ":max_failures is not reached" do
@@ -818,7 +847,7 @@ defmodule ExUnitTest do
         end)
 
       refute output =~ max_failures_reached_msg()
-      assert output =~ "\n8 tests, 2 failures, 2 excluded, 1 skipped\n"
+      assert output =~ "\n6 tests, 2 failures, 1 skipped (2 excluded)\n"
     end
 
     test ":max_failures has been reached" do
@@ -852,7 +881,7 @@ defmodule ExUnitTest do
         end)
 
       assert output =~ max_failures_reached_msg()
-      assert output =~ "\n7 tests, 2 failures, 2 excluded, 2 skipped\n"
+      assert output =~ "\n5 tests, 2 failures, 2 skipped (2 excluded)\n"
     end
 
     # Excluded and skipped tests are detected before setup_all
@@ -886,7 +915,7 @@ defmodule ExUnitTest do
         end)
 
       assert output =~ max_failures_reached_msg()
-      assert output =~ "\n4 tests, 0 failures, 1 excluded, 2 invalid, 1 skipped\n"
+      assert output =~ "\n3 tests, 0 failures, 2 invalid, 1 skipped (1 excluded)\n"
     end
 
     test ":max_failures flushes all async/sync cases" do
@@ -997,6 +1026,24 @@ defmodule ExUnitTest do
       assert length(runs) == 6
     end
 
+    test "repeats tests up to the configured number of times with groups" do
+      defmodule TestGroupedRepeatUntilFailureReached do
+        use ExUnit.Case, async: true, group: :example
+        test __ENV__.line, do: assert(true)
+      end
+
+      configure_and_reload_on_exit(repeat_until_failure: 5)
+
+      output =
+        capture_io(fn ->
+          assert ExUnit.run() == %{total: 1, failures: 0, skipped: 0, excluded: 0}
+        end)
+
+      runs = String.split(output, "Running ExUnit", trim: true)
+      # 6 runs in total, 5 repeats
+      assert length(runs) == 6
+    end
+
     test "stops on failure" do
       {:ok, pid} = Agent.start_link(fn -> 0 end)
       Process.register(pid, :ex_unit_repeat_until_failure_count)
@@ -1059,7 +1106,7 @@ defmodule ExUnitTest do
       end)
 
     assert output =~ "All tests have been excluded.\n"
-    assert output =~ "2 tests, 0 failures, 2 excluded\n"
+    assert output =~ "0 tests, 0 failures (2 excluded)\n"
   end
 
   test "tests are run in compile order (FIFO)" do
@@ -1211,6 +1258,14 @@ defmodule ExUnitTest do
     end
   end
 
+  test "sets process label for each test" do
+    # TODO: Remove check once we support Erlang/OTP 27+
+    if function_exported?(:proc_lib, :get_label, 1) do
+      label = apply(:proc_lib, :get_label, [self()])
+      assert label == {ExUnitTest, :"test sets process label for each test"}
+    end
+  end
+
   defp configure_and_reload_on_exit(opts) do
     old_opts = ExUnit.configuration()
 
