@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule IEx.Evaluator do
   @moduledoc false
 
@@ -51,26 +55,21 @@ defmodule IEx.Evaluator do
     end
   end
 
-  # If parsing fails, this might be a TokenMissingError which we treat in
-  # a special way (to allow for continuation of an expression on the next
-  # line in IEx).
-  #
-  # The first two clauses provide support for the break-trigger allowing to
-  # break out from a pending incomplete expression. See
-  # https://github.com/elixir-lang/elixir/issues/1089 for discussion.
-  @break_trigger "#iex:break\n"
+  @break_trigger ~c"#iex:break\n"
 
   @op_tokens [:or_op, :and_op, :comp_op, :rel_op, :arrow_op, :in_op] ++
                [:three_op, :concat_op, :mult_op]
 
-  @doc false
+  @doc """
+  Default parsing implementation with support for pipes and #iex:break.
+
+  If parsing fails, this might be a TokenMissingError which we treat in
+  a special way (to allow for continuation of an expression on the next
+  line in IEx).
+  """
   def parse(input, opts, parser_state)
 
-  def parse(input, opts, ""), do: parse(input, opts, {"", :other})
-
-  def parse(@break_trigger, _opts, {"", _} = parser_state) do
-    {:incomplete, parser_state}
-  end
+  def parse(input, opts, []), do: parse(input, opts, {[], :other})
 
   def parse(@break_trigger, opts, _parser_state) do
     :elixir_errors.parse_error(
@@ -78,20 +77,20 @@ defmodule IEx.Evaluator do
       opts[:file],
       "incomplete expression",
       "",
-      {~c"", Keyword.get(opts, :line, 1), Keyword.get(opts, :column, 1)}
+      {~c"", Keyword.get(opts, :line, 1), Keyword.get(opts, :column, 1), 0}
     )
   end
 
   def parse(input, opts, {buffer, last_op}) do
-    input = buffer <> input
+    input = buffer ++ input
     file = Keyword.get(opts, :file, "nofile")
     line = Keyword.get(opts, :line, 1)
     column = Keyword.get(opts, :column, 1)
-    charlist = String.to_charlist(input)
 
     result =
-      with {:ok, tokens} <- :elixir.string_to_tokens(charlist, line, column, file, opts),
-           {:ok, adjusted_tokens} <- adjust_operator(tokens, line, column, file, opts, last_op),
+      with {:ok, tokens} <- :elixir.string_to_tokens(input, line, column, file, opts),
+           {:ok, adjusted_tokens, adjusted_op} <-
+             adjust_operator(tokens, line, column, file, opts, last_op),
            {:ok, forms} <- :elixir.tokens_to_quoted(adjusted_tokens, file, opts) do
         last_op =
           case forms do
@@ -99,12 +98,22 @@ defmodule IEx.Evaluator do
             _ -> :other
           end
 
+        forms =
+          if adjusted_op != nil do
+            quote do
+              IEx.Evaluator.assert_no_error!()
+              unquote(forms)
+            end
+          else
+            forms
+          end
+
         {:ok, forms, last_op}
       end
 
     case result do
       {:ok, forms, last_op} ->
-        {:ok, forms, {"", last_op}}
+        {:ok, forms, {[], last_op}}
 
       {:error, {_, _, ""}} ->
         {:incomplete, {input, last_op}}
@@ -115,7 +124,7 @@ defmodule IEx.Evaluator do
           file,
           error,
           token,
-          {charlist, line, column}
+          {input, line, column, 0}
         )
     end
   end
@@ -131,10 +140,20 @@ defmodule IEx.Evaluator do
   defp adjust_operator([{op_type, _, _} | _] = tokens, line, column, file, opts, _last_op)
        when op_type in @op_tokens do
     {:ok, prefix} = :elixir.string_to_tokens(~c"v(-1)", line, column, file, opts)
-    {:ok, prefix ++ tokens}
+    {:ok, prefix ++ tokens, op_type}
   end
 
-  defp adjust_operator(tokens, _line, _column, _file, _opts, _last_op), do: {:ok, tokens}
+  defp adjust_operator(tokens, _line, _column, _file, _opts, _last_op), do: {:ok, tokens, nil}
+
+  @doc """
+  Raises an error if the last iex result was itself an error
+  """
+  def assert_no_error!() do
+    if Process.get(:iex_error, false) do
+      message = "skipping evaluation of expression because pipeline has failed"
+      reraise RuntimeError.exception(message), []
+    end
+  end
 
   @doc """
   Gets a value out of the binding, using the provided
@@ -185,9 +204,13 @@ defmodule IEx.Evaluator do
 
   defp loop(%{server: server, ref: ref} = state) do
     receive do
-      {:eval, ^server, code, counter, parser_state} ->
-        {status, parser_state, state} = parse_eval_inspect(code, counter, parser_state, state)
-        send(server, {:evaled, self(), status, parser_state})
+      {:reader_errored, ^server} ->
+        Process.put(:iex_error, true)
+        loop(state)
+
+      {:eval, ^server, code, counter} ->
+        {status, state} = safe_eval_and_inspect(code, counter, state)
+        send(server, {:evaled, self(), status})
         loop(state)
 
       {:fields_from_env, ^server, ref, receiver, fields} ->
@@ -292,32 +315,22 @@ defmodule IEx.Evaluator do
     end
   end
 
-  defp parse_eval_inspect(code, counter, parser_state, state) do
-    try do
-      {parser_module, parser_fun, args} = IEx.Config.parser()
-      args = [code, [line: counter, file: "iex"], parser_state | args]
-      eval_and_inspect_parsed(apply(parser_module, parser_fun, args), counter, state)
-    catch
-      kind, error ->
-        print_error(kind, error, __STACKTRACE__)
-        {:error, "", state}
-    end
-  end
-
-  defp eval_and_inspect_parsed({:ok, forms, parser_state}, counter, state) do
+  defp safe_eval_and_inspect(forms, counter, state) do
     put_history(state)
     put_whereami(state)
-    state = eval_and_inspect(forms, counter, state)
-    {:ok, parser_state, state}
+    result = eval_and_inspect(forms, counter, state)
+    Process.delete(:iex_error)
+    {:ok, result}
+  catch
+    kind, error ->
+      Process.put(:iex_error, true)
+      print_error(kind, error, __STACKTRACE__)
+      {:error, state}
   after
     Process.delete(:iex_history)
     Process.delete(:iex_whereami)
   end
 
-  defp eval_and_inspect_parsed({:incomplete, parser_state}, _counter, state) do
-    {:incomplete, parser_state, state}
-  end
-
   defp put_history(%{history: history}) do
     Process.put(:iex_history, history)
   end
@@ -406,12 +419,7 @@ defmodule IEx.Evaluator do
 
         _ ->
           banner = Exception.format_banner(kind, blamed, stacktrace)
-
-          if String.contains?(banner, IO.ANSI.reset()) do
-            [banner]
-          else
-            [IEx.color(:eval_error, banner)]
-          end
+          [IEx.color(:eval_error, banner)]
       end
 
     stackdata = Exception.format_stacktrace(prune_stacktrace(stacktrace))
