@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule IEx.Server do
   @moduledoc """
   The IEx.Server.
@@ -10,8 +14,16 @@ defmodule IEx.Server do
 
   """
 
+  @type run_opts :: [
+          prefix: String.t(),
+          env: Macro.Env.t(),
+          binding: keyword(),
+          on_eof: :stop_evaluator | :halt,
+          register: boolean()
+        ]
+
   @doc false
-  defstruct parser_state: "",
+  defstruct parser_state: [],
             counter: 1,
             prefix: "iex",
             on_eof: :stop_evaluator,
@@ -31,7 +43,7 @@ defmodule IEx.Server do
 
   """
   @doc since: "1.8.0"
-  @spec run(keyword) :: :ok
+  @spec run(run_opts) :: :ok
   def run(opts) when is_list(opts) do
     if Keyword.get(opts, :register, true) do
       IEx.Broker.register(self())
@@ -88,7 +100,7 @@ defmodule IEx.Server do
     )
 
     evaluator = start_evaluator(state.counter, Keyword.merge(state.evaluator_options, opts))
-    loop(state, :ok, evaluator, Process.monitor(evaluator), input)
+    loop(state, evaluator, Process.monitor(evaluator), input)
   end
 
   # Starts an evaluator using the provided options.
@@ -117,18 +129,19 @@ defmodule IEx.Server do
     run_without_registration(state, opts, input)
   end
 
-  defp loop(state, status, evaluator, evaluator_ref, input) do
-    :io.setopts(expand_fun: state.expand_fun)
-    input = input || io_get(prompt(status, state.prefix, state.counter))
+  defp loop(state, evaluator, evaluator_ref, input) do
+    %{counter: counter, expand_fun: expand_fun, prefix: prefix, parser_state: parser} = state
+    :io.setopts(expand_fun: expand_fun)
+    input = input || io_get(prompt(prefix, counter), counter, parser)
     wait_input(state, evaluator, evaluator_ref, input)
   end
 
   defp wait_input(state, evaluator, evaluator_ref, input) do
     receive do
-      {:io_reply, ^input, code} when is_binary(code) ->
+      {:io_reply, ^input, {:ok, code, parser_state}} ->
         :io.setopts(expand_fun: fn _ -> {:yes, [], []} end)
-        send(evaluator, {:eval, self(), code, state.counter, state.parser_state})
-        wait_eval(state, evaluator, evaluator_ref)
+        send(evaluator, {:eval, self(), code, state.counter})
+        wait_eval(%{state | parser_state: parser_state}, evaluator, evaluator_ref)
 
       {:io_reply, ^input, :eof} ->
         case state.on_eof do
@@ -136,15 +149,31 @@ defmodule IEx.Server do
           :stop_evaluator -> stop_evaluator(evaluator, evaluator_ref)
         end
 
+      {:io_reply, ^input, {:error, kind, error, stacktrace}} ->
+        send(evaluator, {:reader_errored, self()})
+
+        banner = Exception.format_banner(kind, error, stacktrace)
+
+        banner =
+          if String.contains?(banner, IO.ANSI.reset()) do
+            banner
+          else
+            IEx.color(:eval_error, banner)
+          end
+
+        stackdata = Exception.format_stacktrace(stacktrace)
+        IO.write(:stdio, [banner, ?\n, IEx.color(:stack_info, stackdata)])
+        loop(%{state | parser_state: []}, evaluator, evaluator_ref, nil)
+
       # Triggered by pressing "i" as the job control switch
       {:io_reply, ^input, {:error, :interrupted}} ->
         io_error("** (EXIT) interrupted")
-        loop(%{state | parser_state: ""}, :ok, evaluator, evaluator_ref, nil)
+        loop(%{state | parser_state: []}, evaluator, evaluator_ref, nil)
 
       # Unknown IO message
       {:io_reply, ^input, msg} ->
         io_error("** (EXIT) unknown IO message: #{inspect(msg)}")
-        loop(%{state | parser_state: ""}, :ok, evaluator, evaluator_ref, nil)
+        loop(%{state | parser_state: []}, evaluator, evaluator_ref, nil)
 
       # Triggered when IO dies while waiting for input
       {:DOWN, ^input, _, _, _} ->
@@ -159,10 +188,10 @@ defmodule IEx.Server do
 
   defp wait_eval(state, evaluator, evaluator_ref) do
     receive do
-      {:evaled, ^evaluator, status, parser_state} ->
+      {:evaled, ^evaluator, status} ->
         counter = if(status == :ok, do: state.counter + 1, else: state.counter)
-        state = %{state | counter: counter, parser_state: parser_state}
-        loop(state, status, evaluator, evaluator_ref, nil)
+        state = %{state | counter: counter}
+        loop(state, evaluator, evaluator_ref, nil)
 
       msg ->
         handle_common(msg, state, evaluator, evaluator_ref, nil, fn state ->
@@ -199,7 +228,7 @@ defmodule IEx.Server do
         if take_over?(take_pid, take_ref, state.counter + 1, true) do
           # Since we are in process, also bump the counter
           state = reset_state(bump_counter(state))
-          loop(state, :ok, evaluator, evaluator_ref, input)
+          loop(state, evaluator, evaluator_ref, input)
         else
           callback.(state)
         end
@@ -266,16 +295,14 @@ defmodule IEx.Server do
          input,
          _callback
        ) do
-    if abnormal_reason?(reason) do
-      try do
-        io_error(
-          "** (EXIT from #{inspect(evaluator)}) shell process exited with reason: " <>
-            Exception.format_exit(reason)
-        )
-      catch
-        type, detail ->
-          io_error("** (IEx.Error) #{type} when printing EXIT message: #{inspect(detail)}")
-      end
+    try do
+      io_error(
+        "** (EXIT from #{inspect(evaluator)}) shell process exited with reason: " <>
+          Exception.format_exit(reason)
+      )
+    catch
+      type, detail ->
+        io_error("** (IEx.Error) #{type} when printing EXIT message: #{inspect(detail)}")
     end
 
     rerun(state, [], evaluator, evaluator_ref, input)
@@ -285,11 +312,6 @@ defmodule IEx.Server do
     callback.(state)
   end
 
-  defp abnormal_reason?(:normal), do: false
-  defp abnormal_reason?(:shutdown), do: false
-  defp abnormal_reason?({:shutdown, _}), do: false
-  defp abnormal_reason?(_), do: true
-
   defp take_over?(take_pid, take_ref, take_location, take_whereami, take_opts, counter) do
     evaluator = take_opts[:evaluator] || self()
     message = "Request to pry #{inspect(evaluator)} at #{take_location}#{take_whereami}"
@@ -338,18 +360,18 @@ defmodule IEx.Server do
       prefix: prefix,
       on_eof: on_eof,
       expand_fun: expand_fun,
-      evaluator_options: Keyword.take(opts, [:dot_iex_path])
+      evaluator_options: Keyword.take(opts, [:dot_iex])
     }
   end
 
   # For the state, reset only reset the parser state.
   # The counter will continue going up as the input process is shared.
-  # The opts can also set "dot_iex_path" and the "evaluator" itself,
+  # The opts can also set "dot_iex" and the "evaluator" itself,
   # but those are not stored: they are temporary to whatever is rerunning.
   # Once the rerunning session restarts, we keep the same evaluator_options
   # and rollback to a new evaluator.
   defp reset_state(state) do
-    %{state | parser_state: ""}
+    %{state | parser_state: []}
   end
 
   defp bump_counter(state) do
@@ -358,28 +380,41 @@ defmodule IEx.Server do
 
   ## IO
 
-  defp io_get(prompt) do
+  defp io_get(prompt, counter, parser_state) do
     gl = Process.group_leader()
     ref = Process.monitor(gl)
-    command = {:get_until, :unicode, prompt, __MODULE__, :__parse__, []}
+    command = {:get_until, :unicode, prompt, __MODULE__, :__parse__, [{counter, parser_state}]}
     send(gl, {:io_request, self(), ref, command})
     ref
   end
 
   @doc false
-  def __parse__([], :eof), do: {:done, :eof, []}
-  def __parse__([], chars), do: {:done, List.to_string(chars), []}
+  def __parse__(_, :eof, _parser_state), do: {:done, :eof, []}
 
-  defp prompt(status, prefix, counter) do
-    {mode, prefix} =
-      if Node.alive?() do
-        {prompt_mode(status, :alive), default_prefix(status, prefix)}
-      else
-        {prompt_mode(status, :default), default_prefix(status, prefix)}
-      end
+  def __parse__([], chars, {counter, parser_state} = to_be_unused) do
+    __parse__({counter, parser_state, IEx.Config.parser()}, chars, to_be_unused)
+  end
 
+  def __parse__({counter, parser_state, mfa}, chars, _unused) do
+    {parser_module, parser_fun, args} = mfa
+    args = [chars, [line: counter, file: "iex"], parser_state | args]
+
+    case apply(parser_module, parser_fun, args) do
+      {:ok, forms, parser_state} -> {:done, {:ok, forms, parser_state}, []}
+      {:incomplete, parser_state} -> {:more, {counter, parser_state, mfa}}
+    end
+  catch
+    kind, error ->
+      {:done, {:error, kind, error, __STACKTRACE__}, []}
+  end
+
+  defp prompt(prefix, counter) do
     prompt =
-      apply(IEx.Config, mode, [])
+      if Node.alive?() do
+        IEx.Config.alive_prompt()
+      else
+        IEx.Config.default_prompt()
+      end
       |> String.replace("%counter", to_string(counter))
       |> String.replace("%prefix", to_string(prefix))
       |> String.replace("%node", to_string(node()))
@@ -387,14 +422,6 @@ defmodule IEx.Server do
     [prompt, " "]
   end
 
-  defp default_prefix(:incomplete, _prefix), do: "..."
-  defp default_prefix(_ok_or_error, prefix), do: prefix
-
-  defp prompt_mode(:incomplete, :default), do: :continuation_prompt
-  defp prompt_mode(:incomplete, :alive), do: :alive_continuation_prompt
-  defp prompt_mode(_ok_or_error, :default), do: :default_prompt
-  defp prompt_mode(_ok_or_error, :alive), do: :alive_prompt
-
   defp io_error(result) do
     IO.puts(:stdio, IEx.color(:eval_error, result))
   end
