@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule IEx.MixProject do
   use Mix.Project
 
@@ -18,7 +22,8 @@ defmodule IEx.MixProject do
         inspect: [pretty: true],
         history_size: 20,
         default_prompt: "%prefix(%counter)>",
-        alive_prompt: "%prefix(%node)%counter>"
+        alive_prompt: "%prefix(%node)%counter>",
+        auto_reload: false
       ]
     ]
   end
