@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule IEx.InteractionTest do
@@ -5,7 +9,7 @@ defmodule IEx.InteractionTest do
 
   test "whole output" do
     assert capture_io("IO.puts \"Hello world\"", fn ->
-             IEx.Server.run(dot_iex_path: "")
+             IEx.Server.run(dot_iex: "")
            end) =~
              "Interactive Elixir (#{System.version()}) - press Ctrl+C to exit (type h() ENTER for help)" <>
                "\niex(1)> Hello world\n:ok\niex(2)>"
@@ -93,7 +97,8 @@ defmodule IEx.InteractionTest do
     output = capture_iex(input)
 
     assert output =~ "** (TokenMissingError) token missing on iex:1:"
-    assert output =~ "error: incomplete expression\n"
+    assert output =~ "error:"
+    assert output =~ "incomplete expression\n"
     assert output =~ "iex:1"
   end
 
@@ -140,11 +145,6 @@ defmodule IEx.InteractionTest do
     assert capture_iex("1\n", opts, [], true) == "prompt(1)> 1\nprompt(2)>"
   end
 
-  test "continuation prompt" do
-    opts = [default_prompt: "%prefix(%counter)>", continuation_prompt: "%prefix(%counter)>>>"]
-    assert capture_iex("[\n1\n]\n", opts, [], true) == "iex(1)> ...(1)>>> ...(1)>>> [1]\niex(2)>"
-  end
-
   if IO.ANSI.enabled?() do
     test "color" do
       opts = [colors: [enabled: true, eval_result: [:red]]]
@@ -197,6 +197,16 @@ defmodule IEx.InteractionTest do
              ~r"\*\* \(EXIT from #PID<\d+\.\d+\.\d+>\) shell process exited with reason: {:bye, \[:world\]}"
   end
 
+  test "receive normal exits" do
+    assert capture_iex("spawn_link(fn -> exit(:normal) end); Process.sleep(1000)") =~ ":ok"
+
+    assert capture_iex("spawn_link(fn -> exit(:shutdown) end); Process.sleep(1000)") =~
+             ~r"\*\* \(EXIT from #PID<\d+\.\d+\.\d+>\) shell process exited with reason: shutdown"
+
+    assert capture_iex("spawn_link(fn -> exit({:shutdown, :bye}) end); Process.sleep(1000)") =~
+             ~r"\*\* \(EXIT from #PID<\d+\.\d+\.\d+>\) shell process exited with reason: shutdown: :bye"
+  end
+
   test "receive exit from exception" do
     # use exit/1 to fake an error so that an error message
     # is not sent to the error logger.
@@ -217,11 +227,30 @@ defmodule IEx.InteractionTest do
   end
 
   test "blames function clause error" do
+    import PathHelpers
+
+    write_beam(
+      defmodule ExampleModule do
+        def fun(:valid), do: :ok
+      end
+    )
+
+    content = capture_iex("IEx.InteractionTest.ExampleModule.fun(:invalid)")
+
+    assert content =~
+             "** (FunctionClauseError) no function clause matching in IEx.InteractionTest.ExampleModule.fun/1"
+
+    assert content =~
+             "The following arguments were given to IEx.InteractionTest.ExampleModule.fun/1"
+
+    assert content =~ ":invalid"
+    assert content =~ "def fun(-:valid-)"
+
+    assert content =~
+             ~r"test/iex/interaction_test.exs:\d+: IEx\.InteractionTest\.ExampleModule\.fun/1"
+
     content = capture_iex("Access.fetch(:foo, :bar)")
-    assert content =~ "** (FunctionClauseError) no function clause matching in Access.fetch/2"
-    assert content =~ "The following arguments were given to Access.fetch/2"
-    assert content =~ ":foo"
-    assert content =~ "def fetch(-%module{} = container-, key)"
+
     assert content =~ ~r"\(elixir #{System.version()}\) lib/access\.ex:\d+: Access\.fetch/2"
   end
 
@@ -232,7 +261,7 @@ defmodule IEx.InteractionTest do
       end
     end
 
-    assert capture_iex("foo", parser: {EchoParser, :parse, []}) == "\"foo\""
+    assert capture_iex("foo", parser: {EchoParser, :parse, []}) == "~c\"foo\""
   after
     IEx.configure(parser: {IEx.Evaluator, :parse, []})
   end
@@ -256,7 +285,7 @@ defmodule IEx.InteractionTest do
         my_variable = 42
         """)
 
-      assert capture_iex("{my_fun_single(), my_variable}", [], dot_iex_path: path) ==
+      assert capture_iex("{my_fun_single(), my_variable}", [], dot_iex: path) ==
                "{:single, 42}"
     end
 
@@ -274,7 +303,22 @@ defmodule IEx.InteractionTest do
         write_dot_iex!(tmp_dir, "dot-iex", "import_file \"#{tmp_dir}/dot-iex-1\"\nmy_variable=13")
 
       input = "nested_var\nmy_variable\nmy_fun_nested()"
-      assert capture_iex(input, [], dot_iex_path: path) == "42\n13\n:nested"
+      assert capture_iex(input, [], dot_iex: path) == "42\n13\n:nested"
+    end
+
+    @tag :tmp_dir
+    test "configured .iex", %{tmp_dir: tmp_dir} do
+      path =
+        write_dot_iex!(tmp_dir, "configured-dot-iex", """
+        defmodule ConfiguredDotIEx do
+          def my_fun_single, do: :single
+        end
+        import ConfiguredDotIEx
+        my_variable = 42
+        """)
+
+      assert capture_iex("{my_fun_single(), my_variable}", [dot_iex: path], dot_iex: nil) ==
+               "{:single, 42}"
     end
   end
 
@@ -283,4 +327,41 @@ defmodule IEx.InteractionTest do
     File.write!(path, contents)
     path
   end
+
+  describe "pipeline error safety" do
+    test "syntax error in pipeline prevents subsequent operations" do
+      input = """
+      1, 2, 3]
+      |> Enum.map(&(&1 * 2))
+      """
+
+      output = capture_iex(input)
+      assert output =~ "skipping evaluation of expression because pipeline has failed"
+    end
+
+    test "runtime error in first pipeline step prevents subsequent operations" do
+      input = """
+      :not_a_list
+      |> Enum.map(&(&1 * 2))
+      |> Enum.sum()
+      """
+
+      output = capture_iex(input)
+      assert output =~ "skipping evaluation of expression because pipeline has failed"
+    end
+
+    test "error state is cleared after successful evaluation" do
+      input = """
+      :not_a_list
+      |> Enum.map(&(&1 * 2))
+      |> Enum.sum()
+      [1, 2, 3]
+      |> Enum.sum()
+      """
+
+      output = capture_iex(input)
+      assert output =~ "skipping evaluation of expression because pipeline has failed"
+      assert String.ends_with?(output, "6")
+    end
+  end
 end
