@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Logger do
   @moduledoc ~S"""
   A logger for Elixir applications.
@@ -8,7 +12,7 @@ defmodule Logger do
 
   Overall, you will find that `Logger`:
 
-    * Provides all 7 syslog levels
+    * Provides all 8 syslog levels
       (although debug, info, warning, and error are the most commonly used).
 
     * Supports both message-based and structural logging.
@@ -19,9 +23,12 @@ defmodule Logger do
     * Formats and truncates messages on the client
       to avoid clogging `Logger` handlers.
 
-    * Alternates between sync and async modes to remain
-      performant when required but also apply back-pressure
-      when under stress.
+    * Provides multiple forms of [overload protection](https://www.erlang.org/doc/apps/kernel/logger_chapter.html#protecting-the-handler-from-overload):
+      * keeps track of its message queue and switches to sync mode to apply
+        back pressure or even drop messages
+      * limits the number of logs emitted defaulting to 500 per second
+      * optionally allows to terminate and restart it if the message queue length
+        or memory thresholds are exceeded
 
     * Allows overriding the logging level for a specific module,
       application or process.
@@ -102,8 +109,8 @@ defmodule Logger do
 
   Furthermore, metadata can be set per process with `Logger.metadata/1`.
 
-  Some metadata, however, is always added automatically by Logger
-  whenever possible. Those are:
+  Some metadata, however, may be added automatically by Logger whenever
+  possible. Those are:
 
     * `:application` - the current application
 
@@ -137,13 +144,13 @@ defmodule Logger do
   There are two special metadata keys, `:module` and `:function`, which
   extract the relevant bits from `:mfa`.
 
-  Note that all metadata is optional and may not always be available.
-  The `:mfa`, `:file`, `:line`, and similar metadata are automatically
-  included when using `Logger` macros. `Logger.bare_log/3` does not include
-  any metadata beyond the `:pid` by default. Other metadata, such as
+  The metadata keys above may not always be available. The `:mfa`, `:file`,
+  `:line`, and similar metadata are automatically included when using `Logger`
+  macros, but not when using `Logger.bare_log/3`. Other metadata, such as
   `:crash_reason`, `:initial_call`, and `:registered_name` are available
   only inside behaviours such as GenServer, Supervisor, and others.
 
+  It is also possible to pass metadata on a particular Logger invocation.
   For example, you might wish to include a custom `:error_code` metadata in
   your logs:
 
@@ -196,8 +203,10 @@ defmodule Logger do
       if `:handle_otp_reports` is true.
 
     * `:metadata` - key-value pairs of global primary metadata to be included
-      in all log messages. Defaults to `[]`. This can be overridden at the process
-      level with `metadata/1` or each on log call as desired.
+      in all log messages. Defaults to `[]`. The default formatter writes to
+      standard out and therefore cannot print all metadata. See
+      [`Logger.Formatter`'s documentation](`m:Logger.Formatter#module-metadata`)
+      for more information.
 
   For example, to configure `Logger` to redirect all Erlang messages using a
   `config/config.exs` file:
@@ -226,13 +235,30 @@ defmodule Logger do
           compress_on_rotate: true
         ]
 
-  See [`:logger_std_h`](`:logger_std_h`) for all relevant configuration,
-  including overload protection. Or set `:default_handler` to false to
-  disable the default logging altogether:
+  You can find a complete reference on all handler options
+  [on Erlang/OTP docs](`t::logger_handler.config/0`). Here is Elixir's
+  default configuration for the default handler:
+
+      [
+        # Do not log messages from other nodes
+        filters: [{&:logger_filters.remote_gl/2, :stop}],
+        filter_default: :log,
+        formatter: &Logger.default_formatter/0,
+        level: :all,
+        module: :logger_std_h
+      ]
+
+  The `:config` customizes a specific handler module. The default handler
+  is [`:logger_std_h`](`:logger_std_h`), which logs to standard IO, and you
+  can find all relevant configuration in its module documentation, including
+  information overload protection.
+
+  You may also set `:default_handler` to false to disable the default logging
+  altogether:
 
       config :logger, :default_handler, false
 
-  How to add new handlers is covered in later sections.
+  How to add more handlers besides the default one is covered in later sections.
 
   > #### Keywords or maps {: .tip}
   >
@@ -315,7 +341,8 @@ defmodule Logger do
     * `:translator_inspect_opts` - when translating OTP reports and
       errors, the last message and state must be inspected in the
       error reports. This configuration allow developers to change
-      how much and how the data should be inspected.
+      how much and how the data should be inspected. See `Kernel.inspect/2`
+      for more information on the available options.
 
   For example, to configure the `:level` options in a `config/config.exs`
   file:
@@ -330,10 +357,14 @@ defmodule Logger do
   ## Erlang/OTP handlers
 
   Handlers represent the ability to integrate into the logging system to
-  handle each logged message/event. Elixir's Logger automatically sets a
-  default handler based on Erlang's `:logger_std_h`, which you can configure
-  using the `:default_handler` boot configuration outlined above. You may
-  also attach additional handlers when you boot your application.
+  handle each logged message/event.
+
+  ### Built-in handlers
+
+  Elixir's Logger automatically sets a default handler based on Erlang's
+  `:logger_std_h`, which you can configure using the `:default_handler` boot
+  configuration outlined above. You may also attach additional handlers
+  when you boot your application.
 
   To do so, you must list a series of handlers under the `:logger` key
   of your application configuration. For example, to setup an additional
@@ -355,24 +386,47 @@ defmodule Logger do
 
   Each handler has the shape `{:handler, name, handler_module, config_map}`.
   Once defined, a handler can be explicitly attached in your
-  `c:Application.start/2` callback with `add_handlers/1`:
+  `c:Application.start/2` callback, typically in `lib/my_app/application.ex`
+  with `Logger.add_handlers/1`:
 
       Logger.add_handlers(:my_app)
 
   You can also add, remove, and update handlers at runtime with the help
   of the Erlang's [`:logger`](`:logger`) module.
 
+  ### Custom handlers
+
   You may also develop your own handlers. Handlers run in the same
   process as the process logging the message/event. This gives developers
   flexibility but they should avoid performing any long running action in
   such handlers, as it may slow down the action being executed considerably.
-  At the moment, there is no built-in overload protection for Erlang handlers,
-  so it is your responsibility to implement it.
 
-  Alternatively, you can use the
-  [`:logger_backends`](https://github.com/elixir-lang/logger_backends) project.
-  It sets up a log handler with overload protection and allows incoming events
-  to be dispatched to multiple backends.
+  You must implement the [`:logger_handler`](`:logger_handler`) behaviour
+  to define custom handlers, which has only one required callback:
+
+      defmodule MyApp.CustomHandler do
+        @behaviour :logger_handler
+
+        def log(event, _config) do
+          IO.inspect(event)
+        end
+      end
+
+  Then you must define the handler in your configuration file:
+
+      config :my_app, :logger, [
+        {:handler, :my_handler, MyApp.CustomHandler, _config = %{}}
+      ]
+
+  And attach it on your application start using `Logger.add_handlers/1`,
+  as in the previous section.
+
+  Note there is no built-in overload protection for Erlang handlers,
+  so it is your responsibility to implement it if necessary. One alternative is
+  to use the [`:logger_backends`](https://github.com/elixir-lang/logger_backends)
+  project, which sets up a log handler with overload protection and allows
+  incoming events to be dispatched to multiple backends, although one must be
+  careful for backends to not become a bottleneck during logging.
 
   ### Filtering
 
@@ -382,8 +436,9 @@ defmodule Logger do
       defmodule LogFilter do
         def filter(log_event, _opts) do
           case log_event do
-            %{msg: msg} when is_binary(msg) ->
-              if msg =~ "password" do
+            %{msg: {:string, msg}} ->
+              # msg may be a charlist or a binary
+              if to_string(msg) =~ "password" do
                 :stop
               else
                 :ignore
@@ -395,8 +450,12 @@ defmodule Logger do
         end
       end
 
-  Then, when you start your application, such as in the
-  `c:Application.start/2` callback:
+  It may return `:log` (to log the message), `:stop` (to not log the
+  message), or `:ignore` (to ignore the filter).
+
+  Then you can attach the filter, either as a primary filter (which
+  applies to all handlers), or to a specific handler, when you start
+  your application, such as in the `c:Application.start/2` callback:
 
       :logger.add_primary_filter(:word_filter, {&LogFilter.filter/2, []})
 
@@ -405,7 +464,7 @@ defmodule Logger do
   Prior to Elixir v1.15, custom logging could be achieved with Logger
   backends. The main API for writing Logger backends have been moved to
   the [`:logger_backends`](https://github.com/elixir-lang/logger_backends)
-  project. However, the backends API are still part of Elixir for backwards
+  project. However, the backends API is still part of Elixir for backwards
   compatibility.
 
   Important remarks:
@@ -440,7 +499,7 @@ defmodule Logger do
       in future releases
 
     * `Logger.Backends` only receive `:debug`, `:info`, `:warning`, and `:error`
-      messages. `:notice` maps to `:info`. `:warn` amps to `:warnings`.
+      messages. `:notice` maps to `:info`. `:warn` maps to `:warnings`.
       All others map to `:error`
   """
 
@@ -449,6 +508,30 @@ defmodule Logger do
   @type report :: map() | keyword()
   @type message :: :unicode.chardata() | String.Chars.t() | report()
   @type metadata :: keyword()
+
+  @type configure_opts :: [
+          level: level(),
+          translator_inspect_opts: Inspect.Opts.t(),
+          sync_threshold: non_neg_integer(),
+          discard_threshold: non_neg_integer(),
+          truncate: non_neg_integer() | :infinity,
+          utc_log: boolean()
+        ]
+
+  @type formatter_opts :: [
+          colors: [
+            enabled: boolean(),
+            debug: atom(),
+            info: atom(),
+            warning: atom(),
+            error: atom()
+          ],
+          format: String.t() | {module(), atom()},
+          metadata: :all | [atom()],
+          truncate: pos_integer() | :infinity,
+          utc_log: boolean()
+        ]
+
   @new_erlang_levels [:emergency, :alert, :critical, :warning, :notice]
   @levels [:error, :info, :debug] ++ @new_erlang_levels
   @metadata :logger_level
@@ -494,7 +577,7 @@ defmodule Logger do
   instead.
   """
   @doc since: "1.15.0"
-  @spec default_formatter(keyword) :: {module, :logger.formatter_config()}
+  @spec default_formatter(formatter_opts) :: {module, :logger.formatter_config()}
   def default_formatter(overrides \\ []) when is_list(overrides) do
     Application.get_env(:logger, :default_formatter, [])
     |> Keyword.merge(overrides)
@@ -502,25 +585,24 @@ defmodule Logger do
   end
 
   @doc """
-  Alters the current process metadata according to the given keyword list.
+  Alters the current process metadata according to the given enumerable.
 
-  This function will merge the given keyword list into the existing metadata,
+  This function will merge the given enumerable into the existing metadata,
   with the exception of setting a key to `nil`, which will remove that key
   from the metadata.
 
-  Note not all keys can be set as metadata. The metadata automatically added
-  by Logger, as declared in the module documentation, will always override
-  custom one.
+  Note some metadata keys are reserved and cannot be overridden. See
+  [the module documentation](#module-metadata) for more information.
   """
-  @spec metadata(metadata) :: :ok
-  def metadata(keyword) do
+  @spec metadata(Enumerable.t({atom(), term()})) :: :ok
+  def metadata(enumerable) do
     case :logger.get_process_metadata() do
       :undefined ->
-        reset_metadata(keyword)
+        reset_metadata(enumerable)
 
       map when is_map(map) ->
         metadata =
-          Enum.reduce(keyword, map, fn
+          Enum.reduce(enumerable, map, fn
             {k, nil}, acc -> Map.delete(acc, k)
             {k, v}, acc -> Map.put(acc, k, v)
           end)
@@ -544,11 +626,11 @@ defmodule Logger do
   end
 
   @doc """
-  Resets the current process metadata to the given keyword list.
+  Resets the current process metadata to the given enumerable.
   """
-  @spec reset_metadata(metadata) :: :ok
-  def reset_metadata(keyword \\ []) do
-    :ok = :logger.set_process_metadata(filter_out_nils(keyword))
+  @spec reset_metadata(Enumerable.t({atom(), term()})) :: :ok
+  def reset_metadata(enumerable \\ []) do
+    :ok = :logger.set_process_metadata(filter_out_nils(enumerable))
   end
 
   defp filter_out_nils(keyword) do
@@ -564,7 +646,7 @@ defmodule Logger do
 
       delete_process_level(pid)
   """
-  # TODO: Deprecate me on v1.18
+  # TODO: Deprecate me on v1.20
   @doc deprecated: "Use Logger.delete_process_level(pid) instead"
   @spec enable(pid) :: :ok
   def enable(pid) when pid == self() do
@@ -580,7 +662,7 @@ defmodule Logger do
 
       put_process_level(pid, :none)
   """
-  # TODO: Deprecate me on v1.18
+  # TODO: Deprecate me on v1.20
   @doc deprecated: "Use Logger.put_process_level(pid, :none) instead"
   @spec disable(pid) :: :ok
   def disable(pid) when pid == self() do
@@ -592,7 +674,7 @@ defmodule Logger do
 
   Currently the only accepted PID is `self()`.
   """
-  # TODO: Deprecate me on v1.18
+  # TODO: Deprecate me on v1.20
   @doc deprecated: "Use Logger.get_process_level(pid) instead"
   @spec enabled?(pid) :: boolean
   def enabled?(pid) when pid == self() do
@@ -655,7 +737,7 @@ defmodule Logger do
     :translator_inspect_opts
   ]
   @backend_options [:sync_threshold, :discard_threshold, :truncate, :utc_log]
-  @spec configure(keyword) :: :ok
+  @spec configure(configure_opts) :: :ok
   def configure(options) do
     for {k, v} <- options do
       cond do
@@ -665,10 +747,23 @@ defmodule Logger do
       end
     end
 
-    # TODO: Deprecate passing backend options to configure on Elixir v1.19
     case Keyword.take(options, @backend_options) do
-      [] -> :ok
-      backend_options -> Logger.Backends.Internal.configure(backend_options)
+      [] ->
+        :ok
+
+      backend_options ->
+        for {key, _} <- backend_options do
+          IO.warn_once(
+            {__MODULE__, :configure, key},
+            fn ->
+              "setting #{inspect(key)} in Logger.configure/2 is deprecated, " <>
+                "add the :logger_backends dependency and configure it instead"
+            end,
+            3
+          )
+        end
+
+        Logger.Backends.Internal.configure(backend_options)
     end
 
     :ok
@@ -677,8 +772,11 @@ defmodule Logger do
   @doc """
   Flushes the logger.
 
-  This guarantees all log handlers are flushed. This is useful
-  for testing and it should not be called in production code.
+  This guarantees all logger handlers flush to disk or storage.
+  This is useful for testing but it should be avoided in production,
+  as it could force logger handlers to drop whatever they are doing
+  and flush, even if continuing to buffer would be the most performant
+  option.
   """
   @spec flush :: :ok
   def flush do
@@ -775,8 +873,9 @@ defmodule Logger do
 
   Currently the only accepted PID is `self()`.
 
-  This will take priority over the primary level set, so it can be
-  used to increase or decrease verbosity of some parts of the running system.
+  Different from `put_module_level/2`, the process level doesn't take priority
+  over the global level, but instead works alongside it. Effectively, the higher
+  logger level is used.
   """
   @doc since: "1.15.0"
   @spec put_process_level(pid(), level() | :all | :none) :: :ok
@@ -828,7 +927,7 @@ defmodule Logger do
   @doc """
   Adds a new backend.
   """
-  # TODO: Deprecate this on Elixir v1.19
+  # TODO: Deprecate this on Elixir v1.20
   @doc deprecated: "Use LoggerBackends.add/2 from :logger_backends dependency"
   def add_backend(backend, opts \\ []) do
     Logger.Backends.Internal.add(backend, opts)
@@ -837,7 +936,7 @@ defmodule Logger do
   @doc """
   Removes a backend.
   """
-  # TODO: Deprecate this on Elixir v1.19
+  # TODO: Deprecate this on Elixir v1.20
   @doc deprecated: "Use LoggerBackends.remove/2 from :logger_backends dependency"
   def remove_backend(backend, opts \\ []) do
     Logger.Backends.Internal.remove(backend, opts)
@@ -846,7 +945,7 @@ defmodule Logger do
   @doc """
   Configures the given backend.
   """
-  # TODO: Deprecate this on Elixir v1.19
+  # TODO: Deprecate this on Elixir v1.20
   @doc deprecated: "Use LoggerBackends.configure/2 from :logger_backends dependency"
   def configure_backend(:console, options) when is_list(options) do
     options = Keyword.merge(Application.get_env(:logger, :console, []), options)
@@ -899,19 +998,22 @@ defmodule Logger do
   anonymous functions to `bare_log/3` and they will only be evaluated
   if there is something to be logged.
   """
-  @spec bare_log(level, message | (-> message | {message, keyword}), keyword) :: :ok
+  @spec bare_log(level, message | (-> message | {message, keyword}), metadata) :: :ok
   def bare_log(level, message_or_fun, metadata \\ []) do
-    case __should_log__(level, nil) do
-      nil -> :ok
-      level -> __do_log__(level, message_or_fun, %{}, Map.new(metadata))
+    level = elixir_level_to_erlang_level(level)
+
+    if :logger_config.allow(level) do
+      __do_log__(level, message_or_fun, %{}, Map.new(metadata))
     end
+
+    :ok
   end
 
   @doc false
   def __should_log__(level, module) do
     level = elixir_level_to_erlang_level(level)
 
-    if :logger.allow(level, module) do
+    if :logger_config.allow(level, module) do
       level
     end
   end
