@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Logger.TranslatorTest do
   use Logger.Case
 
@@ -136,23 +140,27 @@ defmodule Logger.TranslatorTest do
   end
 
   setup_all do
-    sasl_reports? = Application.get_env(:logger, :handle_sasl_reports, false)
-    Application.put_env(:logger, :handle_sasl_reports, true)
+    config = Application.get_all_env(:logger)
 
-    # Shutdown the application
-    Logger.App.stop()
+    config
+    |> Keyword.put(:handle_sasl_reports, true)
+    |> put_logger_config()
 
-    # And start it without warnings
-    Application.put_env(:logger, :level, :error)
-    Application.start(:logger)
-    Application.delete_env(:logger, :level)
-    Logger.configure(level: :debug)
+    on_exit(fn -> restore_logger_config(config) end)
+  end
 
-    on_exit(fn ->
-      Application.put_env(:logger, :handle_sasl_reports, sasl_reports?)
-      Logger.App.stop()
-      Application.start(:logger)
-    end)
+  setup context do
+    test_overrides = Map.get(context, :logger_config, [])
+    existing_config = Application.get_all_env(:logger)
+
+    case Keyword.merge(existing_config, test_overrides) do
+      ^existing_config ->
+        :ok
+
+      new_config ->
+        put_logger_config(new_config)
+        on_exit(fn -> restore_logger_config(existing_config) end)
+    end
   end
 
   setup do
@@ -175,8 +183,8 @@ defmodule Logger.TranslatorTest do
            ** (RuntimeError) oops
            """
 
-    assert_receive {:event, {:string, ["GenServer " <> _ | _]}, gen_server_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_server, :terminate}}}, gen_server_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert {%RuntimeError{message: "oops"}, [_ | _]} = gen_server_metadata[:crash_reason]
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
@@ -192,8 +200,8 @@ defmodule Logger.TranslatorTest do
     {:ok, pid} = GenServer.start(MyGenServer, :ok, name: config.test)
     capture_log(:info, fn -> catch_exit(GenServer.call(pid, :error)) end)
 
-    assert_receive {:event, {:string, ["GenServer " <> _ | _]}, gen_server_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_server, :terminate}}}, gen_server_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert gen_server_metadata[:registered_name] == config.test
     assert process_metadata[:registered_name] == config.test
@@ -203,8 +211,8 @@ defmodule Logger.TranslatorTest do
     {:ok, pid} = GenServer.start(MyGenServer, :ok, name: {:global, config.test})
     capture_log(:info, fn -> catch_exit(GenServer.call(pid, :error)) end)
 
-    assert_receive {:event, {:string, ["GenServer " <> _ | _]}, gen_server_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_server, :terminate}}}, gen_server_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert gen_server_metadata[:registered_name] == config.test
     refute Map.has_key?(process_metadata, :registered_name)
@@ -252,8 +260,8 @@ defmodule Logger.TranslatorTest do
            .*
            """s
 
-    assert_receive {:event, {:string, [["GenServer " <> _ | _] | _]}, gen_server_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_server, :terminate}}}, gen_server_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert {%RuntimeError{message: "oops"}, [_ | _]} = gen_server_metadata[:crash_reason]
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
@@ -322,8 +330,8 @@ defmodule Logger.TranslatorTest do
              assert_receive {:DOWN, ^mon, _, _, _}
            end) =~ "Client"
 
-    assert_receive {:event, {:string, [["GenServer " <> _ | _] | _]}, _gen_server_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, _process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_server, :terminate}}}, _gen_server_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, _process_metadata}
   end
 
   test "translates :gen_event crashes" do
@@ -338,7 +346,7 @@ defmodule Logger.TranslatorTest do
            ** (RuntimeError) oops
            """
 
-    assert_receive {:event, {:string, [":gen_event handler " <> _ | _]}, metadata}
+    assert_receive {:event, {:report, %{label: {:gen_event, :terminate}}}, metadata}
     assert {%RuntimeError{message: "oops"}, [_ | _]} = metadata[:crash_reason]
     refute Map.has_key?(metadata, :initial_call)
     refute Map.has_key?(metadata, :registered_name)
@@ -356,7 +364,7 @@ defmodule Logger.TranslatorTest do
            ** (RuntimeError) oops
            """
 
-    assert_receive {:event, {:string, [":gen_event handler " <> _ | _]}, metadata}
+    assert_receive {:event, {:report, %{label: {:gen_event, :terminate}}}, metadata}
     assert {%RuntimeError{message: "oops"}, [_ | _]} = metadata[:crash_reason]
     assert metadata[:registered_name] == config.test
   end
@@ -394,7 +402,7 @@ defmodule Logger.TranslatorTest do
            State: :ok
            """s
 
-    assert_receive {:event, {:string, [[":gen_event handler " <> _ | _] | _]}, metadata}
+    assert_receive {:event, {:report, %{label: {:gen_event, :terminate}}}, metadata}
     assert {%RuntimeError{message: "oops"}, [_ | _]} = metadata[:crash_reason]
   end
 
@@ -408,8 +416,8 @@ defmodule Logger.TranslatorTest do
            ** (RuntimeError) oops
            """
 
-    assert_receive {:event, {:string, [":gen_statem " <> _ | _]}, gen_statem_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_statem, :terminate}}}, gen_statem_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert {%RuntimeError{message: "oops"}, [_ | _]} = gen_statem_metadata[:crash_reason]
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
@@ -425,8 +433,8 @@ defmodule Logger.TranslatorTest do
     {:ok, pid} = :gen_statem.start({:local, config.test}, MyGenStatem, :ok, [])
     capture_log(:info, fn -> catch_exit(:gen_statem.call(pid, :error)) end)
 
-    assert_receive {:event, {:string, [":gen_statem " <> _ | _]}, gen_statem_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_statem, :terminate}}}, gen_statem_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert gen_statem_metadata[:registered_name] == config.test
     assert process_metadata[:registered_name] == config.test
@@ -436,8 +444,8 @@ defmodule Logger.TranslatorTest do
     {:ok, pid} = :gen_statem.start({:global, config.test}, MyGenStatem, :ok, [])
     capture_log(:info, fn -> catch_exit(:gen_statem.call(pid, :error)) end)
 
-    assert_receive {:event, {:string, [":gen_statem " <> _ | _]}, gen_statem_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_statem, :terminate}}}, gen_statem_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert gen_statem_metadata[:registered_name] == config.test
     refute Map.has_key?(process_metadata, :registered_name)
@@ -490,8 +498,8 @@ defmodule Logger.TranslatorTest do
            .*
            """s
 
-    assert_receive {:event, {:string, [[":gen_statem " <> _ | _] | _]}, gen_statem_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_statem, :terminate}}}, gen_statem_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert {%RuntimeError{message: "oops"}, [_ | _]} = gen_statem_metadata[:crash_reason]
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
@@ -566,8 +574,8 @@ defmodule Logger.TranslatorTest do
              assert_receive {:DOWN, ^mon, _, _, _}
            end) =~ "Client"
 
-    assert_receive {:event, {:string, [[":gen_statem " <> _ | _] | _]}, _gen_statem_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, _process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_statem, :terminate}}}, _gen_statem_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, _process_metadata}
   end
 
   test "translates :gen_statem crashes when callback_mode is :handle_event_function" do
@@ -585,8 +593,8 @@ defmodule Logger.TranslatorTest do
            Callback mode: .*, state_enter: false
            """s
 
-    assert_receive {:event, {:string, [[":gen_statem " <> _ | _] | _]}, gen_statem_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_statem, :terminate}}}, gen_statem_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert {%RuntimeError{message: "oops"}, [_ | _]} = gen_statem_metadata[:crash_reason]
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
@@ -614,7 +622,7 @@ defmodule Logger.TranslatorTest do
                Args: \[#PID<\d+\.\d+\.\d+>\]
            """s
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
+    assert_receive {:event, {:report, %{label: {Task.Supervisor, :terminating}}}, task_metadata}
     assert {%RuntimeError{message: "oops"}, [_ | _]} = task_metadata[:crash_reason]
     assert [parent] == task_metadata[:callers]
     refute Map.has_key?(task_metadata, :initial_call)
@@ -635,7 +643,7 @@ defmodule Logger.TranslatorTest do
                Args: \[\]
            """s
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
+    assert_receive {:event, {:report, %{label: {Task.Supervisor, :terminating}}}, task_metadata}
     assert [parent] == task_metadata[:callers]
     assert {%UndefinedFunctionError{function: :undef}, [_ | _]} = task_metadata[:crash_reason]
   end
@@ -655,7 +663,7 @@ defmodule Logger.TranslatorTest do
                Args: \[\]
            """s
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
+    assert_receive {:event, {:report, %{label: {Task.Supervisor, :terminating}}}, task_metadata}
     assert [parent] == task_metadata[:callers]
     assert {%UndefinedFunctionError{function: :undef}, [_ | _]} = task_metadata[:crash_reason]
   end
@@ -683,7 +691,7 @@ defmodule Logger.TranslatorTest do
                Args: \[%ErlangError{.*}\]
            """s
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
+    assert_receive {:event, {:report, %{label: {Task.Supervisor, :terminating}}}, task_metadata}
     assert [parent] == task_metadata[:callers]
     assert {%ErlangError{original: :foo}, [_ | _]} = task_metadata[:crash_reason]
   end
@@ -703,7 +711,7 @@ defmodule Logger.TranslatorTest do
                Args: \[:badarg\]
            """s
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
+    assert_receive {:event, {:report, %{label: {Task.Supervisor, :terminating}}}, task_metadata}
     assert [parent] == task_metadata[:callers]
     assert {%ArgumentError{message: "argument error"}, [_ | _]} = task_metadata[:crash_reason]
   end
@@ -723,7 +731,7 @@ defmodule Logger.TranslatorTest do
                Args: \[:abnormal\]
            """s
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
+    assert_receive {:event, {:report, %{label: {Task.Supervisor, :terminating}}}, task_metadata}
     assert [parent] == task_metadata[:callers]
     assert {:abnormal, [_ | _]} = task_metadata[:crash_reason]
   end
@@ -805,7 +813,7 @@ defmodule Logger.TranslatorTest do
            Ancestors: \[#PID<\d+\.\d+\.\d+>\]
            """s
 
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
 
     assert is_pid(process_metadata[:pid])
     assert is_list(process_metadata[:ancestors])
@@ -849,7 +857,7 @@ defmodule Logger.TranslatorTest do
            Ancestors: \[#PID<\d+\.\d+\.\d+>\]
            """s
 
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
   end
 
@@ -876,7 +884,7 @@ defmodule Logger.TranslatorTest do
            Ancestors: \[#PID<\d+\.\d+\.\d+>\]
            """s
 
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
   end
 
@@ -900,7 +908,7 @@ defmodule Logger.TranslatorTest do
            Ancestors: \[#PID<\d+\.\d+\.\d+>\]
            """s
 
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
   end
 
@@ -948,13 +956,13 @@ defmodule Logger.TranslatorTest do
   test "translates Supervisor progress" do
     {:ok, pid} = Supervisor.start_link([], strategy: :one_for_one)
 
-    assert capture_log(:info, fn ->
+    assert capture_log(:debug, fn ->
              ref = Process.monitor(pid)
              Supervisor.start_child(pid, worker(Task, [__MODULE__, :sleep, [self()]]))
              Process.exit(pid, :normal)
              receive do: ({:DOWN, ^ref, _, _, _} -> :ok)
            end) =~ ~r"""
-           \[info\] Child Task of Supervisor #PID<\d+\.\d+\.\d+> \(Supervisor\.Default\) started
+           \[(debug|info)\] Child Task of Supervisor #PID<\d+\.\d+\.\d+> \(Supervisor\.Default\) started
            Pid: #PID<\d+\.\d+\.\d+>
            Start Call: Task.start_link\(Logger.TranslatorTest, :sleep, \[#PID<\d+\.\d+\.\d+>\]\)
            """
@@ -963,36 +971,36 @@ defmodule Logger.TranslatorTest do
   test "translates Supervisor progress with name" do
     {:ok, pid} = Supervisor.start_link([], strategy: :one_for_one, name: __MODULE__)
 
-    assert capture_log(:info, fn ->
+    assert capture_log(:debug, fn ->
              ref = Process.monitor(pid)
              Supervisor.start_child(pid, worker(Task, [__MODULE__, :sleep, [self()]]))
              Process.exit(pid, :normal)
              receive do: ({:DOWN, ^ref, _, _, _} -> :ok)
            end) =~ ~r"""
-           \[info\] Child Task of Supervisor Logger.TranslatorTest started
+           \[(debug|info)\] Child Task of Supervisor Logger.TranslatorTest started
            """
 
     {:ok, pid} = Supervisor.start_link([], strategy: :one_for_one, name: {:global, __MODULE__})
 
-    assert capture_log(:info, fn ->
+    assert capture_log(:debug, fn ->
              ref = Process.monitor(pid)
              Supervisor.start_child(pid, worker(Task, [__MODULE__, :sleep, [self()]]))
              Process.exit(pid, :normal)
              receive do: ({:DOWN, ^ref, _, _, _} -> :ok)
            end) =~ ~r"""
-           \[info\] Child Task of Supervisor Logger.TranslatorTest started
+           \[(debug|info)\] Child Task of Supervisor Logger.TranslatorTest started
            """
 
     {:ok, pid} =
       Supervisor.start_link([], strategy: :one_for_one, name: {:via, :global, __MODULE__})
 
-    assert capture_log(:info, fn ->
+    assert capture_log(:debug, fn ->
              ref = Process.monitor(pid)
              Supervisor.start_child(pid, worker(Task, [__MODULE__, :sleep, [self()]]))
              Process.exit(pid, :normal)
              receive do: ({:DOWN, ^ref, _, _, _} -> :ok)
            end) =~ ~r"""
-           \[info\] Child Task of Supervisor Logger.TranslatorTest started
+           \[(debug|info)\] Child Task of Supervisor Logger.TranslatorTest started
            """
   end
 
@@ -1043,7 +1051,7 @@ defmodule Logger.TranslatorTest do
            Start Call: Logger.TranslatorTest.undef\(\)
            """s
 
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
+    assert_receive {:event, {:report, %{label: {:supervisor, :start_error}}}, _child_metadata}
   end
 
   test "translates Supervisor reports terminated" do
@@ -1060,9 +1068,12 @@ defmodule Logger.TranslatorTest do
            Start Call: Task.start_link\(Kernel, :exit, \[:stop\]\)
            """
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
-    assert_receive {:event, {:string, ["Child ", "Task" | _]}, _child_task_metadata}
+    assert_receive {:event, {:report, %{label: {Task.Supervisor, :terminating}}}, task_metadata}
+    assert_receive {:event, {:report, %{label: {:supervisor, :shutdown}}}, _child_metadata}
+
+    assert_receive {:event, {:report, %{label: {:supervisor, :child_terminated}}},
+                    _child_task_metadata}
+
     assert {:stop, [_ | _]} = task_metadata[:crash_reason]
   end
 
@@ -1079,9 +1090,12 @@ defmodule Logger.TranslatorTest do
            Start Call: Task.start_link\(Kernel, :exit, \[:stop\]\)
            """
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
-    assert_receive {:event, {:string, ["Child ", "Task" | _]}, _child_task_metadata}
+    assert_receive {:event, {:report, %{label: {Task.Supervisor, :terminating}}}, task_metadata}
+    assert_receive {:event, {:report, %{label: {:supervisor, :shutdown}}}, _child_metadata}
+
+    assert_receive {:event, {:report, %{label: {:supervisor, :child_terminated}}},
+                    _child_task_metadata}
+
     assert {:stop, [_ | _]} = task_metadata[:crash_reason]
   end
 
@@ -1097,7 +1111,7 @@ defmodule Logger.TranslatorTest do
            Start Call: Logger.TranslatorTest.abnormal\(\)
            """
 
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
+    assert_receive {:event, {:report, %{label: {:supervisor, :shutdown_error}}}, _child_metadata}
   end
 
   test "translates Supervisor reports abnormal shutdown on debug" do
@@ -1117,7 +1131,7 @@ defmodule Logger.TranslatorTest do
            Type: :worker
            """
 
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
+    assert_receive {:event, {:report, %{label: {:supervisor, :shutdown_error}}}, _child_metadata}
   end
 
   test "translates DynamicSupervisor reports abnormal shutdown" do
@@ -1136,7 +1150,7 @@ defmodule Logger.TranslatorTest do
            Start Call: Logger.TranslatorTest.abnormal\(\)
            """
 
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
+    assert_receive {:event, {:report, %{label: {:supervisor, :shutdown_error}}}, _child_metadata}
   end
 
   test "translates DynamicSupervisor reports abnormal shutdown including extra_arguments" do
@@ -1157,7 +1171,7 @@ defmodule Logger.TranslatorTest do
            Start Call: Logger.TranslatorTest.abnormal\(:extra, :args\)
            """
 
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
+    assert_receive {:event, {:report, %{label: {:supervisor, :shutdown_error}}}, _child_metadata}
   end
 
   test "translates named DynamicSupervisor reports abnormal shutdown" do
@@ -1176,7 +1190,7 @@ defmodule Logger.TranslatorTest do
            Start Call: Logger.TranslatorTest.abnormal\(\)
            """
 
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
+    assert_receive {:event, {:report, %{label: {:supervisor, :shutdown_error}}}, _child_metadata}
   end
 
   test "translates :supervisor_bridge progress" do
@@ -1205,8 +1219,8 @@ defmodule Logger.TranslatorTest do
            Start Module: Logger.TranslatorTest.MyBridge
            """
 
-    assert_receive {:event, {:string, ["Task " <> _ | _]}, task_metadata}
-    assert_receive {:event, {:string, ["Child of Supervisor " | _]}, _child_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, task_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_server, :terminate}}}, _child_metadata}
     assert {:stop, [_ | _]} = task_metadata[:crash_reason]
   end
 
@@ -1244,9 +1258,11 @@ defmodule Logger.TranslatorTest do
       end)
 
     assert log =~ ~s(Start Call: Logger.TranslatorTest.WeirdFunctionNamesGenServer."start link"/?)
-    assert_receive {:event, {:string, ["GenServer " <> _ | _]}, server_metadata}
-    assert_receive {:event, {:string, ["Process " | _]}, process_metadata}
-    assert_receive {:event, {:string, ["Child " | _]}, _child_metadata}
+    assert_receive {:event, {:report, %{label: {:gen_server, :terminate}}}, server_metadata}
+    assert_receive {:event, {:report, %{label: {:proc_lib, :crash}}}, process_metadata}
+
+    assert_receive {:event, {:report, %{label: {:supervisor, :child_terminated}}},
+                    _child_metadata}
 
     assert {%RuntimeError{message: "oops"}, [_ | _]} = server_metadata[:crash_reason]
     assert {%RuntimeError{message: "oops"}, [_ | _]} = process_metadata[:crash_reason]
@@ -1255,6 +1271,13 @@ defmodule Logger.TranslatorTest do
     :code.delete(WeirdFunctionNamesGenServer)
   end
 
+  @tag logger_config: [handle_otp_reports: false]
+  test "drops events if otp report handling is switched off" do
+    {:ok, pid} = GenServer.start(MyGenServer, :ok)
+    catch_exit(GenServer.call(pid, :error))
+    refute_receive {:event, _, _}, 100
+  end
+
   def task(parent, fun \\ fn -> raise "oops" end) do
     mon = Process.monitor(parent)
     Process.unlink(parent)
@@ -1308,4 +1331,25 @@ defmodule Logger.TranslatorTest do
   defp worker(name, args, opts \\ []) do
     Enum.into(opts, %{id: name, start: {name, opts[:function] || :start_link, args}})
   end
+
+  defp put_logger_config(new_config) do
+    Application.put_all_env(logger: new_config)
+
+    Logger.App.stop()
+
+    # Shutdown the application
+    Logger.App.stop()
+
+    # And start it without warnings
+    Application.put_env(:logger, :level, :error)
+    Application.start(:logger)
+    Application.delete_env(:logger, :level)
+    Logger.configure(level: :debug)
+  end
+
+  defp restore_logger_config(config) do
+    Application.put_all_env(logger: config)
+    Logger.App.stop()
+    Application.start(:logger)
+  end
 end
