@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule LoggerTest do
   use Logger.Case
   require Logger
@@ -186,15 +190,14 @@ defmodule LoggerTest do
     assert Logger.metadata() == [data: true]
     assert Logger.metadata(data: true) == :ok
     assert Logger.metadata() == [data: true]
-    assert Logger.metadata(meta: 1) == :ok
-    metadata = Logger.metadata()
-    assert Enum.sort(metadata) == [data: true, meta: 1]
-    assert Logger.metadata(data: nil) == :ok
+    assert Logger.metadata(%{meta: 1}) == :ok
+    assert Enum.sort(Logger.metadata()) == [data: true, meta: 1]
+    assert Logger.metadata(%{data: nil}) == :ok
     assert Logger.metadata() == [meta: 1]
 
     assert Logger.reset_metadata(meta: 2) == :ok
     assert Logger.metadata() == [meta: 2]
-    assert Logger.reset_metadata(data: true, app: nil) == :ok
+    assert Logger.reset_metadata(%{data: true, app: nil}) == :ok
     assert Logger.metadata() == [data: true]
     assert Logger.reset_metadata() == :ok
     assert Logger.metadata() == []
@@ -346,8 +349,10 @@ defmodule LoggerTest do
              end) == ""
 
       assert capture_log(:info, fn ->
-               assert Logger.debug(raise("not invoked"), []) == :ok
+               assert Logger.debug(send(self(), :something), []) == :ok
              end) == ""
+
+      refute_received :something
     end
 
     test "info/2" do
@@ -360,8 +365,10 @@ defmodule LoggerTest do
              end) == ""
 
       assert capture_log(:notice, fn ->
-               assert Logger.info(raise("not invoked"), []) == :ok
+               assert Logger.info(send(self(), :something), []) == :ok
              end) == ""
+
+      refute_received :something
     end
 
     test "warning/2" do
@@ -374,8 +381,10 @@ defmodule LoggerTest do
              end) == ""
 
       assert capture_log(:error, fn ->
-               assert Logger.warning(raise("not invoked"), []) == :ok
+               assert Logger.warning(send(self(), :something), []) == :ok
              end) == ""
+
+      refute_received :something
     end
 
     test "error/2" do
@@ -388,8 +397,10 @@ defmodule LoggerTest do
              end) == ""
 
       assert capture_log(:critical, fn ->
-               assert Logger.error(raise("not invoked"), []) == :ok
+               assert Logger.error(send(self(), :something), []) == :ok
              end) == ""
+
+      refute_received :something
     end
 
     test "critical/2" do
@@ -402,8 +413,10 @@ defmodule LoggerTest do
              end) == ""
 
       assert capture_log(:alert, fn ->
-               assert Logger.critical(raise("not invoked"), []) == :ok
+               assert Logger.critical(send(self(), :something), []) == :ok
              end) == ""
+
+      refute_received :something
     end
 
     test "alert/2" do
@@ -416,8 +429,10 @@ defmodule LoggerTest do
              end) == ""
 
       assert capture_log(:emergency, fn ->
-               assert Logger.alert(raise("not invoked"), []) == :ok
+               assert Logger.alert(send(self(), :something), []) == :ok
              end) == ""
+
+      refute_received :something
     end
 
     test "emergency/2" do
@@ -430,8 +445,10 @@ defmodule LoggerTest do
              end) == ""
 
       assert capture_log(:none, fn ->
-               assert Logger.emergency(raise("not invoked"), []) == :ok
+               assert Logger.emergency(send(self(), :something), []) == :ok
              end) == ""
+
+      refute_received :something
     end
   end
 
@@ -627,26 +644,9 @@ defmodule LoggerTest do
   end
 
   test "configure/1 sets options" do
-    Logger.configure(sync_threshold: 10)
-    Logger.configure(truncate: 4048)
-    Logger.configure(utc_log: true)
-    Logger.configure(discard_threshold: 10_000)
     Logger.configure(translator_inspect_opts: [limit: 3])
-
-    assert Application.get_env(:logger, :sync_threshold) == 10
-    assert Application.get_env(:logger, :utc_log) == true
-    assert Application.get_env(:logger, :truncate) == 4048
-    assert Application.get_env(:logger, :discard_threshold) == 10_000
     assert Application.get_env(:logger, :translator_inspect_opts) == [limit: 3]
-
-    assert {:ok, %{config: log_data}} = :logger.get_handler_config(Logger)
-    assert log_data.utc_log == true
-    assert log_data.truncate == 4048
   after
-    Logger.configure(sync_threshold: 20)
-    Logger.configure(truncate: 8096)
-    Logger.configure(utc_log: false)
-    Logger.configure(discard_threshold: 500)
     Logger.configure(translator_inspect_opts: [])
   end
 
