@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix do
   @moduledoc ~S"""
   Mix is a build tool that provides tasks for creating, compiling,
@@ -327,20 +331,20 @@ defmodule Mix do
     * `MIX_ARCHIVES` - specifies the directory into which the archives should be installed
       (default: `~/.mix/archives`)
 
-    * `MIX_BUILD_PATH` - sets the project `Mix.Project.build_path/0` config.
-      This option must always point to a subdirectory inside a temporary directory.
-      For instance, never "/tmp" or "_build" but "_build/PROD" or "/tmp/PROD", as
-      required by Mix. This environment variable is used mostly by external build
-      tools. For your CI servers, you likely want to use `MIX_BUILD_ROOT` below.
+    * `MIX_BUILD_PATH` - sets the project `Mix.Project.build_path/0`, including the
+      current environment, such as "/path/to/project/_build/dev". This environment
+      variable is used mostly by external build tools who need enforce a build
+      directory independent of `MIX_ENV` and `MIX_TARGET`. For your CI servers,
+      you likely want to use `MIX_BUILD_ROOT` below.
 
     * `MIX_BUILD_ROOT` - sets the root directory where build artifacts should be
-      written to. For example, "_build". If `MIX_BUILD_PATH` is set, this option
-      is ignored.
+      written to. For example, "/path/to/_build". If `MIX_BUILD_PATH` is set,
+      this option is ignored.
 
     * `MIX_DEBUG` - outputs debug information about each task before running it
 
     * `MIX_DEPS_PATH` - sets the project `Mix.Project.deps_path/0` config for the
-      current project (default: `deps`)
+      current project, as an absolute path, such as `/path/to/project/deps`
 
     * `MIX_ENV` - specifies which environment should be used. See [Environments](#module-environments)
 
@@ -355,12 +359,18 @@ defmodule Mix do
     * `MIX_OS_CONCURRENCY_LOCK` - when set to `0` or `false`, disables mix compilation locking.
       While not recommended, this may be necessary in cases where hard links or TCP sockets are
       not available. When opting for this behaviour, make sure to not start concurrent compilations
-      of the same project.
+      of the same project
+
+    * `MIX_OS_DEPS_COMPILE_PARTITION_COUNT` - when set to a number greater than 1, it enables
+      compilation of dependencies over multiple operating system processes. See `mix help deps.compile`
+      for more information
 
     * `MIX_PATH` - appends extra code paths
 
     * `MIX_PROFILE` - a list of comma-separated Mix tasks to profile the time spent on
-      functions by the process running the task
+      functions by the process running the task, such as `MIX_PROFILE=compile,test`.
+      You can also set `MIX_PROFILE_FLAGS` to control the flags given to profiling,
+      see `mix profile.tprof` for all options. By default, it uses `--no-warmup`.
 
     * `MIX_QUIET` - does not print information messages to the terminal
 
@@ -414,7 +424,26 @@ defmodule Mix do
 
   """
 
-  @mix_install_project __MODULE__.InstallProject
+  @typedoc """
+  Options for `install/2`.
+
+  See the `install/2` function documentation for detailed information about each option.
+  """
+  @type install_opts :: [
+          force: boolean(),
+          verbose: boolean(),
+          consolidate_protocols: boolean(),
+          elixir: String.t(),
+          system_env: Enum.t(),
+          config: keyword(),
+          config_path: String.t() | atom(),
+          lockfile: String.t() | atom(),
+          start_applications: boolean(),
+          compilers: [atom()],
+          runtime_config: keyword()
+        ]
+
+  @mix_install_project Mix.InstallProject
   @mix_install_app :mix_install
   @mix_install_app_string Atom.to_string(@mix_install_app)
 
@@ -709,8 +738,7 @@ defmodule Mix do
     * `:verbose` - if `true`, prints additional debugging information
       (Default: `false`)
 
-    * `:consolidate_protocols` - if `true`, runs protocol
-      consolidation via the `mix compile.protocols` task (Default: `true`)
+    * `:consolidate_protocols` - if `true`, runs protocol consolidation (Default: `true`)
 
     * `:elixir` - if set, ensures the current Elixir version matches the given
       version requirement (Default: `nil`)
@@ -818,7 +846,7 @@ defmodule Mix do
   ## Environment variables
 
   The `MIX_INSTALL_DIR` environment variable configures the directory that
-  caches all `Mix.install/2`. It defaults to the "mix/install" folder in the
+  caches all `Mix.install/2`. It defaults to the `mix/install` folder in the
   default user cache of your operating system. You can use `install_project_dir/0`
   to access the directory of an existing install (alongside other installs):
 
@@ -836,6 +864,10 @@ defmodule Mix do
   if `:force` is enabled.
   """
   @doc since: "1.12.0"
+  @spec install(
+          [atom() | {atom(), String.t()} | {atom(), String.t(), keyword()} | {atom(), keyword()}],
+          install_opts()
+        ) :: :ok
   def install(deps, opts \\ [])
 
   def install(deps, opts) when is_list(deps) and is_list(opts) do
@@ -875,6 +907,7 @@ defmodule Mix do
         config: [],
         config_path: nil,
         consolidate_protocols: true,
+        compilers: [:elixir],
         elixir: nil,
         force: false,
         lockfile: nil,
@@ -889,20 +922,19 @@ defmodule Mix do
     system_env = Keyword.fetch!(opts, :system_env)
     consolidate_protocols? = Keyword.fetch!(opts, :consolidate_protocols)
     start_applications? = Keyword.fetch!(opts, :start_applications)
+    compilers = Keyword.fetch!(opts, :compilers)
 
     id =
       {deps, config, system_env, consolidate_protocols?}
       |> :erlang.term_to_binary()
       |> :erlang.md5()
-      |> Base.encode16(case: :lower)
+      |> Base.url_encode64(padding: false)
 
     force? = System.get_env("MIX_INSTALL_FORCE") in ["1", "true"] or Keyword.fetch!(opts, :force)
 
     case Mix.State.get(:installed) do
       nil ->
-        Application.put_all_env(config, persistent: true)
         System.put_env(system_env)
-
         install_project_dir = install_project_dir(id)
 
         if Keyword.fetch!(opts, :verbose) do
@@ -916,13 +948,18 @@ defmodule Mix do
         dynamic_config = [
           deps: deps,
           consolidate_protocols: consolidate_protocols?,
-          config_path: config_path
+          config_path: config_path,
+          compilers: compilers
         ]
 
-        config = install_project_config(dynamic_config)
+        :ok =
+          Mix.ProjectStack.push(
+            @mix_install_project,
+            [compile_config: config] ++ install_project_config(dynamic_config),
+            "nofile"
+          )
 
         started_apps = Application.started_applications()
-        :ok = Mix.ProjectStack.push(@mix_install_project, config, "nofile")
         build_dir = Path.join(install_project_dir, "_build")
         external_lockfile = expand_path(opts[:lockfile], deps, :lockfile, "mix.lock")
 
@@ -939,9 +976,9 @@ defmodule Mix do
           File.mkdir_p!(install_project_dir)
 
           File.cd!(install_project_dir, fn ->
-            if config_path do
-              Mix.Task.rerun("loadconfig")
-            end
+            # This steps need to be mirror in mix deps.partition
+            Application.put_all_env(config, persistent: true)
+            Mix.Task.rerun("loadconfig")
 
             cond do
               external_lockfile ->
@@ -1068,14 +1105,13 @@ defmodule Mix do
       System.get_env("MIX_INSTALL_DIR") ||
         Path.join(Mix.Utils.mix_cache(), "installs")
 
-    version = "elixir-#{System.version()}-erts-#{:erlang.system_info(:version)}"
+    version = "ex-#{System.version()}-erl-#{:erlang.system_info(:version)}"
     Path.join([install_root, version, cache_id])
   end
 
   defp install_project_config(dynamic_config) do
     [
-      version: "0.1.0",
-      build_embedded: false,
+      version: "1.0.0",
       build_per_environment: true,
       build_path: "_build",
       lockfile: "mix.lock",
@@ -1083,7 +1119,6 @@ defmodule Mix do
       app: @mix_install_app,
       erlc_paths: [],
       elixirc_paths: [],
-      compilers: [],
       prune_code_paths: false
     ] ++ dynamic_config
   end
