@@ -1,8 +1,12 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Compilers.Elixir do
   @moduledoc false
 
-  @manifest_vsn 26
-  @checkpoint_vsn 2
+  @manifest_vsn 28
+  @checkpoint_vsn 4
 
   import Record
 
@@ -46,13 +50,23 @@ defmodule Mix.Compilers.Elixir do
     all_paths = Mix.Utils.extract_files(srcs, [:ex])
 
     {all_modules, all_sources, all_local_exports, old_parents, old_cache_key, old_deps_config,
-     old_project_mtime,
-     old_config_mtime} =
+     old_project_mtime, old_config_mtime, old_protocols_and_impls} =
       parse_manifest(manifest, dest)
 
+    # In case we aborted in the middle of a verification,
+    # we need to delete all modules that we wrote to disk.
+    # In the future, we may want to make it so we only run
+    # the verification again.
+    with {:ok, modules} <- parse_checkpoint(:verify, manifest) do
+      for module <- modules do
+        File.rm(Path.join(dest, Atom.to_string(module) <> ".beam"))
+      end
+    end
+
     # Prepend ourselves early because of __mix_recompile__? checks
     # and also that, in case of nothing compiled, we already need
     # ourselves available in the path.
+    File.mkdir_p!(dest)
     Code.prepend_path(dest)
 
     # If modules have been added or removed from the Erlang compiler,
@@ -129,7 +143,7 @@ defmodule Mix.Compilers.Elixir do
           {false, stale, old_deps_config}
       end
 
-    {stale_modules, stale_exports, all_local_exports} =
+    {stale_modules, stale_exports, all_local_exports, protocols_and_impls} =
       stale_local_deps(local_deps, manifest, stale, modified, all_local_exports)
 
     prev_paths = Map.keys(all_sources)
@@ -148,35 +162,60 @@ defmodule Mix.Compilers.Elixir do
           removed,
           Map.merge(stale_modules, removed_modules),
           Map.merge(stale_exports, removed_modules),
-          dest
+          dest,
+          timestamp
         )
       end
 
     stale = changed -- removed
 
-    {sources, removed_modules} =
+    {sources, stale_exports} =
       update_stale_sources(sources, stale, removed_modules, sources_stats)
 
-    if stale != [] or stale_modules != %{} do
+    consolidation_status =
+      if Mix.Project.umbrella?() do
+        :off
+      else
+        Mix.Compilers.Protocol.status(config_mtime > old_config_mtime, opts)
+      end
+
+    if stale != [] or stale_modules != %{} or removed != [] or deps_changed? or
+         consolidation_status == :force do
       path = opts[:purge_consolidation_path_if_stale]
 
       if is_binary(path) and Code.delete_path(path) do
         purge_modules_in_path(path)
       end
 
-      Mix.Utils.compiling_n(length(stale), :ex)
-      Mix.Project.ensure_structure()
+      if stale != [] do
+        Mix.Utils.compiling_n(length(stale), :ex)
+      end
 
-      # We don't want to cache this path as we will write to it
-      true = Code.prepend_path(dest)
+      Mix.Project.ensure_structure()
       previous_opts = set_compiler_opts(opts)
 
       try do
-        state = {%{}, exports, sources, [], modules, removed_modules}
-        compiler_loop(stale, stale_modules, dest, timestamp, opts, state)
+        consolidation = {consolidation_status, old_protocols_and_impls, protocols_and_impls}
+
+        state = %{
+          modules: %{},
+          exports: exports,
+          sources: sources,
+          changed: [],
+          pending_modules: modules,
+          stale_exports: stale_exports,
+          consolidation: consolidation
+        }
+
+        compiler_loop(manifest, stale, stale_modules, dest, timestamp, opts, state)
       else
         {:ok, %{runtime_warnings: runtime_warnings, compile_warnings: compile_warnings}, state} ->
-          {modules, _exports, sources, _changed, pending_modules, _stale_exports} = state
+          %{
+            modules: modules,
+            sources: sources,
+            pending_modules: pending_modules,
+            consolidation: protocols_and_impls
+          } = state
 
           previous_warnings =
             if Keyword.get(opts, :all_warnings, true),
@@ -197,6 +236,7 @@ defmodule Mix.Compilers.Elixir do
             new_deps_config,
             project_mtime,
             config_mtime,
+            protocols_and_impls,
             timestamp
           )
 
@@ -208,54 +248,26 @@ defmodule Mix.Compilers.Elixir do
           end
 
           Mix.Task.Compiler.notify_modules_compiled(lazy_modules_diff)
-
-          unless_previous_warnings_as_errors(previous_warnings, opts, {:ok, all_warnings})
+          unless_warnings_as_errors(opts, {:ok, all_warnings})
 
         {:error, errors, %{runtime_warnings: r_warnings, compile_warnings: c_warnings}, state} ->
+          {errors, warnings} =
+            if opts[:warnings_as_errors],
+              do: {errors ++ r_warnings ++ c_warnings, []},
+              else: {errors, r_warnings ++ c_warnings}
+
           # In case of errors, we show all previous warnings and all new ones.
-          {_, _, sources, _, _, _} = state
           errors = Enum.map(errors, &diagnostic/1)
-          warnings = Enum.map(r_warnings ++ c_warnings, &diagnostic/1)
+          warnings = Enum.map(warnings, &diagnostic/1)
           all_warnings = Keyword.get(opts, :all_warnings, errors == [])
-          {:error, previous_warnings(sources, all_warnings) ++ warnings ++ errors}
+          {:error, previous_warnings(state.sources, all_warnings) ++ warnings ++ errors}
       after
         Code.compiler_options(previous_opts)
       end
     else
-      # We need to return ok if deps_changed? or stale_modules changed,
-      # even if no code was compiled, because we need to propagate the changed
-      # status to compile.protocols. This will be the case whenever:
-      #
-      #   * the lock file or a config changes
-      #   * any module in a path dependency changes
-      #   * the mix.exs changes
-      #   * the Erlang manifest updates (Erlang files are compiled)
-      #
-      # In the first case, we will consolidate from scratch. In the remaining, we
-      # will only compute the diff with current protocols. In fact, there is no
-      # need to reconsolidate if an Erlang file changes and it doesn't trigger
-      # any other change, but the diff check should be reasonably fast anyway.
-      status = if removed != [] or deps_changed? or stale_modules != %{}, do: :ok, else: :noop
-
-      if status != :noop do
-        write_manifest(
-          manifest,
-          modules,
-          sources,
-          all_local_exports,
-          new_parents,
-          new_cache_key,
-          new_deps_config,
-          project_mtime,
-          config_mtime,
-          timestamp
-        )
-      end
-
       all_warnings = Keyword.get(opts, :all_warnings, true)
       previous_warnings = previous_warnings(sources, all_warnings)
-
-      unless_previous_warnings_as_errors(previous_warnings, opts, {status, previous_warnings})
+      unless_warnings_as_errors(opts, {:noop, previous_warnings})
     end
   end
 
@@ -292,24 +304,6 @@ defmodule Mix.Compilers.Elixir do
     end)
   end
 
-  @doc """
-  Returns protocols and implementations for the given `manifest`.
-  """
-  def protocols_and_impls(paths) do
-    Enum.reduce(paths, {%{}, %{}}, fn path, acc ->
-      {modules, _} = read_manifest(Path.join(path, ".mix/compile.elixir"))
-
-      Enum.reduce(modules, acc, fn
-        {module, module(kind: kind, timestamp: timestamp)}, {protocols, impls} ->
-          case kind do
-            :protocol -> {Map.put(protocols, module, timestamp), impls}
-            {:impl, protocol} -> {protocols, Map.put(impls, module, protocol)}
-            _ -> {protocols, impls}
-          end
-      end)
-    end)
-  end
-
   @doc """
   Reads the manifest for external consumption.
   """
@@ -319,7 +313,7 @@ defmodule Mix.Compilers.Elixir do
     rescue
       _ -> {[], []}
     else
-      {@manifest_vsn, modules, sources, _, _, _, _, _, _} -> {modules, sources}
+      {@manifest_vsn, modules, sources, _, _, _, _, _, _, _} -> {modules, sources}
       _ -> {[], []}
     end
   end
@@ -327,9 +321,9 @@ defmodule Mix.Compilers.Elixir do
   @doc """
   Retrieves all diagnostics from the given manifest.
   """
-  def diagnostics(manifest, dest) do
-    {_, all_sources, _, _, _, _, _, _} = parse_manifest(manifest, dest)
-    previous_warnings(all_sources, false)
+  def diagnostics(manifest) do
+    {_, sources} = read_manifest(manifest)
+    previous_warnings(sources, false)
   end
 
   defp compiler_info_from_force(manifest, all_paths, all_modules, dest) do
@@ -350,8 +344,11 @@ defmodule Mix.Compilers.Elixir do
     {%{}, %{}, all_paths, sources_stats}
   end
 
-  # If any .beam file is missing, the first one will the first to miss,
-  # so we always check that. If there are no modules, then we can rely
+  # If the user does a change, compilation fails, and then they revert
+  # the change, the mtime will have changed but the .beam files will
+  # be missing and the digest is the same, so we need to check if .beam
+  # files are available. Checking the first .beam file is enough, as
+  # they would be all removed. If there are no modules, then we can rely
   # purely on digests.
   defp missing_beam_file?(dest, [mod | _]), do: not File.exists?(beam_path(dest, mod))
   defp missing_beam_file?(_dest, []), do: false
@@ -364,56 +361,59 @@ defmodule Mix.Compilers.Elixir do
          removed,
          stale_modules,
          stale_exports,
-         dest
+         dest,
+         timestamp
        ) do
-    {modules_to_recompile, modules_to_mix_check} =
-      for {module, module(recompile?: recompile?)} <- all_modules, reduce: {[], []} do
-        {modules_to_recompile, modules_to_mix_check} ->
-          cond do
-            Map.has_key?(stale_modules, module) ->
-              {[module | modules_to_recompile], modules_to_mix_check}
-
-            recompile? ->
-              {modules_to_recompile, [module | modules_to_mix_check]}
-
-            true ->
-              {modules_to_recompile, modules_to_mix_check}
-          end
+    {checkpoint_stale_modules, checkpoint_stale_exports} =
+      case parse_checkpoint(:update, manifest) do
+        {:ok, {_, _} = data} -> data
+        :error -> {%{}, %{}}
       end
 
-    _ = Code.ensure_all_loaded(modules_to_mix_check)
-
-    modules_to_recompile =
-      modules_to_recompile ++
-        for {:ok, {module, true}} <-
-              Task.async_stream(
-                modules_to_mix_check,
-                fn module ->
-                  {module,
-                   function_exported?(module, :__mix_recompile__?, 0) and
-                     module.__mix_recompile__?()}
-                end,
-                ordered: false,
-                timeout: :infinity
-              ) do
-          module
-        end
-
-    {checkpoint_stale_modules, checkpoint_stale_exports, checkpoint_modules} =
-      parse_checkpoint(manifest)
-
-    modules_to_recompile =
-      Map.merge(checkpoint_modules, Map.from_keys(modules_to_recompile, true))
-
     stale_modules = Map.merge(checkpoint_stale_modules, stale_modules)
     stale_exports = Map.merge(checkpoint_stale_exports, stale_exports)
 
+    # Once we added semantic recompilation, the following can happen:
+    #
+    # 1. The user changes config/mix.exs/__mix_recompile__?
+    # 2. We detect the change, remove .beam files and start recompilation
+    # 3. Recompilation fails
+    # 4. The user reverts the change
+    # 5. The compiler no longer recompiles and the .beam files are missing
+    #
+    # Therefore, it is important for us to checkpoint any state that may
+    # have lead to a compilation and which can now be reverted.
     if map_size(stale_modules) != map_size(checkpoint_stale_modules) or
-         map_size(stale_exports) != map_size(checkpoint_stale_exports) or
-         map_size(modules_to_recompile) != map_size(checkpoint_modules) do
-      write_checkpoint(manifest, stale_modules, stale_exports, modules_to_recompile)
+         map_size(stale_exports) != map_size(checkpoint_stale_exports) do
+      write_checkpoint(:update, manifest, {stale_modules, stale_exports})
     end
 
+    # We don't need to store those in the checkpoint because
+    # these changes come from modules and, when they are stale,
+    # we remove the .beam files and touch sources.
+    modules_to_mix_check =
+      for {module, module(recompile?: true)} <- all_modules,
+          not Map.has_key?(stale_modules, module),
+          do: module
+
+    _ = Code.ensure_all_loaded(modules_to_mix_check)
+
+    modules_to_recompile =
+      for {:ok, {module, true}} <-
+            Task.async_stream(
+              modules_to_mix_check,
+              fn module ->
+                {module,
+                 function_exported?(module, :__mix_recompile__?, 0) and
+                   module.__mix_recompile__?()}
+              end,
+              ordered: false,
+              timeout: :infinity
+            ),
+          into: %{} do
+        {module, true}
+      end
+
     sources_stats =
       for path <- new_paths,
           into: mtimes_and_sizes(all_sources),
@@ -422,20 +422,30 @@ defmodule Mix.Compilers.Elixir do
     # Sources that have changed on disk or
     # any modules associated with them need to be recompiled
     changed =
-      for {source,
-           source(external: external, size: size, mtime: mtime, digest: digest, modules: modules)} <-
-            all_sources,
-          {last_mtime, last_size} = Map.fetch!(sources_stats, source),
-          # If the user does a change, compilation fails, and then they revert
-          # the change, the mtime will have changed but the .beam files will
-          # be missing and the digest is the same, so we need to check if .beam
-          # files are available.
-          size != last_size or
-            Enum.any?(modules, &Map.has_key?(modules_to_recompile, &1)) or
+      Enum.flat_map(all_sources, fn
+        {source,
+         source(external: external, size: size, mtime: mtime, digest: digest, modules: modules)} ->
+          {last_mtime, last_size} = Map.fetch!(sources_stats, source)
+
+          cond do
             Enum.any?(external, &stale_external?(&1, sources_stats)) or
-            (last_mtime > mtime and
-               (missing_beam_file?(dest, modules) or digest_changed?(source, digest))),
-          do: source
+                has_any_key?(modules_to_recompile, modules) ->
+              # Mark the source as changed so the combination of a timestamp
+              # plus removed beam files (which are removed by update_stale_entries)
+              # causes it to be recompiled
+              File.touch!(source, timestamp + 1)
+              [source]
+
+            size != last_size or
+              has_any_key?(stale_modules, modules) or
+                (last_mtime > mtime and
+                   (missing_beam_file?(dest, modules) or digest_changed?(source, digest))) ->
+              [source]
+
+            true ->
+              []
+          end
+      end)
 
     changed = new_paths ++ changed
 
@@ -606,7 +616,7 @@ defmodule Mix.Compilers.Elixir do
   end
 
   defp has_any_key?(map, enumerable) do
-    Enum.any?(enumerable, &Map.has_key?(map, &1))
+    map != %{} and Enum.any?(enumerable, &Map.has_key?(map, &1))
   end
 
   defp stale_local_deps(local_deps, manifest, stale_modules, modified, deps_exports) do
@@ -619,8 +629,8 @@ defmodule Mix.Compilers.Elixir do
     for %{app: app, opts: opts} <- local_deps,
         manifest = Path.join([opts[:build], ".mix", base]),
         Mix.Utils.last_modified(manifest) > modified,
-        reduce: {stale_modules, stale_modules, deps_exports} do
-      {modules, exports, deps_exports} ->
+        reduce: {stale_modules, stale_modules, deps_exports, protocols_and_impls()} do
+      {modules, exports, deps_exports, protocols_and_impls} ->
         {manifest_modules, manifest_sources} = read_manifest(manifest)
 
         dep_modules =
@@ -641,7 +651,10 @@ defmodule Mix.Compilers.Elixir do
         {exports, new_exports} =
           for {module, _} <- dep_modules, reduce: {exports, []} do
             {exports, new_exports} ->
-              export = exports_md5(module, false)
+              export =
+                if Code.ensure_loaded?(module) and function_exported?(module, :__info__, 1) do
+                  module.__info__(:exports_md5)
+                end
 
               # If the exports are the same, then the API did not change,
               # so we do not mark the export as stale. Note this has to
@@ -673,7 +686,11 @@ defmodule Mix.Compilers.Elixir do
         modules = modules |> Map.merge(dep_modules) |> Map.merge(removed)
         exports = Map.merge(exports, removed)
         deps_exports = Map.put(deps_exports, app, new_exports)
-        {modules, exports, deps_exports}
+
+        protocols_and_impls =
+          protocols_and_impls_from_modules(manifest_modules, protocols_and_impls)
+
+        {modules, exports, deps_exports, protocols_and_impls}
     end
   end
 
@@ -711,40 +728,24 @@ defmodule Mix.Compilers.Elixir do
   defp fixpoint_non_compile_modules([], modules, true, pending_sources),
     do: fixpoint_non_compile_modules(pending_sources, modules, false, [])
 
-  defp exports_md5(module, use_attributes?) do
-    cond do
-      function_exported?(module, :__info__, 1) ->
-        module.__info__(:exports_md5)
-
-      use_attributes? ->
-        defs = :lists.sort(Module.definitions_in(module, :def))
-        defmacros = :lists.sort(Module.definitions_in(module, :defmacro))
-
-        struct =
-          case Module.get_attribute(module, :__struct__) do
-            %{} = entry -> {entry, List.wrap(Module.get_attribute(module, :enforce_keys))}
-            _ -> nil
-          end
-
-        {defs, defmacros, struct} |> :erlang.term_to_binary() |> :erlang.md5()
-
-      true ->
-        nil
-    end
-  end
-
   defp remove_and_purge(beam, module) do
     _ = File.rm(beam)
-    :code.purge(module)
-    :code.delete(module)
+
+    if Code.loaded?(module) do
+      :code.purge(module)
+      :code.delete(module)
+    end
   end
 
   defp purge_modules_in_path(path) do
     with {:ok, beams} <- File.ls(path) do
       Enum.each(beams, fn beam ->
         module = beam |> Path.rootname() |> String.to_atom()
-        :code.purge(module)
-        :code.delete(module)
+
+        if Code.loaded?(module) do
+          :code.purge(module)
+          :code.delete(module)
+        end
       end)
     end
   end
@@ -879,7 +880,7 @@ defmodule Mix.Compilers.Elixir do
 
   ## Manifest handling
 
-  @default_manifest {%{}, %{}, %{}, [], nil, nil, 0, 0}
+  @default_manifest {%{}, %{}, %{}, [], nil, nil, 0, 0, {%{}, %{}}}
 
   # Similar to read_manifest, but for internal consumption and with data migration support.
   defp parse_manifest(manifest, compile_path) do
@@ -890,9 +891,9 @@ defmodule Mix.Compilers.Elixir do
         @default_manifest
     else
       {@manifest_vsn, modules, sources, local_exports, parent, cache_key, deps_config,
-       project_mtime, config_mtime} ->
+       project_mtime, config_mtime, protocols_and_impls} ->
         {modules, sources, local_exports, parent, cache_key, deps_config, project_mtime,
-         config_mtime}
+         config_mtime, protocols_and_impls}
 
       # {vsn, %{module => record}, sources, ...} v22-?
       # {vsn, [module_record], sources, ...} v5-v21
@@ -922,9 +923,7 @@ defmodule Mix.Compilers.Elixir do
         end
 
       for {module, _} <- data do
-        File.rm(beam_path(compile_path, module))
-        :code.purge(module)
-        :code.delete(module)
+        remove_and_purge(beam_path(compile_path, module), module)
       end
     rescue
       _ ->
@@ -946,28 +945,24 @@ defmodule Mix.Compilers.Elixir do
          deps_config,
          project_mtime,
          config_mtime,
+         protocols_and_impls,
          timestamp
        ) do
-    if modules == %{} and sources == %{} do
-      File.rm(manifest)
-    else
-      File.mkdir_p!(Path.dirname(manifest))
+    File.mkdir_p!(Path.dirname(manifest))
 
-      term =
-        {@manifest_vsn, modules, sources, exports, parents, cache_key, deps_config, project_mtime,
-         config_mtime}
+    term =
+      {@manifest_vsn, modules, sources, exports, parents, cache_key, deps_config, project_mtime,
+       config_mtime, protocols_and_impls}
 
-      manifest_data = :erlang.term_to_binary(term, [:compressed])
-      File.write!(manifest, manifest_data)
-      File.touch!(manifest, timestamp)
-      delete_checkpoint(manifest)
-
-      # Since Elixir is a dependency itself, we need to touch the lock
-      # so the current Elixir version, used to compile the files above,
-      # is properly stored.
-      Mix.Dep.ElixirSCM.update()
-    end
+    manifest_data = :erlang.term_to_binary(term, [:compressed])
+    File.write!(manifest, manifest_data)
+    File.touch!(manifest, timestamp)
+    delete_checkpoints(manifest)
 
+    # Since Elixir is a dependency itself, we need to touch the lock
+    # so the current Elixir version, used to compile the files above,
+    # is properly stored.
+    Mix.Dep.ElixirSCM.update(Mix.Project.config()[:build_scm])
     :ok
   end
 
@@ -975,45 +970,31 @@ defmodule Mix.Compilers.Elixir do
     Path.join(compile_path, Atom.to_string(module) <> ".beam")
   end
 
-  # Once we added semantic recompilation, the following can happen:
-  #
-  # 1. The user changes config/mix.exs/__mix_recompile__?
-  # 2. We detect the change, remove .beam files and start recompilation
-  # 3. Recompilation fails
-  # 4. The user reverts the change
-  # 5. The compiler no longer recompiles and the .beam files are missing
-  #
-  # Therefore, it is important for us to checkpoint any state that may
-  # have lead to a compilation and which can now be reverted.
-
-  defp parse_checkpoint(manifest) do
+  defp parse_checkpoint(type, manifest) when type in [:update, :verify] do
     try do
-      (manifest <> ".checkpoint") |> File.read!() |> :erlang.binary_to_term()
+      (manifest <> ".#{type}.cp") |> File.read!() |> :erlang.binary_to_term()
     rescue
-      _ ->
-        {%{}, %{}, %{}}
+      _ -> :error
     else
-      {@checkpoint_vsn, stale_modules, stale_exports, recompile_modules} ->
-        {stale_modules, stale_exports, recompile_modules}
-
-      _ ->
-        {%{}, %{}, %{}}
+      {@checkpoint_vsn, data} -> {:ok, data}
+      _ -> :error
     end
   end
 
-  defp write_checkpoint(manifest, stale_modules, stale_exports, recompile_modules) do
+  defp write_checkpoint(type, manifest, data) when type in [:update, :verify] do
     File.mkdir_p!(Path.dirname(manifest))
-    term = {@checkpoint_vsn, stale_modules, stale_exports, recompile_modules}
-    checkpoint_data = :erlang.term_to_binary(term, [:compressed])
-    File.write!(manifest <> ".checkpoint", checkpoint_data)
+    term = {@checkpoint_vsn, data}
+    checkpoint_data = :erlang.term_to_binary(term)
+    File.write!(manifest <> ".#{type}.cp", checkpoint_data)
   end
 
-  defp delete_checkpoint(manifest) do
-    File.rm(manifest <> ".checkpoint")
+  defp delete_checkpoints(manifest) do
+    File.rm(manifest <> ".update.cp")
+    File.rm(manifest <> ".verify.cp")
   end
 
-  defp unless_previous_warnings_as_errors(previous_warnings, opts, {status, all_warnings}) do
-    if previous_warnings != [] and opts[:warnings_as_errors] do
+  defp unless_warnings_as_errors(opts, {status, all_warnings}) do
+    if all_warnings != [] and opts[:warnings_as_errors] do
       message = "Compilation failed due to warnings while using the --warnings-as-errors option"
       IO.puts(:stderr, message)
       {:error, all_warnings}
@@ -1043,17 +1024,25 @@ defmodule Mix.Compilers.Elixir do
   ## Compiler loop
   # The compiler is invoked in a separate process so we avoid blocking its main loop.
 
-  defp compiler_loop(stale, stale_modules, dest, timestamp, opts, state) do
+  defp compiler_loop(manifest, stale, stale_modules, dest, timestamp, opts, state) do
     ref = make_ref()
     parent = self()
-    threshold = opts[:long_compilation_threshold] || 10
+    compilation_threshold = opts[:long_compilation_threshold] || 10
+    verification_threshold = opts[:long_verification_threshold] || 10
     profile = opts[:profile]
-    verbose = opts[:verbose] || false
-    warnings_as_errors = opts[:warnings_as_errors] || false
+    verbose = Keyword.get(opts, :verbose, false)
+    verification = Keyword.get(opts, :verification, true)
+
+    if not verification and not Mix.debug?() do
+      Mix.shell().error("--no-verification flag is only recommended with MIX_DEBUG=1")
+    end
 
     pid =
       spawn_link(fn ->
         compile_opts = [
+          after_compile: fn ->
+            compiler_call(parent, ref, {:after_compile, manifest, opts})
+          end,
           each_cycle: fn ->
             compiler_call(parent, ref, {:each_cycle, stale_modules, dest, timestamp})
           end,
@@ -1063,16 +1052,24 @@ defmodule Mix.Compilers.Elixir do
           each_module: fn file, module, _binary ->
             compiler_call(parent, ref, {:each_module, file, module, System.os_time(:second)})
           end,
-          each_long_compilation: fn file ->
+          each_long_compilation: fn file, pid ->
             Mix.shell().info(
-              "Compiling #{Path.relative_to(file, File.cwd!())} (it's taking more than #{threshold}s)"
+              "Compiling #{Path.relative_to(file, File.cwd!())} " <>
+                "(it's taking more than #{compilation_threshold}s)#{debug_stacktrace(pid)}"
             )
           end,
-          long_compilation_threshold: threshold,
-          profile: profile,
+          each_long_verification: fn module, pid ->
+            Mix.shell().info(
+              "Verifying #{inspect(module)} " <>
+                "(it's taking more than #{verification_threshold}s)#{debug_stacktrace(pid)}"
+            )
+          end,
+          long_compilation_threshold: compilation_threshold,
+          long_verification_threshold: verification_threshold,
           beam_timestamp: timestamp,
           return_diagnostics: true,
-          warnings_as_errors: warnings_as_errors
+          profile: profile,
+          verification: verification
         ]
 
         response = Kernel.ParallelCompiler.compile_to_path(stale, dest, compile_opts)
@@ -1092,6 +1089,11 @@ defmodule Mix.Compilers.Elixir do
 
   defp compiler_loop(ref, pid, state, cwd) do
     receive do
+      {^ref, {:after_compile, manifest, opts}} ->
+        {response, state} = after_compile(manifest, state, opts)
+        send(pid, {ref, response})
+        compiler_loop(ref, pid, state, cwd)
+
       {^ref, {:each_cycle, stale_modules, dest, timestamp}} ->
         {response, state} = each_cycle(stale_modules, dest, timestamp, state)
         send(pid, {ref, response})
@@ -1108,9 +1110,13 @@ defmodule Mix.Compilers.Elixir do
         # Read the relevant module information and unblock the compiler
         kind = detect_kind(module)
         external = Module.get_attribute(module, :external_resource)
-        new_export = exports_md5(module, true)
+        new_export = Module.get_attribute(module, :exports_md5)
+        recompile? = Module.defines?(module, {:__mix_recompile__?, 0}, :def)
         send(pid, {ref, :ok})
-        state = each_module(file, module, kind, external, new_export, state, timestamp, cwd)
+
+        state =
+          each_module(file, module, kind, external, new_export, recompile?, state, timestamp, cwd)
+
         compiler_loop(ref, pid, state, cwd)
 
       {^ref, {:ok, _modules, info}} ->
@@ -1121,14 +1127,45 @@ defmodule Mix.Compilers.Elixir do
     end
   end
 
-  defp each_cycle(stale_modules, compile_path, timestamp, state) do
-    {modules, _exports, sources, changed, pending_modules, stale_exports} = state
+  defp after_compile(manifest, state, opts) do
+    %{modules: modules, pending_modules: pending_modules, consolidation: consolidation} = state
+    write_checkpoint(:verify, manifest, Map.keys(modules))
+    consolidation = maybe_consolidate(consolidation, modules, pending_modules, opts)
+    {:ok, %{state | consolidation: consolidation}}
+  end
 
+  defp each_cycle(stale_modules, dest, timestamp, state) do
+    %{
+      modules: modules,
+      sources: sources,
+      changed: changed,
+      pending_modules: pending_modules,
+      stale_exports: stale_exports
+    } = state
+
+    # At this point, we may have additional files to compile.
+    # There are two potential sources:
+    #
+    # * We need to go through all exports that we have confirmed that changed.
+    #   When we first compile, we store all removed/changed modules as stale
+    #   exports. Then, if they are not compiled again, or compiled with a
+    #   different exports MD5, they remain as stale, causing the next cycle.
+    #
+    # * In case a module is defined in two places, we add all sources to changed
     {pending_modules, exports, changed} =
-      update_stale_entries(pending_modules, sources, changed, %{}, stale_exports, compile_path)
+      update_stale_entries(pending_modules, sources, changed, %{}, stale_exports, dest)
 
-    # For each changed file, mark it as changed.
-    # If compilation fails mid-cycle, they will be picked next time around.
+    state = %{
+      state
+      | changed: [],
+        exports: exports,
+        pending_modules: pending_modules
+    }
+
+    # Those files have been changed transitively, so we mark them as changed
+    # in case compilation fails mid-cycle. The combination of the outdated
+    # timestamp plus the missing BEAM files (which were removed in
+    # update_stale_entries above) will cause them to be recompiled next time.
     for file <- changed do
       File.touch!(file, timestamp)
     end
@@ -1163,36 +1200,30 @@ defmodule Mix.Compilers.Elixir do
         end)
 
       runtime_paths =
-        Enum.map(runtime_modules, &{&1, Path.join(compile_path, Atom.to_string(&1) <> ".beam")})
+        Enum.map(runtime_modules, &{&1, Path.join(dest, Atom.to_string(&1) <> ".beam")})
 
-      state = {modules, exports, sources, [], pending_modules, stale_exports}
+      state = %{state | sources: sources}
       {{:runtime, runtime_paths, []}, state}
     else
       Mix.Utils.compiling_n(length(changed), :ex)
 
-      # If we have a compile time dependency to a module, as soon as its file
-      # change, we will detect the compile time dependency and recompile. However,
-      # the whole goal of pending exports is to delay this decision, so we need to
-      # track which modules were removed and start them as our pending exports and
-      # remove the pending exports as we notice they have not gone stale.
-      {sources, removed_modules} =
-        Enum.reduce(changed, {sources, %{}}, fn file, {acc_sources, acc_modules} ->
+      # Now we need to detect the new stale_exports.
+      # This is a simplified version of update_stale_sources.
+      {sources, %{}} =
+        Enum.reduce(changed, {sources, stale_exports}, fn file, {acc_sources, acc_modules} ->
           source(size: size, digest: digest, modules: modules) = Map.fetch!(acc_sources, file)
           acc_modules = Enum.reduce(modules, acc_modules, &Map.put(&2, &1, true))
-
-          # Define empty records for the sources that needs
-          # to be recompiled (but were not changed on disk)
           {Map.replace!(acc_sources, file, source(size: size, digest: digest)), acc_modules}
         end)
 
-      state = {modules, exports, sources, [], pending_modules, removed_modules}
+      state = %{state | sources: sources, stale_exports: stale_exports}
       {{:compile, changed, []}, state}
     end
   end
 
   defp each_file(file, references, verbose, state, cwd) do
     {compile_references, export_references, runtime_references, compile_env} = references
-    {modules, exports, sources, changed, pending_modules, stale_exports} = state
+    %{sources: sources} = state
 
     file = Path.relative_to(file, cwd)
 
@@ -1219,12 +1250,18 @@ defmodule Mix.Compilers.Elixir do
         compile_env: compile_env
       )
 
-    sources = Map.replace!(sources, file, source)
-    {modules, exports, sources, changed, pending_modules, stale_exports}
+    %{state | sources: Map.replace!(sources, file, source)}
   end
 
-  defp each_module(file, module, kind, external, new_export, state, timestamp, cwd) do
-    {modules, exports, sources, changed, pending_modules, stale_exports} = state
+  defp each_module(file, module, kind, external, new_export, recompile?, state, timestamp, cwd) do
+    %{
+      modules: modules,
+      exports: exports,
+      sources: sources,
+      changed: changed,
+      pending_modules: pending_modules,
+      stale_exports: stale_exports
+    } = state
 
     file = Path.relative_to(file, cwd)
     external = process_external_resources(external, cwd)
@@ -1263,7 +1300,7 @@ defmodule Mix.Compilers.Elixir do
         sources: module_sources,
         export: new_export,
         timestamp: timestamp,
-        recompile?: function_exported?(module, :__mix_recompile__?, 0)
+        recompile?: recompile?
       )
 
     modules = Map.put(modules, module, entry)
@@ -1277,7 +1314,7 @@ defmodule Mix.Compilers.Elixir do
         %{} -> changed
       end
 
-    {modules, exports, sources, changed, pending_modules, stale_exports}
+    %{state | modules: modules, changed: changed, sources: sources, stale_exports: stale_exports}
   end
 
   defp detect_kind(module) do
@@ -1306,4 +1343,61 @@ defmodule Mix.Compilers.Elixir do
       {Path.relative_to(file, cwd), Mix.Utils.last_modified_and_size(file), digest}
     end
   end
+
+  defp debug_stacktrace(pid) do
+    with true <- Mix.debug?(),
+         {:current_stacktrace, stacktrace} <- Process.info(pid, :current_stacktrace) do
+      [?\n, Exception.format_stacktrace(stacktrace)]
+    else
+      _ -> ""
+    end
+  end
+
+  ## Consolidation
+
+  @doc """
+  Returns protocols and implementations for the given `manifest`.
+  """
+  def protocols_and_impls_from_paths(paths) do
+    Enum.reduce(paths, protocols_and_impls(), fn path, acc ->
+      {modules, _} = read_manifest(Path.join(path, ".mix/compile.elixir"))
+      protocols_and_impls_from_modules(modules, acc)
+    end)
+  end
+
+  defp protocols_and_impls_from_modules(modules, protocols_and_impls) do
+    Enum.reduce(modules, protocols_and_impls, fn
+      {module, module(kind: kind, timestamp: timestamp)}, {protocols, impls} ->
+        case kind do
+          :protocol -> {Map.put(protocols, module, timestamp), impls}
+          {:impl, protocol} -> {protocols, Map.put(impls, module, protocol)}
+          _ -> {protocols, impls}
+        end
+    end)
+  end
+
+  defp protocols_and_impls(), do: {%{}, %{}}
+
+  defp maybe_consolidate({:off, _, _}, _, _, _) do
+    protocols_and_impls()
+  end
+
+  defp maybe_consolidate(
+         {on_or_force, old_protocols_and_impls, protocols_and_impls},
+         modules,
+         pending_modules,
+         opts
+       ) do
+    protocols_and_impls = protocols_and_impls_from_modules(modules, protocols_and_impls)
+    protocols_and_impls = protocols_and_impls_from_modules(pending_modules, protocols_and_impls)
+
+    Mix.Compilers.Protocol.compile(
+      on_or_force == :force,
+      old_protocols_and_impls,
+      protocols_and_impls,
+      opts
+    )
+
+    protocols_and_impls
+  end
 end
