@@ -1,8 +1,18 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Generator do
   @moduledoc """
   Conveniences for working with paths and generating content.
   """
 
+  @type generator_opts :: [
+          force: boolean(),
+          quiet: boolean(),
+          format_elixir: boolean()
+        ]
+
   @doc ~S"""
   Creates a file with the given contents.
 
@@ -13,6 +23,7 @@ defmodule Mix.Generator do
 
     * `:force` - forces creation without a shell prompt
     * `:quiet` - does not log command output
+    * `:format_elixir` (since v1.18.0) - if `true`, apply formatter to the generated file
 
   ## Examples
 
@@ -21,7 +32,7 @@ defmodule Mix.Generator do
       true
 
   """
-  @spec create_file(Path.t(), iodata, keyword) :: boolean()
+  @spec create_file(Path.t(), iodata, generator_opts) :: boolean()
   def create_file(path, contents, opts \\ []) when is_binary(path) do
     log(:green, :creating, Path.relative_to_cwd(path), opts)
 
@@ -58,7 +69,7 @@ defmodule Mix.Generator do
       true
 
   """
-  @spec create_directory(Path.t(), keyword) :: true
+  @spec create_directory(Path.t(), quiet: boolean()) :: true
   def create_directory(path, options \\ []) when is_binary(path) do
     log(:green, "creating", Path.relative_to_cwd(path), options)
     File.mkdir_p!(path)
@@ -75,6 +86,7 @@ defmodule Mix.Generator do
 
     * `:force` - forces copying without a shell prompt
     * `:quiet` - does not log command output
+    * `:format_elixir` (since v1.18.0) - if `true`, apply formatter to the generated file
 
   ## Examples
 
@@ -84,7 +96,7 @@ defmodule Mix.Generator do
 
   """
   @doc since: "1.9.0"
-  @spec copy_file(Path.t(), Path.t(), keyword) :: boolean()
+  @spec copy_file(Path.t(), Path.t(), generator_opts) :: boolean()
   def copy_file(source, target, options \\ []) do
     create_file(target, File.read!(source), options)
   end
@@ -112,7 +124,7 @@ defmodule Mix.Generator do
 
   """
   @doc since: "1.9.0"
-  @spec copy_template(Path.t(), Path.t(), keyword, keyword) :: boolean()
+  @spec copy_template(Path.t(), Path.t(), keyword, generator_opts) :: boolean()
   def copy_template(source, target, assigns, options \\ []) do
     create_file(target, EEx.eval_file(source, assigns: assigns), options)
   end
@@ -160,7 +172,7 @@ defmodule Mix.Generator do
   end
 
   defp log(color, command, message, opts) do
-    unless opts[:quiet] do
+    if !opts[:quiet] do
       Mix.shell().info([color, "* #{command} ", :reset, message])
     end
   end
