@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Project do
   @moduledoc """
   Defines and manipulates Mix projects.
@@ -126,6 +130,67 @@ defmodule Mix.Project do
   makes sure Elixir is not added as a dependency to the generated `.app` file or
   to the escript generated with `mix escript.build`, and so on.
 
+  ## Umbrella projects
+
+  Umbrella projects are a convenience to help you organize and manage multiple
+  applications. While it provides a degree of separation between applications,
+  those applications are not fully decoupled, as they share the same configuration
+  and the same dependencies.
+
+  In an umbrella project, you have an `apps/` folder where you store each application.
+  Then, instead of each app in the umbrella having its own configuration, build cache,
+  lockfile and so, they all point to the parent project by specifying the following
+  configuration in their `mix.exs`:
+
+      build_path: "../../_build",
+      config_path: "../../config/config.exs",
+      deps_path: "../../deps",
+      lockfile: "../../mix.lock",
+
+  The pattern of keeping multiple applications in the same repository is known as
+  [monorepo](https://en.wikipedia.org/wiki/Monorepo). Umbrella projects maximize
+  this pattern by providing conveniences to compile, test and run multiple
+  applications at once. When an umbrella application needs to depend on another
+  one, it can be done by passing the `in_umbrella: true` option to your dependency.
+  If an umbrella application `:foo` depends on its sibling `:bar`, you can specify
+  this dependency in `foo`'s `mix.exs` file as:
+
+      {:bar, in_umbrella: true}
+
+  ### Undoing umbrellas
+
+  Using umbrella projects can impact how you design and write your software and,
+  as time passes, they may turn out to be the wrong choice.
+  If you find yourself in a position where you want to use different configurations
+  in each application for the same dependency or use different dependency versions,
+  then it is likely your codebase has grown beyond what umbrellas can provide.
+
+  If you find yourself in this situation, you have two options:
+
+    1. Convert everything into a single Mix project, which can be done in steps.
+       First move all files in `lib`, `test`, `priv`, and friends into a single
+       application, while still keeping the overall umbrella structure and
+       `mix.exs` files. For example, if your umbrellas has three applications,
+       `foo`, `bar` and `baz`, where `baz` depends on both `foo` and `bar`,
+       move all source to `baz`. Then remove `foo` and `bar` one by one,
+       updating any configuration and removing references to the `:foo` and
+      `:bar` application names. Until you have only a single application.
+
+    2. Remove umbrella structure while keeping them as distinct applications.
+       This is done by moving applications outside of the umbrella
+       project's `apps/` directory and updating the projects' `mix.exs` files
+       to no longer set the `build_path`, `config_path`, `deps_path`, and
+       `lockfile` configurations, guaranteeing each of them have their own
+       build and dependency structure.
+
+  Keep in mind that umbrellas are one of many options for managing private
+  packages within your organization. You might:
+
+    1. Have multiple directories inside the same repository and using `:path`
+       dependencies (which is essentially the monorepo pattern)
+    2. Use private Git repositories and Mix' ability to fetch Git dependencies
+    3. Publishing packages to a private [Hex.pm](https://hex.pm/) organization
+
   ## Invoking this module
 
   This module contains many functions that return project information and
@@ -154,6 +219,22 @@ defmodule Mix.Project do
 
   """
 
+  @type build_structure_opts :: [
+          symlink_ebin: boolean(),
+          source: String.t()
+        ]
+
+  @typedoc """
+  Options for dependency traversal functions.
+
+  These options control how dependency trees are traversed and filtered
+  in functions like `deps_scms/1`, `deps_paths/1`, and `deps_tree/1`.
+  """
+  @type deps_traversal_opts :: [
+          depth: pos_integer(),
+          parents: [atom()]
+        ]
+
   @doc false
   defmacro __using__(_) do
     quote do
@@ -536,7 +617,7 @@ defmodule Mix.Project do
 
   """
   @doc since: "1.10.0"
-  @spec deps_scms(keyword) :: %{optional(atom) => Mix.SCM.t()}
+  @spec deps_scms(deps_traversal_opts) :: %{optional(atom) => Mix.SCM.t()}
   def deps_scms(opts \\ []) when is_list(opts) do
     traverse_deps(opts, fn %{scm: scm} -> scm end)
   end
@@ -557,7 +638,7 @@ defmodule Mix.Project do
       #=> %{foo: "deps/foo", bar: "custom/path/dep"}
 
   """
-  @spec deps_paths(keyword) :: %{optional(atom) => Path.t()}
+  @spec deps_paths(deps_traversal_opts) :: %{optional(atom) => Path.t()}
   def deps_paths(opts \\ []) when is_list(opts) do
     traverse_deps(opts, fn %{opts: opts} -> opts[:dest] end)
   end
@@ -579,7 +660,7 @@ defmodule Mix.Project do
 
   """
   @doc since: "1.15.0"
-  @spec deps_tree(keyword) :: %{optional(atom) => [atom]}
+  @spec deps_tree(deps_traversal_opts) :: %{optional(atom) => [atom]}
   def deps_tree(opts \\ []) when is_list(opts) do
     traverse_deps(opts, fn %{deps: deps} -> Enum.map(deps, & &1.app) end)
   end
@@ -652,26 +733,23 @@ defmodule Mix.Project do
   (which defaults to `"_build"`) and a subdirectory. The subdirectory
   is built based on two factors:
 
-    * If `:build_per_environment` is set, the subdirectory is the value
-      of `Mix.env/0` (which can be set via `MIX_ENV`). Otherwise it is
-      set to "shared".
+    * If `:build_per_environment` is set (the default), the subdirectory
+      is the value of `Mix.env/0` (which can be set via `MIX_ENV`).
+      Otherwise it is set to "shared".
 
     * If `Mix.target/0` is set (often via the `MIX_TARGET` environment
       variable), it will be used as a prefix to the subdirectory.
 
-  Finally, the environment variables `MIX_BUILD_ROOT` and `MIX_BUILD_PATH`
-  can be used to change the result of this function. `MIX_BUILD_ROOT`
-  overwrites only the root `"_build"` directory while keeping the
-  subdirectory as is. It may be useful to change it for caching reasons,
-  typically during Continuous Integration (CI). `MIX_BUILD_PATH` overrides
-  the build path altogether and it typically used by other build tools
-  that invoke the `mix` CLI.
+  The behaviour of this function can be modified by two environment
+  variables, `MIX_BUILD_ROOT` and `MIX_BUILD_PATH`, see [the Mix
+  documentation for more information](Mix.html#environment-variables).
 
   > #### Naming differences {: .info}
   >
   > Ideally the configuration option `:build_path` would be called
-  > `:build_root`, as it would fully mirror the environment variable.
-  > However, its name is preserved for backwards compatibility.
+  > `:build_root`, as it only sets the root component of the build
+  > path but not the subdirectory. However, its name is preserved
+  > for backwards compatibility.
 
   ## Examples
 
@@ -842,8 +920,11 @@ defmodule Mix.Project do
 
     * `:symlink_ebin` - symlink ebin instead of copying it
 
+    * `:source` - the source directory to copy from.
+      Defaults to the current working directory.
+
   """
-  @spec build_structure(keyword, keyword) :: :ok
+  @spec build_structure(keyword, build_structure_opts) :: :ok
   def build_structure(config \\ config(), opts \\ []) do
     source = opts[:source] || File.cwd!()
     target = app_path(config)
@@ -877,7 +958,7 @@ defmodule Mix.Project do
 
   `opts` are the same options that can be passed to `build_structure/2`.
   """
-  @spec ensure_structure(keyword, keyword) :: :ok
+  @spec ensure_structure(keyword, build_structure_opts) :: :ok
   def ensure_structure(config \\ config(), opts \\ []) do
     if File.exists?(app_path(config)) do
       :ok
@@ -983,7 +1064,6 @@ defmodule Mix.Project do
   defp default_config do
     [
       aliases: [],
-      build_embedded: false,
       build_per_environment: true,
       build_scm: Mix.SCM.Path,
       config_path: "config/config.exs",
