@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Release do
   @moduledoc """
   Defines the release structure and convenience for assembling releases.
@@ -64,6 +68,23 @@ defmodule Mix.Release do
           steps: [(t -> t) | :assemble, ...]
         }
 
+  @typedoc """
+  Options for stripping BEAM files.
+  """
+  @type strip_beam_opts :: [
+          keep: [String.t()],
+          compress: boolean()
+        ]
+
+  @typedoc """
+  Erlang/OTP sys.config structure.
+
+  A list of tuples where each tuple contains an application name and its
+  configuration as a keyword list. This is the standard format for Erlang
+  application configuration.
+  """
+  @type sys_config :: [{application(), keyword()}]
+
   @default_apps [kernel: :permanent, stdlib: :permanent, elixir: :permanent, sasl: :permanent]
   @safe_modes [:permanent, :temporary, :transient]
   @unsafe_modes [:load, :none]
@@ -75,7 +96,7 @@ defmodule Mix.Release do
   def from_config!(name, config, overrides) do
     {name, apps, opts} = find_release(name, config)
 
-    unless Atom.to_string(name) =~ ~r/^[a-z][a-z0-9_]*$/ do
+    if not (Atom.to_string(name) =~ ~r/^[a-z][a-z0-9_]*$/) do
       Mix.raise(
         "Invalid release name. A release name must start with a lowercase ASCII letter, " <>
           "followed by lowercase ASCII letters, numbers, or underscores, got: #{inspect(name)}"
@@ -153,7 +174,7 @@ defmodule Mix.Release do
       version_path: Path.join([path, "releases", version]),
       erts_source: erts_source,
       erts_version: erts_version,
-      applications: loaded_apps,
+      applications: Map.delete(loaded_apps, :erts),
       boot_scripts: %{start: start_boot, start_clean: start_clean_boot},
       config_providers: config_providers,
       options: opts,
@@ -441,7 +462,7 @@ defmodule Mix.Release do
 
   In case there are no config providers, it doesn't change `sys_config`.
   """
-  @spec make_sys_config(t, keyword(), Config.Provider.config_path()) ::
+  @spec make_sys_config(t, sys_config(), Config.Provider.config_path()) ::
           :ok | {:error, String.t()}
   def make_sys_config(release, sys_config, config_provider_path) do
     {sys_config, runtime_config?} =
@@ -851,10 +872,13 @@ defmodule Mix.Release do
         source_file = Path.join(source, file)
         target_file = Path.join(target, file)
 
-        with true <- is_list(strip_options) and String.ends_with?(file, ".beam"),
-             {:ok, binary} <- strip_beam(File.read!(source_file), strip_options) do
-          File.write!(target_file, binary)
-        else
+        case Path.extname(file) do
+          ".beam" ->
+            process_beam_file(source_file, target_file, strip_options)
+
+          ".app" ->
+            process_app_file(source_file, target_file)
+
           _ ->
             # Use File.cp!/3 to preserve file mode for any executables stored
             # in the ebin directory.
@@ -868,6 +892,23 @@ defmodule Mix.Release do
     end
   end
 
+  defp process_beam_file(source_file, target_file, strip_options) do
+    with true <- is_list(strip_options),
+         {:ok, binary} <- strip_beam(File.read!(source_file), strip_options) do
+      File.write!(target_file, binary)
+    else
+      _ -> File.cp!(source_file, target_file)
+    end
+  end
+
+  defp process_app_file(source_file, target_file) do
+    with {:ok, [{:application, app, info}]} <- :file.consult(source_file) do
+      File.write!(target_file, :io_lib.format("~tp.~n", [{:application, app, info}]))
+    else
+      _ -> File.cp!(source_file, target_file)
+    end
+  end
+
   @doc """
   Strips a beam file for a release.
 
@@ -876,8 +917,25 @@ defmodule Mix.Release do
 
   The exact chunks that are kept are not documented and may change in
   future versions.
+
+  ## Options
+
+    * `:keep` - a list of additional chunk names (as strings) to keep in the
+      stripped BEAM file beyond those required by Erlang/Elixir
+
+    * `:compress` - when `true`, the resulting BEAM file will be compressed
+      using gzip. Defaults to `false`
+
+  ## Examples
+
+      # Strip with default options
+      Mix.Release.strip_beam(beam_binary)
+
+      # Keep additional chunks and compress
+      Mix.Release.strip_beam(beam_binary, keep: ["Docs", "ChunkName"], compress: true)
+
   """
-  @spec strip_beam(binary(), keyword()) :: {:ok, binary()} | {:error, :beam_lib, term()}
+  @spec strip_beam(binary(), strip_beam_opts()) :: {:ok, binary()} | {:error, :beam_lib, term()}
   def strip_beam(binary, options \\ []) when is_list(options) do
     chunks_to_keep = options[:keep] |> List.wrap() |> Enum.map(&String.to_charlist/1)
     all_chunks = Enum.uniq(@additional_chunks ++ :beam_lib.significant_chunks() ++ chunks_to_keep)
