@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Compile.All do
   use Mix.Task.Compiler
 
@@ -12,12 +16,19 @@ defmodule Mix.Tasks.Compile.All do
   @impl true
   def run(args) do
     Mix.Project.get!()
+
     config = Mix.Project.config()
 
+    Mix.Project.with_build_lock(config, fn ->
+      do_run(config, args)
+    end)
+  end
+
+  defp do_run(config, args) do
     # Compute the app cache if it is stale and we are
     # not compiling from a dependency.
     app_cache =
-      unless "--from-mix-deps-compile" in args do
+      if "--from-mix-deps-compile" not in args do
         Mix.AppLoader.stale_cache(config)
       end
 
@@ -43,13 +54,15 @@ defmodule Mix.Tasks.Compile.All do
       Code.delete_paths(current_paths -- loaded_paths)
     end
 
+    Code.prepend_paths(loaded_paths -- current_paths, cache: true)
+
     # Add the current compilation path. compile.elixir and compile.erlang
     # will also add this path, but only if they run, so we always add it
     # here too. Furthermore, we don't cache it as we may still write to it.
     compile_path = to_charlist(Mix.Project.compile_path())
-    Code.prepend_paths([compile_path | loaded_paths -- current_paths], cache: true)
+    Code.prepend_path(compile_path)
 
-    result =
+    {status, diagnostics} =
       if "--no-compile" in args do
         Mix.Task.reenable("compile.all")
         {:noop, []}
@@ -58,15 +71,19 @@ defmodule Mix.Tasks.Compile.All do
         Mix.Project.build_structure(config)
 
         config
-        |> Mix.Tasks.Compile.compilers()
-        |> compile(args, :noop, [])
+        |> Mix.Task.Compiler.compilers()
+        |> Mix.Task.Compiler.run(args)
       end
 
+    if status == :error and "--return-errors" not in args do
+      exit({:shutdown, 1})
+    end
+
     if app_cache do
       Mix.AppLoader.write_cache(app_cache, Map.new(loaded_modules))
     end
 
-    unless "--no-app-loading" in args do
+    if "--no-app-loading" not in args do
       app = config[:app]
 
       with {:ok, properties} <- Mix.AppLoader.load_app(app, "#{compile_path}/#{app}.app"),
@@ -77,38 +94,9 @@ defmodule Mix.Tasks.Compile.All do
       end
     end
 
-    result
-  end
-
-  defp compile([], _, status, diagnostics) do
     {status, diagnostics}
   end
 
-  defp compile([compiler | rest], args, status, diagnostics) do
-    {new_status, new_diagnostics} = run_compiler(compiler, args)
-    diagnostics = diagnostics ++ new_diagnostics
-
-    case new_status do
-      :error ->
-        if "--return-errors" not in args do
-          exit({:shutdown, 1})
-        end
-
-        {:error, diagnostics}
-
-      :ok ->
-        compile(rest, args, :ok, diagnostics)
-
-      :noop ->
-        compile(rest, args, status, diagnostics)
-    end
-  end
-
-  defp run_compiler(compiler, args) do
-    result = Mix.Task.Compiler.normalize(Mix.Task.run("compile.#{compiler}", args), compiler)
-    Enum.reduce(Mix.ProjectStack.pop_after_compiler(compiler), result, & &1.(&2))
-  end
-
   def project_apps(config) do
     project = Mix.Project.get!()
 
