@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Compile.App do
   use Mix.Task.Compiler
 
@@ -96,8 +100,8 @@ defmodule Mix.Tasks.Compile.App do
 
     * `:reliable_dir_mtime` - this task relies on the operating system
       changing the mtime on a directory whenever a file is added or removed.
-      You can set this option to false if your system does not provide
-      reliable mtimes. Defaults to false on Windows.
+      You can set this option to `false` if your system does not provide
+      reliable mtimes. Defaults to `false` on Windows.
 
   ## Phases
 
