@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Compile do
   use Mix.Task
 
@@ -16,16 +20,11 @@ defmodule Mix.Tasks.Compile do
 
   ## Configuration
 
-    * `:build_embedded` - this option was used to copy all code and
-      priv content to the `_build` directory. However, this option no
-      longer has an effect as Elixir will now copy those at release time
-
     * `:compilers` - compilers to run, defaults to `Mix.compilers/0`,
       which are `[:erlang, :elixir, :app]`.
 
     * `:consolidate_protocols` - when `true`, runs protocol
-      consolidation via the `mix compile.protocols` task. The default
-      value is `true`.
+      consolidation after compiling Elixir. The default value is `true`.
 
     * `:build_path` - the directory where build artifacts
       should be written to. This option is intended only for
@@ -62,6 +61,8 @@ defmodule Mix.Tasks.Compile do
     * `--erl-config` - path to an Erlang term file that will be loaded as Mix config
     * `--force` - forces compilation
     * `--list` - lists all enabled compilers
+    * `--listeners` - starts Mix listeners (they are started by default,
+      unless `--no-listeners` or `--no-deps-check` are given)
     * `--no-app-loading` - does not load .app resource file after compilation
     * `--no-archives-check` - skips checking of archives
     * `--no-compile` - does not actually compile, only loads code and perform checks
@@ -141,7 +142,12 @@ defmodule Mix.Tasks.Compile do
     {res, diagnostics} =
       Mix.Task.run("compile.all", args)
       |> List.wrap()
-      |> Enum.map(&Mix.Task.Compiler.normalize(&1, :all))
+      |> Enum.map(
+        &case &1 do
+          :noop -> {:noop, []}
+          {status, diagnostics} -> {status, diagnostics}
+        end
+      )
       |> Enum.reduce({:noop, []}, &merge_diagnostics/2)
 
     config = Mix.Project.config()
@@ -158,26 +164,22 @@ defmodule Mix.Tasks.Compile do
       Code.prepend_paths(loaded_paths -- :code.get_path())
     end
 
-    consolidate_protocols? =
-      config[:consolidate_protocols] and "--no-protocol-consolidation" not in args
-
     res =
       cond do
         "--no-compile" in args ->
           Mix.Task.reenable("compile")
           :noop
 
-        consolidate_protocols? and reconsolidate_protocols?(res) ->
-          Mix.Task.run("compile.protocols", args)
-          :ok
+        Mix.Project.umbrella?(config) ->
+          Mix.Compilers.Protocol.umbrella(args, res)
 
         true ->
           res
       end
 
-    with true <- consolidate_protocols?,
-         path = Mix.Project.consolidation_path(config),
-         {:ok, protocols} <- File.ls(path) do
+    path = Mix.Project.consolidation_path(config)
+
+    with {:ok, protocols} <- File.ls(path) do
       # We don't cache consolidation path as we may write to it
       Code.prepend_path(path)
       Enum.each(protocols, &load_protocol/1)
@@ -191,7 +193,7 @@ defmodule Mix.Tasks.Compile do
   end
 
   defp first_line(doc) do
-    String.split(doc, "\n", parts: 2) |> hd |> String.trim() |> String.trim_trailing(".")
+    String.split(doc, "\n", parts: 2) |> hd() |> String.trim() |> String.trim_trailing(".")
   end
 
   defp merge_diagnostics({status1, diagnostics1}, {status2, diagnostics2}) do
@@ -215,12 +217,6 @@ defmodule Mix.Tasks.Compile do
   @deprecated "Use Mix.Task.Compiler.manifests/0 instead"
   defdelegate manifests, to: Mix.Task.Compiler
 
-  ## Consolidation handling
-
-  defp reconsolidate_protocols?(:ok), do: true
-  defp reconsolidate_protocols?(:noop), do: not Mix.Tasks.Compile.Protocols.consolidated?()
-  defp reconsolidate_protocols?(:error), do: false
-
   defp load_protocol(file) do
     case file do
       "Elixir." <> _ ->
