@@ -1,10 +1,14 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Compile.Yecc do
   use Mix.Task.Compiler
   alias Mix.Compilers.Erlang
 
   @recursive true
   @manifest "compile.yecc"
-  @switches [force: :boolean, all_warnings: :boolean]
+  @switches [force: :boolean, verbose: :boolean, all_warnings: :boolean]
 
   # These options can't be controlled with :yecc_options.
   @forced_opts [report: true, return: true]
@@ -26,6 +30,7 @@ defmodule Mix.Tasks.Compile.Yecc do
     * `--all-warnings` (`--no-all-warnings`) - prints all warnings, including previous compilations
       (default is true except on errors)
     * `--force` - forces compilation regardless of modification times
+    * `--verbose` - prints verbose output
 
   ## Configuration
 
@@ -52,7 +57,7 @@ defmodule Mix.Tasks.Compile.Yecc do
 
     yecc_options = project[:yecc_options] || []
 
-    unless is_list(yecc_options) do
+    if not is_list(yecc_options) do
       Mix.raise(":yecc_options should be a list of options, got: #{inspect(yecc_options)}")
     end
 
@@ -66,7 +71,7 @@ defmodule Mix.Tasks.Compile.Yecc do
 
   defp preload(project) do
     # TODO: Remove me in Elixir v2.0
-    unless :yecc in List.wrap(project[:compilers]) do
+    if :yecc not in List.wrap(project[:compilers]) do
       Mix.shell().error(
         "warning: in order to compile .yrl files, you must add \"compilers: [:yecc] ++ Mix.compilers()\" to the \"def project\" section of #{project[:app]}'s mix.exs"
       )
