@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Deps.Clean do
   use Mix.Task
 
@@ -63,13 +67,19 @@ defmodule Mix.Tasks.Deps.Clean do
           )
       end
 
-    do_clean(apps_to_clean, loaded_deps, build_path, deps_path, opts[:build])
+    Mix.Project.with_build_lock(fn ->
+      clean_build(apps_to_clean, build_path)
+    end)
 
-    if opts[:unlock] do
-      Mix.Task.run("deps.unlock", args)
-    else
-      :ok
-    end
+    Mix.Project.with_deps_lock(fn ->
+      clean_source(apps_to_clean, loaded_deps, deps_path, opts[:build])
+
+      if opts[:unlock] do
+        Mix.Task.run("deps.unlock", args)
+      else
+        :ok
+      end
+    end)
   end
 
   defp checked_deps(build_path, deps_path) do
@@ -108,11 +118,9 @@ defmodule Mix.Tasks.Deps.Clean do
     end
   end
 
-  defp do_clean(apps, deps, build_path, deps_path, build_only?) do
+  defp clean_build(apps, build_path) do
     shell = Mix.shell()
 
-    local = for %{scm: scm, app: app} <- deps, not scm.fetchable?(), do: Atom.to_string(app)
-
     Enum.each(apps, fn app ->
       shell.info("* Cleaning #{app}")
 
@@ -122,7 +130,13 @@ defmodule Mix.Tasks.Deps.Clean do
       |> Path.wildcard()
       |> maybe_warn_for_invalid_path(app)
       |> Enum.map(&(&1 |> File.rm_rf() |> maybe_warn_failed_file_deletion()))
+    end)
+  end
 
+  defp clean_source(apps, deps, deps_path, build_only?) do
+    local = for %{scm: scm, app: app} <- deps, not scm.fetchable?(), do: Atom.to_string(app)
+
+    Enum.each(apps, fn app ->
       # Remove everything from the source directory of dependencies.
       # Skip this step if --build option is specified or if
       # the dependency is local, i.e., referenced using :path.
