@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Deps.Compile do
   use Mix.Task
 
@@ -31,6 +35,24 @@ defmodule Mix.Tasks.Deps.Compile do
   recompiled without propagating those changes upstream. To ensure
   `b` is included in the compilation step, pass `--include-children`.
 
+  ## Compiling dependencies across multiple OS processes
+
+  If you set the environment variable `MIX_OS_DEPS_COMPILE_PARTITION_COUNT`
+  to a number greater than 1, Mix will start multiple operating system
+  processes to compile your dependencies concurrently.
+
+  While Mix and Rebar compile all files within a given project in parallel,
+  enabling this environment variable can still yield useful gains in several
+  cases, such as when compiling dependencies with native code, dependencies
+  that must download assets, or dependencies where the compilation time is not
+  evenly distributed (for example, one file takes much longer to compile than
+  all others).
+
+  While most configuration in Mix is done via command line flags, this particular
+  environment variable exists because the best number will vary per machine
+  (and often per project too). The environment variable also makes it more accessible
+  to enable concurrent compilation in CI and also during `Mix.install/2` commands.
+
   ## Command line options
 
     * `--force` - force compilation of deps
@@ -53,7 +75,6 @@ defmodule Mix.Tasks.Deps.Compile do
     end
 
     Mix.Project.get!()
-
     config = Mix.Project.config()
 
     Mix.Project.with_build_lock(config, fn ->
@@ -71,86 +92,82 @@ defmodule Mix.Tasks.Deps.Compile do
 
   @doc false
   def compile(deps, options \\ []) do
-    shell = Mix.shell()
-    config = Mix.Project.deps_config()
     Mix.Task.run("deps.precompile")
+    force? = Keyword.get(options, :force, false)
 
-    compiled =
+    deps =
       deps
       |> reject_umbrella_children(options)
       |> reject_local_deps(options)
-      |> Enum.map(fn %Mix.Dep{app: app, status: status, opts: opts, scm: scm} = dep ->
-        check_unavailable!(app, scm, status)
-        maybe_clean(dep, options)
 
-        compiled? =
-          cond do
-            not is_nil(opts[:compile]) ->
-              do_compile(dep, config)
+    count = System.get_env("MIX_OS_DEPS_COMPILE_PARTITION_COUNT", "0") |> String.to_integer()
 
-            Mix.Dep.mix?(dep) ->
-              do_mix(dep, config)
+    compiled? =
+      if count > 1 and length(deps) > 1 do
+        Mix.shell().info("mix deps.compile running across #{count} OS processes")
+        Mix.Tasks.Deps.Partition.server(deps, count, force?)
+      else
+        config = Mix.Project.deps_config()
+        true in Enum.map(deps, &compile_single(&1, force?, config))
+      end
 
-            Mix.Dep.make?(dep) ->
-              do_make(dep, config)
-
-            dep.manager == :rebar3 ->
-              do_rebar3(dep, config)
-
-            true ->
-              shell.error(
-                "Could not compile #{inspect(app)}, no \"mix.exs\", \"rebar.config\" or \"Makefile\" " <>
-                  "(pass :compile as an option to customize compilation, set it to \"false\" to do nothing)"
-              )
-
-              false
-          end
-
-        if compiled? do
-          build_path = Mix.Project.build_path(config)
-
-          lazy_message = fn ->
-            info = %{
-              app: dep.app,
-              scm: dep.scm,
-              manager: dep.manager,
-              os_pid: System.pid()
-            }
-
-            {:dep_compiled, info}
-          end
-
-          Mix.Sync.PubSub.broadcast(build_path, lazy_message)
-        end
-
-        # We should touch fetchable dependencies even if they
-        # did not compile otherwise they will always be marked
-        # as stale, even when there is nothing to do.
-        fetchable? = touch_fetchable(scm, opts[:build])
-
-        compiled? and fetchable?
-      end)
-
-    if true in compiled, do: Mix.Task.run("will_recompile"), else: :ok
+    if compiled?, do: Mix.Task.run("will_recompile"), else: :ok
   end
 
-  defp maybe_clean(dep, opts) do
+  @doc false
+  def compile_single(%Mix.Dep{} = dep, force?, config) do
+    %{app: app, status: status, opts: opts, scm: scm} = dep
+    check_unavailable!(app, scm, status)
+
     # If a dependency was marked as fetched or with an out of date lock
     # or missing the app file, we always compile it from scratch.
-    if Keyword.get(opts, :force, false) or Mix.Dep.compilable?(dep) do
+    if force? or Mix.Dep.compilable?(dep) do
       File.rm_rf!(Path.join([Mix.Project.build_path(), "lib", Atom.to_string(dep.app)]))
     end
-  end
 
-  defp touch_fetchable(scm, path) do
-    if scm.fetchable?() do
-      path = Path.join(path, ".mix")
-      File.mkdir_p!(path)
-      File.touch!(Path.join(path, "compile.fetch"))
-      true
-    else
-      false
+    compiled? =
+      cond do
+        not is_nil(opts[:compile]) ->
+          do_compile(dep, config)
+
+        Mix.Dep.mix?(dep) ->
+          do_mix(dep, config)
+
+        Mix.Dep.make?(dep) ->
+          do_make(dep, config)
+
+        dep.manager == :rebar3 ->
+          do_rebar3(dep, config)
+
+        true ->
+          Mix.shell().error(
+            "Could not compile #{inspect(app)}, no \"mix.exs\", \"rebar.config\" or \"Makefile\" " <>
+              "(pass :compile as an option to customize compilation, set it to \"false\" to do nothing)"
+          )
+
+          false
+      end
+
+    if compiled? do
+      config
+      |> Mix.Project.build_path()
+      |> Mix.Sync.PubSub.broadcast(fn ->
+        info = %{
+          app: dep.app,
+          scm: dep.scm,
+          manager: dep.manager,
+          os_pid: System.pid()
+        }
+
+        {:dep_compiled, info}
+      end)
     end
+
+    # We should touch fetchable dependencies even if they
+    # did not compile otherwise they will always be marked
+    # as stale, even when there is nothing to do.
+    fetchable? = touch_fetchable(scm, opts[:build])
+    compiled? and fetchable?
   end
 
   defp check_unavailable!(app, scm, {:unavailable, path}) do
@@ -172,6 +189,15 @@ defmodule Mix.Tasks.Deps.Compile do
     :ok
   end
 
+  defp touch_fetchable(scm, path) do
+    if scm.fetchable?() do
+      Mix.Dep.ElixirSCM.update(Path.join(path, ".mix"), scm)
+      true
+    else
+      false
+    end
+  end
+
   defp do_mix(dep, _config) do
     Mix.Dep.in_dependency(dep, fn _ ->
       config = Mix.Project.config()
@@ -201,9 +227,9 @@ defmodule Mix.Tasks.Deps.Compile do
     end)
   end
 
-  defp do_rebar3(%Mix.Dep{opts: opts} = dep, config) do
-    if not Mix.Rebar.available?(:rebar3) do
-      handle_rebar_not_found(dep)
+  defp do_rebar3(%Mix.Dep{opts: opts, manager: manager} = dep, config) do
+    if not Mix.Rebar.available?(manager) do
+      Mix.Tasks.Local.Rebar.run(["--force"])
     end
 
     dep_path = opts[:dest]
@@ -228,6 +254,7 @@ defmodule Mix.Tasks.Deps.Compile do
     env = [
       # REBAR_BARE_COMPILER_OUTPUT_DIR is only honored by rebar3 >= 3.14
       {"REBAR_BARE_COMPILER_OUTPUT_DIR", build_path},
+      {"REBAR_SKIP_PROJECT_PLUGINS", "true"},
       {"REBAR_CONFIG", config_path},
       {"REBAR_PROFILE", "prod"},
       {"TERM", "dumb"}
@@ -258,36 +285,10 @@ defmodule Mix.Tasks.Deps.Compile do
     |> Mix.Rebar.serialize_config()
   end
 
-  defp handle_rebar_not_found(%Mix.Dep{app: app, manager: manager}) do
-    shell = Mix.shell()
-
-    shell.info(
-      "Could not find \"#{manager}\", which is needed to build dependency #{inspect(app)}"
-    )
-
-    install_question =
-      "Shall I install #{manager}? (if running non-interactively, " <>
-        "use \"mix local.rebar --force\")"
-
-    if not shell.yes?(install_question) do
-      error_message =
-        "Could not find \"#{manager}\" to compile " <>
-          "dependency #{inspect(app)}, please ensure \"#{manager}\" is available"
-
-      Mix.raise(error_message)
-    end
-
-    Mix.Tasks.Local.Rebar.run(["--force"])
-
-    if not Mix.Rebar.available?(manager) do
-      Mix.raise("\"#{manager}\" installation failed")
-    end
-  end
-
   defp do_make(dep, config) do
     command = make_command(dep)
     shell_cmd!(dep, config, command, [{"IS_DEP", "1"}])
-    build_structure(dep, config)
+    build_symlink_structure(dep, config)
     true
   end
 
@@ -316,7 +317,7 @@ defmodule Mix.Tasks.Deps.Compile do
   defp do_compile(%Mix.Dep{opts: opts} = dep, config) do
     if command = opts[:compile] do
       shell_cmd!(dep, config, command)
-      build_structure(dep, config)
+      build_symlink_structure(dep, config)
       true
     else
       false
@@ -340,7 +341,7 @@ defmodule Mix.Tasks.Deps.Compile do
     [env: env, cd: opts[:dest]]
   end
 
-  defp build_structure(%Mix.Dep{opts: opts}, config) do
+  defp build_symlink_structure(%Mix.Dep{opts: opts}, config) do
     config = Keyword.put(config, :deps_app_path, opts[:build])
     Mix.Project.build_structure(config, symlink_ebin: true, source: opts[:dest])
     Code.prepend_path(Path.join(opts[:build], "ebin"), cache: true)
