@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Format do
   use Mix.Task
 
@@ -29,6 +33,10 @@ defmodule Mix.Tasks.Format do
       to be used by this task. For example, `["mix.exs", "{config,lib,test}/**/*.{ex,exs}"]`.
       Patterns are expanded with `Path.wildcard/2`.
 
+    * `:excludes` (a list of paths and patterns) - specifies the files to exclude from the
+      list of inputs to this task. For example, `["config/runtime.exs", "test/**/*.{ex,exs}"]`.
+      Patterns are expanded with `Path.wildcard/2`.
+
     * `:plugins` (a list of modules) (since v1.13.0) - specifies a list of
       modules to customize how the formatter works. See the "Plugins" section
       below for more information.
@@ -76,7 +84,7 @@ defmodule Mix.Tasks.Format do
     * `--stdin-filename` - path to the file being formatted on stdin.
       This is useful if you are using plugins to support custom filetypes such
       as `.heex`. Without passing this flag, it is assumed that the code being
-      passed via stdin is valid Elixir code. Defaults to "stdin.exs".
+      passed via stdin is valid Elixir code. Defaults to `stdin.exs`.
 
     * `--migrate` - enables the `:migrate` option, which should be able to
       automatically fix some deprecation warnings but changes the AST.
@@ -221,13 +229,17 @@ defmodule Mix.Tasks.Format do
 
   @doc """
   Returns which features this plugin should plug into.
+
+  It receives all options specified in `.formatter.exs`.
   """
-  @callback features(Keyword.t()) :: [sigils: [atom()], extensions: [binary()]]
+  @callback features(keyword) :: [sigils: [atom()], extensions: [binary()]]
 
   @doc """
   Receives a string to be formatted with options and returns said string.
+
+  It receives all options specified in `.formatter.exs`.
   """
-  @callback format(String.t(), Keyword.t()) :: String.t()
+  @callback format(String.t(), keyword) :: String.t()
 
   @impl true
   def run(all_args) do
@@ -352,6 +364,16 @@ defmodule Mix.Tasks.Format do
     plugins
   end
 
+  @typedoc """
+  Options for `formatter_for_file/2`.
+  """
+  @type formatter_for_file_opts :: [
+          deps_paths: %{atom() => String.t()},
+          dot_formatter: String.t(),
+          plugin_loader: ([module()] -> [module()]),
+          root: String.t()
+        ]
+
   @doc """
   Returns a formatter function and the formatter options to
   be used for the given file.
@@ -385,6 +407,8 @@ defmodule Mix.Tasks.Format do
     * `:root` - use the given root as the current working directory.
   """
   @doc since: "1.13.0"
+  @spec formatter_for_file(String.t(), formatter_for_file_opts) ::
+          {(String.t() -> String.t()), keyword()}
   def formatter_for_file(file, opts \\ []) do
     cwd = Keyword.get_lazy(opts, :root, &File.cwd!/0)
     {dot_formatter, formatter_opts} = eval_dot_formatter(cwd, opts)
@@ -498,6 +522,12 @@ defmodule Mix.Tasks.Format do
   defp eval_deps_opts(deps, opts) do
     deps_paths = opts[:deps_paths] || Mix.Project.deps_paths()
 
+    if not is_map(deps_paths) do
+      Mix.raise(
+        "Expected :deps_paths to return a map of dependency paths, got: #{inspect(deps_paths)}"
+      )
+    end
+
     for dep <- deps,
         dep_path = assert_valid_dep_and_fetch_path(dep, deps_paths),
         dep_dot_formatter = Path.join(dep_path, ".formatter.exs"),
@@ -608,9 +638,16 @@ defmodule Mix.Tasks.Format do
       Mix.raise("Expected :inputs or :subdirectories key in #{dot_formatter}")
     end
 
+    excluded_files =
+      formatter_opts[:excludes]
+      |> List.wrap()
+      |> Enum.flat_map(&Path.wildcard(Path.expand(&1, cwd), match_dot: true))
+      |> MapSet.new()
+
     map =
       for input <- List.wrap(formatter_opts[:inputs]),
           file <- Path.wildcard(Path.expand(input, cwd), match_dot: true),
+          file not in excluded_files,
           do: {file, {dot_formatter, formatter_opts}},
           into: %{}
 
@@ -673,7 +710,7 @@ defmodule Mix.Tasks.Format do
       size = byte_size(sub)
 
       case file do
-        <<prefix::binary-size(size), dir_separator, _::binary>>
+        <<prefix::binary-size(^size), dir_separator, _::binary>>
         when prefix == sub and dir_separator in [?\\, ?/] ->
           recur_formatter_opts_for_file(file, sub, formatter_opts_and_subs)
 
@@ -693,8 +730,11 @@ defmodule Mix.Tasks.Format do
     do: path |> Path.expand() |> Path.wildcard(match_dot: true) |> Enum.filter(&File.regular?/1)
 
   defp elixir_format(content, formatter_opts) do
-    case Code.format_string!(content, formatter_opts) do
-      [] -> ""
+    content
+    |> Code.format_string!(formatter_opts)
+    |> IO.iodata_to_binary()
+    |> case do
+      "" -> ""
       formatted_content -> IO.iodata_to_binary([formatted_content, ?\n])
     end
   end
@@ -803,7 +843,6 @@ defmodule Mix.Tasks.Format do
   end
 
   @doc false
-  @spec text_diff_format(String.t(), String.t()) :: iolist()
   def text_diff_format(old, new, opts \\ [])
 
   def text_diff_format(code, code, _opts), do: []
