@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Release do
   use Mix.Task
 
@@ -12,7 +16,7 @@ defmodule Mix.Tasks.Release do
   Once a release is assembled, it can be packaged and deployed to a
   target, as long as the target runs on the same operating system (OS)
   distribution and version as the machine running the `mix release`
-  command.
+  command. Windows releases also require Microsoft Visual C++ Runtime.
 
   A release can be configured in your `mix.exs` file under the `:releases`
   key inside `def project`:
@@ -167,16 +171,16 @@ defmodule Mix.Tasks.Release do
       $ bin/RELEASE_NAME daemon
 
   In daemon mode, the system is started on the background via
-  [`run_erl`](https://www.erlang.org/doc/man/run_erl.html). You may also
-  want to enable [`heart`](https://www.erlang.org/doc/man/heart.html)
+  [`run_erl`](https://www.erlang.org/doc/apps/erts/run_erl_cmd.html).
+  You may also want to enable [`:heart`](`:heart`)
   in daemon mode so it automatically restarts the system in case
   of crashes. See the generated `releases/RELEASE_VSN/env.sh` file.
 
-  The daemon will write all of its standard output to the "tmp/log/"
+  The daemon will write all of its standard output to the `tmp/log/`
   directory in the release root. You can watch the log file by doing
   `tail -f tmp/log/erlang.log.1` or similar. Once files get too large,
   the index suffix will be incremented. A developer can also attach
-  to the standard input of the daemon by invoking "to_erl tmp/pipe/"
+  to the standard input of the daemon by invoking `to_erl tmp/pipe/`
   from the release root. However, note that attaching to the system
   should be done with extreme care, since the usual commands for
   exiting an Elixir system, such as hitting Ctrl+C twice or Ctrl+\\,
@@ -191,8 +195,8 @@ defmodule Mix.Tasks.Release do
 
   While daemons are not available on Windows, it is possible to install a
   released system as a service on Windows with the help of
-  [`erlsrv`](https://www.erlang.org/doc/man/erlsrv.html). This can be done by
-  running:
+  [`erlsrv`](https://www.erlang.org/doc/apps/erts/erlsrv_cmd.html).
+  This can be done by running:
 
       $ bin/RELEASE_NAME install
 
@@ -242,8 +246,8 @@ defmodule Mix.Tasks.Release do
   but usually separate, and often there are many targets (either multiple
   instances, or the release is deployed to heterogeneous environments).
 
-  To deploy straight from a host to a separate target without cross-compilation,
-  the following must be the same between the host and the target:
+  To deploy straight from a host to a separate target, the following must be
+  the same between the host and the target:
 
     * Target architecture (for example, x86_64 or ARM)
     * Target vendor + operating system (for example, Windows, Linux, or Darwin/macOS)
@@ -251,23 +255,24 @@ defmodule Mix.Tasks.Release do
 
   This is often represented in the form of target triples, for example,
   `x86_64-unknown-linux-gnu`, `x86_64-unknown-linux-musl`, `x86_64-apple-darwin`.
+  If you are building on a MacBook (`x86_64-apple-darwin`) and trying to deploy
+  to a typical Ubuntu machine (`x86_64-unknown-linux-gnu`), the release will not
+  work. Instead you should build the release on a `x86_64-unknown-linux-gnu` host.
 
-  So to be more precise, to deploy straight from a host to a separate target,
-  the Erlang Runtime System (ERTS), and any native dependencies (NIFs), must
-  be compiled for the same target triple. If you are building on a MacBook
-  (`x86_64-apple-darwin`) and trying to deploy to a typical Ubuntu machine
-  (`x86_64-unknown-linux-gnu`), the release will not work. Instead you should
-  build the release on a `x86_64-unknown-linux-gnu` host. As we will see, this
-  can be done in multiple ways, such as releasing on the target itself, or by
-  using virtual machines or containers, usually as part of your release pipeline.
+  Typically, different versions of Erlang VM and Elixir are available for
+  different targets via package managers, precompiled artifacts, and similar.
+  However, to deploy from a host to a separate target, you must also guarantee
+  that any dependency with NIFs (Natively-Implemented Functions) are compiled
+  for the same triplet. As we will see, this can be done in different ways,
+  such as releasing on the target itself, or by using virtual machines or
+  containers, usually as part of your release pipeline.
 
   In addition to matching the target triple, it is also important that the
   target has all of the system packages that your application will need at
   runtime. A common one is the need for OpenSSL when building an application
-  that uses `:crypto` or `:ssl`, which is dynamically linked to ERTS. The other
-  common source for native dependencies like this comes from dependencies
-  containing NIFs (natively-implemented functions) which may expect to
-  dynamically link to libraries they use.
+  that uses `:crypto` or `:ssl`, which is dynamically linked to the Erlang VM.
+  Project dependencies containing NIFs (natively-implemented functions) may
+  also dynamically link to system libraries, so check those accordingly.
 
   Of course, some operating systems and package managers can differ between
   versions, so if your goal is to have full compatibility between host and
@@ -276,26 +281,7 @@ defmodule Mix.Tasks.Release do
   some systems, especially so with package managers that try to create fully
   reproducible environments (Nix, Guix).
 
-  Similarly, when creating a stand-alone package and release for Windows, note
-  the Erlang Runtime System has a dependency to some Microsoft libraries
-  (Visual C++ Redistributable Packages for Visual Studio 2013). These libraries
-  are installed (if not present before) when Erlang is installed but it is not
-  part of the standard Windows environment. Deploying a stand-alone release on
-  a computer without these libraries will result in a failure when trying to
-  run the release. One way to solve this is to download and install these
-  Microsoft libraries the first time a release is deployed (the Erlang installer
-  version 10.6 ships with “Microsoft Visual C++ 2013 Redistributable - 12.0.30501”).
-
-  Alternatively, you can also bundle the compiled object files in the release,
-  as long as they were compiled for the same target. If doing so, you need to
-  update `LD_LIBRARY_PATH` environment variable with the paths containing the
-  bundled objects on Unix-like systems or the `$PATH` environment variable on
-  Windows systems.
-
-  Currently, there is no official way to cross-compile a release from one
-  target triple to another, due to the complexities involved in the process.
-
-  ### Techniques
+  ### Using matching host and target
 
   There are a couple of ways to guarantee that a release is built on a host with
   the same properties as the target. A simple option is to fetch the source,
@@ -334,6 +320,8 @@ defmodule Mix.Tasks.Release do
   production machines can fetch the deployment from the network storage
   and run `bin/my_app start`.
 
+  ### Using images
+
   Another mechanism to automate deployments is to use images, such as
   Amazon Machine Images, or container platforms, such as Docker.
   For instance, you can use Docker to run locally a system with the
@@ -342,9 +330,14 @@ defmodule Mix.Tasks.Release do
   a complete image and/or container with the operating system, all
   dependencies as well as the releases.
 
-  In other words, there are multiple ways systems can be deployed and
-  releases can be automated and incorporated into all of them as long
-  as you remember to build the system in the same target triple.
+  However, when building such images on your machine, those technologies
+  use emulation which may not interplay well with Erlang VM's JIT
+  (just-in time) compiler. To address this, you can set this environment
+  variable on your build stage:
+
+      ENV ERL_AFLAGS "+JMsingle true"
+
+  ## Shutting down
 
   Once a system is deployed, shutting down the system can be done by
   sending SIGINT/SIGTERM to the system, which is what most containers,
@@ -448,12 +441,12 @@ defmodule Mix.Tasks.Release do
           ]
 
     * `:rel_templates_path` - the path to find template files that are copied to
-      the release, such as "vm.args.eex", "remote.vm.args.eex", "env.sh.eex"
-      (or "env.bat.eex"), and "overlays". Defaults to "rel" in the project root.
+      the release, such as `vm.args.eex`, `remote.vm.args.eex`, `env.sh.eex`
+      (or `env.bat.eex`), and `overlays`. Defaults to `"rel"` in the project root.
 
     * `:overlays` - a list of directories with extra files to be copied
       as is to the release. The "overlays" directory at `:rel_templates_path`
-      is always included in this list by default (typically at "rel/overlays").
+      is always included in this list by default (typically at `"rel/overlays"`).
       See the "Overlays" section for more information.
 
     * `:steps` - a list of steps to execute when assembling the release. See
@@ -486,7 +479,7 @@ defmodule Mix.Tasks.Release do
   the release is assembled. This can be easily done by placing such
   files in the `rel/overlays` directory. Any file in there is copied
   as is to the release root. For example, if you have placed a
-  "rel/overlays/Dockerfile" file, the "Dockerfile" will be copied as
+  `rel/overlays/Dockerfile` file, the "Dockerfile" will be copied as
   is to the release root.
 
   If you want to specify extra overlay directories, you can do so
@@ -633,8 +626,8 @@ defmodule Mix.Tasks.Release do
 
   ### Runtime configuration
 
-  To enable runtime configuration in your release, all you need to do is
-  to create a file named `config/runtime.exs`:
+  To enable runtime configuration in your release, create a file named
+  `config/runtime.exs`:
 
       import Config
       config :my_app, :secret_key, System.fetch_env!("MY_APP_SECRET_KEY")
@@ -689,7 +682,10 @@ defmodule Mix.Tasks.Release do
       You can configure the "tmp" directory by setting the `RELEASE_TMP` environment
       variable, either explicitly or inside your `releases/RELEASE_VSN/env.sh`
       (or `env.bat` on Windows). Defaults to `true` if using the deprecated
-      `config/releases.exs`, `false` otherwise.
+      `config/releases.exs`, `false` otherwise. Be careful of which libraries you
+      load when setting this option to true, if a library is loaded early during
+      configuration and it includes native code, it may not actually be able to
+      restart cleanly.
 
     * `:prune_runtime_sys_config_after_boot` - if `:reboot_system_after_config`
       is set, every time your system boots, the release will write a config file
@@ -1276,7 +1272,7 @@ defmodule Mix.Tasks.Release do
     cond do
       path ->
         msg = "#{path} to configure the release at runtime"
-        Mix.shell().info([:green, "* using ", :reset, msg])
+        info(release, [:green, "* using ", :reset, msg])
         File.cp!(path, Path.join(version_path, "runtime.exs"))
         init = {:system, "RELEASE_ROOT", "/releases/#{release.version}/runtime.exs"}
         opts = [path: init, env: Mix.env(), target: Mix.target(), imports: :disabled]
@@ -1357,7 +1353,7 @@ defmodule Mix.Tasks.Release do
   end
 
   defp info(release, message) do
-    unless release.options[:quiet] do
+    if !release.options[:quiet] do
       Mix.shell().info(message)
     end
   end
@@ -1451,16 +1447,7 @@ defmodule Mix.Tasks.Release do
 
   defp cli_for(:windows, release) do
     {"env.bat", &env_bat_template(release: &1),
-     [{"#{release.name}.bat", cli_bat_template(release: release) |> maybe_replace_werl()}]}
-  end
-
-  defp maybe_replace_werl(contents) do
-    # TODO: Remove me when we require Erlang/OTP 26+
-    if :erlang.system_info(:otp_release) >= ~c"26" do
-      String.replace(contents, "--werl", "")
-    else
-      contents
-    end
+     [{"#{release.name}.bat", cli_bat_template(release: release)}]}
   end
 
   defp elixir_cli_for(:unix, release) do
