@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Run do
   use Mix.Task
 
@@ -45,6 +49,8 @@ defmodule Mix.Tasks.Run do
   ## Command-line options
 
     * `--eval`, `-e` - evaluates the given code
+    * `--listeners` - starts Mix listeners (they are started by default,
+      unless `--no-listeners` or `--no-deps-check` are given)
     * `--require`, `-r` - executes the given pattern/file
     * `--parallel`, `-p` - makes all requires parallel
     * `--preload-modules` - preloads all modules defined in applications
@@ -59,6 +65,19 @@ defmodule Mix.Tasks.Run do
 
   """
 
+  @typedoc """
+  Options for the internal `run/5` function.
+
+  These options are typically parsed from command-line arguments.
+  """
+  @type run_opts :: [
+          {:parallel_require, String.t()}
+          | {:eval, String.t()}
+          | {:require, String.t()}
+          | {:parallel, boolean()}
+          | {:config, String.t()}
+        ]
+
   @impl true
   def run(args) do
     {opts, head} =
@@ -93,7 +112,7 @@ defmodule Mix.Tasks.Run do
   @doc false
   @spec run(
           OptionParser.argv(),
-          keyword,
+          run_opts,
           OptionParser.argv(),
           (String.t() -> term()),
           (String.t() -> term())
@@ -172,7 +191,7 @@ defmodule Mix.Tasks.Run do
     require_runner =
       if opts[:parallel] do
         fn files ->
-          case Kernel.ParallelCompiler.require(files) do
+          case Kernel.ParallelCompiler.require(files, return_diagnostics: true) do
             {:ok, _, _} -> :ok
             {:error, _, _} -> exit({:shutdown, 1})
           end
