@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Test do
   use Mix.Task
 
@@ -113,7 +117,7 @@ defmodule Mix.Tasks.Test do
       and imports (but not local functions). You can press `n` for the next line
       and `c` for the next test. This automatically sets `--trace`
 
-    * `--color` - enables color in the output
+    * `--color` - enables color in ExUnit formatting results
 
     * `--cover` - runs coverage tool. See "Coverage" section below
 
@@ -146,6 +150,12 @@ defmodule Mix.Tasks.Test do
     * `--max-failures` - the suite stops evaluating tests when this number of test
       failures is reached. It runs all tests if omitted
 
+    * `--max-requires` - sets the maximum number of test files to compile in parallel.
+      Setting this to 1 will compile test files sequentially.
+
+    * `-n`, `--name-pattern` *(since v1.19.0)* - only run tests with names that match
+      the given regular expression
+
     * `--no-archives-check` - does not check archives
 
     * `--no-color` - disables color in the output
@@ -185,11 +195,12 @@ defmodule Mix.Tasks.Test do
       `--seed 0` disables randomization so the tests in a single file will always be ran
       in the same order they were defined in
 
-    * `--slowest` - prints timing information for the N slowest tests.
-      Automatically sets `--trace` and `--preload-modules`
+    * `--slowest` - prints timing information for the N slowest tests. Includes time spent in
+     `ExUnit.Callbacks.setup/1`. Automatically sets `--trace` and `--preload-modules`
 
     * `--slowest-modules` *(since v1.17.0)* - prints timing information for the N slowest
-      modules. Automatically sets `--trace` and `--preload-modules`
+      modules. Includes time spent in `ExUnit.Callbacks.setup/1`. Automatically sets
+      `--trace` and `--preload-modules`
 
     * `--stale` - runs only tests which reference modules that changed since the
       last time tests were ran with `--stale`. You can read more about this option
@@ -201,8 +212,12 @@ defmodule Mix.Tasks.Test do
       Note that in trace mode test timeouts will be ignored as timeout is set to `:infinity`
 
     * `--warnings-as-errors` *(since v1.12.0)* - treats compilation warnings (from loading the
-      test suite) as errors and returns a non-zero exit status if the test suite would otherwise
-      pass. Note that failures reported by `--warnings-as-errors` cannot be retried with the
+      test suite) as errors and returns an exit status of 1 if the test suite would otherwise
+      pass. If the test suite fails and also include warnings as errors, the exit
+      status returned will be the value of the `--exit-status` option, which
+      defaults to `2`, plus one. Therefore in the default case, this will be exit status `3`.
+
+      Note that failures reported by `--warnings-as-errors` cannot be retried with the
       `--failed` flag.
 
       This option only applies to test files. To treat warnings as errors during compilation and
@@ -217,17 +232,37 @@ defmodule Mix.Tasks.Test do
       mechanism. See the "Coverage" section for more information
 
     * `:test_elixirc_options` - the compiler options to used when
-      loading/compiling test files. By default it disables the debug chunk
-      and docs chunk
+      loading/compiling test files. By default it disables the debug chunk,
+      docs chunk, and module type inference
 
     * `:test_paths` - list of paths containing test files. Defaults to
       `["test"]` if the `test` directory exists, otherwise, it defaults to `[]`.
       It is expected that all test paths contain a `test_helper.exs` file
 
-    * `:test_pattern` - a pattern to load test files. Defaults to `*_test.exs`
+    * `:test_pattern` - a pattern to find potential test files. Defaults to `"*.{ex,exs}"`.
 
-    * `:warn_test_pattern` - a pattern to match potentially misnamed test files
-      and display a warning. Defaults to `*_test.ex`
+      Once files are found, they are filtered in two ways:
+
+        * Files that match `:test_load_filters` are loaded
+        * Files that match `:test_ignore_filters` are ignored
+        * All other files emit a warning
+
+      In Elixir versions earlier than 1.19.0, this option defaulted to `*_test.exs`.
+
+    * `:test_load_filters` - a list of files, regular expressions or one-arity
+      functions to restrict which files matched by the `:test_pattern` are loaded.
+      Defaults to `[&String.ends_with?(&1, "_test.exs")]`. Paths are relative to
+      the project root and separated by `/`, even on Windows.
+
+    * `:test_ignore_filters` - a list of files, regular expressions or one-arity
+      functions to ignore by the loader. Any file that is not loaded and not
+      ignored will emit a warning.
+
+      Mix ignores files ending in `_helper.exs` by default, as well as any file
+      included in the project's `:elixirc_paths`. You may choose to disable all
+      warnings by ignoring all files with `[fn _ -> true end]`.
+
+      Paths are relative to the project root and separated by `/`, even on Windows.
 
   ## Coloring
 
@@ -439,9 +474,11 @@ defmodule Mix.Tasks.Test do
     trace: :boolean,
     max_cases: :integer,
     max_failures: :integer,
+    max_requires: :integer,
     include: :keep,
     exclude: :keep,
     seed: :integer,
+    name_pattern: :regex,
     only: :keep,
     compile: :boolean,
     start: :boolean,
@@ -464,11 +501,13 @@ defmodule Mix.Tasks.Test do
     repeat_until_failure: :integer
   ]
 
+  @aliases [b: :breakpoints, n: :name_pattern]
+
   @cover [output: "cover", tool: Mix.Tasks.Test.Coverage]
 
   @impl true
   def run(args) do
-    {opts, files} = OptionParser.parse!(args, strict: @switches, aliases: [b: :breakpoints])
+    {opts, files} = OptionParser.parse!(args, strict: @switches, aliases: @aliases)
 
     if not Mix.Task.recursing?() do
       do_run(opts, args, files)
@@ -507,7 +546,7 @@ defmodule Mix.Tasks.Test do
   defp do_run(opts, args, files) do
     _ = Mix.Project.get!()
 
-    unless System.get_env("MIX_ENV") || Mix.env() == :test do
+    if System.get_env("MIX_ENV") == nil && Mix.env() != :test do
       Mix.raise("""
       "mix test" is running in the \"#{Mix.env()}\" environment. If you are \
       running tests from within another command, you can either:
@@ -542,6 +581,7 @@ defmodule Mix.Tasks.Test do
     Mix.Task.run("compile", args -- ["--warnings-as-errors"])
 
     project = Mix.Project.config()
+
     {partitions, opts} = Keyword.pop(opts, :partitions)
     partitioned? = is_integer(partitions) and partitions > 1
 
@@ -563,7 +603,14 @@ defmodule Mix.Tasks.Test do
     # before requiring test_helper.exs so that the configuration is
     # available in test_helper.exs
     Mix.shell().print_app()
-    app_start_args = if opts[:slowest], do: ["--preload-modules" | args], else: args
+
+    app_start_args =
+      if opts[:slowest] || opts[:slowest_modules] do
+        ["--preload-modules" | args]
+      else
+        args
+      end
+
     Mix.Task.run("app.start", app_start_args)
 
     # The test helper may change the Mix.shell(), so revert it whenever we raise and after suite
@@ -587,52 +634,98 @@ defmodule Mix.Tasks.Test do
     {ex_unit_opts, allowed_files} = process_ex_unit_opts(opts)
     ExUnit.configure(ex_unit_opts)
 
+    warnings_as_errors? = Keyword.get(opts, :warnings_as_errors, false)
+    exit_status = Keyword.fetch!(ex_unit_opts, :exit_status)
+
+    # Prepare and extract all files to require and run
     test_paths = project[:test_paths] || default_test_paths()
-    Enum.each(test_paths, &require_test_helper(shell, &1))
-    ExUnit.configure(merge_helper_opts(ex_unit_opts))
+    test_pattern = project[:test_pattern] || "*.{ex,exs}"
 
-    # Finally parse, require and load the files
-    test_files = if files != [], do: parse_file_paths(files), else: test_paths
-    test_pattern = project[:test_pattern] || "*_test.exs"
-    warn_test_pattern = project[:warn_test_pattern] || "*_test.ex"
+    # Warn about deprecated warn configuration
+    if project[:warn_test_pattern] do
+      Mix.shell().info("""
+      warning: the `:warn_test_pattern` configuration is deprecated and will be ignored. \
+      Use `:test_load_filters` and `:test_ignore_filters` instead.
+      """)
+    end
 
-    files_with_matched_path = Mix.Utils.extract_files(test_files, test_pattern)
+    {test_files, test_opts} =
+      if files != [], do: ExUnit.Filters.parse_paths(files), else: {test_paths, []}
+
+    # get a list of all files in the test folders, which we filter by the test_load_filters
+    {potential_test_files, directly_included_test_files} = extract_files(test_files, test_pattern)
+
+    {load_files, _ignored_files, warn_files} =
+      classify_test_files(shell, potential_test_files, project)
+
+    # ensure that files given as direct argument to mix test are loaded,
+    # even if the test_load_filters don't match
+    load_files =
+      if files != [],
+        do: Enum.uniq(load_files ++ directly_included_test_files),
+        else: load_files
 
     matched_test_files =
-      files_with_matched_path
+      load_files
       |> filter_to_allowed_files(allowed_files)
       |> filter_by_partition(shell, partitions)
 
-    display_warn_test_pattern(
-      test_files,
-      test_pattern,
-      files_with_matched_path,
-      warn_test_pattern
-    )
+    warn_files != [] && warn_misnamed_test_files(warn_files)
 
-    case CT.require_and_run(matched_test_files, test_paths, test_elixirc_options, opts) do
-      {:ok, %{excluded: excluded, failures: failures, total: total}} ->
+    try do
+      helper_warned? = Enum.any?(test_paths, &require_test_helper(shell, &1))
+      # test_opts always wins because those are given via args
+      ExUnit.configure(ex_unit_opts |> merge_helper_opts() |> Keyword.merge(test_opts))
+
+      {CT.require_and_run(matched_test_files, test_paths, test_elixirc_options, opts),
+       helper_warned?}
+    catch
+      kind, reason ->
+        # Also mark the whole suite as failed
+        file = get_manifest_path(opts)
+        ExUnit.Filters.fail_all!(file)
+        :erlang.raise(kind, reason, __STACKTRACE__)
+    else
+      {{:ok, %{excluded: excluded, failures: failures, warnings?: warnings?, total: total}},
+       helper_warned?} ->
         Mix.shell(shell)
         cover && cover.()
 
         cond do
+          warnings_as_errors? and (warnings? or helper_warned?) and failures == 0 ->
+            message =
+              "\nERROR! Test suite aborted after successful execution due to warnings while using the --warnings-as-errors option"
+
+            IO.puts(:stderr, IO.ANSI.format([:red, message]))
+
+            System.at_exit(fn _ ->
+              exit({:shutdown, 1})
+            end)
+
           failures > 0 and opts[:raise] ->
             raise_with_shell(shell, "\"mix test\" failed")
 
+          warnings_as_errors? and warnings? and failures > 0 ->
+            System.at_exit(fn _ ->
+              exit({:shutdown, exit_status + 1})
+            end)
+
           failures > 0 ->
             System.at_exit(fn _ ->
-              exit({:shutdown, Keyword.fetch!(ex_unit_opts, :exit_status)})
+              exit({:shutdown, exit_status})
             end)
 
           excluded == total and Keyword.has_key?(opts, :only) ->
-            message = "The --only option was given to \"mix test\" but no test was executed"
-            raise_or_error_at_exit(shell, message, opts)
+            nothing_executed(shell, "--only", opts)
+
+          excluded == total and Keyword.has_key?(opts, :name_pattern) ->
+            nothing_executed(shell, "--name-pattern", opts)
 
           true ->
             :ok
         end
 
-      :noop ->
+      {:noop, _} ->
         cond do
           opts[:stale] ->
             Mix.shell().info("No stale tests")
@@ -649,11 +742,43 @@ defmodule Mix.Tasks.Test do
     end
   end
 
+  # similar to Mix.Utils.extract_files/2, but returns a list of directly included test files,
+  # that should be not filtered by the test_load_filters, e.g.
+  # mix test test/some_file.exs
+  defp extract_files(paths, pattern) do
+    {files, directly_included} =
+      for path <- paths, reduce: {[], []} do
+        {acc, directly_included} ->
+          case :elixir_utils.read_file_type(path) do
+            {:ok, :directory} ->
+              {[Path.wildcard("#{path}/**/#{pattern}") | acc], directly_included}
+
+            {:ok, :regular} ->
+              {[path | acc], [path | directly_included]}
+
+            _ ->
+              {acc, directly_included}
+          end
+      end
+
+    files =
+      files
+      |> List.flatten()
+      |> Enum.uniq()
+
+    {files, directly_included}
+  end
+
   defp raise_with_shell(shell, message) do
     Mix.shell(shell)
     Mix.raise(message)
   end
 
+  defp nothing_executed(shell, option, opts) do
+    message = "The #{option} option was given to \"mix test\" but no test was executed"
+    raise_or_error_at_exit(shell, message, opts)
+  end
+
   defp raise_or_error_at_exit(shell, message, opts) do
     cond do
       opts[:raise] ->
@@ -668,16 +793,79 @@ defmodule Mix.Tasks.Test do
     end
   end
 
-  defp display_warn_test_pattern(test_files, test_pattern, matched_test_files, warn_test_pattern) do
-    files = Mix.Utils.extract_files(test_files, warn_test_pattern) -- matched_test_files
+  defp classify_test_files(shell, potential_test_files, project) do
+    test_load_filters = project[:test_load_filters] || [&String.ends_with?(&1, "_test.exs")]
 
-    for file <- files do
-      Mix.shell().info(
-        "warning: #{file} does not match #{inspect(test_pattern)} and won't be loaded"
-      )
+    test_ignore_filters = get_test_ignore_filters(shell, project)
+
+    {to_load, to_ignore, to_warn} =
+      for file <- potential_test_files, reduce: {[], [], []} do
+        {to_load, to_ignore, to_warn} ->
+          cond do
+            any_file_matches?(file, test_load_filters) ->
+              {[file | to_load], to_ignore, to_warn}
+
+            any_file_matches?(file, test_ignore_filters) ->
+              {to_load, [file | to_ignore], to_warn}
+
+            true ->
+              {to_load, to_ignore, [file | to_warn]}
+          end
+      end
+
+    # get the files back in the original order
+    {Enum.reverse(to_load), Enum.reverse(to_ignore), Enum.reverse(to_warn)}
+  end
+
+  defp get_test_ignore_filters(shell, project) do
+    elixirc_paths = project[:elixirc_paths] || []
+
+    case Keyword.get(project, :test_ignore_filters, []) do
+      list when is_list(list) ->
+        # ignore any _helper.exs files and files that are compiled (test support files)
+        [
+          &String.ends_with?(&1, "_helper.exs"),
+          fn file -> Enum.any?(elixirc_paths, &String.starts_with?(file, &1)) end
+        ] ++ list
+
+      other ->
+        raise_with_shell(
+          shell,
+          "Invalid configuration for :test_ignore_filters, expected a list, got: #{inspect(other)}"
+        )
     end
   end
 
+  defp any_file_matches?(file, filters) do
+    Enum.any?(filters, fn filter ->
+      case filter do
+        regex when is_struct(regex, Regex) ->
+          Regex.match?(regex, file)
+
+        binary when is_binary(binary) ->
+          file == binary
+
+        fun when is_function(fun, 1) ->
+          fun.(file)
+      end
+    end)
+  end
+
+  defp warn_misnamed_test_files(ignored) do
+    Mix.shell().info("""
+    warning: the following files do not match any of the configured \
+    `:test_load_filters` / `:test_ignore_filters`:
+
+    #{Enum.join(ignored, "\n")}
+
+    This might indicate a typo in a test file name (for example, \
+    using "foo_tests.exs" instead of "foo_test.exs").
+
+    See the configuration for `:test_pattern` under `mix help test` \
+    for more information.
+    """)
+  end
+
   @option_keys [
     :breakpoints,
     :trace,
@@ -706,7 +894,7 @@ defmodule Mix.Tasks.Test do
       opts
       |> filter_opts(:include)
       |> filter_opts(:exclude)
-      |> filter_opts(:only)
+      |> filter_only_and_name_pattern()
       |> formatter_opts()
       |> color_opts()
       |> exit_status_opts()
@@ -719,7 +907,10 @@ defmodule Mix.Tasks.Test do
   defp merge_helper_opts(opts) do
     # The only options that are additive from app env are the excludes
     value = List.wrap(Application.get_env(:ex_unit, :exclude, []))
-    Keyword.update(opts, :exclude, value, &Enum.uniq(&1 ++ value))
+
+    opts
+    |> Keyword.update(:exclude, value, &Enum.uniq(&1 ++ value))
+    |> Keyword.put_new_lazy(:failures_manifest_path, fn -> get_manifest_path([]) end)
   end
 
   defp default_opts(opts) do
@@ -727,33 +918,33 @@ defmodule Mix.Tasks.Test do
     [autorun: false] ++ opts
   end
 
-  defp parse_file_paths(file_paths) do
-    {parsed_file_paths, ex_unit_opts} = ExUnit.Filters.parse_paths(file_paths)
-    ExUnit.configure(ex_unit_opts)
-    parsed_file_paths
-  end
-
   defp parse_filters(opts, key) do
     if Keyword.has_key?(opts, key) do
       ExUnit.Filters.parse(Keyword.get_values(opts, key))
+    else
+      []
     end
   end
 
-  defp filter_opts(opts, :only) do
-    if filters = parse_filters(opts, :only) do
-      opts
-      |> Keyword.update(:include, filters, &(filters ++ &1))
-      |> Keyword.update(:exclude, [:test], &[:test | &1])
-    else
-      opts
+  defp filter_only_and_name_pattern(opts) do
+    only = parse_filters(opts, :only)
+    name_patterns = opts |> Keyword.get_values(:name_pattern) |> Enum.map(&{:test, &1})
+
+    case only ++ name_patterns do
+      [] ->
+        opts
+
+      filters ->
+        opts
+        |> Keyword.update(:include, filters, &(filters ++ &1))
+        |> Keyword.update(:exclude, [:test], &[:test | &1])
     end
   end
 
   defp filter_opts(opts, key) do
-    if filters = parse_filters(opts, key) do
-      Keyword.put(opts, key, filters)
-    else
-      opts
+    case parse_filters(opts, key) do
+      [] -> opts
+      filters -> Keyword.put(opts, key, filters)
     end
   end
 
@@ -772,23 +963,27 @@ defmodule Mix.Tasks.Test do
 
   @manifest_file_name ".mix_test_failures"
 
-  defp manifest_opts(opts) do
-    opts =
-      Keyword.put_new(
-        opts,
-        :failures_manifest_path,
-        Path.join(Mix.Project.manifest_path(), @manifest_file_name)
-      )
+  defp get_manifest_path(opts) do
+    opts[:failures_manifest_path] ||
+      Application.get_env(:ex_unit, :failures_manifest_path) ||
+      Path.join(Mix.Project.manifest_path(), @manifest_file_name)
+  end
 
-    manifest_file = Keyword.get(opts, :failures_manifest_path)
+  defp manifest_opts(opts) do
+    manifest_file = get_manifest_path(opts)
 
     if opts[:failed] do
       if opts[:stale] do
         Mix.raise("Combining --failed and --stale is not supported.")
       end
 
-      {allowed_files, failed_ids} = ExUnit.Filters.failure_info(manifest_file)
-      {Keyword.put(opts, :only_test_ids, failed_ids), allowed_files}
+      case ExUnit.Filters.failure_info(manifest_file) do
+        {allowed_files, failed_ids} ->
+          {Keyword.put(opts, :only_test_ids, failed_ids), allowed_files}
+
+        :all ->
+          {opts, nil}
+      end
     else
       {opts, nil}
     end
@@ -852,7 +1047,8 @@ defmodule Mix.Tasks.Test do
     file = Path.join(dir, "test_helper.exs")
 
     if File.exists?(file) do
-      Code.require_file(file)
+      {_result, warnings} = Code.with_diagnostics([log: true], fn -> Code.require_file(file) end)
+      warnings != []
     else
       raise_with_shell(
         shell,
