@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 defmodule Mix.Tasks.Xref do
   use Mix.Task
 
@@ -27,16 +31,47 @@ defmodule Mix.Tasks.Xref do
   touching one file in a project causes a large subset of the project
   to recompile. The most common cause of these problems are the so-called
   "compile-connected" files. Those are files you depend on at compile-time
-  (for example, by invoking its macro or using it in the body of amodule)
+  (for example, by invoking its macro or using it in the body of a module)
   which also have their own dependencies.
 
-  Therefore, if your goal is to reduce recompilations, the first step is to run:
+  The most harmful form of compile-connected dependencies are the ones
+  that are also in a cycle. Imagine you have files `lib/a.ex`, `lib/b.ex`,
+  and `lib/c.ex` with the following dependencies:
+
+      lib/a.ex
+      └── lib/b.ex (compile)
+            └── lib/c.ex
+                  └── lib/a.ex
+
+  Because you have a compile-time dependency, any of the files `lib/a.ex`,
+  `lib/b.ex`, and `lib/c.ex` depend on will cause the whole cycle to
+  recompile. Therefore, your first priority to reduce compile times is
+  to remove such cycles. You can spot them by running:
+
+      $ mix xref graph --format cycles --label compile-connected
+
+  Whenever you find a compile-time dependency, such as `lib/a.ex` pointing
+  to `lib/b.ex`, there are two ways to remove them:
+
+    1. Run `mix xref trace lib/a.ex` to understand where and how `lib/a.ex`
+       depends on `lib/b.ex` at compile time and address it
+
+    2. Or run `mix xref trace lib/b.ex` and make sure it does not depend on
+       any other module in your project because a compile dependency makes
+       those runtime dependencies also compile time by transitivity
+
+  We outline all options for `mix xref trace` and the types of dependencies
+  over the following sections.
+
+  If you don't have compile cycles in your project, that's a good beginning,
+  but you want to avoid any compile-connected dependencies in general, as they
+  may become cycles in the future. To verify the general health of your project,
+  you may run:
 
       $ mix xref graph --format stats --label compile-connected
 
-  This command will show general information about the project, but
-  focus on compile-connected dependencies. In the stats, you will see
-  the following report:
+  This command will show general information about the project, but focus on
+  compile-connected dependencies. In the stats, you will see the following report:
 
       Top 10 files with most incoming dependencies:
         * lib/livebook_web.ex (97)
@@ -58,24 +93,16 @@ defmodule Mix.Tasks.Xref do
 
   The trouble here is precisely that, if any of the files in the latter
   command changes, all of the files in the first command will be recompiled,
-  because compile time dependencies are transitive.
-
-  Having compile time dependencies is a common feature in Elixir projects.
-  However, the modules you depend on at compile-time must avoid runtime
-  dependencies within the same project. You can understand all of the
-  dependencies of a given file by running:
-
-      $ mix xref trace lib/livebook_web.ex
-
-  The command above will output three types of dependencies, which we
-  detail next.
+  because compile time dependencies are transitive. As we did with cycles,
+  you can use `mix xref trace` to understand why and how these dependencies
+  exist.
 
   ### Dependency types
 
   Elixir tracks three types of dependencies between modules: compile,
   exports, and runtime. If a module has a compile time dependency on
   another module, the caller module has to be recompiled whenever the
-  callee changes (or any runtime dependency of the callee changes).
+  callee changes (or any dependency of the callee changes).
   Let's see an example:
 
       # lib/a.ex
@@ -156,10 +183,9 @@ defmodule Mix.Tasks.Xref do
 
       lib/b.ex:2: require A (export)
       lib/b.ex:3: call A.macro/0 (compile)
-      lib/b.ex:4: import A.macro/0 (compile)
+      lib/b.ex:4: import call A.macro/0 (compile)
       lib/b.ex:5: call A.fun/0 (compile)
-      lib/b.ex:6: call A.fun/0 (compile)
-      lib/b.ex:6: import A.fun/0 (compile)
+      lib/b.ex:6: import call A.fun/0 (compile)
       lib/b.ex:7: call A.macro/0 (compile)
       lib/b.ex:8: call A.fun/0 (runtime)
       lib/b.ex:9: struct A (export)
@@ -224,18 +250,33 @@ defmodule Mix.Tasks.Xref do
 
       * `stats` - prints general statistics about the graph;
 
-      * `cycles` - prints all cycles in the graph;
+      * `cycles` - prints all strongly connected cycles in the graph;
 
-      * `dot` - produces a DOT graph description in `xref_graph.dot` in the
-        current directory. Warning: this will override any previously generated file
+      * `dot` - produces a DOT graph description, by default written to `xref_graph.dot`
+        in the current directory.  See the documentation for the `--output` option to
+        learn how to control where the file is written and other related details.
 
-    * `--output` *(since v1.15.0)* - can be set to one of
+      * `json` *(since v1.19.0)* - produces a JSON file, by default written to
+        `xref_graph.json` in the current directory.  See the documentation for the
+        `--output` option to learn how to control where the file is written and other
+        related details.
+
+        The JSON format is always a two level map of maps. The top level keys
+        specify source files, with their values containing maps whose keys specify
+        sink files and whose values specify the type of relationship, which will
+        be one of `compile`, `export` or `runtime`. Files which have no dependencies
+        will be present in the top level map, and will have empty maps for values.
+
+    * `--output` *(since v1.15.0)* - can be used to override the location of
+      the files created by the `dot` and `json` formats. It can be set to
 
       * `-` - prints the output to standard output;
 
       * a path - writes the output graph to the given path
 
-      Defaults to `xref_graph.dot` in the current directory.
+      If the output file already exists then it will be renamed in place
+      to have a `.bak` suffix, possibly overwriting any existing `.bak` file.
+      If this rename fails a fatal exception will be thrown.
 
   The `--source` and `--sink` options are particularly useful when trying to understand
   how the modules in a particular file interact with the whole system. You can combine
@@ -333,34 +374,29 @@ defmodule Mix.Tasks.Xref do
 
   ### Understanding the printed cycle
 
-  If you run `mix xref graph --format cycle`, Elixir will print cycles
+  If you run `mix xref graph --format cycles`, Elixir will print cycles
   of shape:
 
       Cycle of length 3:
 
           lib/c.ex
           lib/b.ex
-          lib/a.ex
+          lib/a.ex (compile)
 
-  The cycles are given in order: `c.ex` depends on `b.ex` which depends
-  on `a.ex` which depends on `c.ex`. In particular, you want to avoid
-  cycles with compile dependencies in there. You can find those cycles
-  with:
+  More precisely, `xref` is printing the [strongly connected components](https://en.wikipedia.org/wiki/Strongly_connected_component)
+  in the dependency graph, which is (roughly speaking) the largest set of
+  files which are part of a dependency cycle involving these files.
+
+  For this reason, files may have multiple relationships between them, and
+  therefore the cycles are not printed in order. The label reflects the
+  highest type of relationship between the given file and any other file in
+  the cycle. In the example above, it means `lib/a.ex` depends on something
+  else in the cycle at compile-time.  Those are exactly the type of dependencies
+  we want to avoid, and you can ask `mix xref` to only print graphs with
+  with compile dependencies in them by passing the `--label` flag:
 
       $ mix xref graph --format cycles --label compile-connected
 
-  Which may look like this:
-
-      Cycle of length 3:
-
-          lib/c.ex
-          lib/b.ex (compile)
-          lib/a.ex
-
-  This means `c.ex` depends on `b.ex` at compile time. Any compile dependency
-  in a cycle is by definition a compile-connected dependency, which must be
-  generally avoided, as explained earlier in the module documentation.
-
   ## Shared options
 
   Those options are shared across all modes:
@@ -712,7 +748,7 @@ defmodule Mix.Tasks.Xref do
       shell.info([
         Exception.format_file_line(Path.relative_to_cwd(file), line),
         ?\s,
-        Atom.to_string(type),
+        trace_type(type),
         ?\s,
         format_module_or_mfa(module_or_mfa),
         " (#{mode})"
@@ -722,6 +758,9 @@ defmodule Mix.Tasks.Xref do
     end
   end
 
+  defp trace_type(:import), do: "import call"
+  defp trace_type(other), do: Atom.to_string(other)
+
   defp trace_label(nil), do: nil
   defp trace_label("compile"), do: :compile
   defp trace_label("export"), do: :export
@@ -905,20 +944,20 @@ defmodule Mix.Tasks.Xref do
     if files == [], do: nil, else: files
   end
 
-  defp write_graph(file_references, filter, opts) do
-    {file_references, aliases} = merge_groups(file_references, Keyword.get_values(opts, :group))
+  defp write_graph(all_references, filter, opts) do
+    {all_references, aliases} = merge_groups(all_references, Keyword.get_values(opts, :group))
 
-    file_references =
-      exclude(file_references, get_files(:exclude, opts, file_references, aliases))
+    all_references =
+      exclude(all_references, get_files(:exclude, opts, all_references, aliases))
 
-    sources = get_files(:source, opts, file_references, aliases)
-    sinks = get_files(:sink, opts, file_references, aliases)
+    sources = get_files(:source, opts, all_references, aliases)
+    sinks = get_files(:sink, opts, all_references, aliases)
 
     file_references =
       cond do
-        sinks -> sink_tree(file_references, sinks)
-        sources -> source_tree(file_references, sources)
-        true -> file_references
+        sinks -> sink_tree(all_references, sinks)
+        sources -> source_tree(all_references, sources)
+        true -> all_references
       end
 
     {found, count} =
@@ -927,17 +966,22 @@ defmodule Mix.Tasks.Xref do
           {roots, callback, count} =
             roots_and_callback(file_references, filter, sources, sinks, opts)
 
-          path = Keyword.get(opts, :output, "xref_graph.dot")
+          file_spec =
+            Mix.Utils.write_dot_graph!(
+              "xref_graph.dot",
+              "xref graph",
+              Enum.sort(roots),
+              callback,
+              opts
+            )
 
-          Mix.Utils.write_dot_graph!(path, "xref graph", Enum.sort(roots), callback, opts)
-
-          if path != "-" do
-            png_path = (path |> Path.rootname() |> Path.basename()) <> ".png"
+          if file_spec != "-" do
+            png_file_spec = (file_spec |> Path.rootname() |> Path.basename()) <> ".png"
 
             """
-            Generated #{inspect(path)} in the current directory. To generate a PNG:
+            Generated "#{Path.relative_to_cwd(file_spec)}". To generate a PNG:
 
-               dot -Tpng #{inspect(path)} -o #{inspect(png_path)}
+               dot -Tpng #{inspect(file_spec)} -o #{inspect(png_file_spec)}
 
             For more options see http://www.graphviz.org/.
             """
@@ -954,12 +998,31 @@ defmodule Mix.Tasks.Xref do
         "cycles" ->
           {:cycles, print_cycles(file_references, filter, opts)}
 
+        "json" ->
+          {roots, callback, count} =
+            roots_and_callback(file_references, filter, sources, sinks, opts)
+
+          file_spec =
+            Mix.Utils.write_json_tree!("xref_graph.json", Enum.sort(roots), callback, opts)
+
+          if file_spec != "-" do
+            Mix.shell().info("Generated \"#{file_spec}\".")
+          end
+
+          {:references, count}
+
         other when other in [nil, "plain", "pretty"] ->
           {roots, callback, count} =
             roots_and_callback(file_references, filter, sources, sinks, opts)
 
           Mix.Utils.print_tree(Enum.sort(roots), callback, opts)
 
+          if sources do
+            # We compute the tree again in case sinks are also given
+            file_references = source_tree(all_references, sources)
+            print_sources_cycles(file_references, sources, opts)
+          end
+
           {:references, count}
 
         other ->
@@ -1146,12 +1209,11 @@ defmodule Mix.Tasks.Xref do
     cycles =
       graph
       |> :digraph_utils.cyclic_strong_components()
-      |> Enum.reduce([], &inner_cycles(graph, Enum.sort(&1), &2))
-      |> Enum.map(&{length(&1), &1})
+      |> Enum.map(&{length(&1), add_labels(&1, graph)})
 
     cycles =
       if min = opts[:min_cycle_size] do
-        Enum.filter(cycles, &(elem(&1, 0) > min))
+        Enum.filter(cycles, &(elem(&1, 0) >= min))
       else
         cycles
       end
@@ -1169,20 +1231,29 @@ defmodule Mix.Tasks.Xref do
   defp cycle_filter_fn(:compile_connected), do: cycle_filter_fn(:compile)
   defp cycle_filter_fn(filter), do: fn {_node, type} -> type == filter end
 
-  defp inner_cycles(_graph, [], acc), do: acc
+  defp add_labels(vertices, graph) do
+    set = MapSet.new(vertices)
 
-  defp inner_cycles(graph, [v | vertices], acc) do
-    cycle = :digraph.get_cycle(graph, v)
-    inner_cycles(graph, vertices -- cycle, [label_cycle(cycle, graph) | acc])
+    vertices
+    |> Enum.map(fn v ->
+      {v,
+       graph
+       |> :digraph.out_neighbours(v)
+       |> cycle_label(set, graph, v, false)}
+    end)
+    |> Enum.sort()
   end
 
-  defp label_cycle([from, to | cycle], graph) do
-    {_edge, _v1, _v2, label} = :digraph.edge(graph, {from, to})
-    [{to, label} | label_cycle([to | cycle], graph)]
+  defp cycle_label([out | outs], set, graph, v, export?) do
+    case out in set && :digraph.edge(graph, {v, out}) do
+      {_, _, _, :compile} -> :compile
+      {_, _, _, :export} -> cycle_label(outs, set, graph, v, true)
+      _ -> cycle_label(outs, set, graph, v, export?)
+    end
   end
 
-  defp label_cycle([_from], _graph) do
-    []
+  defp cycle_label([], _set, _graph, _v, export?) do
+    if export?, do: :export, else: nil
   end
 
   defp print_cycles(references, filter, opts) do
@@ -1213,6 +1284,29 @@ defmodule Mix.Tasks.Xref do
     end)
   end
 
+  defp print_sources_cycles(references, sources, opts) do
+    with_digraph(references, fn graph ->
+      shell = Mix.shell()
+
+      graph
+      |> cycles(:compile, opts)
+      |> Enum.sort(:desc)
+      |> Enum.each(fn {length, cycle} ->
+        if source = Enum.find(sources, &List.keymember?(cycle, &1, 0)) do
+          shell.info("""
+
+          WARNING: Source #{source} is part of a cycle of #{length} nodes \
+          and this cycle has a compile dependency. Therefore source and the \
+          whole cycle will recompile whenever any of the files they depend \
+          on change. Run "mix xref graph --format stats --label compile-connected" \
+          to print compilation cycles and "mix help xref" for information on \
+          removing them\
+          """)
+        end
+      end)
+    end)
+  end
+
   ## Helpers
 
   defp apps(opts) do
