@@ -1,3 +1,10 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
+# NOTE: As this is a utils file it should not contain hard coded files
+# everything should be parameterized.
+
 defmodule Mix.Utils do
   @moduledoc false
 
@@ -257,15 +264,108 @@ defmodule Mix.Utils do
     |> Enum.uniq()
   end
 
+  @doc """
+  Handles writing the contents to either STDOUT or to a file, as specified
+  by the :output keyword in opts, defaulting to the provided default_file_spec.
+
+  If the resolved file specification is "-" then the contents is written to STDOUT,
+  otherwise if the file already exists it is renamed with a ".bak" suffix before
+  the contents is written.  The underlying IO operations will throw an exception
+  if there is an error.
+
+  Returns the name of the file written to, or "-" if the output was to STDOUT.
+  This function is made public mostly for testing.
+  """
+  @spec write_according_to_opts!(Path.t(), iodata(), write_opts) :: Path.t()
+  def write_according_to_opts!(default_file_spec, contents, opts) do
+    file_spec = Keyword.get(opts, :output, default_file_spec)
+
+    if file_spec == "-" do
+      IO.write(contents)
+    else
+      if File.exists?(file_spec) do
+        new_file_spec = "#{file_spec}.bak"
+        File.rename!(file_spec, new_file_spec)
+      end
+
+      File.write!(file_spec, contents)
+    end
+
+    # return the file_spec just in case the caller has a use for it.
+    file_spec
+  end
+
+  @doc """
+  Outputs the given tree according to the callback as a two level
+  map of maps in JSON format.
+
+  The callback will be invoked for each node and it
+  must return a `{printed, children}` tuple.
+
+  If the `:output` option is `-` then prints to standard output,
+  see write_according_to_opts!/3 for details.
+  """
+  @spec write_json_tree!(Path.t(), [node], (node -> {formatted_node, [node]}), write_opts) ::
+          Path.t()
+        when node: term()
+  def write_json_tree!(default_file_spec, nodes, callback, opts \\ []) do
+    src_map = build_json_tree(_src_map = %{}, nodes, callback)
+    write_according_to_opts!(default_file_spec, JSON.encode_to_iodata!(src_map), opts)
+  end
+
+  defp build_json_tree(src_map, [], _callback), do: src_map
+
+  defp build_json_tree(src_map, nodes, callback) do
+    Enum.reduce(nodes, src_map, fn node, src_map ->
+      {{name, _}, children} = callback.(node)
+
+      if Map.has_key?(src_map, name) do
+        src_map
+      else
+        sink_map =
+          Enum.reduce(children, %{}, fn {name, info}, sink_map ->
+            info = if info == nil, do: "runtime", else: Atom.to_string(info)
+            Map.put(sink_map, name, info)
+          end)
+
+        Map.put(src_map, name, sink_map)
+        |> build_json_tree(children, callback)
+      end
+    end)
+  end
+
   @type formatted_node :: {name :: String.Chars.t(), edge_info :: String.Chars.t()}
 
+  @typedoc """
+  Options for `write_according_to_opts!/3`, `write_json_tree!/4`, and `write_dot_graph!/5`.
+  """
+  @type write_opts :: [
+          output: String.t()
+        ]
+
+  @typedoc """
+  Options for `print_tree/3`.
+  """
+  @type print_tree_opts :: [
+          format: String.t()
+        ]
+
+  @typedoc """
+  Options for `read_path/2`.
+  """
+  @type read_path_opts :: [
+          timeout: pos_integer(),
+          sha512: String.t()
+        ]
+
   @doc """
   Prints the given tree according to the callback.
 
   The callback will be invoked for each node and it
   must return a `{printed, children}` tuple.
   """
-  @spec print_tree([node], (node -> {formatted_node, [node]}), keyword) :: :ok when node: term()
+  @spec print_tree([node], (node -> {formatted_node, [node]}), print_tree_opts) :: :ok
+        when node: term()
   def print_tree(nodes, callback, opts \\ []) do
     pretty? =
       case Keyword.get(opts, :format) do
@@ -309,7 +409,7 @@ defmodule Mix.Utils do
   end
 
   defp depth(_pretty?, []), do: ""
-  defp depth(pretty?, depth), do: Enum.reverse(depth) |> tl |> Enum.map(&entry(pretty?, &1))
+  defp depth(pretty?, depth), do: Enum.reverse(depth) |> tl() |> Enum.map(&entry(pretty?, &1))
 
   defp entry(false, true), do: "|   "
   defp entry(false, false), do: "    "
@@ -329,25 +429,21 @@ defmodule Mix.Utils do
   The callback will be invoked for each node and it
   must return a `{printed, children}` tuple.
 
-  If `path` is `-`, prints the output to standard output.
+  If the `:output` option is `-` then prints to standard output,
+  see write_according_to_opts!/3 for details.
   """
   @spec write_dot_graph!(
           Path.t(),
           String.t(),
           [node],
           (node -> {formatted_node, [node]}),
-          keyword
-        ) :: :ok
+          write_opts
+        ) :: Path.t()
         when node: term()
-  def write_dot_graph!(path, title, nodes, callback, _opts \\ []) do
+  def write_dot_graph!(default_file_spec, title, nodes, callback, opts \\ []) do
     {dot, _} = build_dot_graph(make_ref(), nodes, MapSet.new(), callback)
     contents = ["digraph ", quoted(title), " {\n", dot, "}\n"]
-
-    if path == "-" do
-      IO.write(contents)
-    else
-      File.write!(path, contents)
-    end
+    write_according_to_opts!(default_file_spec, contents, opts)
   end
 
   defp build_dot_graph(_parent, [], seen, _callback), do: {[], seen}
@@ -552,16 +648,14 @@ defmodule Mix.Utils do
   ## Options
 
     * `:sha512` - checks against the given SHA-512 checksum. Returns
-      `{:checksum, message}` in case it fails. This option is required
-      for URLs unless the `:unsafe_uri` is given (WHICH IS NOT RECOMMENDED
-      unless another security mechanism is in place, such as private keys)
+      `{:checksum, message}` in case it fails
 
     * `:timeout` - times out the request after the given milliseconds.
       Returns `{:remote, timeout_message}` if it fails. Defaults to 60
       seconds
 
   """
-  @spec read_path(String.t(), keyword) ::
+  @spec read_path(String.t(), read_path_opts) ::
           {:ok, binary}
           | :badpath
           | {:remote, String.t()}
@@ -572,8 +666,7 @@ defmodule Mix.Utils do
       url?(path) ->
         task =
           Task.async(fn ->
-            with :ok <- require_checksum(opts),
-                 {:ok, binary} <- read_httpc(path) do
+            with {:ok, binary} <- read_httpc(path) do
               checksum(binary, opts)
             end
           end)
@@ -597,19 +690,6 @@ defmodule Mix.Utils do
 
   @checksums [:sha512]
 
-  defp require_checksum(opts) do
-    cond do
-      Keyword.take(opts, @checksums) != [] ->
-        :ok
-
-      Keyword.get(opts, :unsafe_uri) ->
-        :ok
-
-      true ->
-        {:checksum, "fetching from URIs require a checksum to be given"}
-    end
-  end
-
   defp checksum(binary, opts) do
     Enum.find_value(@checksums, {:ok, binary}, fn hash ->
       with expected when expected != nil <- opts[hash],
@@ -794,4 +874,37 @@ defmodule Mix.Utils do
 
     [proxy_auth: {user, pass}]
   end
+
+  @doc """
+  Returns the user id of the currently running process.
+
+  The user id is obtained by creating a temporary file and checking
+  it's UID. Note that the UID may be `nil` on non-Unix systems.
+  """
+  def detect_user_id!() do
+    case Mix.State.fetch(:user_id) do
+      {:ok, uid} ->
+        uid
+
+      :error ->
+        dir = System.tmp_dir!()
+        File.mkdir_p!(dir)
+        rand = :crypto.strong_rand_bytes(3) |> Base.url_encode64()
+        path = Path.join(dir, "mix_user_check_#{System.os_time()}_#{rand}")
+
+        uid =
+          try do
+            File.touch!(path)
+            File.stat!(path)
+          else
+            %{uid: :undefined} -> nil
+            %{uid: uid} -> uid
+          after
+            File.rm(path)
+          end
+
+        Mix.State.put(:user_id, uid)
+        uid
+    end
+  end
 end
