@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../../test_helper.exs", __DIR__)
 
 defmodule Mix.Local.InstallerTest do
@@ -26,6 +30,8 @@ defmodule Mix.Local.InstallerTest do
     assert config[:app] == :git_repo
     assert config[:deps_path] =~ ~r/mix-local-installer-fetcher-.*\/deps/
     assert config[:lockfile] =~ ~r/mix-local-installer-fetcher-.*\/mix.lock/
+  after
+    purge([GitRepo.MixProject, Mix.Local.Installer.MixProject])
   end
 
   test "parse_args Git" do
