@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../../test_helper.exs", __DIR__)
 
 defmodule Mix.Tasks.Compile.ElixirTest do
@@ -6,13 +10,29 @@ defmodule Mix.Tasks.Compile.ElixirTest do
   import ExUnit.CaptureIO
   alias Mix.Task.Compiler.Diagnostic
 
+  @old_time {{2010, 1, 1}, {0, 0, 0}}
+  @elixir_otp_version {System.version(), :erlang.system_info(:otp_release)}
+
   def trace(event, env) do
     send(__MODULE__, {event, env})
     :ok
   end
 
-  @old_time {{2010, 1, 1}, {0, 0, 0}}
-  @elixir_otp_version {System.version(), :erlang.system_info(:otp_release)}
+  defp mtime(path) do
+    File.stat!(path).mtime
+  end
+
+  defp mark_as_old!(path) do
+    mtime = mtime(path)
+    File.touch!(path, @old_time)
+    mtime
+  end
+
+  defp purge_protocol(module) do
+    :code.del_path(:filename.absname(~c"_build/dev/lib/sample/consolidated"))
+    :code.purge(module)
+    :code.delete(module)
+  end
 
   test "compiles a project without per environment build" do
     Mix.ProjectStack.post_config(build_per_environment: false)
@@ -128,7 +148,7 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       refute_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-      assert File.stat!("_build/dev/lib/sample/.mix/compile.elixir").mtime > @old_time
+      assert mtime("_build/dev/lib/sample/.mix/compile.elixir") > @old_time
 
       ensure_touched(__ENV__.file, "_build/dev/lib/sample/.mix/compile.elixir")
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
@@ -245,7 +265,7 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       refute_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-      assert File.stat!("_build/dev/lib/sample/.mix/compile.elixir").mtime > @old_time
+      assert mtime("_build/dev/lib/sample/.mix/compile.elixir") > @old_time
     end)
   after
     Application.put_env(:elixir, :dbg_callback, {Macro, :dbg, []})
@@ -468,6 +488,9 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       recompile = fn ->
         Mix.ProjectStack.pop()
         Mix.Project.push(MixTest.Case.Sample)
+        File.touch!("_build/dev/lib/sample/.mix/compile.elixir", @old_time)
+        ensure_touched("mix.lock")
+        ensure_touched("_build/dev/lib/sample/.mix/compile.lock")
         Mix.Tasks.WillRecompile.run([])
         Mix.Tasks.Compile.Elixir.run(["--verbose"])
       end
@@ -477,44 +500,36 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       %{"logger": :unused}
       """)
 
-      ensure_touched("mix.lock")
-      File.touch!("_build/dev/lib/sample/.mix/compile.elixir", @old_time)
       assert recompile.() == {:ok, []}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       refute_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-      assert File.stat!("_build/dev/lib/sample/.mix/compile.elixir").mtime > @old_time
+      assert mtime("_build/dev/lib/sample/.mix/compile.elixir") > @old_time
 
       # Changing lock recompiles
       File.write!("mix.lock", """
       %{"logger": :another}
       """)
 
-      ensure_touched("mix.lock")
-      File.touch!("_build/dev/lib/sample/.mix/compile.elixir", @old_time)
       assert recompile.() == {:ok, []}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       refute_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-      assert File.stat!("_build/dev/lib/sample/.mix/compile.elixir").mtime > @old_time
+      assert mtime("_build/dev/lib/sample/.mix/compile.elixir") > @old_time
 
       # Removing a lock recompiles
       File.write!("mix.lock", """
       %{}
       """)
 
-      ensure_touched("mix.lock")
-      File.touch!("_build/dev/lib/sample/.mix/compile.elixir", @old_time)
       assert recompile.() == {:ok, []}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       refute_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-      assert File.stat!("_build/dev/lib/sample/.mix/compile.elixir").mtime > @old_time
+      assert mtime("_build/dev/lib/sample/.mix/compile.elixir") > @old_time
 
       # Adding an unknown dependency returns :ok but does not recompile
       File.write!("mix.lock", """
       %{"unknown": :unknown}
       """)
 
-      # We use ensure_touched because an outdated manifest would recompile anyway.
-      ensure_touched("mix.lock", "_build/dev/lib/sample/.mix/compile.elixir")
       assert recompile.() == {:ok, []}
       refute_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       refute_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
@@ -584,9 +599,7 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       Mix.Tasks.Compile.run([])
       assert Mix.Dep.ElixirSCM.read() == {:ok, @elixir_otp_version, Mix.SCM.Path}
 
-      assert File.stat!("_build/dev/lib/sample/.mix/compile.elixir_scm").mtime >
-               @old_time
-
+      assert mtime("_build/dev/lib/sample/.mix/compile.elixir_scm") > @old_time
       refute File.exists?("_build/dev/lib/sample/consolidated/.to_be_removed")
     end)
   end
@@ -608,8 +621,7 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       Mix.Tasks.Compile.run([])
       assert Mix.Dep.ElixirSCM.read() == {:ok, @elixir_otp_version, Mix.SCM.Path}
 
-      assert File.stat!("_build/dev/lib/sample/.mix/compile.elixir_scm").mtime >
-               @old_time
+      assert mtime("_build/dev/lib/sample/.mix/compile.elixir_scm") > @old_time
     end)
   end
 
@@ -720,38 +732,6 @@ defmodule Mix.Tasks.Compile.ElixirTest do
     end)
   end
 
-  test "purges consolidation path if asked" do
-    in_fixture("no_mixfile", fn ->
-      File.write!("lib/a.ex", """
-      defmodule A do
-        defstruct []
-      end
-
-      defimpl Inspect, for: A do
-        def inspect(_, _), do: "sample"
-      end
-      """)
-
-      Mix.Project.push(MixTest.Case.Sample)
-      assert Mix.Tasks.Compile.run([]) == {:ok, []}
-      assert inspect(struct(A, [])) == "sample"
-
-      purge([A, B, Inspect.A])
-      Mix.Task.clear()
-
-      assert capture_io(:stderr, fn ->
-               {:ok, [_]} = Mix.Tasks.Compile.run(["--force"])
-             end) =~
-               "the Inspect protocol has already been consolidated"
-
-      purge([A, B, Inspect.A])
-      Mix.Task.clear()
-      consolidation = Mix.Project.consolidation_path()
-      args = ["--force", "--purge-consolidation-path-if-stale", consolidation]
-      assert Mix.Tasks.Compile.run(args) == {:ok, []}
-    end)
-  end
-
   test "recompiles mtime changed files if content changed but not length" do
     in_fixture("no_mixfile", fn ->
       Mix.Project.push(MixTest.Case.Sample)
@@ -953,13 +933,13 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       assert_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
       purge([A, B])
 
-      capture_io(:stderr, fn ->
-        File.write!("lib/a.ex", "defmodule B, do: :not_ok")
-        assert {:ok, [_ | _]} = Mix.Tasks.Compile.Elixir.run(["--verbose"])
-        assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
-        assert_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-        purge([A, B])
-      end)
+      assert capture_io(:stderr, fn ->
+               File.write!("lib/a.ex", "defmodule B, do: :not_ok")
+               assert {:ok, [_]} = Mix.Tasks.Compile.Elixir.run(["--verbose"])
+               assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
+               assert_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
+               purge([A, B])
+             end) =~ "redefining module B"
 
       capture_io(:stderr, fn ->
         File.write!("lib/a.ex", "defmodule A, do: :ok")
@@ -985,6 +965,10 @@ defmodule Mix.Tasks.Compile.ElixirTest do
         @external_resource "lib/a.eex"
         @external_resource #{inspect(tmp)}
         def a, do: :ok
+
+        if System.get_env("RAISE_EXTERNAL_RESOURCE") do
+          raise "oops"
+        end
       end
       """)
 
@@ -998,7 +982,16 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       File.touch!("lib/a.eex", {{2038, 1, 1}, {0, 0, 0}})
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:noop, []}
 
+      # Force recompilation but fail compilation
       force_recompilation("lib/a.eex")
+      System.put_env("RAISE_EXTERNAL_RESOURCE", "1")
+
+      assert capture_io(:stderr, fn ->
+               assert {:error, [_]} = Mix.Tasks.Compile.Elixir.run(["--verbose"])
+             end) =~ "oops"
+
+      # Delete failure reason, it will now compile for good
+      System.delete_env("RAISE_EXTERNAL_RESOURCE")
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       refute_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
@@ -1215,7 +1208,7 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       assert Mix.Tasks.Compile.Elixir.run(["--verbose", "--ignore-module-conflict"]) == {:ok, []}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       assert_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-      refute function_exported?(A, :one, 0)
+      refute Code.ensure_loaded?(A) and function_exported?(A, :one, 0)
 
       Mix.shell().flush()
       purge([A])
@@ -1224,7 +1217,7 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       Mix.Tasks.Compile.Elixir.run(["--verbose"])
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       refute_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-      assert function_exported?(A, :one, 0)
+      assert Code.ensure_loaded?(A) and function_exported?(A, :one, 0)
     end)
   end
 
@@ -1293,19 +1286,10 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       end
       """)
 
-      File.write!("lib/c.ex", """
-      defmodule C do
-        @compile {:autoload, false}
-
-        def __mix_recompile__?(), do: true
-      end
-      """)
-
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
-      assert_received {:mix_shell, :info, ["Compiling 3 files (.ex)"]}
+      assert_received {:mix_shell, :info, ["Compiling 2 files (.ex)"]}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
       assert_received {:mix_shell, :info, ["Compiled lib/b.ex"]}
-      assert_received {:mix_shell, :info, ["Compiled lib/c.ex"]}
 
       # Mix recompile should work even if the compile path
       # was removed and the module purged
@@ -1315,6 +1299,7 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
       assert_received {:mix_shell, :info, ["Compiling 1 file (.ex)"]}
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
+      Mix.shell().flush()
 
       File.rm!("lib/a.ex")
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:ok, []}
@@ -1343,7 +1328,7 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       assert_received {:mix_shell, :info, ["Compiled lib/a.ex"]}
 
       assert Mix.Tasks.Compile.Elixir.run(["--verbose"]) == {:noop, []}
-      assert File.stat!("lib/a.ex").mtime == @old_time
+      assert mtime("lib/a.ex") == @old_time
 
       Agent.update(:mix_recompile_raise, fn _ -> true end)
 
@@ -1689,6 +1674,16 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       assert output =~ "B.foo/0 is undefined or private"
       assert output =~ "B.bar/0 is undefined or private"
       assert output =~ "AFTER_VERIFY"
+
+      # But we can also disable verification if we want to
+      output =
+        capture_io(:stderr, fn ->
+          Mix.Tasks.Compile.Elixir.run(["--no-verification", "--force"])
+        end)
+
+      refute output =~ "B.foo/0 is undefined or private"
+      refute output =~ "B.bar/0 is undefined or private"
+      refute output =~ "AFTER_VERIFY"
     end)
   end
 
@@ -1728,4 +1723,183 @@ defmodule Mix.Tasks.Compile.ElixirTest do
       assert Mix.Tasks.Compile.Elixir.run(["--no-optional-deps"]) == {:ok, []}
     end)
   end
+
+  describe "consolidation protocols" do
+    test "with local protocols", context do
+      in_tmp(context.test, fn ->
+        Mix.Project.push(MixTest.Case.Sample)
+
+        File.mkdir_p!("lib")
+        assert Mix.Task.run("compile")
+
+        # Define a local protocol
+        File.write!("lib/protocol.ex", """
+        defprotocol Compile.Protocol do
+          def foo(a, b)
+        end
+        """)
+
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+        mark_as_old!("_build/dev/lib/sample/consolidated/Elixir.Compile.Protocol.beam")
+
+        # Implement a local protocol
+        File.write!("lib/impl.ex", """
+        defimpl Compile.Protocol, for: Integer do
+          def foo(a, b), do: a + b
+        end
+        """)
+
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+
+        assert mark_as_old!("_build/dev/lib/sample/consolidated/Elixir.Compile.Protocol.beam") !=
+                 @old_time
+
+        # Delete a local implementation
+        File.rm!("lib/impl.ex")
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+
+        assert mark_as_old!("_build/dev/lib/sample/consolidated/Elixir.Compile.Protocol.beam") !=
+                 @old_time
+
+        # Delete a local protocol
+        File.rm!("lib/protocol.ex")
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+        refute File.regular?("_build/dev/lib/sample/consolidated/Elixir.Compile.Protocol.beam")
+      end)
+    end
+
+    test "compiles after converting a protocol into a standard module", context do
+      in_tmp(context.test, fn ->
+        Mix.Project.push(MixTest.Case.Sample)
+
+        File.mkdir_p!("lib")
+        Mix.Task.run("compile")
+        purge_protocol(Compile.Protocol)
+
+        # Define a local protocol
+        File.write!("lib/protocol.ex", """
+        defprotocol Compile.Protocol do
+          def foo(a)
+        end
+
+        defimpl Compile.Protocol, for: Integer do
+          def foo(a), do: a
+        end
+        """)
+
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+        mark_as_old!("_build/dev/lib/sample/consolidated/Elixir.Compile.Protocol.beam")
+        File.rm!("lib/protocol.ex")
+
+        # Define a standard module
+        File.write!("lib/protocol.ex", """
+        defmodule Compile.Protocol do
+        end
+        """)
+
+        purge_protocol(Compile.Protocol)
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+
+        # Delete a local protocol
+        File.rm!("lib/protocol.ex")
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+        refute File.regular?("_build/dev/lib/sample/consolidated/Elixir.Compile.Protocol.beam")
+      end)
+    end
+
+    test "with deps protocols", context do
+      in_tmp(context.test, fn ->
+        Mix.Project.push(MixTest.Case.Sample)
+
+        File.mkdir_p!("lib")
+        Mix.Task.run("compile")
+        purge_protocol(String.Chars)
+        mark_as_old!("_build/dev/lib/sample/consolidated/Elixir.String.Chars.beam")
+
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:noop, []}
+        assert mtime("_build/dev/lib/sample/consolidated/Elixir.String.Chars.beam") == @old_time
+
+        # Implement a deps protocol
+        File.write!("lib/struct.ex", """
+        defmodule Compile.Protocol.Struct do
+          defstruct a: nil
+          defimpl String.Chars do
+            def to_string(_), do: "ok"
+          end
+        end
+        """)
+
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+
+        assert mark_as_old!("_build/dev/lib/sample/consolidated/Elixir.String.Chars.beam") !=
+                 @old_time
+
+        # Delete the local implementation
+        File.rm!("lib/struct.ex")
+        assert Mix.Tasks.Compile.Elixir.run([]) == {:ok, []}
+
+        assert mark_as_old!("_build/dev/lib/sample/consolidated/Elixir.String.Chars.beam") !=
+                 @old_time
+      end)
+    end
+
+    test "keep relative path to their source" do
+      in_fixture("no_mixfile", fn ->
+        Mix.Project.push(MixTest.Case.Sample)
+        Mix.Task.run("compile")
+
+        # Load consolidated
+        :code.add_patha(~c"_build/dev/lib/sample/consolidated")
+        :code.purge(Enumerable)
+        :code.delete(Enumerable)
+
+        try do
+          Enumerable.impl_for!(:oops)
+        rescue
+          Protocol.UndefinedError ->
+            assert [{_, _, _, [file: ~c"lib/enum.ex"] ++ _} | _] = __STACKTRACE__
+        else
+          _ ->
+            flunk("Enumerable.impl_for!/1 should have failed")
+        after
+          purge_protocol(Enumerable)
+        end
+      end)
+    end
+
+    test "purges consolidation path if asked" do
+      in_fixture("no_mixfile", fn ->
+        File.write!("lib/a.ex", """
+        defmodule A do
+          defstruct []
+        end
+
+        defimpl Collectable, for: A do
+          def into(_), do: fn _, _  -> raise "oops" end
+        end
+        """)
+
+        Mix.Project.push(MixTest.Case.Sample)
+
+        # At this point, it may or may not emit warnings, but we don't care,
+        # we validate both scenarios next.
+        assert {:ok, _} = Mix.Tasks.Compile.run([])
+        assert is_function(Collectable.into(struct(A, [])), 2)
+
+        purge([A, B, Collectable.A])
+        Mix.Task.clear()
+
+        assert capture_io(:stderr, fn ->
+                 {:ok, [_]} = Mix.Tasks.Compile.run(["--force"])
+               end) =~
+                 "the Collectable protocol has already been consolidated"
+
+        purge([A, B, Collectable.A])
+        Mix.Task.clear()
+        consolidation = Mix.Project.consolidation_path()
+        args = ["--force", "--purge-consolidation-path-if-stale", consolidation]
+        assert Mix.Tasks.Compile.run(args) == {:ok, []}
+      end)
+    end
+  end
 end
