@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../../test_helper.exs", __DIR__)
 
 defmodule Mix.Tasks.DepsGitTest do
@@ -166,13 +170,13 @@ defmodule Mix.Tasks.DepsGitTest do
 
       # Compile the dependencies
       Mix.Tasks.Deps.Compile.run([])
-      assert File.exists?("_build/dev/lib/deps_on_git_repo/.mix/compile.fetch")
-      assert File.exists?("_build/dev/lib/git_repo/.mix/compile.fetch")
+      assert File.exists?("_build/dev/lib/deps_on_git_repo/.mix/compile.elixir_scm")
+      assert File.exists?("_build/dev/lib/git_repo/.mix/compile.elixir_scm")
 
       # Now update children and make sure it propagates
       Mix.Tasks.Deps.Update.run(["git_repo"])
-      refute File.exists?("_build/dev/lib/deps_on_git_repo/.mix/compile.fetch")
-      refute File.exists?("_build/dev/lib/git_repo/.mix/compile.fetch")
+      refute File.exists?("_build/dev/lib/deps_on_git_repo/.mix/compile.elixir_scm")
+      refute File.exists?("_build/dev/lib/git_repo/.mix/compile.elixir_scm")
 
       # Clear tasks to recompile Git repo but unload it so...
       purge([GitRepo])
@@ -207,7 +211,7 @@ defmodule Mix.Tasks.DepsGitTest do
     purge([GitRepo, GitRepo.MixProject])
   end
 
-  test "all dependencies are up to date" do
+  test "all dependencies have been fetched" do
     in_fixture("no_mixfile", fn ->
       Mix.Project.push(GitApp)
 
@@ -216,7 +220,7 @@ defmodule Mix.Tasks.DepsGitTest do
       assert_received {:mix_shell, :info, [^message]}
 
       Mix.Tasks.Deps.Get.run([])
-      assert_received {:mix_shell, :info, ["All dependencies are up to date"]}
+      assert_received {:mix_shell, :info, ["All dependencies have been fetched"]}
     end)
   after
     purge([GitRepo, GitRepo.MixProject])
@@ -404,7 +408,7 @@ defmodule Mix.Tasks.DepsGitTest do
       assert File.exists?("deps/git_repo/lib/git_repo.ex")
       assert File.read!("mix.lock") =~ last
       assert File.read!("deps/deps_on_git_repo/lib/deps_on_git_repo.ex") =~ "GitRepo.hello()"
-      refute File.exists?("_build/dev/lib/git_repo/.mix/compile.fetch")
+      refute File.exists?("_build/dev/lib/git_repo/.mix/compile.elixir_scm")
     end)
   after
     purge([GitRepo, GitRepo.MixProject])
@@ -437,7 +441,7 @@ defmodule Mix.Tasks.DepsGitTest do
       assert File.exists?("deps/git_repo/lib/git_repo.ex")
       assert File.read!("mix.lock") =~ last
       assert File.read!("deps/deps_on_git_repo/lib/deps_on_git_repo.ex") =~ "GitRepo.hello()"
-      refute File.exists?("_build/dev/lib/git_repo/.mix/compile.fetch")
+      refute File.exists?("_build/dev/lib/git_repo/.mix/compile.elixir_scm")
     end)
   after
     purge([GitRepo, GitRepo.MixProject])
