@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../../test_helper.exs", __DIR__)
 
 defmodule Mix.Tasks.NewTest do
@@ -16,7 +20,6 @@ defmodule Mix.Tasks.NewTest do
       assert_file("hello_world/.gitignore")
 
       assert_file("hello_world/lib/hello_world.ex", ~r/defmodule HelloWorld do/)
-
       assert_file("hello_world/test/test_helper.exs", ~r/ExUnit.start()/)
 
       assert_file("hello_world/test/hello_world_test.exs", fn file ->
@@ -127,7 +130,23 @@ defmodule Mix.Tasks.NewTest do
     end)
   end
 
-  test "new inside umbrella" do
+  test "new inside umbrella root" do
+    in_fixture("umbrella_dep/deps/umbrella", fn ->
+      Mix.Tasks.New.run(["apps/hello_world"])
+
+      assert_file("apps/hello_world/mix.exs", fn file ->
+        assert file =~ "deps_path: \"../../deps\""
+        assert file =~ "lockfile: \"../../mix.lock\""
+      end)
+
+      # Ensure formatting is setup and consistent.
+      File.cd!("apps/hello_world", fn ->
+        Mix.Tasks.Format.run(["--check-formatted"])
+      end)
+    end)
+  end
+
+  test "new inside umbrella apps" do
     in_fixture("umbrella_dep/deps/umbrella", fn ->
       File.cd!("apps", fn ->
         Mix.Tasks.New.run(["hello_world"])
@@ -160,6 +179,18 @@ defmodule Mix.Tasks.NewTest do
                      Mix.Tasks.New.run(["007invalid"])
                    end
 
+      assert_raise Mix.Error,
+                   ~r"Cannot use application name \"tools\" because it is already used by Erlang/OTP or Elixir",
+                   fn ->
+                     Mix.Tasks.New.run(["tools"])
+                   end
+
+      assert_raise Mix.Error,
+                   ~r"Cannot use application name \"pa\" because it is already used by Erlang/OTP or Elixir",
+                   fn ->
+                     Mix.Tasks.New.run(["pa"])
+                   end
+
       assert_raise Mix.Error,
                    ~r"followed by lowercase ASCII letters, numbers, or underscores",
                    fn ->
@@ -185,18 +216,6 @@ defmodule Mix.Tasks.NewTest do
       end
     end)
 
-    in_tmp("new with an already taken application name", fn ->
-      assert_raise Mix.Error, ~r"Module name \w+ is already taken", fn ->
-        Mix.Tasks.New.run(["mix"])
-      end
-    end)
-
-    in_tmp("new with an already taken application name from the app option", fn ->
-      assert_raise Mix.Error, ~r"Module name \w+ is already taken", fn ->
-        Mix.Tasks.New.run(["valid", "--app", "mix"])
-      end
-    end)
-
     in_tmp("new with an already taken module name from the module options", fn ->
       assert_raise Mix.Error, ~r"Module name \w+ is already taken", fn ->
         Mix.Tasks.New.run(["valid", "--module", "Mix"])
