@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../../test_helper.exs", __DIR__)
 
 defmodule Mix.Tasks.TestTest do
@@ -5,53 +9,52 @@ defmodule Mix.Tasks.TestTest do
 
   describe "ex_unit_opts/1" do
     test "returns ex unit options" do
-      assert ex_unit_opts_from_given(unknown: "ok", seed: 13) == [seed: 13]
+      assert filtered_ex_unit_opts(unknown: "ok", seed: 13) == [seed: 13]
     end
 
     test "returns includes and excludes" do
       included = [include: [:focus, key: "val"]]
-      assert ex_unit_opts_from_given(include: "focus", include: "key:val") == included
+      assert filtered_ex_unit_opts(include: "focus", include: "key:val") == included
 
       excluded = [exclude: [:focus, key: "val"]]
-      assert ex_unit_opts_from_given(exclude: "focus", exclude: "key:val") == excluded
+      assert filtered_ex_unit_opts(exclude: "focus", exclude: "key:val") == excluded
     end
 
     test "translates :only into includes and excludes" do
-      assert ex_unit_opts_from_given(only: "focus") == [include: [:focus], exclude: [:test]]
+      assert filtered_ex_unit_opts(only: "focus") == [include: [:focus], exclude: [:test]]
 
       only = [include: [:focus, :special], exclude: [:test]]
-      assert ex_unit_opts_from_given(only: "focus", include: "special") == only
+      assert filtered_ex_unit_opts(only: "focus", include: "special") == only
+    end
+
+    test "translates :name_pattern into includes and excludes" do
+      assert [include: [test: hello_regex, test: world_regex], exclude: [:test]] =
+               filtered_ex_unit_opts(name_pattern: ~r/hello/, name_pattern: ~r/world/)
+
+      assert Regex.match?(hello_regex, "hello")
+      refute Regex.match?(hello_regex, "world")
+      refute Regex.match?(world_regex, "hello")
+      assert Regex.match?(world_regex, "world")
     end
 
     test "translates :color into list containing an enabled key-value pair" do
-      assert ex_unit_opts_from_given(color: false) == [colors: [enabled: false]]
-      assert ex_unit_opts_from_given(color: true) == [colors: [enabled: true]]
+      assert filtered_ex_unit_opts(color: false) == [colors: [enabled: false]]
+      assert filtered_ex_unit_opts(color: true) == [colors: [enabled: true]]
     end
 
     test "translates :formatter into list of modules" do
-      assert ex_unit_opts_from_given(formatter: "A.B") == [formatters: [A.B]]
+      assert filtered_ex_unit_opts(formatter: "A.B") == [formatters: [A.B]]
     end
 
     test "accepts custom :exit_status" do
-      assert {:exit_status, 5} in ex_unit_opts(exit_status: 5)
+      assert {:exit_status, 5} in ex_unit_opts(exit_status: 5, failures_manifest_path: "foo.bar")
     end
 
     test "includes some default options" do
-      assert ex_unit_opts([]) == [
+      assert ex_unit_opts(failures_manifest_path: "foo.bar") == [
                autorun: false,
                exit_status: 2,
-               failures_manifest_path:
-                 Path.join(Mix.Project.manifest_path(), ".mix_test_failures")
-             ]
-    end
-
-    test "respect failures_manifest_path option" do
-      custom_manifest_file = Path.join(Mix.Project.manifest_path(), ".mix_test_failures_custom")
-
-      assert ex_unit_opts(failures_manifest_path: custom_manifest_file) == [
-               autorun: false,
-               exit_status: 2,
-               failures_manifest_path: custom_manifest_file
+               failures_manifest_path: "foo.bar"
              ]
     end
 
@@ -60,8 +63,9 @@ defmodule Mix.Tasks.TestTest do
       ex_unit_opts
     end
 
-    defp ex_unit_opts_from_given(passed) do
-      passed
+    defp filtered_ex_unit_opts(opts) do
+      opts
+      |> Keyword.put(:failures_manifest_path, "foo.bar")
       |> ex_unit_opts()
       |> Keyword.drop([:failures_manifest_path, :autorun, :exit_status])
     end
@@ -144,6 +148,7 @@ defmodule Mix.Tasks.TestTest do
   end
 
   describe "--cover" do
+    @describetag :cover
     test "reports the coverage of each app's modules in an umbrella" do
       in_fixture("umbrella_test", fn ->
         # This fixture by default results in coverage above the default threshold
@@ -155,12 +160,12 @@ defmodule Mix.Tasks.TestTest do
         assert output =~ """
                Generating cover results ...
 
-               Percentage | Module
-               -----------|--------------------------
-                  100.00% | Bar.Protocol
-                  100.00% | Bar.Protocol.BitString
-               -----------|--------------------------
-                  100.00% | Total
+               | Percentage | Module                 |
+               |------------|------------------------|
+               |    100.00% | Bar.Protocol           |
+               |    100.00% | Bar.Protocol.BitString |
+               |------------|------------------------|
+               |    100.00% | Total                  |
                """
 
         assert output =~ "1 test, 0 failures"
@@ -169,11 +174,11 @@ defmodule Mix.Tasks.TestTest do
         assert output =~ """
                Generating cover results ...
 
-               Percentage | Module
-               -----------|--------------------------
-                  100.00% | Foo
-               -----------|--------------------------
-                  100.00% | Total
+               | Percentage | Module     |
+               |------------|------------|
+               |    100.00% | Foo        |
+               |------------|------------|
+               |    100.00% | Total      |
                """
 
         # We skip a test in bar to force coverage below the default threshold
@@ -200,15 +205,15 @@ defmodule Mix.Tasks.TestTest do
                Importing cover results: apps/bar/cover/default.coverdata
                Importing cover results: apps/foo/cover/default.coverdata
 
-               Percentage | Module
-               -----------|--------------------------
-                  100.00% | Bar
-                  100.00% | Bar.Ignore
-                  100.00% | Bar.Protocol
-                  100.00% | Bar.Protocol.BitString
-                  100.00% | Foo
-               -----------|--------------------------
-                  100.00% | Total
+               | Percentage | Module                 |
+               |------------|------------------------|
+               |    100.00% | Bar                    |
+               |    100.00% | Bar.Ignore             |
+               |    100.00% | Bar.Protocol           |
+               |    100.00% | Bar.Protocol.BitString |
+               |    100.00% | Foo                    |
+               |------------|------------------------|
+               |    100.00% | Total                  |
                """
       end)
     end
@@ -247,7 +252,7 @@ defmodule Mix.Tasks.TestTest do
         # Of the passing tests, 1 is tagged with `@tag :foo`.
         # But only the failing test with that tag should run.
         output = mix(["test", "--failed", "--only", "foo"])
-        assert output =~ "2 tests, 1 failure, 1 excluded"
+        assert output =~ "1 test, 1 failure (1 excluded)"
 
         # Run again to give it a chance to record as passed
         System.put_env("PASS_FAILING_TESTS", "true")
@@ -268,6 +273,18 @@ defmodule Mix.Tasks.TestTest do
     after
       System.delete_env("PASS_FAILING_TESTS")
     end
+
+    test "marks the whole suite as failed on compilation error" do
+      in_fixture("test_failed", fn ->
+        File.write!("test/passing_and_failing_test_failed.exs", "raise ~s(oops)")
+
+        output = mix(["test"])
+        assert output =~ "** (RuntimeError) oops"
+
+        output = mix(["test", "--failed"])
+        assert output =~ "** (RuntimeError) oops"
+      end)
+    end
   end
 
   describe "--listen-on-stdin" do
@@ -279,67 +296,14 @@ defmodule Mix.Tasks.TestTest do
 
         Port.command(port, "\n")
 
-        assert receive_until_match(port, "No stale tests", "") =~ "Restarting..."
-      end)
-    end
-
-    @tag :unix
-    test "does not exit on compilation failure" do
-      in_fixture("test_stale", fn ->
-        File.write!("lib/b.ex", """
-        defmodule B do
-          def f, do: error_not_a_var
-        end
-        """)
-
-        port = mix_port(~w[test --listen-on-stdin])
-
-        assert receive_until_match(port, "error", "") =~ "lib/b.ex"
-
-        File.write!("lib/b.ex", """
-        defmodule B do
-          def f, do: A.f
-        end
-        """)
-
-        Port.command(port, "\n")
-
-        assert receive_until_match(port, "0 failures", "") =~ "2 tests"
-
-        File.write!("test/b_test_stale.exs", """
-        defmodule BTest do
-          use ExUnit.Case
-
-          test "f" do
-            assert B.f() == error_not_a_var
-          end
-        end
-        """)
-
-        Port.command(port, "\n")
-
-        message = "undefined variable \"error_not_a_var\""
-        assert receive_until_match(port, message, "") =~ "test/b_test_stale.exs"
-
-        File.write!("test/b_test_stale.exs", """
-        defmodule BTest do
-          use ExUnit.Case
-
-          test "f" do
-            assert B.f() == :ok
-          end
-        end
-        """)
-
-        Port.command(port, "\n")
-
-        assert receive_until_match(port, "0 failures", "") =~ "2 tests"
+        assert receive_until_match(port, "Restarting...", "") =~ "Restarting..."
       end)
     end
   end
 
   describe "--partitions" do
-    test "splits tests into partitions" do
+    @tag :cover
+    test "splits tests into partitions (with coverage)" do
       in_fixture("test_stale", fn ->
         assert mix(["test", "--partitions", "3", "--cover"], [{"MIX_TEST_PARTITION", "1"}]) =~
                  "1 test, 0 failures"
@@ -358,12 +322,12 @@ defmodule Mix.Tasks.TestTest do
                Importing cover results: cover/1.coverdata
                Importing cover results: cover/2.coverdata
 
-               Percentage | Module
-               -----------|--------------------------
-                  100.00% | A
-                  100.00% | B
-               -----------|--------------------------
-                  100.00% | Total
+               | Percentage | Module     |
+               |------------|------------|
+               |    100.00% | A          |
+               |    100.00% | B          |
+               |------------|------------|
+               |    100.00% | Total      |
 
                Generated HTML coverage results in \"cover\" directory
                """
@@ -544,13 +508,13 @@ defmodule Mix.Tasks.TestTest do
                Including tags: [location: {"test/bar_tests.exs", 5}]
                """
 
-        assert output =~ "4 tests, 0 failures, 3 excluded\n"
+        assert output =~ "1 test, 0 failures (3 excluded)\n"
       end)
     end
   end
 
   describe "--warnings-as-errors" do
-    test "fail on warning in tests" do
+    test "fail with exit status 1 if warning in tests but tests pass" do
       in_fixture("test_stale", fn ->
         msg =
           "Test suite aborted after successful execution due to warnings while using the --warnings-as-errors option"
@@ -571,10 +535,42 @@ defmodule Mix.Tasks.TestTest do
         end
         """)
 
-        output = mix(["test", "--warnings-as-errors", "test/warning_test_stale.exs"])
+        {output, exit_status} =
+          mix_code(["test", "--warnings-as-errors", "test/warning_test_stale.exs"])
+
         assert output =~ "variable \"unused_compile_var\" is unused"
         assert output =~ "variable \"unused_test_var\" is unused"
         assert output =~ msg
+        assert exit_status == 1
+      end)
+    end
+
+    test "fail with --exit-status + 1 if warning in tests and tests fail" do
+      in_fixture("test_stale", fn ->
+        File.write!("test/warning_test_warnings_as_errors_and_failures.exs", """
+        defmodule WarningsAsErrorsAndFailuresTest do
+          use ExUnit.Case
+
+          test "warning and failure" do
+            unused_test_var = 1
+            assert false
+          end
+        end
+        """)
+
+        {output, exit_status} =
+          mix_code([
+            "test",
+            "--warnings-as-errors",
+            "--exit-status",
+            "42",
+            "test/warning_test_warnings_as_errors_and_failures.exs"
+          ])
+
+        assert output =~ "variable \"unused_test_var\" is unused"
+        assert output =~ "1 failure"
+
+        assert exit_status == 43
       end)
     end
 
@@ -597,6 +593,25 @@ defmodule Mix.Tasks.TestTest do
         assert output =~ "2 failures"
       end)
     end
+
+    test "fail with exit status 1 if warning in test_helper.exs" do
+      in_fixture("test_stale", fn ->
+        File.write!("test/test_helper.exs", """
+        unused_var = 123
+
+        ExUnit.start()
+        """)
+
+        msg =
+          "Test suite aborted after successful execution due to warnings while using the --warnings-as-errors option"
+
+        {output, exit_status} = mix_code(["test", "--warnings-as-errors"])
+
+        assert output =~ "variable \"unused_var\" is unused"
+        assert output =~ msg
+        assert exit_status == 1
+      end)
+    end
   end
 
   describe "--exit-status" do
@@ -609,6 +624,127 @@ defmodule Mix.Tasks.TestTest do
     end
   end
 
+  describe "--max-requires" do
+    test "runs tests with --max-requires 1" do
+      # this is only a smoke test to ensure that tests run with --max-requires 1
+      # it does not test the concurrency behavior
+      in_fixture("test_stale", fn ->
+        output = mix(["test", "--max-requires", "1"])
+        assert output =~ "0 failures"
+      end)
+    end
+  end
+
+  describe "test_load_filters and test_ignore_filters" do
+    test "warns for files that are not loaded and don't match test_ignore_filters" do
+      in_tmp("test_warn", fn ->
+        File.write!("mix.exs", """
+        defmodule TestWarn.MixProject do
+          use Mix.Project
+
+          def project do
+            [
+              app: :test_warn,
+              version: "0.0.1",
+              test_load_filters: [~r/.*_tests\.exs/],
+              test_ignore_filters: [
+                "test/test_helper.exs",
+                ~r/ignored_regex/,
+                fn file -> file == "test/ignored_file.exs" end
+              ]
+            ]
+          end
+        end
+        """)
+
+        File.mkdir!("test")
+
+        File.write!("test/a_tests.exs", """
+        defmodule ATests do
+          use ExUnit.Case
+
+          test "dummy" do
+            assert true
+          end
+        end
+        """)
+
+        File.write!("test/test_helper.exs", "ExUnit.start()")
+        File.touch("test/a_missing.exs")
+        File.touch("test/a_tests.ex")
+        File.touch("test/ignored_file.exs")
+        File.touch("test/ignored_regex.exs")
+        File.write!("test/other_file.txt", "this is not a test file")
+
+        output = mix(["test"])
+
+        # This test relies on the files present in the test_warn fixture.
+        #
+        # We test that we don't warn about a_tests.exs, as it already matches the load pattern.
+        # Similarly, we ignore the empty but present ignored_file.exs and ignored_regex.exs.
+        # other_file.txt does not match the test_pattern and is ignored from the beginning.
+        #
+        # Therefore, we expect to get a warning for a_missing.exs and a_tests.ex.
+        assert output =~ """
+               the following files do not match any of the configured `:test_load_filters` / `:test_ignore_filters`:
+
+               test/a_missing.exs
+               test/a_tests.ex
+
+               This might indicate a typo\
+               """
+
+        # the dummy test ran successfully
+        assert output =~ "1 test, 0 failures"
+      end)
+    end
+
+    test "does not warn when test_ignore_filters are disabled" do
+      in_tmp("test_warn", fn ->
+        File.write!("mix.exs", """
+        defmodule TestWarn.MixProject do
+          use Mix.Project
+
+          def project do
+            [
+              app: :test_warn,
+              version: "0.0.1",
+              test_load_filters: [~r/.*_tests\.exs/],
+              test_ignore_filters: [fn _ -> true end]
+            ]
+          end
+        end
+        """)
+
+        File.mkdir!("test")
+
+        File.write!("test/a_tests.exs", """
+        defmodule ATests do
+          use ExUnit.Case
+
+          test "dummy" do
+            assert true
+          end
+        end
+        """)
+
+        File.write!("test/test_helper.exs", "ExUnit.start()")
+        File.touch("test/a_missing.exs")
+        File.touch("test/a_tests.ex")
+        File.touch("test/ignored_file.exs")
+        File.touch("test/ignored_regex.exs")
+        File.write!("test/other_file.txt", "this is not a test file")
+
+        output = mix(["test"])
+
+        refute output =~ "the following files do not match"
+
+        # the dummy test ran successfully
+        assert output =~ "1 test, 0 failures"
+      end)
+    end
+  end
+
   defp receive_until_match(port, expected, acc) do
     receive do
       {^port, {:data, output}} ->
@@ -619,6 +755,13 @@ defmodule Mix.Tasks.TestTest do
         else
           receive_until_match(port, expected, acc)
         end
+    after
+      15_000 ->
+        raise """
+        nothing received from port after 15s.
+        Expected: #{inspect(expected)}
+        Got: #{inspect(acc)}
+        """
     end
   end
 
