@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("../test_helper.exs", __DIR__)
 
 defmodule Mix.Tasks.Cheers do
@@ -8,6 +12,8 @@ defmodule Mix.UtilsTest do
   doctest Mix.Utils
 
   setup do
+    Mix.ensure_application!(:inets)
+
     # Store state before test
     mix_home = System.get_env("MIX_HOME")
 
@@ -84,7 +90,7 @@ defmodule Mix.UtilsTest do
     end)
   end
 
-  @tag unix: true
+  @tag :unix
   test "symlink or copy erases wrong symlinks" do
     in_fixture("archive", fn ->
       File.mkdir_p!("_build/archive")
@@ -100,16 +106,19 @@ defmodule Mix.UtilsTest do
     assert Mix.Utils.proxy_config("http://example.com") == []
 
     System.put_env("http_proxy", "http://nopass@example.com")
-    assert Mix.Utils.proxy_config("http://example.com") == [proxy_auth: {'nopass', ''}]
+    assert Mix.Utils.proxy_config("http://example.com") == [proxy_auth: {~c"nopass", ~c""}]
 
     System.put_env("HTTP_PROXY", "http://my:proxy@example.com")
-    assert Mix.Utils.proxy_config("http://example.com") == [proxy_auth: {'my', 'proxy'}]
+    assert Mix.Utils.proxy_config("http://example.com") == [proxy_auth: {~c"my", ~c"proxy"}]
 
     System.put_env("https_proxy", "https://another:proxy@example.com")
-    assert Mix.Utils.proxy_config("https://example.com") == [proxy_auth: {'another', 'proxy'}]
+    assert Mix.Utils.proxy_config("https://example.com") == [proxy_auth: {~c"another", ~c"proxy"}]
 
     System.put_env("HTTPS_PROXY", "https://example.com")
     assert Mix.Utils.proxy_config("https://example.com") == []
+  after
+    System.delete_env("http_proxy")
+    System.delete_env("https_proxy")
   end
 
   # 10.0.0.0 is a non-routable address
@@ -129,7 +138,7 @@ defmodule Mix.UtilsTest do
     end
 
     @tag :unix
-    test "falls back to XDG_DATA_HOME/mix" do
+    test "falls back to XDG_DATA_HOME/mix when MIX_XDG is set" do
       System.put_env("MIX_XDG", "1")
       assert Mix.Utils.mix_home() == :filename.basedir(:user_data, "mix", %{os: :linux})
     end
@@ -147,7 +156,7 @@ defmodule Mix.UtilsTest do
     end
 
     @tag :unix
-    test "falls back to XDG_CONFIG_HOME/mix" do
+    test "falls back to XDG_CONFIG_HOME/mix when MIX_XDG is set" do
       System.put_env("MIX_XDG", "1")
       assert Mix.Utils.mix_config() == :filename.basedir(:user_config, "mix", %{os: :linux})
     end
@@ -157,6 +166,68 @@ defmodule Mix.UtilsTest do
     end
   end
 
+  describe "mix_cache/0" do
+    @tag :unix
+    test "prefers XDG_CACHE_HOME/mix when MIX_XDG is set" do
+      System.put_env("MIX_XDG", "1")
+      assert Mix.Utils.mix_cache() == :filename.basedir(:user_cache, "mix", %{os: :linux})
+    end
+
+    test "falls back to user cache dir" do
+      assert Mix.Utils.mix_cache() == :filename.basedir(:user_cache, "mix")
+    end
+  end
+
+  describe "write_according_to_opts/3" do
+    test "verify that file writes with backups work as expected" do
+      test_out = "test.out"
+      test_out_bak = "test.out.bak"
+      hello_world = "Hello World!"
+      new_hello_world = "New Hello World!"
+      default_out = "default.out"
+
+      # ignore any error from this call.
+      in_tmp("write to default file", fn ->
+        # no optional override - write to the specified default file
+        assert Mix.Utils.write_according_to_opts!(test_out, [hello_world], []) == test_out
+        assert File.read!(test_out) == hello_world
+
+        # no optional override - write to the specified default file again, with old file backed up
+        assert Mix.Utils.write_according_to_opts!(test_out, [new_hello_world], []) == test_out
+        assert File.read!(test_out) == new_hello_world
+        assert File.read!(test_out_bak) == hello_world
+      end)
+
+      in_tmp("write to optional file override", fn ->
+        # with optional override - write to the specified default file
+        assert Mix.Utils.write_according_to_opts!(default_out, [hello_world], output: test_out) ==
+                 test_out
+
+        assert File.read!(test_out) == hello_world
+
+        # with optional override - write to the specified default file again, with old file backed up
+        assert Mix.Utils.write_according_to_opts!(default_out, [new_hello_world],
+                 output: test_out
+               ) ==
+                 test_out
+
+        assert File.read!(test_out) == new_hello_world
+        assert File.read!(test_out_bak) == hello_world
+      end)
+    end
+
+    test "verify that writing to STDOUT works as expected" do
+      output =
+        ExUnit.CaptureIO.capture_io(fn ->
+          Mix.Utils.write_according_to_opts!("the_file.txt", ["some text"], output: "-")
+        end)
+
+      assert output == "some text"
+
+      refute File.exists?("the_file.txt")
+    end
+  end
+
   defp assert_ebin_symlinked_or_copied(result) do
     case result do
       {:ok, paths} ->
@@ -170,7 +241,7 @@ defmodule Mix.UtilsTest do
               "ebin" |> Path.expand() |> String.to_charlist()
 
             _ ->
-              '../../ebin'
+              ~c"../../ebin"
           end
 
         {:ok, actual_link} = :file.read_link("_build/archive/ebin")
