@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 Code.require_file("test_helper.exs", __DIR__)
 
 defmodule MixTest do
@@ -326,7 +330,7 @@ defmodule MixTest do
       end
     end
 
-    test "restore dir", %{tmp_dir: tmp_dir} do
+    test "using restore dir", %{tmp_dir: tmp_dir} do
       with_cleanup(fn ->
         Mix.install([
           {:git_repo, git: fixture_path("git_repo")}
@@ -393,6 +397,76 @@ defmodule MixTest do
       System.delete_env("MIX_INSTALL_RESTORE_PROJECT_DIR")
     end
 
+    test "using restore dir removes consolidated files when not needed", %{tmp_dir: tmp_dir} do
+      with_cleanup(fn ->
+        Mix.install([
+          {:install_test, path: Path.join(tmp_dir, "install_test")}
+        ])
+
+        install_project_dir = Mix.install_project_dir()
+        build_lib_path = Path.join([install_project_dir, "_build", "dev", "lib"])
+        build_project_path = Path.join(build_lib_path, "mix_install")
+
+        assert File.ls!(build_project_path) |> Enum.sort() == [".mix", "consolidated", "ebin"]
+
+        System.put_env("MIX_INSTALL_RESTORE_PROJECT_DIR", install_project_dir)
+      end)
+
+      with_cleanup(fn ->
+        Mix.install(
+          [
+            {:install_test, path: Path.join(tmp_dir, "install_test")}
+          ],
+          consolidate_protocols: false
+        )
+
+        install_project_dir = Mix.install_project_dir()
+        build_lib_path = Path.join([install_project_dir, "_build", "dev", "lib"])
+        build_project_path = Path.join(build_lib_path, "mix_install")
+
+        assert File.ls!(build_project_path) |> Enum.sort() == [".mix", "ebin"]
+      end)
+    after
+      System.delete_env("MIX_INSTALL_RESTORE_PROJECT_DIR")
+    end
+
+    test "custom compilers", %{tmp_dir: tmp_dir} do
+      File.mkdir_p!("#{tmp_dir}/install_test/lib/mix/tasks/compile/")
+
+      File.write!("#{tmp_dir}/install_test/lib/mix/tasks/compile/install_test.ex", """
+      defmodule Mix.Tasks.Compile.InstallTest do
+        use Mix.Task.Compiler
+
+        def run(_args) do
+          Mix.shell().info("Hello from custom compiler!")
+
+          :noop
+        end
+      end
+      """)
+
+      Mix.install(
+        [
+          {:install_test, path: Path.join(tmp_dir, "install_test")}
+        ],
+        compilers: [:elixir, :install_test]
+      )
+
+      assert File.dir?(Path.join(tmp_dir, "installs"))
+
+      assert Protocol.consolidated?(InstallTest.Protocol)
+
+      assert_received {:mix_shell, :info, ["==> install_test"]}
+      assert_received {:mix_shell, :info, ["Compiling 3 files (.ex)"]}
+      assert_received {:mix_shell, :info, ["Generated install_test app"]}
+      assert_received {:mix_shell, :info, ["==> mix_install"]}
+      assert_received {:mix_shell, :info, ["Hello from custom compiler!"]}
+      refute_received _
+
+      assert List.keyfind(Application.started_applications(), :install_test, 0)
+      assert apply(InstallTest, :hello, []) == :world
+    end
+
     test "installed?", %{tmp_dir: tmp_dir} do
       refute Mix.installed?()
 
