@@ -1,3 +1,7 @@
+# SPDX-License-Identifier: Apache-2.0
+# SPDX-FileCopyrightText: 2021 The Elixir Team
+# SPDX-FileCopyrightText: 2012 Plataformatec
+
 home = Path.expand("../tmp/.home", __DIR__)
 File.mkdir_p!(home)
 System.put_env("HOME", home)
@@ -6,9 +10,6 @@ mix = Path.expand("../tmp/.mix", __DIR__)
 File.mkdir_p!(mix)
 System.put_env("MIX_HOME", mix)
 
-System.delete_env("XDG_DATA_HOME")
-System.delete_env("XDG_CONFIG_HOME")
-
 # Load protocols to make sure they are not unloaded during tests
 [Collectable, Enumerable, Inspect, String.Chars, List.Chars]
 |> Enum.each(& &1.__protocol__(:module))
@@ -18,9 +19,7 @@ System.delete_env("XDG_CONFIG_HOME")
 Mix.start()
 Mix.shell(Mix.Shell.Process)
 Application.put_env(:mix, :colors, enabled: false)
-
-Logger.remove_backend(:console)
-Application.put_env(:logger, :backends, [])
+Application.put_env(:logger, :default_handler, false)
 
 ## Setup ExUnit
 
@@ -44,20 +43,16 @@ cover_exclude =
     []
   end
 
+Code.require_file("../../elixir/scripts/cover_record.exs", __DIR__)
+CoverageRecorder.maybe_record("mix")
+
 ExUnit.start(
   trace: !!System.get_env("TRACE"),
   exclude: epmd_exclude ++ os_exclude ++ git_exclude ++ line_exclude ++ cover_exclude,
-  include: line_include
+  include: line_include,
+  assert_receive_timeout: String.to_integer(System.get_env("ELIXIR_ASSERT_TIMEOUT", "300"))
 )
 
-# Clear environment variables that may affect tests
-System.delete_env("http_proxy")
-System.delete_env("https_proxy")
-System.delete_env("HTTP_PROXY")
-System.delete_env("HTTPS_PROXY")
-System.delete_env("MIX_ENV")
-System.delete_env("MIX_TARGET")
-
 defmodule MixTest.Case do
   use ExUnit.CaseTemplate
 
@@ -182,7 +177,10 @@ defmodule MixTest.Case do
   end
 
   def ensure_touched(file, current) when is_binary(current) do
-    ensure_touched(file, File.stat!(current).mtime)
+    case File.stat(current) do
+      {:ok, %{mtime: mtime}} -> ensure_touched(file, mtime)
+      {:error, _} -> File.touch!(file)
+    end
   end
 
   def ensure_touched(file, current) when is_tuple(current) do
@@ -230,11 +228,11 @@ defmodule MixTest.Case do
     File.write!(file, File.read!(file) <> "\n")
   end
 
-  defp mix_executable do
+  def mix_executable do
     Path.expand("../../../bin/mix", __DIR__)
   end
 
-  defp elixir_executable do
+  def elixir_executable do
     Path.expand("../../../bin/elixir", __DIR__)
   end
 
@@ -250,6 +248,20 @@ defmodule MixTest.Case do
   end
 end
 
+# Prepare and clear environment variables
+System.put_env(
+  "MIX_OS_DEPS_COMPILE_PARTITION_ELIXIR_EXECUTABLE",
+  MixTest.Case.elixir_executable()
+)
+
+# Clear environment variables that may affect tests
+Enum.each(
+  ~w(http_proxy https_proxy HTTP_PROXY HTTPS_PROXY) ++
+    ~w(MIX_ENV MIX_OS_DEPS_COMPILE_PARTITION_COUNT MIX_TARGET) ++
+    ~w(XDG_DATA_HOME XDG_CONFIG_HOME),
+  &System.delete_env/1
+)
+
 ## Set up Rebar fixtures
 
 rebar3_source = System.get_env("REBAR3") || Path.expand("fixtures/rebar3", __DIR__)
@@ -419,8 +431,7 @@ if not File.dir?(target) do
   File.mkdir_p!(Path.join(target, "ebin"))
   File.mkdir_p!(Path.join(target, "src"))
 
-  # This is used to test that the built-in ebin is ignored
-  # when build_embedded is true.
+  # This is used to test that the built-in ebin is ignored.
   File.write!(Path.join(target, "ebin/.unused"), """
   """)
 
